/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope;

import org.apache.commons.cli.ParseException;
import org.apache.commons.io.FilenameUtils;

import org.apache.commons.lang3.time.DurationFormatUtils;
import ru.ispras.fortress.solver.Environment;
import ru.ispras.fortress.solver.Solver;
import ru.ispras.fortress.solver.SolverId;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Configuration;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.EngineEventHandler;
import ru.ispras.retrascope.basis.EngineRegistry;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.Parameters;
import ru.ispras.retrascope.basis.ToolChain;
import ru.ispras.retrascope.engine.cfg.extractor.iface.CfgInterfaceExtractor;
import ru.ispras.retrascope.engine.cfg.generator.test.random.CfgRandomTestGenerator;
import ru.ispras.retrascope.engine.cfg.printer.graphml.CfgGraphMlPrinter;
import ru.ispras.retrascope.engine.cfg.printer.graphml.depend.CfgDependencyGraphMlPrinter;
import ru.ispras.retrascope.engine.cfg.transformer.cgaa.CfgCgaaTransformer;
import ru.ispras.retrascope.engine.cgaa.printer.graphml.CgaaGraphMlPrinter;
import ru.ispras.retrascope.engine.cgaa.transformer.efsm.CgaaEfsmTransformer;
import ru.ispras.retrascope.engine.cgaa.transformer.hldd.CgaaHlddTransformer;
import ru.ispras.retrascope.engine.efsm.extractor.assertion.EfsmTransitionAssertExtractor;
import ru.ispras.retrascope.engine.efsm.generator.test.EfsmTestGenerator;
import ru.ispras.retrascope.engine.efsm.generator.test.fate.EfsmFatePlusTestGenerator;
import ru.ispras.retrascope.engine.efsm.generator.test.fate.EfsmFateTestGenerator;
import ru.ispras.retrascope.engine.efsm.printer.graphml.EfsmGraphMlPrinter;
import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulator;
import ru.ispras.retrascope.engine.hldd.printer.smv.HlddSmvPrinter;
import ru.ispras.retrascope.engine.hldd.printer.smv.assertion.HlddAssertSmvPrinter;
import ru.ispras.retrascope.engine.smv.launcher.SmvModelCheckerLauncher;
import ru.ispras.retrascope.engine.smv.test.parser.SmvTestParser;
import ru.ispras.retrascope.engine.test.printer.testbench.verilog.TestVerilogTestbenchPrinter;
import ru.ispras.retrascope.engine.test.printer.testbench.vhdl.TestVhdlTestbenchPrinter;
import ru.ispras.retrascope.parser.verilog.VerilogParser;
import ru.ispras.retrascope.parser.vhdl.VhdlParser;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.io.File;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.logging.Level;

/**
 * The main class of the project.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class Retrascope {

  /* The command line format. */
  public static final String FORMAT = "[options] files";

  /* The name of the solver path property. */
  public static final String SOLVER_PROPERTY = "smt.solver.path";

  /* The name of the solver path environment variable. */
  public static final String SOLVER_VARIABLE = "SMT_SOLVER_PATH";

  /* The command line parameters. */
  public static final Parameter TARGET_PARAM =
      new Parameter("target", true, "Set a target entity");
  public static final Parameter ENGINE_PARAM =
      new Parameter("engine", true, "Set a subset of engines");
  public static final Parameter LOG_PARAM =
      new Parameter("log", true, "Set a log file");
  public static final Parameter LOG_LEVEL =
      new Parameter("log-level", true, "Set a log level");
  public static final Parameter HELP_PARAM =
      new Parameter("help", false, "Show this message");
  public static final Parameter SOLVER_DEBUG =
      new Parameter("solver-debug", false, "Set debug mode for SMT solver");
  public static final Parameter TOOL_DEBUG_PARAM =
      new Parameter("tool-debug-file", true, "Set debug mode and save info to debug log file");

  public static final Parameters PARAMETERS = new Parameters();

  /* The tool configuration. */
  private static final Configuration CONFIG = new Configuration();

  /* The engine registry. */
  private static final EngineRegistry ENGINE_REGISTRY = new EngineRegistry();

  static {

    /* Register parameters. */

    PARAMETERS.addParameter(TARGET_PARAM);
    PARAMETERS.addParameter(ENGINE_PARAM);
    PARAMETERS.addParameter(LOG_PARAM);
    PARAMETERS.addParameter(LOG_LEVEL);
    PARAMETERS.addParameter(HELP_PARAM);
    PARAMETERS.addParameter(SOLVER_DEBUG);
    PARAMETERS.addParameter(TOOL_DEBUG_PARAM);

    /* Register engines. */

    // HDL Parsers.
    ENGINE_REGISTRY.add(new VerilogParser());
    ENGINE_REGISTRY.add(new VhdlParser());
    // CFG Engines.
    ENGINE_REGISTRY.add(new CfgGraphMlPrinter());
    ENGINE_REGISTRY.add(new CfgDependencyGraphMlPrinter());
    ENGINE_REGISTRY.add(new CfgCgaaTransformer());
    ENGINE_REGISTRY.add(new CfgInterfaceExtractor());
    ENGINE_REGISTRY.add(new CfgRandomTestGenerator());
    // CGAA Engines.
    ENGINE_REGISTRY.add(new CgaaGraphMlPrinter());
    ENGINE_REGISTRY.add(new CgaaEfsmTransformer());
    ENGINE_REGISTRY.add(new CgaaHlddTransformer());
    // EFSM Engines.
    ENGINE_REGISTRY.add(new EfsmGraphMlPrinter());
    ENGINE_REGISTRY.add(new EfsmFateTestGenerator());
    ENGINE_REGISTRY.add(new EfsmFatePlusTestGenerator());
    ENGINE_REGISTRY.add(new EfsmTestGenerator());
    ENGINE_REGISTRY.add(new EfsmSimulator());
    ENGINE_REGISTRY.add(new EfsmTransitionAssertExtractor());
    // HLDD Engines.
    ENGINE_REGISTRY.add(new HlddSmvPrinter());
    ENGINE_REGISTRY.add(new HlddAssertSmvPrinter());
    // SMV Engines.
    ENGINE_REGISTRY.add(new SmvModelCheckerLauncher());
    ENGINE_REGISTRY.add(new SmvTestParser());
    // testbench engines
    ENGINE_REGISTRY.add(new TestVerilogTestbenchPrinter());
    ENGINE_REGISTRY.add(new TestVhdlTestbenchPrinter());
  }

  private Retrascope() {
    /* Empty. */
  }

  /**
   * The tool run information.
   */
  public static class ToolRun {
    // The tool chain.
    private final ToolChain toolChain;
    // The source entities.
    private final Map<EntityType, Entity> sources;

    /**
     * Constructs a tool run.
     *
     * @param toolChain the tool chain to be used.
     * @param sources the source entities to be analyzed.
     */
    ToolRun(final ToolChain toolChain, final Map<EntityType, Entity> sources) {
      this.toolChain = toolChain;
      this.sources = sources;
    }

    /**
     * Construct a tool run with unspecified tool chain.
     */
    ToolRun() {
      this(null, null);
    }

    /**
     * Returns the tool chain to be used.
     *
     * @return the tool chain.
     */
    ToolChain getEngine() {
      return toolChain;
    }

    /**
     * Returns the sources entities to be analyzed.
     *
     * @return the source entities.
     */
    public Map<EntityType, Entity> getSources() {
      return sources;
    }

    /**
     * Starts the tool chain with the given source entities.
     *
     * @param eventHandler the engine event handler.
     */
    void start(final EngineEventHandler eventHandler) {
      if (toolChain != null) {
        toolChain.start(sources, eventHandler);
      }
    }
  }

  /**
   * Returns the engine registry.
   *
   * @return the engine registry.
   */
  public static EngineRegistry getEngineRegistry() {
    return ENGINE_REGISTRY;
  }

  /**
   * Returns the tool configuration.
   *
   * @return the tool configuration
   */
  public static Configuration getConfiguration() {
    return CONFIG;
  }

  /**
   * Analyzes the command line and returns the run information.
   *
   * @param args the command line.
   * @return the run information.
   */
  private static ToolRun getRun(final String... args) {
    final Arguments arguments = new Arguments(PARAMETERS);

    try {
      // Parse the command line.
      arguments.setCommandLine(args);
    } catch (ParseException e) {
      return new ToolRun();
    }

    // Fill the configuration.
    CONFIG.configure(args);

    // Get the target entity types.
    final Set<EntityType> targets = new HashSet<>();

    if (arguments.contains(TARGET_PARAM)) {
      final StringTokenizer tokenizer =
          new StringTokenizer(arguments.getValue(TARGET_PARAM), File.pathSeparator);

      while (tokenizer.hasMoreTokens()) {
        targets.add(EntityType.get(tokenizer.nextToken()));
      }
    }

    // Get the engines to be used.
    final Set<Engine> engines = new HashSet<>();

    if (arguments.contains(ENGINE_PARAM)) {
      final StringTokenizer tokenizer =
          new StringTokenizer(arguments.getValue(ENGINE_PARAM), File.pathSeparator);

      while (tokenizer.hasMoreTokens()) {
        final String engineId = tokenizer.nextToken();
        final Engine engine = ENGINE_REGISTRY.getEngine(engineId);

        if (engine == null) {
          Log.getLogger()
              .log(LogLevel.ERROR, String.format("The engine '%s' is unknown", engineId));
        } else {
          engines.add(engine);
        }
      }
    }

    // Get the list of file names.
//    final String[] files = arguments.getArguments();
    final String[] files = getFileNamesOnly(arguments);

    // Construct the map of source entities.
    final Map<EntityType, Entity> sources = getSourceEntities(arguments, files);

    // Get the tool chain for the specified target.
    final ToolChain toolChain = ENGINE_REGISTRY.getEngine(targets, engines, sources.keySet());

    return new ToolRun(toolChain, sources);
  }

  private static String[] getFileNamesOnly(final Arguments arguments) {
    final String[] commandLine = arguments.getCommandLine();
    final String[] files = arguments.getArguments();
    final ArrayList<String> excluded = new ArrayList<>();
    final ArrayList<String> result = new ArrayList<>();

    for (int i = 1; i < commandLine.length; i++) {
      if (commandLine[i - 1].startsWith("--")) {
        excluded.add(commandLine[i]);
      }
    }

    for (final String file : files) {
      if (!excluded.contains(file)) {
        result.add(file);
      }
    }

    return result.toArray(new String[result.size()]);
  }

  private static Map<EntityType, Entity> getSourceEntities(
      final Arguments arguments,
      final String ... files) {

    final Map<EntityType, Entity> sources = new HashMap<>();

    // Add the arguments as one of the entities.
    sources.put(arguments.getEntityType(), arguments);

    // Divide the file names into groups according to the extension.
    for (final String file : files) {
      final String extension = FilenameUtils.getExtension(file);
      final EntityType source = EntityType.get(extension);

      FileNames fileNames = (FileNames) sources.get(source);

      if (fileNames == null) {
        fileNames = new FileNames(extension);
      }

      fileNames.add(file);
      sources.put(source, fileNames);
    }
    return sources;
  }

  /**
   * Configures the Fortress environment.
   */
  private static void configureSolverPath() {
    String solverPath = System.getProperties().getProperty(SOLVER_PROPERTY);

    if (solverPath == null) {
      solverPath = System.getenv(SOLVER_VARIABLE);
    }

    if (solverPath == null) {
      Log.getLogger().log(LogLevel.ERROR,
          String.format("Neither property '%s' nor variable '%s' is set",
              SOLVER_PROPERTY, SOLVER_VARIABLE));
    }

    final Solver z3Solver = SolverId.Z3_TEXT.getSolver();
    z3Solver.setSolverPath(solverPath);
  }

  /**
   * Configures and runs the tool chain.
   *
   * @param args the command line.
   */
  public static void main(final String ... args) {
    main(args, null);
  }

  /**
   * Configures and runs the tool chain.
   *
   * @param args the command line.
   * @param eventHandler the engine event handler.
   */
  private static void main(final String[] args, final EngineEventHandler eventHandler) {
    final Arguments arguments = new Arguments(PARAMETERS);

    try {
      // Configure the solver path.
      configureSolverPath();
      // Parse the command line.
      arguments.setCommandLine(args);
    } catch (ParseException e) {
      PARAMETERS.help(FORMAT);
    }

    Log.setLevel(getLogLevel(arguments));

    if (arguments.contains(LOG_PARAM)) {
      Log.addFileOutput(arguments.getValue(LOG_PARAM));
    }

    if (arguments.contains(SOLVER_DEBUG)) {
      Environment.setDebugMode(true);
    }

    // Handle the help option.
    if (arguments.contains(HELP_PARAM)) {
      PARAMETERS.help(FORMAT);
    }

    // Handle the target option.
    Log.getLogger().log(LogLevel.INFO, "Retrascope is starting");
    final Date startDate = new Date(System.currentTimeMillis());

    final ToolRun run = getRun(args);

    if (run.getEngine() == null) {
      Log.getLogger().log(LogLevel.ERROR,
          String.format("The targets '%s' with engines '%s' are unreachable",
          arguments.contains(TARGET_PARAM) ? arguments.getValue(TARGET_PARAM) : "unknown",
          arguments.contains(ENGINE_PARAM) ? arguments.getValue(ENGINE_PARAM) : "unknown"));
    } else {
      run.start(eventHandler);
    }

    final Date stopDate = new Date(System.currentTimeMillis());
    Log.getLogger().log(LogLevel.INFO, "Retrascope is shutting down");

    Log.getLogger().log(LogLevel.INFO, "Duration: " + getDifference(startDate, stopDate));
    Log.closeFileHandlers();
  }

  private static Level getLogLevel(final Arguments arguments) {
    return arguments.contains(TOOL_DEBUG_PARAM)
      ? LogLevel.DEBUG
      : arguments.contains(LOG_LEVEL)
      ? LogLevel.parse(arguments.getValue(LOG_LEVEL))
      : LogLevel.INFO;
  }

  private static String getDifference(final Date startDate, final Date stopDate) {
    final long diffTime = stopDate.getTime() - startDate.getTime();

    return DurationFormatUtils.formatDurationHMS(diffTime);
  }
}
