/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.basis;

import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * This class implements a tool chain, which is an {@link Engine} engine composed of smaller ones.
 * Processing of an input entity is done by one-by-one execution of the chain's engines.
 *
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public class ToolChain extends Engine {

  // The list of the engines to be applied.
  private final List<Engine> engines = new ArrayList<>();

  /**
   * This class implements internal event handler.
   */
  private class InternalEventHandler implements EngineEventHandler {

    // The index the engine in the tool chain.
    private final int index;

    // The overall number of engines in the tool chain.
    private final int count;

    /**
     * Constructs an internal event handler.
     *
     * @param index the index of the engine in the tool chain.
     * @param count the overall number of engines in the tool chain.
     */
    InternalEventHandler(final int index, final int count) {
      this.index = index;
      this.count = count;
    }

    private int getIndex() {
      return this.index;
    }

    private int getCount() {
      return this.count;
    }

    @Override
    public void onSetProgress(final Engine engine, final int progress) {
      setProgress((100 * getIndex() + progress) / getCount());
    }
  }

  /**
   * Constructs a tool chain with a given id.
   *
   * @param toolChainId the tool chain identifier.
   */
  private ToolChain(final String toolChainId) {
    super(toolChainId);
  }

  /**
   * Constructs a tool chain with the default id.
   */
  public ToolChain() {
    this("");
  }

  @Override
  public Entity start(final Map<EntityType, Entity> inputs) {
    // The output entity.
    Entity output = null;

    // Create a map of constructed entities.
    final Map<EntityType, Entity> results = new HashMap<>(inputs);

    // Process all registered engines one-by-one.
    for (int i = 0; i < engines.size(); i++) {
      // Get the engine.
      final Engine engine = engines.get(i);
      // Get the engine input types.
      final Set<EntityType> inputTypes = engine.getInputTypes();

      // Compose the engine's arguments.
      final Map<EntityType, Entity> arguments = compose(inputs, results, inputTypes);

      // Apply the engine with the given arguments.
      Log.getLogger().log(LogLevel.INFO, String.format("Running: %s%n", engine.getId()));
      Log.getLogger().log(LogLevel.INFO, String.format("Options: %s%n", arguments.toString()));
      output = engine.start(arguments, new InternalEventHandler(i, engines.size()));

      final EntityType outputType = output.getEntityType();

      // Update the tool chain's results.
      final Entity result = results.get(outputType);

      if (result == null) {
        // Put new results into the map.
        Log.getLogger().log(LogLevel.INFO, String.format("Storing: %s%n", outputType.toString()));
        results.put(output.getEntityType(), output);
      } else {
        // Merge new and old results.
        Log.getLogger().log(LogLevel.INFO, String.format("Merging: %s%n", outputType.toString()));
        result.add(output);
      }
    }

    // Return result of the last engine.
    return output;
  }

  private Map<EntityType, Entity> compose(final Map<EntityType, Entity> inputs,
      final Map<EntityType, Entity> results, final Set<EntityType> inputTypes) {

    final Map<EntityType, Entity> arguments = new HashMap<>();
    final EntityType args = Arguments.TYPE;
    arguments.put(args, inputs.get(args));

    for (final EntityType inputType : inputTypes) {
      arguments.put(inputType, results.get(inputType));
    }

    return arguments;
  }

  /**
   * Returns all engines of the tool chain.
   *
   * @return the engines of the tool chain.
   */
  public List<Engine> getEngines() {
    return engines;
  }

  /**
   * Registers the given engine in the tool chain taking into account data dependencies.
   * The implementation is not effective, but performance is not required.
   *
   * @param engine the engine to be registered.
   */
  public void addEngine(final Engine engine) {
    int shouldBeBefore = -1;
    final Collection<Engine> shouldBeAfter = new LinkedHashSet<>();

    // The engine is registered in a proper position of the list.
    for (int i = 0; i < engines.size(); i++) {
      final Engine current = engines.get(i);

      // The engine should be called before its output used.
      if (current.uses(engine.getOutputType()) && shouldBeBefore == -1) {
        shouldBeBefore = i;
      }

      // The engine should be called after its input is defined.
      if (engine.uses(current.getOutputType())) {
        shouldBeAfter.add(current);
      }
    }

    // There are no engines that use the output of the engine begin added (X).
    if (shouldBeBefore == -1) {

      // Append X to the end of the list.
      engines.add(engine);

      // There is an engine (U) that uses the output of the engine being added (X).
    } else {

      // Insert X just before U.
      engines.add(shouldBeBefore, engine);

      for (final Engine afterEngine : shouldBeAfter) {

        // There is an engine (D) that defines the input of the engine being added (X).
        if (engines.indexOf(afterEngine) > shouldBeBefore) {
          // Remove D and add it once again (it should be inserted into a proper position).
          engines.remove(afterEngine);
          addEngine(afterEngine);
        }
      }
    }
  }
}
