/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.basis;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.walker.ModelVisitor;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;

import java.util.Arrays;
import java.util.Map;

/**
 * The basic engine that applies several walkers on the target model.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class ModelWalkerEngine extends Engine {

  private ModelWalker[] walkers;

  /**
   * Constructs object with the specified identifier, input/output types,
   * walkers.
   *
   * @param engineId Engine identifier.
   * @param inputType Engine model input type.
   * @param outputType Engine model output type.
   * @param walkers Model walkers.
   *
   * @throws IllegalArgumentException when walkers array is {@code null}.
   *
   */
  protected ModelWalkerEngine(
      final String engineId,
      final EntityType inputType,
      final EntityType outputType,
      final ModelWalker... walkers) {

    super(engineId);

    InvariantChecks.checkNotNull(walkers);

    this.addInputType(inputType);
    this.setOutputType(outputType);

    this.walkers = walkers;
  }

  /**
   * Initializes engine with the specified inputs.
   *
   * @param inputs Engine inputs.
   */
  protected abstract void initialize(final Map<EntityType, Entity> inputs);

  @Override
  public Entity start(final Map<EntityType, Entity> inputs) {

    initialize(inputs);

    final EntityType inputModelType = getInputTypes().iterator().next();
    final Model model = (Model) inputs.get(inputModelType);

    for (final ModelWalker walker : getWalkers()) {
      walker.start(model);
    }
    return getOutput();
  }

  /**
   * Returns engine output entity.
   *
   * @return Output entity.
   */
  protected abstract Entity getOutput();

  /**
   * Return engine walkers.
   *
   * @return Internal walkers.
   */
  protected ModelWalker[] getWalkers() {
    return Arrays.copyOf(this.walkers, this.walkers.length);
  }

  /**
   * Sets the specified walker sequence.
   *
   * @param walkers New sequence of walkers.
   */
  protected final void setWalkers(final ModelWalker... walkers) {
    this.walkers = walkers;
  }

  /**
   * Returns initial visitor.
   *
   * @return Visitor.
   */
  protected final ModelVisitor getInitVisitor() {
    return this.walkers[0].getVisitor();
  }
}
