/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.basis;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.State;
import ru.ispras.retrascope.model.basis.memory.HashMapMemory;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.TestCoverage;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The class representing an abstract simulator.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public abstract class Simulator<T extends DecStatement> {

  protected final T simulatedModel;

  private HashMapMemory inputsMemory;
  protected HashMapMemory innerVariablesMemory;
  protected HashMapMemory outputsMemory;

  /**
   * Constructs a simulator for the specified model.
   *
   * @param simulatedModel the simulated model
   */
  protected Simulator(final T simulatedModel) {
    this.simulatedModel = simulatedModel;

    this.inputsMemory = new HashMapMemory(simulatedModel.getInputNames());
    this.innerVariablesMemory = new HashMapMemory(simulatedModel.getRegisterNames());
    this.outputsMemory = new HashMapMemory(simulatedModel.getOutputNames());
  }

  /**
   * Returns a model which is simulated by this simulator.
   *
   * @return the simulated model
   */
  public abstract T getSimulatedModel();

  /**
   * Simulates processing of the specified input vector and returns produced output vector.
   *
   * @param inputVector the vector to be processed
   *
   * @return the output vector
   *
   * @throws IllegalArgumentException if {@code inputVector} is {@code null}
   */
  public abstract Vector simulateVector(final Vector inputVector);

  /**
   * Simulates processing of the specified input vector and returns produced output vector.
   *
   * @param inputSequence the vector to be processed
   *
   * @return the output vector
   *
   * @throws IllegalArgumentException if {@code inputSequence} is {@code null}
   */
  public abstract Sequence simulateSequence(final Sequence inputSequence);

  /**
   * Resets this simulator, that is, returns it at the same state as it had immediately after
   * construction. Default implementation only sets variables of this simulator to their initial
   * values.
   */
  protected void reset() {
    this.resetInputs();

    for (final String register : this.simulatedModel.getRegisterNames()) {
      final NodeValue initialValue = (NodeValue) this.simulatedModel.getInitialValue(register);
      this.innerVariablesMemory.setValue(register, initialValue);
    }

    for (final String output : this.simulatedModel.getOutputNames()) {
      this.outputsMemory.setValue(output, (NodeValue) this.simulatedModel.getInitialValue(output));
    }
  }

  /**
   * Sets all inputs of this simulator to their initial values.
   */
  private void resetInputs() {
    for (final String input : this.simulatedModel.getInputNames()) {
      final NodeValue initialValue = (NodeValue) this.simulatedModel.getInitialValue(input);
      this.inputsMemory.setValue(input, initialValue);
    }
  }

  /**
   * Returns a test coverage achieved by this simulator since the last reset (if any).
   *
   * @return the test coverage
   */
  public abstract TestCoverage getCoverage();

  /**
   * Resets a test coverage achieved by this simulator.
   */
  public abstract void resetCoverage();

  /**
   * Returns a state of an underlying model which this simulator is currently in.
   *
   * @return the current state of this simulator
   */
  public abstract State getCurrentState();

  protected Node substituteVariables(final Node expression,
      final Map<String, NodeValue> bufferedVariables) {

    final Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(expression));
    final List<BoundVariable> bindingList = new ArrayList<>(usedVariables.size());

    for (final NodeVariable usedVariable : usedVariables) {
      final String variableName = usedVariable.getName();
      NodeValue variableValue = null;

      if (bufferedVariables.containsKey(variableName)) {
        variableValue = bufferedVariables.get(variableName);
      } else {
        if (this.hasVariable(variableName)) {
          variableValue = this.getVariableValue(variableName);
        } /*else {

          throw RetrascopeRuntimeException
              .makeException(LogMessage.CANT_FIND, variableName + " variable");
        }*/
      }

      if (variableValue != null) {
        bindingList.add(NodeBinding.bindVariable(usedVariable, variableValue));
      }
    }

    return Transformer.substituteAllBindings(new NodeBinding(expression, bindingList));
  }

  protected void readVector(final Vector vector) {
    InvariantChecks.checkNotNull(vector);

    for (final String inputName : this.simulatedModel.getInputNames()) {
      this.inputsMemory.setValue(inputName, (NodeValue) vector.getValue(inputName));
    }

  }

  /**
   * Returns output vector that is produced by this object.
   *
   * @return output vector
   */
  protected Vector getOutputVector() {
    final Transaction transaction = new Transaction();

    for (final String variableName: this.simulatedModel.getOutputNames()) {
      transaction.addVariable(variableName, this.outputsMemory.getValue(variableName));
    }

    return new Vector(transaction);
  }

  /**
   * Returns a collection of inner variables' names.
   *
   * @return inner variables names
   */
  public Set<String> getInnerVariableNames() {
    return this.innerVariablesMemory.getValues().keySet();
  }

  /**
   * Returns a collection of output variables' names.
   *
   * @return output variables names
   */
  public Set<String> getOutputVariableNames() {
    return this.outputsMemory.getValues().keySet();
  }

  /**
   * Checks rather this operates with input variable that has the specified name.
   *
   * @param variableName Input variable's name.
   *
   * @return {@code true} when it operates with input variable
   *         that has the specified name, {@code false} otherwise.
   */
  protected boolean hasInput(final String variableName) {
    return this.inputsMemory.containsVariable(variableName);
  }

  /**
   * Checks rather this operates with inner variable that has the specified name.
   *
   * @param variableName Inner variable's name.
   *
   * @return {@code true} when it operates with inner variable
   *         that has the specified name, {@code false} otherwise.
   */
  protected boolean hasInnerVariable(final String variableName) {
    return this.innerVariablesMemory.containsVariable(variableName);
  }

  /**
   * Checks rather this operates with output variable that has the specified name.
   *
   * @param variableName Output variable's name.
   *
   * @return {@code true} when it operates with output variable
   *         that has the specified name, {@code false} otherwise.
   */
  protected boolean hasOutput(final String variableName) {
    return this.outputsMemory.containsVariable(variableName);
  }

  /**
   * Checks rather this object operates with variable of the specified name.
   *
   * @param name Variable name.
   * @return {@code true} when this object operates with variable that has the specified name,
   *         {@code false} otherwise.
   */
  boolean hasVariable(final String name) {
    return hasInput(name) || hasInnerVariable(name) || hasOutput(name);
  }

  /**
   * Returns value of variable with the specified name.
   *
   * @param variableName variable name
   *
   * @return variable value, if this operates with it,
   *         {@code null} otherwise
   */
  public NodeValue getVariableValue(final String variableName) {
    if (this.hasInput(variableName)) {
      return this.inputsMemory.getValue(variableName);
    } else {
      if (this.hasInnerVariable(variableName)) {
        return this.innerVariablesMemory.getValue(variableName);
      } else {
        if (this.hasOutput(variableName)) {
          return this.outputsMemory.getValue(variableName);
        } else {
          return null;
        }
      }
    }
  }

  /**
   * The snapshot of simulation.
   *
   * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
   *
   * @see Simulator
   */
  public class SimulationSnapshot {
    private final HashMapMemory inputsMemory;
    private final HashMapMemory outputsMemory;
    private final HashMapMemory innerVariablesMemory;

    protected SimulationSnapshot() {
      this.inputsMemory = new HashMapMemory(Simulator.this.inputsMemory);
      this.innerVariablesMemory = new HashMapMemory(Simulator.this.innerVariablesMemory);
      this.outputsMemory = new HashMapMemory(Simulator.this.outputsMemory);
    }
  }

  /**
   * Returns a snapshot representing a current state of the simulation.
   *
   * @return the snapshot
   */
  public SimulationSnapshot makeSimulationSnapshot() {
    return new SimulationSnapshot();
  }

  /**
   * Restores a simulation state from the specified snapshot.
   *
   * @param snapshot the snapshot
   */
  protected void loadSnapshot(final SimulationSnapshot snapshot) {
    this.inputsMemory = new HashMapMemory(snapshot.inputsMemory);
    this.innerVariablesMemory = new HashMapMemory(snapshot.innerVariablesMemory);
    this.outputsMemory = new HashMapMemory(snapshot.outputsMemory);
  }
}
