/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg;

import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.IntegerRange;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.AbstractMap;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The CGAA model visitor that transforms the target model into an SSA form
 * and transforms non-concurrent basic block nodes into concurrent non-ranged ones
 * and compresses basic block sequences.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgSsaVisitor extends CfgEmptyVisitor {

  /* The context that is formed by blocking assignments. */
  private Map<NodeVariable, Node> blockingContext;

  /* The context that is formed by concurrent assignments. */
  private Map<RangedVariable, Node> concurrentContext;

  /**
   * Default constructor.
   */
  CfgSsaVisitor() {
    super();
  }

  @Override
  public void onBlockBegin(final CfgBlockStatement block) {

    this.blockingContext = new LinkedHashMap<>();
    this.concurrentContext = new LinkedHashMap<>();

    addToContexts(block);

    if (!this.concurrentContext.isEmpty()) {
      considerBlockingContext();
    }

    block.removeAssignments();

    block.addAssignments(getConcurrentAssignments());
    block.setConcurrent();

    this.blockingContext.clear();
    this.concurrentContext.clear();
  }

  private void addToContexts(final CfgBlockStatement block) {

    addToContexts(block.getAssignments(), block.isConcurrent());

    /* Mapping of statements to be removed to their parents. */
    final Map<Statement, Statement> parentsOfRemoved = new LinkedHashMap<>();

    final Collection<Statement> blockSuccessors = StatementUtils.getSuccessors(block);

    if (!blockSuccessors.isEmpty()) {

      Statement current = blockSuccessors.iterator().next();

      while (current.getType() == CfgStatementType.BASIC_BLOCK) {

        final CfgBlockStatement currentBlock = (CfgBlockStatement) current;

        addToContexts(currentBlock.getAssignments(), currentBlock.isConcurrent());

        parentsOfRemoved.put(current, current.getParent());

        final Collection<Statement> currentSuccessors = StatementUtils.getSuccessors(current);

        if (currentSuccessors.isEmpty()) {
          break;

        } else {
          current = currentSuccessors.iterator().next();
        }
      }
    }
    removeDescentsFromProcess(parentsOfRemoved);
  }

  private void addToContexts(final List<Assignment> list, final boolean concurrent) {

    if (concurrent) {
      addToConcurrentContext(list);

    } else {
      addToBlockingContext(list);
    }
  }

  private void addToConcurrentContext(final List<Assignment> list) {

    for (final Assignment assignment : list) {

      this.concurrentContext.put(assignment.getTarget(), assignment.getValue());
    }
  }

  private void addToBlockingContext(final List<Assignment> list) {

    for (final Assignment assignment : list) {

      final RangedVariable target = assignment.getTarget();
      final Node value =
          target.isRanged()
              ? RangedVariableUtils.transformNodeWithRanged(target, assignment.getValue())
              : assignment.getValue();

      this.blockingContext.put(target.getVariable(), applyBlockingContext(value));
    }
  }

  private void considerBlockingContext() {

    final Map<RangedVariable, Node> newConcurrentContext = new LinkedHashMap<>();

    for (final Map.Entry<RangedVariable, Node> entry : this.concurrentContext.entrySet()) {

      final RangedVariable target = entry.getKey();
      final Node value = entry.getValue();

      if (target.isRanged()) {

        final Range range = target.getRange();
        final Node young = applyBlockingContext(range.getYoung());
        final Node old = applyBlockingContext(range.getOld());

        target.setRange(new Range(young, old));
      }

      newConcurrentContext.put(target, applyBlockingContext(value));
    }
    this.concurrentContext.clear();
    this.concurrentContext.putAll(newConcurrentContext);
  }

  private List<Assignment> getConcurrentAssignments() {

    final List<Assignment> concurrentAssignments = new LinkedList<>();
    addToConcurrent(concurrentAssignments, makeConcurrentFromBlocking());

    final List<Assignment> noSimilarTargetConcurrent =
        modifySimilarTargetAssignments();

    final List<Assignment> applyBlockingResults =
        applyBlocking(noSimilarTargetConcurrent, concurrentAssignments);

    addToConcurrent(concurrentAssignments, applyBlockingResults);

    return concurrentAssignments;
  }

  private List<Assignment> applyBlocking(
      final List<Assignment> targetList,
      final List<Assignment> blockingList) {

    final List<Assignment> result = new LinkedList<>();

    final Map<NodeVariable, Node> blockingMap = getAssignmentMap(blockingList);

    for (final Assignment target : targetList) {

      Node value = target.getValue();
      final Collection<NodeVariable> used = ExprUtils.getVariables(value);

      for (final NodeVariable variable : used) {

        if (blockingMap.containsKey(variable)) {

          value = Transformer.substitute(value, variable.getName(), blockingMap.get(variable));
        }
      }
      target.setValue(value);
      result.add(target);
    }
    return result;
  }

  private Map<NodeVariable, Node> getAssignmentMap(final List<Assignment> list) {

    final Map<NodeVariable, Node> result = new LinkedHashMap<>();

    for (final Assignment assignment : list) {
      result.put(assignment.getVariable(), assignment.getValue());
    }
    return result;
  }

  private List<Assignment> makeConcurrentFromBlocking() {

    final List<Assignment> concurrent = new LinkedList<>();
    final Map<NodeVariable, Node> currentContext = new LinkedHashMap<>();

    for (final Map.Entry<NodeVariable, Node> entry : this.blockingContext.entrySet()) {

      Node assignValue = entry.getValue();
      final Collection<NodeVariable> used = ExprUtils.getVariables(assignValue);

      for (final NodeVariable usedVariable : used) {

        if (currentContext.containsKey(usedVariable)) {

          assignValue =
              Transformer.substitute(
                  assignValue, usedVariable.getName(), currentContext.get(usedVariable));
        }
      }
      if (currentContext.containsKey(entry.getKey())) {

        currentContext.remove(entry.getKey());
      }
      currentContext.put(entry.getKey(), assignValue);
    }

    for (final Map.Entry<NodeVariable, Node> contextEntry : currentContext.entrySet()) {

      concurrent.add(new Assignment(contextEntry.getKey(), contextEntry.getValue()));
    }

    return concurrent;
  }

  private void addToConcurrent(
      final List<Assignment> result,
      final List<Assignment> more) {

    final Map<NodeVariable, Assignment> map = new LinkedHashMap<>();

    for (final Assignment assignment : result) {

      final NodeVariable variable = assignment.getVariable();

      if (map.containsKey(variable)) {

        map.remove(variable);
      }
      map.put(variable, assignment);
    }

    for (final Assignment assignment : more) {

      final NodeVariable variable = assignment.getVariable();

      if (map.containsKey(variable)) {

        map.remove(variable);
      }
      map.put(variable, assignment);
    }
    result.clear();

    for (final Map.Entry<NodeVariable, Assignment> entry : map.entrySet()) {
      result.add(entry.getValue());
    }
  }

  private List<Assignment> modifySimilarTargetAssignments() {

    final List<List<Assignment>> similarTargetList = new LinkedList<>();

    for (final Map.Entry<RangedVariable, Node> contextEntry : this.concurrentContext.entrySet()) {

      final NodeVariable variable = contextEntry.getKey().getVariable();
      final int subListIndex = getSubListIndex(similarTargetList, variable);
      final Assignment assignment = new Assignment(contextEntry.getKey(), contextEntry.getValue());

      if (subListIndex >= 0) {

        final List<Assignment> related = new LinkedList<>();
        related.addAll(similarTargetList.get(subListIndex));
        related.add(assignment);
        similarTargetList.remove(subListIndex);
        similarTargetList.add(subListIndex, related);

      } else {
        similarTargetList.add(Collections.singletonList(assignment));
      }
    }

    final List<Assignment> result = new LinkedList<>();

    for (final List<Assignment> similarTarget : similarTargetList) {

      result.add(similarTarget.size() == 1
          ? similarTarget.get(0)
          : compressAssignments(similarTarget));
    }
    return result;
  }

  private Assignment compressAssignments(final List<Assignment> list) {

    final List<Assignment> filtered = filterWithNonRanged(list);

    if (filtered.size() == 1) {
      return filtered.get(0);
    }

    final NodeVariable variable = filtered.get(0).getVariable();

    if (variable.isType(DataTypeId.BIT_VECTOR)) {

      return compressBitVectorAssignments(variable, filtered);

    } else if (variable.isType(DataTypeId.MAP)) {

      return compressMapAssignments(variable, filtered);
    }
    throw new RetrascopeRuntimeException("Can't process assignments: " + filtered);
  }

  private Assignment compressBitVectorAssignments(
      final NodeVariable variable,
      final List<Assignment> list) {

    final List<Map.Entry<Assignment, IntegerRange>> entryList = getIntegerRangeList(list);
    final List<IntegerRange> disjointRanges = IntegerRange.divide(getIntegerRanges(entryList));
    final Map<Assignment, Set<IntegerRange>> related = getRelatedRanges(disjointRanges, entryList);
    final List<Assignment> disjointAssignments = getDisjointAssignments(related);

    return new Assignment(variable, concatenate(disjointAssignments));
  }

  private Node concatenate(final List<Assignment> list) {

    class AssignmentComparator implements Comparator<Assignment> {

      @Override
      public int compare(final Assignment a1, final Assignment a2) {

        final IntegerRange a1Range = createIntegerRange(a1.getTarget().getRange());
        final IntegerRange a2Range = createIntegerRange(a2.getTarget().getRange());

        assert a1Range != null && a2Range != null;

        return a2Range.compareTo(a1Range);
      }
    }

    Collections.sort(list, new AssignmentComparator());
    final List<Node> values = getSubValues(list);

    return new NodeOperation(StandardOperation.BVCONCAT, values);
  }

  private List<Node> getSubValues(final List<Assignment> list) {
    final List<Node> result = new LinkedList<>();

    for (final Assignment assignment : list) {
      result.add(assignment.getValue());
    }

    return result;
  }

  private List<Assignment> getDisjointAssignments(final Map<Assignment, Set<IntegerRange>> map) {

    final List<Assignment> result = new LinkedList<>();

    for (final Map.Entry<Assignment, Set<IntegerRange>> entry : map.entrySet()) {

      final Assignment assignment = entry.getKey();
      final RangedVariable target = assignment.getTarget();
      final Range assignRange = target.getRange();

      for (final IntegerRange range : entry.getValue()) {

        final IntegerRange intAssignRange = createIntegerRange(assignRange);

        assert  intAssignRange != null;

        /* If disjoint range coincides with own, do not modify assignment. */
        result.add(
            range.equals(intAssignRange)
                ? assignment
                : getAssignment(assignment, range, intAssignRange));
      }
    }
    return result;
  }

  private Assignment getAssignment(
      final Assignment assignment,
      final IntegerRange range,
      final IntegerRange intAssignRange) {

    final int rangeYoung = range.getMin().intValue();
    final int rangeOld = range.getMax().intValue();
    final int assignYoung = intAssignRange.getMin().intValue();
    final int assignOld = intAssignRange.getMax().intValue();

    Node subRangeValue;

    if (rangeYoung == assignYoung && assignOld > rangeOld) {

      subRangeValue =
          new NodeOperation(
              StandardOperation.BVEXTRACT,
              NodeValue.newInteger(rangeOld - rangeYoung),
              NodeValue.newInteger(0), assignment.getValue());

    } else if (rangeOld == assignOld && assignYoung < rangeYoung) {

      subRangeValue =
          new NodeOperation(
              StandardOperation.BVEXTRACT,
              NodeValue.newInteger(assignOld - assignYoung),
              NodeValue.newInteger(rangeYoung - assignYoung),
              assignment.getValue());

    } else {

      subRangeValue =
          new NodeOperation(
              StandardOperation.BVEXTRACT,
              NodeValue.newInteger(rangeOld - assignYoung),
              NodeValue.newInteger(rangeYoung - assignYoung),
              assignment.getValue());
    }
    final RangedVariable newVariable =
        new RangedVariable(
            assignment.getVariable(),
        new Range(rangeYoung, rangeOld));

    return new Assignment(newVariable, subRangeValue);
  }

  private Map<Assignment, Set<IntegerRange>> getRelatedRanges(
      final List<IntegerRange> disjointRanges,
      final List<Map.Entry<Assignment, IntegerRange>> entryList) {

    final Map<Assignment, Set<IntegerRange>> result = new LinkedHashMap<>();

    for (final IntegerRange disjointRange : disjointRanges) {

      for (final Map.Entry<Assignment, IntegerRange> entry : entryList) {

        if (disjointRange.overlaps(entry.getValue())) {

          final Assignment related = entry.getKey();

          if (result.containsKey(related)) {

            result.get(related).add(disjointRange);

          } else {
            result.put(related, new LinkedHashSet<>(Collections.singletonList(disjointRange)));
          }
        }
      }
    }
    return result;
  }

  private Collection<IntegerRange> getIntegerRanges(
      final List<Map.Entry<Assignment, IntegerRange>> entries) {

    final List<IntegerRange> result = new LinkedList<>();

    for (final Map.Entry<Assignment, IntegerRange> entry : entries) {
      result.add(entry.getValue());
    }
    return result;
  }

  private List<Map.Entry<Assignment, IntegerRange>> getIntegerRangeList(
      final List<Assignment> list) {

    final List<Map.Entry<Assignment, IntegerRange>> result = new LinkedList<>();

    for (final Assignment assignment : list) {

      result.add(
          new AbstractMap.SimpleEntry<>(
              assignment,
              createIntegerRange(assignment.getTarget().getRange())));
    }

    return result;
  }

  private IntegerRange createIntegerRange(final Range range) {

    final Node old = range.getOld();
    final Node young = range.getYoung();

    if (ExprUtils.isKind(Node.Kind.VALUE, old, young)) {

      return new IntegerRange(
          ((NodeValue)TransformerUtils.toInt(young)).getInteger(),
          ((NodeValue)TransformerUtils.toInt(old)).getInteger());

    } else {

      Log.printError(
          String.format(
              "%s either 'old' class=%s or 'young' class=%s is incorrect.",
              LogMessage.UNSUPPORTED_TYPE,
              old.getClass().getName(),
              young.getClass().getName()));
    }
    return null;
  }

  private Assignment compressMapAssignments(
      final NodeVariable variable,
      final List<Assignment> list) {

    Node value = null;

    for (final Assignment assignment : list) {

      if (assignment.ranged()) {

        final Range range = assignment.getTarget().getRange();

        if (!range.getOld().equals(range.getYoung())) {

          throw RetrascopeRuntimeException.makeException(
              LogMessage.CANT_PARSE,
              assignment.getDescription());
        }

        value = (value == null)
            ? new NodeOperation(
                StandardOperation.STORE,
                variable,
                range.getOld(),
                assignment.getValue())
            : new NodeOperation(
                StandardOperation.STORE,
                value,
                range.getOld(),
                assignment.getValue());
      }
    }

    return new Assignment(variable, value);
  }

  private List<Assignment> filterWithNonRanged(final List<Assignment> list) {

    final List<Assignment> filtered = new LinkedList<>();

    for (final Assignment assignment : list) {

      if (assignment.ranged()) {

        final Assignment sameTargetAssignment = getSameTarget(filtered, assignment.getTarget());

        if (sameTargetAssignment != null) {

          filtered.remove(sameTargetAssignment);
        }
      } else {
        filtered.clear();
      }
      filtered.add(
          assignment.ranged() ? assignment : createFullRangeAssignment(assignment));
    }
    return filtered;
  }

  private Assignment createFullRangeAssignment(final Assignment assignment) {

    final NodeVariable variable = assignment.getVariable();
    final Range fullRange = new Range(0, variable.getDataType().getSize() - 1);

    return new Assignment(new RangedVariable(variable, fullRange), assignment.getValue());
  }

  private Assignment getSameTarget(
      final List<Assignment> list,
      final RangedVariable target) {

    for (final Assignment assignment : list) {

      if (assignment.getTarget().equals(target)) {
        return assignment;
      }
    }
    return null;
  }

  private int getSubListIndex(
      final List<List<Assignment>> list,
      final NodeVariable variable) {

    for (final List<Assignment> subList : list) {

      if (subList.get(0).getVariable().equals(variable)) {
        return list.indexOf(subList);
      }
    }
    return -1;
  }

  private Node applyBlockingContext(final Node node) {
    Node result = node;

    for (final Map.Entry<NodeVariable, Node> entry : this.blockingContext.entrySet()) {
      result = Transformer.substitute(result, entry.getKey().getName(), entry.getValue());
    }
    return result;
  }

  private void removeDescentsFromProcess(final Map<Statement, Statement> toBeRemoved) {

    for (final Map.Entry<Statement, Statement> entry : toBeRemoved.entrySet()) {
      entry.getValue().remove(entry.getKey());
    }
  }
}
