/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.extractor.iface;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.DeclarationContainer;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;

import java.util.Collection;
import java.util.Map;

/**
 * Interface extractor for control flow graph model.
 *
 * <p>Engine that extracts {@link ru.ispras.retrascope.result.cfginterface.CfgInterface} interface
 * from {@link ru.ispras.retrascope.model.cfg.CfgModel} control flow graph model.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgInterfaceExtractor extends Engine {

  /**
   *  The engine identifier.
   */
  public static final String ENGINE_ID = "cfg-cfginterface-extractor";

  /**
   * Default constructor.
   */
  public CfgInterfaceExtractor() {
    super(CfgInterfaceExtractor.ENGINE_ID);

    addInputType(CfgModel.TYPE);
    setOutputType(CfgInterface.TYPE);
  }

  @Override public Entity start(final Map<EntityType, Entity> inputs) {

    final CfgModel cfgModel = (CfgModel)inputs.get(CfgModel.TYPE);

    return getInterface(cfgModel);
  }

  private CfgInterface getInterface(final CfgModel cfgModel) {
    InvariantChecks.checkNotNull(cfgModel);

    final CfgInterface cfgInterface = new CfgInterface();
    final Collection<Module> modules = cfgModel.getModules();

    if (!modules.isEmpty()) {

      for (final Module module : modules) {

        addSignalInterface(cfgInterface, (CfgModule) module);

        final Collection<Module> subModules = module.getModules();

        if (!subModules.isEmpty()) {

          getSubModulesInterface(cfgInterface, subModules);
        }
      }
    }
    return cfgInterface;
  }

  /* Fills interface object by modules' data. */
  private void getSubModulesInterface(
      final CfgInterface cfgInterface,
      final Collection<Module> modules) {

    for (final Module module : modules) {

      addSignalInterface(cfgInterface, (CfgModule) module);

      final Collection<Module> subModules = module.getModules();

      if (!subModules.isEmpty()) {

        getSubModulesInterface(cfgInterface, subModules);
      }
    }
  }

  private void addSignalInterface(final CfgInterface cfgInterface, final CfgModule module) {

    final DeclarationContainer moduleInterface = new DeclarationContainer();
    final Map<NodeVariable, Declaration> declarations = module.getDeclarationMap();
    initInterface(moduleInterface, declarations);

    cfgInterface.addInterface(module.getFullName(), moduleInterface);
  }

  private void initInterface(
      final DeclarationContainer moduleInterface,
      final Map<NodeVariable, Declaration> map) {

    for (final Map.Entry<NodeVariable, Declaration> entry : map.entrySet()) {

      final NodeVariable variable = entry.getKey();

      if (isSignal(variable)) {

        final Declaration declaration = entry.getValue();

        if (declaration.getInvariant() == null) {

          moduleInterface.declareVariable(variable);

        } else {

          moduleInterface.declareVariable(variable, declaration.getInvariant());
        }
      } else {

        moduleInterface.declareVariable(variable);
      }
    }
  }

  private static boolean isSignal(final NodeVariable variable) {

    final Object userData = variable.getUserData();

    return userData != null
        && userData instanceof VariableDescriptor
        && VariableType.isSignal(((VariableDescriptor)userData).getVariableType());
  }
}
