/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.EdgeType;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.engine.cfg.CfgSsaTransformer;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseData;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cgaa.CgaaModel;
import ru.ispras.retrascope.model.cgaa.Phase;
import ru.ispras.retrascope.util.NodeUtils;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Clocked guarded atomic actions decision diagram builder.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgCgaaTransformerBuilder {

  /* The mapping of new module to old processes. */
  private final Map<CfgModule, List<CfgProcess>> moduleMap;

  private final Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> processPhaseMap;

  /**
   * Constructs object with the specified mapping.
   *
   * @param map Mapping of modules to their processes.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  CfgCgaaTransformerBuilder(final Map<CfgModule, List<CfgProcess>> map) {
    InvariantChecks.checkNotNull(map);

    this.moduleMap = map;
    this.processPhaseMap = new LinkedHashMap<>();
  }

  private Map<CfgModule, List<CfgProcess>> getModuleMap() {
    return Collections.unmodifiableMap(this.moduleMap);
  }

  /**
   * Creates clocked guarded atomic action model.
   *
   * <p>Builds clocked guarded atomic actions decision diagram
   * from {@link CfgModel} control flow graph model representation.</p>
   *
   * @return Guarded actions decision diagram.
   */
  public CgaaModel build() {

    final CfgModel cfgModel = new CfgModel();

    for (final CfgModule module : getModuleMap().keySet()) {

      final CfgModule gaddModule = buildModule(module);

      if (module.getParent() == null) {

        cfgModel.addModule(gaddModule);
      }
    }

    return new CgaaModel(cfgModel, this.processPhaseMap);
  }

  private CfgModule buildModule(final CfgModule module) {

    /* Apply transformations of the current module's process. */
    final List<CfgProcess> processes = getModuleMap().get(module);

    for (final CfgProcess process : processes) {

      applyToProcess(module, process);
    }

    return module;
  }

  private void applyToProcess(final CfgModule module, final CfgProcess cfgProcess) {

    /* Create new process container. */
    final CfgProcess process = new CfgProcess(cfgProcess);
    module.add(process);

    final CfgProcessPhasePlacer phasePlacer = new CfgProcessPhasePlacer(cfgProcess);

    /* Phase variable. */

    final int maxPhase = phasePlacer.getMaxPhase();
    final int size = (int) (Math.log(maxPhase) / Math.log(2)) + 1;
    final String processId = makeNoNumberHash(process.hashCode());
    final NodeVariable phaseVar = NodeVariable.newBitVector(processId + "_PHASE", size);
    process.addMetaInfo(MetaInfoType.PHASE_VAR_NAME, phaseVar.getName());
    final VariableDescriptor phaseDescriptor = new VariableDescriptor(VariableType.REG);
    phaseVar.setUserData(phaseDescriptor);
    final Integer initPhase = 0;
    final NodeValue initPhaseVal = NodeUtils.newBitVector(initPhase, size);
    final NodeValue maxPhaseVal = NodeUtils.newBitVector(maxPhase, size);

    final Range invariant = new Range(initPhaseVal, maxPhaseVal);
    process.declareVariable(phaseVar, initPhaseVal, invariant);

    /* Associate phases with old process statements. */

    final Graph<Statement, PhaseEdge> phaseGraph = phasePlacer.getPhaseGraph();

    final Graph<Phase, PhaseSeqEdge> phaseSeqGraph = makePhaseSeqGraph(phaseGraph);
    this.processPhaseMap.put(process, phaseSeqGraph);

    final CfgSwitchStatement phaseVarSwitch = new CfgSwitchStatement(phaseVar);
    process.add(phaseVarSwitch);

    final Collection<Statement> switchVertices = phaseGraph.getVertices();

    for (final Statement switchVertex : switchVertices) {

      final Collection<PhaseEdge> outEdges = phaseGraph.getOutEdges(switchVertex);

      /* Calculate possible values of phase variable to reach this switch. */
      final Set<NodeValue> phaseValues = new LinkedHashSet<>();
      final Collection<PhaseEdge> inEdges = phaseGraph.getInEdges(switchVertex);

      for (final PhaseEdge inEdge : inEdges) {

        final Phase inEdgePhase = inEdge.getPhase();
        phaseValues.add(NodeUtils.newBitVector(inEdgePhase.getValue(), size));
      }

      if (switchVertex.equals(cfgProcess)) {

        InvariantChecks.checkTrue(outEdges.size() == 1);
        final PhaseEdge onlyEdge = outEdges.iterator().next();

        final Phase phase = onlyEdge.getPhase();
        final List<Statement> blocks = onlyEdge.getPathBlocks();

        final NodeValue newPhaseVal = NodeUtils.newBitVector(phase.getValue(), size);
        final EventList processEvents = cfgProcess.getEventList();
        final CfgCaseStatement phaseCase = new CfgCaseStatement(processEvents, initPhaseVal);

        for (final Statement block : blocks) {
          phaseCase.add(block.deepCopy());
        }

        final Statement onlyEdgeDst = phaseGraph.getDest(onlyEdge);
        final boolean processIsDst = onlyEdgeDst.equals(cfgProcess);
        phaseCase.add(new CfgBlockStatement(phaseVar, processIsDst ? initPhaseVal : newPhaseVal));
        phaseVarSwitch.add(phaseCase);

      } else {

        final CfgSwitchStatement vertex = (CfgSwitchStatement) switchVertex;
        final CfgCaseStatement phaseCase = new CfgCaseStatement(phaseValues);
        phaseVarSwitch.add(phaseCase);

        final CfgSwitchStatement newSwitch = new CfgSwitchStatement(vertex.getCondition());
        phaseCase.add(newSwitch);

        for (final PhaseEdge outEdge : outEdges) {
          final CfgCaseData data = outEdge.getCaseData();
          final Phase edgePhase = outEdge.getPhase();
          final List<Statement> blocks = outEdge.getPathBlocks();
          final CfgCaseStatement newCase =
              data.isValueFree()
                  ? new CfgCaseStatement(data.isValueFree())
                      : data.hasEvents()
                          ? new CfgCaseStatement(data.getEventList(), data.getValues())
                              : new CfgCaseStatement(data.getValues());

          for (final Statement block : blocks) {
            newCase.add(block.deepCopy());
          }

          final Statement dst = phaseGraph.getDest(outEdge);
          final boolean dstIsProcess = dst.equals(cfgProcess);
          final NodeValue phase =
              dstIsProcess ? initPhaseVal : NodeUtils.newBitVector(edgePhase.getValue(), size);
          newCase.add(new CfgBlockStatement(phaseVar, phase));
          newSwitch.add(newCase);
        }
      }
    }

    CfgSsaTransformer.toSsa(process);
  }

  private Graph<Phase, PhaseSeqEdge> makePhaseSeqGraph(final Graph<Statement, PhaseEdge> graph) {

    final Graph<Phase, PhaseSeqEdge> phaseSeqGraph = new DirectedSparseMultigraph<>();

    final Collection<PhaseEdge> edges = graph.getEdges();

    for (final PhaseEdge edge : edges) {

      final Phase phase = makePhaseVertex(phaseSeqGraph, edge);

      final Statement src = graph.getSource(edge);
      final Collection<PhaseEdge> inEdges = graph.getInEdges(src);

      for (final PhaseEdge inEdge : inEdges) {

        final Phase inEdgePhase = makePhaseVertex(phaseSeqGraph, inEdge);
        phaseSeqGraph.addEdge(new PhaseSeqEdge(), inEdgePhase, phase, EdgeType.DIRECTED);
      }

      final Statement dst = graph.getDest(edge);
      final Collection<PhaseEdge> outEdges = graph.getOutEdges(dst);

      for (final PhaseEdge outEdge : outEdges) {

        final Phase outEdgePhase = makePhaseVertex(phaseSeqGraph, outEdge);
        phaseSeqGraph.addEdge(new PhaseSeqEdge(), phase, outEdgePhase, EdgeType.DIRECTED);
      }
    }

    return phaseSeqGraph;
  }

  private Phase makePhaseVertex(
      final Graph<Phase, PhaseSeqEdge> phaseSeqGraph,
      final PhaseEdge edge) {

    final Phase phase = edge.getPhase();

    if (!phaseSeqGraph.containsVertex(phase)) {
      phaseSeqGraph.addVertex(phase);
    }
    return phase;
  }

  private String makeNoNumberHash(final int hashCode) {
    final Map<Integer, String> map = new LinkedHashMap<>();
    map.put(0, "a");
    map.put(1, "b");
    map.put(2, "c");
    map.put(3, "d");
    map.put(4, "e");
    map.put(5, "f");
    map.put(6, "g");
    map.put(7, "h");
    map.put(8, "i");
    map.put(9, "j");

    String noNumberHash = Integer.toString(hashCode);

    for (final Map.Entry<Integer, String> entry : map.entrySet()) {
      noNumberHash = noNumberHash.replaceAll(String.valueOf(entry.getKey()), entry.getValue());
    }

    return noNumberHash;
  }
}
