/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.transformer.cgaa;

import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.EdgeType;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseData;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgDfsWalker;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgVisitor;

import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * The component that performs Control Flow Graph (CFG) model statements placement
 * upon different phases for the specified process.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CfgProcessPhasePlacer {

  /**
   * The phase dependencies graph. The graph vertices are process\switch statements.
   * The graph edges are annotated with phase values and some other related data (blocks between
   * statements, case statement data, if exists, etc.).
   */
  private Graph<Statement, PhaseEdge> phaseDepGraph;

  /* Phase value. */
  private Integer phase;

  /**
   * The phase placement heuristic is implemented here.
   */
  class PhasePlacementVisitor extends CfgEmptyVisitor {

    private List<CfgBlockStatement> blocks;
    private ModuleProcess process;
    private Integer currentPhase;
    private Map<CfgCaseStatement, Integer> phaseMap;
    private Deque<Statement> srcDeque;
    private CfgCaseData incompleteEdgeData;
    private Status status;

    PhasePlacementVisitor(final Integer phase) {
      setStatus(Status.OK);
      this.currentPhase = phase;
    }

    Integer getPhase() {
      return this.currentPhase;
    }

    private List<CfgBlockStatement> getBlocks() {
      return this.blocks;
    }

    private Deque<Statement> getSrcDeque() {
      return this.srcDeque;
    }

    private Statement getSrcPeek() {
      return getSrcDeque().peek();
    }

    private CfgCaseData getIncompleteData() {
      return this.incompleteEdgeData;
    }

    @Override
    public Status getStatus() {
      return this.status;
    }

    void setStatus(final Status status) {
      this.status = status;
    }

    @Override
    public void onProcessBegin(final ModuleProcess process) {

      this.process = process;
      this.blocks = new LinkedList<>();
      this.currentPhase = 1;
      this.phaseMap = new LinkedHashMap<>();
      phaseDepGraph = new DirectedSparseMultigraph<>();
      phaseDepGraph.addVertex(process);
      this.srcDeque = new LinkedList<>();
      getSrcDeque().push(process);
    }

    @Override
    public void onProcessEnd(final ModuleProcess process) {
      getSrcDeque().pop();
    }

    @Override
    public void onBlockBegin(final CfgBlockStatement block) {
      getBlocks().add(block);
    }

    @Override
    public void onCaseBegin(final CfgCaseStatement caseNode) {

      if (this.phaseMap.containsKey(caseNode)) {

        setStatus(Status.SKIP);

      } else {

        this.currentPhase++;
        this.phaseMap.put(caseNode, getPhase());
        this.incompleteEdgeData = caseNode.getCfgCaseData();
        final Statement parent = caseNode.getParent();

        if (!getSrcPeek().equals(parent)) {
          getSrcDeque().push(parent);
        }
      }
    }

    @Override
    public void onSwitchBegin(final CfgSwitchStatement switchNode) {

      final boolean alreadyVisitedNode = phaseDepGraph.containsVertex(switchNode);

      if (!alreadyVisitedNode) {
        phaseDepGraph.addVertex(switchNode);
      }

      final Statement src = getSrcPeek();
      final PhaseEdge edge = new PhaseEdge(getPhase(), getBlocks(), getIncompleteData());

      if (!phaseDepGraph.containsEdge(edge)) {

        phaseDepGraph.addEdge(edge, src, switchNode, EdgeType.DIRECTED);
        getBlocks().clear();
        this.incompleteEdgeData = null;
      }

      if (alreadyVisitedNode) {
        setStatus(Status.SKIP);
      }
    }

    @Override
    public void onSwitchEnd(final CfgSwitchStatement switchNode) {

      if (getStatus().equals(Status.SKIP)) {
        setStatus(Status.OK);
      }

      if (getSrcPeek().equals(switchNode)) {
        getSrcDeque().pop();
      }
    }

    @Override
    public void onStatementEnd(final Statement statement) {

      if (StatementUtils.isLast(statement)) {

        final Statement src = getSrcPeek();
        final PhaseEdge edge = new PhaseEdge(0, new LinkedList<>(getBlocks()), getIncompleteData());

        phaseDepGraph.addEdge(edge, src, this.process, EdgeType.DIRECTED);
        getBlocks().clear();
      }
    }
  }

  /**
   * Constructs the phase placement component for the specified statement.
   * <p>The constructor call includes the phase placement procedure itself.</p>
   * @param statement The statement to that the phase placement is done.
   */
  CfgProcessPhasePlacer(final Statement statement) {
    InvariantChecks.checkNotNull(statement);

    this.phase = 0;

    final CfgVisitor visitor = new PhasePlacementVisitor(this.phase);
    final ModelWalker walker = new CfgDfsWalker(visitor, WalkerMode.DFS);
    walker.start(statement);
    this.phase = ((PhasePlacementVisitor)visitor).getPhase();
  }

  /**
   * Returns the phase dependency tree.
   * @return The phase dependency tree.
   */
  Graph<Statement, PhaseEdge> getPhaseGraph() {
    return this.phaseDepGraph;
  }

  Integer getMaxPhase() {
    return this.phase;
  }
}
