/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import edu.uci.ics.jung.graph.Graph;
import org.apache.commons.cli.ParseException;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.ModelWalkerEngine;
import ru.ispras.retrascope.engine.cfg.transformer.cgaa.PhaseSeqEdge;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgDfsWalker;
import ru.ispras.retrascope.model.cfg.CfgModel;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.cfg.CfgVisitor;
import ru.ispras.retrascope.model.cgaa.CgaaModel;
import ru.ispras.retrascope.model.cgaa.Phase;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeOrthogonalizer;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.ReadWriteUtils;
import ru.ispras.retrascope.util.SolverUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

/**
 * The transformer of clocked guarded atomic actions container
 * {@link ru.ispras.retrascope.model.cgaa.CgaaModel} into
 * extended finite state machine {@link ru.ispras.retrascope.model.efsm.EfsmModel}.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CgaaEfsmTransformer extends ModelWalkerEngine {

  static class StateVariableContainer {

    final Map<NodeVariable, Integer> stateNumMap;
    final int stateMaxNum;
    final boolean complete;

    StateVariableContainer(
        final Map<NodeVariable, Integer> stateNumMap,
        final int stateMaxNum,
        final boolean complete) {
      this.stateNumMap = stateNumMap;
      this.stateMaxNum = stateMaxNum;
      this.complete = complete;
    }

    Integer getNum(final NodeVariable variable) {
      return this.stateNumMap.get(variable);
    }

    int getStateMaxNum() {
      return this.stateMaxNum;
    }

    boolean hasVariable(final NodeVariable variable) {
      return this.stateNumMap.containsKey(variable);
    }

    boolean isComplete() {
      return this.complete;
    }
  }

  /**
   * The engine identifier.
   */
  public static final String ENGINE_ID = "cgaa-efsm-transformer";

  private static final String STATE_VAR_INFO = "Names of state-like variables";

  /**
   * State-like variables names command line parameter.
   */
  public static final Parameter STATE_VARIABLES = new Parameter("state-vars", true, STATE_VAR_INFO);

  private static final String STATE_MAX_INFO = "Maximum number of EFSM states.";

  /**
   * State maximum number parameter.
   */
  public static final Parameter STATE_MAX_NUM =
      new Parameter("state-max-num", true, STATE_MAX_INFO);

  private static final int DEFAULT_STATE_MAX_NUM = 100;

  /* The model that is to be transformed into EFSM. */
  private CfgModel cfgModel;

  /* Map from CGAA model processes to their phase sequence graphs. */
  private Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> processSeqGraphMap;

  /* The clock-like variables detector. */
  private CgaaClockDetector clockDetector;

  /* The state-like variables extractor. */
  private CgaaStateVarDetector stateDetector;

  /* The state-like variables statistics keeping visitor. */
  private CgaaStatisticsKeeper statKeeper;

  /* The states expressions' extractor. */
  private CgaaPathExtractor pathExtractor;

  /* Auxiliary mapping between EFSM names and processes that are extracted from. */
  private final Map<String, ModuleProcess> efsmProcessMap = new LinkedHashMap<>();

  /* Back-ends. */
  private final EfsmInitStateBackend initStateBackend;

  /* Debug mode flag. */
  private boolean debugMode;

  /* File for debug output. */
  private File debugFile;

  /* Maximum number of EFSM states. */
  private int stateMaxNum;

  /* Model component that is under construction. */
  private Efsm efsm;

  /* States to be checked as possible sources for transition under construction. */
  private Queue<EfsmState> canBeSrcStates;

  private Set<EfsmState> visited;

  private final Map<Integer, LinkedHashSet<CgaaPath>> srcPathMap;

  /* Debug file default name. */
  private static final String DEFAULT_DBG_FILENAME = ENGINE_ID + ".log";

  /**
   * Default constructor.
   */
  public CgaaEfsmTransformer() {
    super(
        CgaaEfsmTransformer.ENGINE_ID,
        CgaaModel.TYPE,
        EfsmModel.TYPE,
        new CfgDfsWalker(new CgaaClockDetector(), WalkerMode.DFS_NO_RPT),
        new CfgDfsWalker(new CgaaStateVarDetector(), WalkerMode.DFS),
        new CfgDfsWalker(new CgaaStatisticsKeeper(), WalkerMode.DFS),
        new CfgDfsWalker(new CgaaPathExtractor(), WalkerMode.DFS));

    /* Engine command line parameters. */
    addParameter(Retrascope.TOOL_DEBUG_PARAM);
    addParameter(STATE_VARIABLES);
    addParameter(STATE_MAX_NUM);

    this.srcPathMap = new LinkedHashMap<>();

    /* Back-end registration. */
    this.initStateBackend = new EfsmInitStateBackend();
    getInitStateBackend().setEfsmCfgProcessMap(this.efsmProcessMap);
    addBackend(getInitStateBackend());
  }

  private EfsmInitStateBackend getInitStateBackend() {
    return this.initStateBackend;
  }

  private CfgModel getCfgModel() {
    return this.cfgModel;
  }

  private void setCfgModel(final CfgModel cfgModel) {
    InvariantChecks.checkNotNull(cfgModel);

    this.cfgModel = cfgModel;
  }

  private CgaaClockDetector getClockDetector() {
    return this.clockDetector;
  }

  private void setClockDetector(final CgaaClockDetector clockDetector) {
    InvariantChecks.checkNotNull(clockDetector);

    this.clockDetector = clockDetector;
  }

  private CgaaStateVarDetector getStateDetector() {
    return this.stateDetector;
  }

  private void setStateDetector(final CgaaStateVarDetector stateDetector) {
    InvariantChecks.checkNotNull(stateDetector);

    this.stateDetector = stateDetector;
  }

  private CgaaStatisticsKeeper getStatKeeper() {
    return this.statKeeper;
  }

  private void setStatKeeper(final CgaaStatisticsKeeper statKeeper) {
    InvariantChecks.checkNotNull(statKeeper);

    this.statKeeper = statKeeper;
  }

  private CgaaPathExtractor getPathExtractor() {
    return this.pathExtractor;
  }

  private void setPathExtractor(final CgaaPathExtractor pathExtractor) {
    InvariantChecks.checkNotNull(pathExtractor);

    this.pathExtractor = pathExtractor;
  }

  private boolean isDebugMode() {
    return this.debugMode;
  }

  private void setDebugMode(final boolean debugMode) {
    this.debugMode = debugMode;
  }

  private File getDebugFile() {
    return this.debugFile;
  }

  private void setDebugFile(final File file) {
    this.debugFile = file;
  }

  private int getStateMaxNum() {
    return this.stateMaxNum;
  }

  private void setStateMaxNum(final int num) {
    this.stateMaxNum = num;
  }

  @Override
  public void initialize(final Map<EntityType, Entity> inputs) {

    /* Extract control flow graph from wrapper. */

    final CgaaModel cgaaModel = (CgaaModel) inputs.get(CgaaModel.TYPE);
    inputs.remove(CgaaModel.TYPE);
    final CfgModel internalCfgModel = cgaaModel.getModel();
    setCfgModel(internalCfgModel);

    this.processSeqGraphMap = cgaaModel.getProcessPhaseMap();

    /* treat internal model as a CGAA. */
    inputs.put(CgaaModel.TYPE, internalCfgModel);

    initWalkers();

    /* Elaborate command line parameters. */
    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);

    try {
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());

      if (engineArgs.contains(STATE_VARIABLES)) {

        final String[] stateNames = engineArgs.getValue(STATE_VARIABLES).split(File.pathSeparator);
        getStateDetector().setStateNames(stateNames);
      }

      if (engineArgs.contains(Retrascope.TOOL_DEBUG_PARAM)) {

        setDebugMode(true);
        final String dbgFileName =
            engineArgs.getValue(Retrascope.TOOL_DEBUG_PARAM, DEFAULT_DBG_FILENAME);
        setDebugFile(new File(dbgFileName));
      }

      setStateMaxNum(
          Integer.parseInt(
              engineArgs.getValue(STATE_MAX_NUM, Integer.toString(DEFAULT_STATE_MAX_NUM))));

    } catch (final ParseException e) {
      throw new RetrascopeRuntimeException(e);
    }
  }

  private void initWalkers() {

    setClockDetector(new CgaaClockDetector());
    setStateDetector(new CgaaStateVarDetector());
    setStatKeeper(new CgaaStatisticsKeeper());
    setPathExtractor(new CgaaPathExtractor());

    setWalkers(new CfgDfsWalker(getClockDetector(), WalkerMode.DFS_NO_RPT),
        new CfgDfsWalker(getStateDetector(), WalkerMode.DFS),
        new CfgDfsWalker(getStatKeeper(), WalkerMode.DFS),
        new CfgDfsWalker(getPathExtractor(), WalkerMode.DFS_NO_RPT));

    setDebugMode(false);

    /* Create mappings for internal components. */
    createClockMap();
    createStateMap();

    getStateDetector().setPhaseSeqMap(this.processSeqGraphMap);
  }

  private void createStateMap() {

    final Map<ModuleProcess, Set<NodeVariable>> map = new LinkedHashMap<>();
    getStateDetector().setStateMap(map);
    getStatKeeper().setStateMap(map);
    getPathExtractor().setStateMap(map);
  }

  private void createClockMap() {

    final Map<ModuleProcess, Set<NodeVariable>> map = new LinkedHashMap<>();
    getClockDetector().setClockMap(map);
    getPathExtractor().setClockMap(map);
  }

  @Override
  public Entity getOutput() {

    final EfsmModel model = new EfsmModel();

    /* Create CGAA model visitor that transforms modules. */
    final CfgVisitor moduleVisitor = new CgaaModuleVisitor(model);

    final ModelWalker walker = new CfgDfsWalker(moduleVisitor, WalkerMode.DFS_NO_RPT);
    walker.start(getCfgModel());

    final Map<CfgModule, EfsmModule> moduleEfsmMap =
        ((CgaaModuleVisitor)moduleVisitor).getModuleMap();

    final Map<ModuleProcess, Map<Integer, LinkedHashSet<CgaaPath>>> pathMap =
        getPathExtractor().getProcessDataMap();

    final Map<ModuleProcess, Set<NodeVariable>> stateVariableMap =
        getStateDetector().getStateLikeVarMap();

    Log.printInfo("======================================");
    int efsmCount = 0;

    for (final Map.Entry<ModuleProcess, Map<Integer, LinkedHashSet<CgaaPath>>> entry
        : pathMap.entrySet()) {

      final ModuleProcess keyProcess = entry.getKey();
      final Statement keyParent = keyProcess.getParent();

      /* Check that Process parent is Module object. */
      assert keyParent instanceof CfgModule
          : LogMessage.makeMessage(LogMessage.WRONG_TYPE, keyParent.getClass().getName());

      final EfsmModule parentContainer = moduleEfsmMap.get(keyParent);

      /* Print clock-like variables. */

      final List<String> clockNames = new LinkedList<>();

      final Map<ModuleProcess, Set<NodeVariable>> clockMap = getClockDetector().getClockMap();
      final Set<NodeVariable> clockVars = clockMap.get(keyProcess);

      for (final NodeVariable var : clockVars) {
        clockNames.add(var.getName());
      }

      Log.printInfo(String.format(
          "Clock-like variables: %s.", clockNames.isEmpty() ? "<none>" : clockNames));
      Log.printInfo(String.format("Transforming the process '%s'.", keyProcess.getName()));

      this.srcPathMap.putAll(entry.getValue());

      this.efsm = new Efsm(parentContainer, keyProcess.getId());
      this.efsm.addMetaInfo(
          MetaInfoType.PHASE_VAR_NAME, keyProcess.getMetaInfo(MetaInfoType.PHASE_VAR_NAME.name()));

      this.efsmProcessMap.put(this.efsm.getName(), keyProcess);
      this.efsm.declareVariables(keyProcess.getDeclarationMap());

      for (final NodeVariable clockVariable : clockVars) {
        this.efsm.addClock(clockVariable);
      }

      final Set<NodeVariable> stateVariables = stateVariableMap.get(keyProcess);

      createEfsm(stateVariables);

      if (keyProcess.isInitial()) {
        parentContainer.addInitialEfsm(this.efsm);
      } else {
        parentContainer.addNonInitialEfsm(this.efsm);
      }

      removeIsolatedStates();

      Log.printInfo(
          String.format("Model states number: %d.", this.efsm.getStateNum()));

      Log.printInfo(
          String.format("Model transitions number: %d.", this.efsm.getTransitionNum()));
      Log.printInfo("======================================");

      efsmCount++;
      this.srcPathMap.clear();
    }
    Log.printInfo(String.format("The number of extracted models: %d.", efsmCount));

    /* If debug log mode is enabled, collect & write statistical info to log file. */
    if (isDebugMode()) {

      final EfsmDebugInfo debugInfo = new EfsmDebugInfo(0, 0, 0);

      final List<String> topNames = new LinkedList<>();

      for (final Module efsmModule : model.getModules()) {

        getModuleInfo((EfsmModule) efsmModule, debugInfo);
        topNames.add(efsmModule.getName());
      }
      writeToDebugFile(
          String.format(
              "modules=%s states=%d transitions=%d efsms=%d",
              StringUtils.join(topNames, File.pathSeparator),
              debugInfo.getStateNum(),
              debugInfo.getTransitionNum(),
              debugInfo.getModelNum()));
    }

    return model;
  }

  private void removeIsolatedStates() {

    final Set<EfsmState> isolated = new LinkedHashSet<>();

    for (final EfsmState state : this.efsm.getStates()) {

      final Collection<EfsmTransition> incoming = EfsmUtils.getIncoming(this.efsm, state);
      final Collection<EfsmTransition> outgoing = EfsmUtils.getOutgoing(this.efsm, state);

      if (incoming.isEmpty() && outgoing.isEmpty()) {
        isolated.add(state);
      }
    }
    for (final EfsmState isolatedState : isolated) {
      this.efsm.remove(isolatedState);
    }
  }

  private void getModuleInfo(final EfsmModule efsmModule, final EfsmDebugInfo info) {

    final Collection<Module> nestedModules = efsmModule.getModules();
    for (final Module subModule : nestedModules) {
      getModuleInfo((EfsmModule) subModule, info);
    }

    final Set<ModuleProcess> efsms = new LinkedHashSet<>();
    efsms.addAll(efsmModule.getProcesses());

    for (final ModuleProcess efsm : efsms) {

      info.setModelNum(info.getModelNum() + 1);
      info.setStateNum(info.getStateNum() + ((Efsm)efsm).getStateNum());
      info.setTransitionNum(info.getTransitionNum() + ((Efsm)efsm).getTransitionNum());
    }
  }

  private void createEfsm(final Set<NodeVariable> stateVariables) {

    final List<Node> states = createStates(stateVariables);

    if (states.isEmpty()) {
      states.add(SolverUtils.TRUE);
    }

    printStateVarInfo(states);
    createTransitions();
  }

  private void printStateVarInfo(final List<Node> states) {

    final Collection<String> names = new LinkedHashSet<>();

    for (final Node state : states) {

      this.efsm.addState(state);
      names.addAll(NodeUtils.getVariableNames(ExprUtils.getVariables(state)));
    }

    Log.printInfo(String.format("State-like variables: %s.", names.isEmpty() ? "<none>" : names));
  }

  private void createTransitions() {

    this.canBeSrcStates = new LinkedList<>(this.efsm.getStates());
    this.visited = new LinkedHashSet<>();

    while (!this.canBeSrcStates.isEmpty()) {

      final EfsmState src = this.canBeSrcStates.poll();
      this.visited.add(src);
      final Integer srcPhase = src.getPhase();
      final LinkedHashSet<CgaaPath> phasePaths = this.srcPathMap.get(srcPhase);

      createTransitionsFrom(src, phasePaths);
    }
    this.visited.clear();
  }

  private void createTransitionsFrom(final EfsmState srcState, final Set<CgaaPath> paths) {

    for (final CgaaPath path : paths) {

      /* When the stabilization splits the state, skip the rest. */
      if (!this.efsm.containsState(srcState)) {
        break;
      }

      final Node preCond = NodeUtils.getConjunction(path.getState());
      final Node pathGuardNode = Node.AND(path.getGuardNode(), preCond);
      final Node clockNode = path.getClockNode();

      final Node srcNode = srcState.getNode();
      final Node guardSatCond = getNodesWithInvariants(pathGuardNode, clockNode, srcNode);
      final boolean guardSat = SolverUtils.isSat(guardSatCond);

      if (!guardSat) {
        continue;
      }

      addTransitionsFromSrc(srcState, path.getDstPhase(), path.getGuardedAction());
    }
  }

  private Node getNodesWithInvariants(final Node pathGuard, final Node clock, final Node srcState) {

    return clock == null
        ? ExprUtils.getConjunction(pathGuard, srcState, getNodeVarInvariants(pathGuard, srcState))
        : ExprUtils.getConjunction(
            clock,
            pathGuard,
            srcState,
            getNodeVarInvariants(clock, pathGuard, srcState));
  }

  private void addTransitionsFromSrc(
      final EfsmState src,
      final Integer dstPhase,
      final GuardedAction guardedAction) {

    final Action action = guardedAction.getAction();
    final Set<EfsmState> dstStates = getDstStates(src, dstPhase, action);

    if (dstStates.size() > 1) {

      for (final EfsmState dst : dstStates) {

        final CfgBlockStatement block = action.toBlock();
        final Node wpCond = CfgUtils.getWeakestPreCondition(dst.getNode(), block);
        final Node guard =
            Node.AND(guardedAction.getGuardNode(), TransformerUtils.standardize(wpCond));

        final GuardedAction transitionData =
            new GuardedAction(guard, action, guardedAction.getClocks());
        this.efsm.addTransition(src, dst, transitionData);
      }

    } else if (dstStates.size() == 1) {
      this.efsm.addTransition(src, dstStates.iterator().next(), guardedAction);
    }
  }

  private List<Node> createStates(final Set<NodeVariable> stateVariables) {

    final StateVariableContainer container = createStateContainer(stateVariables);

    if (container.isComplete() && container.getStateMaxNum() <= getStateMaxNum()) {

      final List<List<Node>> states = new LinkedList<>();

      for (final NodeVariable stateVariable : stateVariables) {

        if (container.hasVariable(stateVariable)) {
          final Integer varValueNum = container.getNum(stateVariable);

          states.add(createVarEquations(stateVariable, varValueNum));
        }
      }

      return conjunctions(cartesianProduct(states));

    } else {

      final Collection<CgaaPath> cgaaPaths = new LinkedHashSet<>();
      final Collection<LinkedHashSet<CgaaPath>> paths = this.srcPathMap.values();

      for (final Set<CgaaPath> pathSet : paths) {
        cgaaPaths.addAll(pathSet);
      }

      final Map<NodeVariable, Collection<Collection<Node>>> stateVarNodes = splitByVars(cgaaPaths);
      final List<List<Node>> ortNodes = new LinkedList<>();

      for (final Collection<Collection<Node>> nodes : stateVarNodes.values()) {
        ortNodes.add(NodeOrthogonalizer.orthogonalizeNodes(nodes));
      }

      return conjunctions(cartesianProduct(ortNodes));
    }
  }

  private List<Node> createVarEquations(final NodeVariable variable, final Integer valueNum) {

    final List<Node> equations = new LinkedList<>();

    for (int i = 0; i < valueNum; i++) {

      final Node value = TypeConversion.coerce(NodeValue.newInteger(i), variable.getDataType());
      equations.add(NodeUtils.getEquation(variable, value));
    }

    return equations;
  }

  private StateVariableContainer createStateContainer(final Set<NodeVariable> stateVariables) {

    final Map<NodeVariable, Integer> valueNumMap = new LinkedHashMap<>();
    int stateMaxNum = 1;
    boolean complete = true;

    for (final NodeVariable variable : stateVariables) {

      final Statement topParent = StatementUtils.getTopParent(this.efsm);
      final DecStatement decStatement = StatementUtils.getVariableContainer(topParent, variable);
      final Declaration declaration = decStatement.getDeclaration(variable);

      if (declaration.containsInvariant()) {

        final Range invariant = declaration.getInvariant();
        final Node maxNode = invariant.getOld();
        final Node minNode = invariant.getYoung();

        if (ExprUtils.isKind(Node.Kind.VALUE, maxNode, minNode)) {

          final NodeValue maxValue = (NodeValue) maxNode;
          final NodeValue minValue = (NodeValue) minNode;

          final BigInteger maxInt = maxValue.getInteger();
          final BigInteger minInt = minValue.getInteger();

          final int valueNum = getValueNum(maxInt, minInt);
          valueNumMap.put(variable, valueNum);

          stateMaxNum = stateMaxNum * valueNum;
        }
      } else if (constrainedDataType(variable)) {

        final DataType dataType = variable.getDataType();
        final BigInteger maxInt = getMaxInt(dataType);
        final BigInteger minInt = getMinInt(dataType);

        final int valueNum = getValueNum(maxInt, minInt);
        valueNumMap.put(variable, valueNum);

        stateMaxNum = stateMaxNum * valueNum;
      } else {
        complete = false;
        break;
      }
    }

    return new StateVariableContainer(valueNumMap, stateMaxNum, complete);
  }

  private int getValueNum(final BigInteger max, final BigInteger min) {

    return (max.subtract(min)).intValue() + 1;
  }

  private BigInteger getMinInt(final DataType dataType) {

    final DataTypeId typeId = dataType.getTypeId();

    switch (typeId) {
      case LOGIC_BOOLEAN:
        return BigInteger.ZERO;

      case BIT_VECTOR:
        BitVector minVector = BitVector.newEmpty(dataType.getSize());
        minVector.reset();
        return new BigInteger(minVector.toByteArray());

      default:
        throw RetrascopeRuntimeException.makeException(LogMessage.UNSUPPORTED_TYPE, typeId);
    }
  }

  private BigInteger getMaxInt(final DataType dataType) {

    final DataTypeId typeId = dataType.getTypeId();

    switch (typeId) {
      case LOGIC_BOOLEAN:
        return BigInteger.ONE;

      case BIT_VECTOR:
        BitVector minVector = BitVector.newEmpty(dataType.getSize());
        minVector.setAll();
        return new BigInteger(minVector.toByteArray());

      default:
        throw RetrascopeRuntimeException.makeException(LogMessage.UNSUPPORTED_TYPE, typeId);
    }
  }

  private boolean constrainedDataType(final NodeVariable variable) {

    return ExprUtils.isType(DataType.BOOLEAN, variable)
        || ExprUtils.isType(DataTypeId.BIT_VECTOR, variable);
  }

  private List<Node> conjunctions(final List<List<Node>> lists) {

    final List<Node> conjunctions = new LinkedList<>();

    for (final List<Node> list : lists) {

      if (!list.isEmpty()) {

        final Node conjunction = NodeUtils.getConjunction(list);

        if (SolverUtils.areCompatible(conjunction, getNodeVarInvariants(conjunction))) {
          conjunctions.add(conjunction);
        }
      }
    }

    return conjunctions;
  }

  private List<List<Node>> cartesianProduct(final List<List<Node>> lists) {
    final List<List<Node>> resultLists = new LinkedList<>();

    if (lists.isEmpty()) {

      resultLists.add(new LinkedList<Node>());
      return resultLists;

    } else {

      final List<Node> firstList = lists.get(0);
      final List<List<Node>> remainingLists = cartesianProduct(lists.subList(1, lists.size()));

      for (final Node condition : firstList) {

        for (final List<Node> remainingList : remainingLists) {

          final List<Node> resultList = new LinkedList<>();
          resultList.add(condition);
          resultList.addAll(remainingList);
          resultLists.add(resultList);
        }
      }
    }
    return resultLists;
  }

  private Map<NodeVariable, Collection<Collection<Node>>> splitByVars(
      final Collection<CgaaPath> paths) {

    final Map<NodeVariable, Collection<Collection<Node>>> varNodeMap = new LinkedHashMap<>();

    for (final CgaaPath path : paths) {

      final Set<Node> state = path.getState();
      final Map<NodeVariable, Collection<Node>> pathNodeMap = new LinkedHashMap<>();

      for (final Node node : state) {

        final Collection<NodeVariable> variables = ExprUtils.getVariables(node);

        if (!variables.isEmpty()) {

          InvariantChecks.checkTrue(variables.size() == 1);
          final NodeVariable variable = variables.iterator().next();

          if (pathNodeMap.containsKey(variable)) {

            pathNodeMap.get(variable).add(node);

          } else {

            final Collection<Node> set = new LinkedHashSet<>();
            set.add(node);
            pathNodeMap.put(variable, set);
          }
        }
      }

      for (final Map.Entry<NodeVariable, Collection<Node>> varEntry : pathNodeMap.entrySet()) {

        if (varNodeMap.containsKey(varEntry.getKey())) {

          varNodeMap.get(varEntry.getKey()).add(varEntry.getValue());

        } else {

          final Collection<Collection<Node>> pathCollections = new LinkedHashSet<>();
          pathCollections.add(varEntry.getValue());
          varNodeMap.put(varEntry.getKey(), pathCollections);
        }
      }
    }

    return varNodeMap;
  }

  private Node getNodeVarInvariants(final Node ... nodes) {

    final ModuleProcess process = this.efsmProcessMap.get(this.efsm.getName());
    final Map<NodeVariable, Node> invariants = new LinkedHashMap<>();

    for (final Node node : nodes) {

      final Collection<NodeVariable> variables = ExprUtils.getVariables(node);

      for (final NodeVariable variable : variables) {

        if (!invariants.containsKey(variable)) {

          final Node invariant =
              process == null ? null : CgaaEfsmTransformerUtils.getInvariant(process, variable);

          if (invariant != null) {
            invariants.put(variable, invariant);
          }
        }
      }
    }
    return invariants.isEmpty() ? SolverUtils.TRUE : NodeUtils.getConjunction(invariants.values());
  }

  private Set<EfsmState> getDstStates(final EfsmState src, final Integer dst, final Action action) {

    final Node srcNode = src.getNode();
    final Set<EfsmState> destinations = new LinkedHashSet<>();

    final Set<NodeVariable> actionDefines = action.getDefines();
    final Set<NodeVariable> stateVars = new LinkedHashSet<>(ExprUtils.getVariables(srcNode));

    if (CollectionUtils.areIntersectedSets(actionDefines, stateVars)) {

      final Node srcInvariants = getNodeVarInvariants(srcNode);
      final Node preCond = Node.AND(srcNode, srcInvariants);

      final CfgBlockStatement block = action.toBlock();

      for (final EfsmState dstState : this.efsm.getStates()) {

        if (dstState.getPhase().equals(dst)) {

          final Node dstNode = dstState.getNode();
          final Node postCond = Node.AND(dstNode, getNodeVarInvariants(dstNode));
          final CfgBlockStatement actionBlock = action.toBlock();
          final Node spCond = CfgUtils.getStrongestPostCondition(preCond, postCond, actionBlock);

          if (SolverUtils.isSat(spCond)) {
            destinations.add(dstState);
          }
        }
      }

      if (destinations.isEmpty()) {

        final Node postCond = CfgUtils.getPostCondition(block);
        final Set<NodeVariable> currentEfsmVars = this.efsm.getStateVariables();
        final Pair<Node, Node> stateOtherPair = NodeUtils.splitByVar(postCond, currentEfsmVars);
        final EfsmState nextPhaseState = createNewPhaseState(dst, stateOtherPair.getLeft());
        destinations.add(nextPhaseState);
      }
    } else {

      final EfsmState nextPhaseState = createNewPhaseState(dst, srcNode);
      destinations.add(nextPhaseState);
    }

    return destinations;
  }

  private EfsmState createNewPhaseState(Integer dst, Node srcNode) {
    final EfsmState nextPhaseState = new EfsmState(srcNode, dst);

    if (!this.efsm.containsState(nextPhaseState)) {

      this.efsm.addState(nextPhaseState);

      if (!this.canBeSrcStates.contains(nextPhaseState)
          && !this.visited.contains(nextPhaseState)) {

        this.canBeSrcStates.add(nextPhaseState);
      }
    }
    return nextPhaseState;
  }

  private void writeToDebugFile(final String string) {

    try {

      final FileOutputStream stream = new FileOutputStream(getDebugFile(), true);
      final BufferedWriter writer = ReadWriteUtils.newBufferedWriter(stream);
      writer.write(string);
      writer.newLine();
      writer.close();

    } catch (final IOException e) {
      throw new RetrascopeRuntimeException(e);
    }
  }
}
