/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.DeclarationUtils;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

/**
 * Auxiliary methods for extended finite state machine construction.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaEfsmTransformerUtils {

  private CgaaEfsmTransformerUtils() {
    /* Empty. */
  }

  /**
   * Returns invariants of the specified variables that are defined/used in the specified process.
   *
   * @param process process
   * @param variables variables
   *
   * @return collection of variable invariants
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public static Collection<Node> getInvariants(
      final ModuleProcess process,
      final Collection<NodeVariable> variables) {

    InvariantChecks.checkNotNull(process);
    InvariantChecks.checkNotNull(variables);

    /* The returned collection of invariants. */
    final List<Node> invariants = new LinkedList<>();

    for (final NodeVariable var : variables) {

      final Node invariant = getInvariant(process, var);

      if (invariant != null) {
        invariants.add(invariant);
      }
    }

    return invariants;
  }

  /**
   * Returns an invariant of the specified variable that is defined/used in the specified process.
   *
   * @param process The process in which the variable is defined/used.
   * @param var The variable whicn invariant is returned.
   * @return The invariant of the specified variable that is defined/used in the specified process,
   *         or {@code null} if there is no such.
   */
  public static Node getInvariant(
      final ModuleProcess process,
      final NodeVariable var) {
    InvariantChecks.checkNotNull(process);
    InvariantChecks.checkNotNull(var);

    Node invariant = null;

    /* The current variable declarations' container. */
    Statement current = process;

    do {

      if (current instanceof DecStatement
          && ((DecStatement)current).declares(var)) {

        final Declaration declaration = ((DecStatement) current).getDeclaration(var);
        invariant = DeclarationUtils.getInvariant(declaration, var);
      }

      current = current.getParent();

    } while (current != null);

    return invariant;
  }
}
