/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmModule;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * CGAA model visitor component for EFSM transformer.
 *
 * <p>Creates {@link EfsmModule} for every {@link CfgModule} component
 * of {@link ru.ispras.retrascope.model.cfg.CfgModel} and makes
 * similar parent-child connections between them.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaModuleVisitor extends CfgEmptyVisitor {

  private final EfsmModel efsmModel;
  private final Map<CfgModule, EfsmModule> moduleMap;

  /**
   * Constructs object with the specified model.
   *
   * <p>The specified model can be modified by this visitor.</p>
   *
   * @param efsmModel model to be modified
   */
  CgaaModuleVisitor(final EfsmModel efsmModel) {
    super();

    this.efsmModel = efsmModel;
    this.moduleMap = new LinkedHashMap<>();
  }

  /**
   * Returns model.
   *
   * @return model
   */
  public EfsmModel getEfsmModel() {
    return this.efsmModel;
  }

  /**
   * Returns a relation between CGAA modules and EFSM containers.
   *
   * @return mapping from modules to containers
   */
  Map<CfgModule, EfsmModule> getModuleMap() {
    return this.moduleMap;
  }

  @Override
  public void onModuleBegin(final Module module) {

    final EfsmModule efsmModule = new EfsmModule(module);
    final Statement moduleParent = module.getParent();

    if (moduleParent instanceof CfgModule) {

      efsmModule.setParent(this.moduleMap.get(moduleParent));

    } else {

      this.efsmModel.addModule(efsmModule);
    }
    this.moduleMap.put((CfgModule) module, efsmModule);
  }
}
