/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgUtils;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.SolverUtils;

import java.math.BigInteger;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
* Component that extracts data for constructing EFSM states
* and transitions from clocked guarded atomic actions model.
 * <p>NOTE: All the constructed actions are of concurrent kind.</p>
*
* @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
*/
final class CgaaPathExtractor extends MapCfgVisitor {

  /* Processes and corresponding execution paths. */
  private Map<ModuleProcess, Map<Integer, LinkedHashSet<CgaaPath>>> processDataMap;

  /* Fields that are used throughout walking process. */

  /* The current process. */
  private CfgProcess process;

  /* The data containers for current process. */
  private List<CgaaPath> dataList;

  /* The current state expression. */
  private List<Node> stateExpressions;

  /* The current list of events. */
  private EventList eventList;

  /* The current guard. */
  private List<Node> guard;

  /* The current action. */
  private List<CfgBlockStatement> action;

  private NodeVariable phaseVariable;

  private Set<Integer> srcPhases;

  private Integer dstPhase;

  /* Cases that used into state-like variables' depending expressions. */
  private List<CfgCaseStatement> stateCases;

  /* Diagram paths number (for statistics keeping). */
  private int pathNum;

  CgaaPathExtractor() {
    super();
  }

  /**
   * Returns mapping of processes into path-related data.
   *
   * @return mapping of processes into path-related data
   */
  Map<ModuleProcess, Map<Integer, LinkedHashSet<CgaaPath>>> getProcessDataMap() {

    Log.printInfo("Execution path number: " + this.pathNum);
    return this.processDataMap;
  }

  @Override
  public void onRootBegin() {

    this.processDataMap = new LinkedHashMap<>();
    this.stateExpressions = new LinkedList<>();
    this.eventList = new EventList();
    this.guard = new LinkedList<>();
    this.action = new LinkedList<>();
    this.stateCases = new LinkedList<>();
    this.pathNum = 0;
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {

    this.process = (CfgProcess) process;
    this.dataList = new LinkedList<>();

    final Enum<?> phaseNameKey = MetaInfoType.PHASE_VAR_NAME;

    if (process.hasMetaInfo(phaseNameKey)) {

      final String phaseVarName = process.getMetaInfo(phaseNameKey);
      this.phaseVariable = process.getVariable(phaseVarName);
      this.srcPhases = new LinkedHashSet<>();
    }
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {

    final Map<Integer, LinkedHashSet<CgaaPath>> srcPathMap = new LinkedHashMap<>();

    for (final CgaaPath path : this.dataList) {

      final Set<Integer> pathSrcPhases = path.getSrcPhases();

      for (final Integer pathSrcPhase : pathSrcPhases) {

        final LinkedHashSet<CgaaPath> paths;

        if (srcPathMap.containsKey(pathSrcPhase)) {

          paths = srcPathMap.get(pathSrcPhase);

        } else {

          paths = new LinkedHashSet<>();
          srcPathMap.put(pathSrcPhase, paths);
        }
        paths.add(path);
      }
    }

    this.processDataMap.put(this.process, srcPathMap);
    this.eventList.clear();
    this.phaseVariable = null;
  }

  @Override
  public void onCaseBegin(final CfgCaseStatement caseNode) {

    final CfgSwitchStatement caseParent = (CfgSwitchStatement) caseNode.getParent();

    if (phaseVarDepend(caseParent)) {

      final Set<NodeValue> caseValues = caseNode.getValues();

      for (final NodeValue caseValue : caseValues) {

        /* Phase variable values are bit vectors, so they should be converted. */
        final NodeValue bvToInt = (NodeValue) TypeConversion.coerce(caseValue, DataType.INTEGER);
        final BigInteger integerValue = bvToInt.getInteger();

        this.srcPhases.add(integerValue.intValue());
      }
    }

    /* Construct a condition expression for the current case node. */
    final Node expr = CfgUtils.getExpression(caseNode, caseParent);

    if (stateVarDepend(expr)) {

      this.stateExpressions.add(expr);
      this.stateCases.add(caseNode);

    } else {

      this.guard.add(expr);
    }

    /* If switch has event, add it as clocks. */
    if (caseNode.hasEvents()) {

      this.eventList.addEvents(caseNode.getEventList());
    }
  }

  private boolean phaseVarDepend(final CfgSwitchStatement switchStatement) {

    final Node parentNode = switchStatement.getCondition();
    final Collection<NodeVariable> parentNodeUses = ExprUtils.getVariables(parentNode);

    return parentNodeUses.contains(this.phaseVariable);
  }

  private boolean stateVarDepend(final Node expr) {

    final Collection<NodeVariable> exprVariables = ExprUtils.getVariables(expr);

    if (exprVariables.isEmpty()) {
      return false;
    }

    final Set<NodeVariable> exprVarSet = new LinkedHashSet<>(exprVariables);
    final Set<NodeVariable> stateVariables = getStateLikeVarMap().get(this.process);
    return CollectionUtils.areIntersectedSets(exprVarSet, stateVariables);
  }

  @Override
  public void onCaseEnd(final CfgCaseStatement caseNode) {

    if (StatementUtils.isLast(caseNode)) {
      processLast();
    }

    if (phaseVarDepend((CfgSwitchStatement) caseNode.getParent())) {
      this.srcPhases.clear();
    }

    if (this.stateCases.contains(caseNode)) {

      this.stateExpressions.remove(this.stateExpressions.size() - 1);
      this.stateCases.remove(caseNode);

    } else {
      this.guard.remove(this.guard.size() - 1);
    }

    if (caseNode.hasEvents()) {

      final EventList caseList = caseNode.getEventList();

      for (final Event event : caseList) {
        this.eventList.remove(event);
      }
    }
  }

  @Override
  public void onBlockBegin(final CfgBlockStatement block) {

    this.action.add(block);

    if (StatementUtils.isLast(block)) {
      processLast();
    }
  }

  private void processLast() {

    /* Action must contain at least an assignment to phase variable. */
    InvariantChecks.checkFalse(this.action.isEmpty());

    final List<Assignment> assignments = new LinkedList<>();

    for (final CfgBlockStatement block : this.action) {

      final Set<NodeVariable> blockDefines = block.getDefines();

      if (blockDefines.contains(this.phaseVariable)) {

        for (final Assignment assignment : block.getAssignments()) {

          final Set<NodeVariable> assignmentDefines = assignment.getDefines();

          if (assignmentDefines.contains(this.phaseVariable)) {

            final Node value = assignment.getValue();
            this.dstPhase = getPhaseValue((NodeValue)value);
            break;
          }
        }
      }
      assignments.addAll(block.getAssignments());
    }
    final Action action = new Action(assignments, true);

    final Set<Node> states = getNodes(this.stateExpressions);
    final Set<Node> guard = getNodes(this.guard);

    final GuardedAction guardedAction = new GuardedAction(guard, action);

    if (!this.eventList.isEmpty()) {
      guardedAction.addClocks(this.eventList);
    }

    final Node condition = guardedAction.getClockGuardNode();

    if (SolverUtils.isSat(condition)) {

      final CgaaPath container =
          new CgaaPath(states, guardedAction, this.srcPhases, this.dstPhase);
      this.pathNum++;
      this.dataList.add(container);
    }
    this.action.clear();
  }

  private static Integer getPhaseValue(final NodeValue value) {

    /* Phase variable values are bit vectors, so they should be converted. */
    final NodeValue bvToInt = (NodeValue) TypeConversion.coerce(value, DataType.INTEGER);
    final BigInteger integerValue = bvToInt.getInteger();
    return integerValue.intValue();
  }

  private Set<Node> getNodes(final List<Node> list) {

    final Set<Node> nodes = new LinkedHashSet<>();

    if (list.isEmpty()) {
      nodes.add(SolverUtils.TRUE);

    } else {
      nodes.addAll(list);
    }
    return nodes;
  }
}
