/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import edu.uci.ics.jung.algorithms.shortestpath.DijkstraShortestPath;
import edu.uci.ics.jung.graph.Graph;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.cfg.transformer.cgaa.PhaseSeqEdge;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgStatement;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cgaa.Phase;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;

import java.math.BigInteger;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Component that detects state-like variables in clocked guarded atomic actions model.
 * <p>An heuristic-based approach is used for state-like variable detection. The following
 * conditions are obligatory for the variable to be treated as state-like:
 * <ul>
 *   <li>Not an input signal</li>
 *   <li>Exists at least one path in model where this variable in used in a control part
 *   and defined in a data part</li>
 *   <li>Every switch expression that uses the state-like variable,
 *   depends exactly from this variable and can not depend from other variables.</li>
 * </ul></p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class CgaaStateVarDetector extends MapCfgVisitor {

  /* The process under analysis. */
  private ModuleProcess process;

  /* The control part of current path. */
  private List<CfgStatement> controlPath;

  /* The data part of current path. */
  private List<CfgBlockStatement> dataPath;

  /* The switch nodes expressions. */
  private Set<Node> switchNodes;

  /* Mapping to variables by which key variables are defined. */
  private Map<NodeVariable, Collection<NodeVariable>> depVarMap;

  /* Mapping from CGAA model processes to their phase sequence graphs. */
  private Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> phaseSeqMap;

  /* Phase sequence graph for the elaborated process. */
  private Graph<Phase, PhaseSeqEdge> phaseSeqGraph;

  /* Mapping from variables to phases at which they are defined. */
  private Map<NodeVariable, Set<Phase>> defPhaseMap;

  /* Mapping from variables to phases at which they are used in control parts. */
  private Map<NodeVariable, Set<Phase>> controlUsePhaseMap;

  /* The state-like variables. */
  private List<NodeVariable> stateLikeVars;

  /* Variables that cannot be state-like. */
  private Set<NodeVariable> forbidden;

  /* Names of user-defined state variables. */
  private String[] userDefinedNames;

  private Status status;

  /**
   * The default constructor.
   */
  CgaaStateVarDetector() {
    super();

    setStatus(Status.OK);
  }

  /**
   * Grants an access to the specified mapping for this component.
   * @param phaseSeqMap The mapping an access to this object needs.
   */
  void setPhaseSeqMap(Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> phaseSeqMap) {
    this.phaseSeqMap = phaseSeqMap;
  }

  private boolean inControlPath(final CfgStatement statement) {
    return this.controlPath.contains(statement);
  }

  private boolean inDataPath(final CfgBlockStatement blockStatement) {
    return this.dataPath.contains(blockStatement);
  }

  private void setStatus(final Status status) {
    this.status = status;
  }

  /**
   * Sets user-defined names of state-like variables.
   *
   * @param names state-like variables names
   */
  void setStateNames(final String... names) {
    this.userDefinedNames = names;
  }

  @Override
  public Status getStatus() {
    return this.status;
  }

  @Override
  public void onRootBegin() {

    this.controlPath = new LinkedList<>();
    this.dataPath = new LinkedList<>();
    this.stateLikeVars = new LinkedList<>();
  }

  @Override
  public void onRootEnd() {
    setStatus(Status.OK);
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {

    this.process = process;
    this.phaseSeqGraph = this.phaseSeqMap.get(process);
    this.depVarMap = new LinkedHashMap<>();
    this.forbidden = new LinkedHashSet<>();
    this.defPhaseMap = new LinkedHashMap<>();
    this.controlUsePhaseMap = new LinkedHashMap<>();

    /* Phase variable should not be detected as a state-like one. */
    final String phaseVarName = getPhaseVarName();

    if (phaseVarName != null) {
      this.forbidden.add(process.getVariable(phaseVarName));
    }

    if (emptyArray(this.userDefinedNames)) {

      this.switchNodes = new LinkedHashSet<>();

    } else {
      final Set<NodeVariable> stateLikeVars = new LinkedHashSet<>();

      for (final String name : this.userDefinedNames) {

        final DecStatement variableContainer = StatementUtils.getVariableContainer(process, name);

        if (variableContainer != null) {
          stateLikeVars.add(variableContainer.getVariable(name));
        }
      }
      getStateLikeVarMap().put(process, stateLikeVars);
      setStatus(Status.SKIP);
    }
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {

    if (emptyArray(this.userDefinedNames)) {

      final Set<NodeVariable> defined = this.defPhaseMap.keySet();
      final Set<NodeVariable> controlUsed = this.controlUsePhaseMap.keySet();

      final Set<NodeVariable> both =
          CollectionUtils.complementSets(
              CollectionUtils.intersectSets(defined, controlUsed), this.forbidden);

      for (final NodeVariable variable : both) {
        checkStateLikeCandidate(variable);
      }

      /* Remove forbidden variables from the collection of state-like variables. */
      for (final NodeVariable forbiddenVar : this.forbidden) {

        if (this.stateLikeVars.contains(forbiddenVar)) {
          this.stateLikeVars.remove(forbiddenVar);
        }
      }

      /* Check that all the module switch nodes
       * use no more than one state-like variable.
       */
      checkStateVariables();

      /*
        Check that all the state-like variables for this process depend exactly from each other.
       */
      checkVariableDependency();

      getStateLikeVarMap().put(process, new LinkedHashSet<>(this.stateLikeVars));
      this.stateLikeVars.clear();
    }
  }

  private void checkStateLikeCandidate(final NodeVariable variable) {

    final Set<Phase> defPhases = this.defPhaseMap.get(variable);
    final Set<Phase> controlUsePhases = this.controlUsePhaseMap.get(variable);

    for (final Phase defPhase : defPhases) {

      for (final Phase usePhase : controlUsePhases) {

        final DijkstraShortestPath<Phase, PhaseSeqEdge> pathInstance =
            new DijkstraShortestPath<>(this.phaseSeqGraph);

        final List<PhaseSeqEdge> path = pathInstance.getPath(usePhase, defPhase);

        if (path != null) {
          this.stateLikeVars.add(variable);
          return;
        }
      }
    }
  }

  private boolean emptyArray(final String[] strings) {
    return strings == null || strings.length == 0;
  }

  private void checkStateVariables() {

    for (final Node node : this.switchNodes) {

      final Collection<NodeVariable> nodeVariables = ExprUtils.getVariables(node);
      final Collection<NodeVariable> stateVariables = new LinkedHashSet<>();
      boolean hasNonState = false;

      for (final NodeVariable variable : nodeVariables) {

        if (fromStateVars(variable)) {
          stateVariables.add(variable);

        } else if (!hasNonState) {
          hasNonState = true;
        }
      }

      if (stateVariables.size() > 1 || hasNonState) {
        this.stateLikeVars.removeAll(stateVariables);
      }
    }
  }

  private boolean fromStateVars(final NodeVariable variable) {
    return this.stateLikeVars.contains(variable);
  }

  private void checkVariableDependency() {

    final Set<NodeVariable> stateVariableCopy = new LinkedHashSet<>();
    stateVariableCopy.addAll(this.stateLikeVars);

    for (final NodeVariable stateCandidate : stateVariableCopy) {

      if (this.depVarMap.containsKey(stateCandidate)) {

        final Collection<NodeVariable> depVariables = this.depVarMap.get(stateCandidate);

        for (final NodeVariable dep : depVariables) {

          if (!this.stateLikeVars.contains(dep)) {
            this.stateLikeVars.remove(stateCandidate);
          }
        }
      }
    }
  }

  @Override
  public void onBlockBegin(final CfgBlockStatement block) {

    this.dataPath.add(block);

    for (final Assignment assignment : block.getAssignments()) {

      final Set<NodeVariable> uses = assignment.getUses();
      final Set<NodeVariable> defines = assignment.getDefines();

      for (final NodeVariable define : defines) {

        if (this.depVarMap.containsKey(define)) {
          this.depVarMap.get(define).addAll(uses);

        } else {
          this.depVarMap.put(define, uses);
        }
      }
    }

    processTerminalStatement(block);
  }

  @Override
  public void onBlockEnd(final CfgBlockStatement block) {
    removeFromDataPath(block);
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {

    this.controlPath.add(switchNode);
    this.switchNodes.add(switchNode.getCondition());
  }

  @Override
  public void onSwitchEnd(final CfgSwitchStatement switchNode) {

    removeFromControlPath(switchNode);
  }

  @Override
  public void onCaseBegin(final CfgCaseStatement caseStatement) {

    this.controlPath.add(caseStatement);

    processTerminalStatement(caseStatement);
  }

  @Override
  public void onCaseEnd(final CfgCaseStatement caseStatement) {

    removeFromControlPath(caseStatement);
  }

  private void removeFromControlPath(final CfgStatement statement) {

    if (inControlPath(statement)) {

      this.controlPath.remove(statement);

    } else {

      throw RetrascopeRuntimeException.makeException(
          LogMessage.CANT_FIND,
          statement.getDescription());
    }
  }

  private void removeFromDataPath(final CfgBlockStatement statement) {

    if (inDataPath(statement)) {

      this.dataPath.remove(statement);

    } else {

      throw RetrascopeRuntimeException.makeException(
          LogMessage.CANT_FIND,
          statement.getDescription());
    }
  }

  private void processTerminalStatement(final Statement statement) {

    if (StatementUtils.isLast(statement)) {

      final Phase pathPhase = getPhase();

      /* Get variables that are defined in data parts of graph paths. */
      final Set<NodeVariable> definedDataVars = getDefinedVariables(this.dataPath);

      /* Get variables that are used in control parts of graph paths. */
      final List<NodeVariable> usedControlVars = getUsedVariables(this.controlPath);

      /* The weak heuristic is applied. */
      for (final NodeVariable var : definedDataVars) {

        if (noDescriptor(var)) {
          continue;
        }

        final VariableType type = ((VariableDescriptor) var.getUserData()).getVariableType();

        if (!VariableType.isSignal(type)) {

          if (!this.defPhaseMap.containsKey(var)) {
            this.defPhaseMap.put(var, new LinkedHashSet<Phase>());
          }
          this.defPhaseMap.get(var).add(pathPhase);
        }
      }

      for (final NodeVariable var : usedControlVars) {

        if (noDescriptor(var)) {
          continue;
        }

        final VariableType type = ((VariableDescriptor) var.getUserData()).getVariableType();

        if (!VariableType.isSignal(type)) {

          if (!this.controlUsePhaseMap.containsKey(var)) {
            this.controlUsePhaseMap.put(var, new LinkedHashSet<Phase>());
          }

          this.controlUsePhaseMap.get(var).add(pathPhase);
        }
      }
    }
  }

  private Phase getPhase() {

    Phase dataPathPhase = null;

    final String phaseVarName = getPhaseVarName();

    for (final CfgBlockStatement block : this.dataPath) {

      final List<Assignment> assignments = block.getAssignments();

      for (final Assignment assignment : assignments) {

        final RangedVariable target = assignment.getTarget();
        final NodeVariable targetVariable = target.getVariable();
        final String targetName = targetVariable.getName();

        if (targetName.equals(phaseVarName)) {

          final Node value = assignment.getValue();

          if (value instanceof NodeValue) {

            /* Phase variable values are bit vectors, so they should be converted. */

            final NodeValue bvToInt = (NodeValue) TypeConversion.coerce(value, DataType.INTEGER);
            final BigInteger integerValue = bvToInt.getInteger();

            dataPathPhase = new Phase(integerValue.intValue());
          }
        }
      }
    }

    InvariantChecks.checkTrue(dataPathPhase != null);

    return dataPathPhase;
  }

  private String getPhaseVarName() {

    String name = null;

    final Enum<?> phaseMetaInfoKey = MetaInfoType.PHASE_VAR_NAME;

    if (this.process.hasMetaInfo(phaseMetaInfoKey)) {
      name = this.process.getMetaInfo(phaseMetaInfoKey);
    }

    return name;
  }

  private static boolean noDescriptor(final NodeVariable var) {
    return var.getUserData() == null || !(var.getUserData() instanceof VariableDescriptor);
  }

  private Set<NodeVariable> getDefinedVariables(final List<CfgBlockStatement> dataPath) {

    /* Get those defined variables that do not depend on input signals. */

    final Set<NodeVariable> defined = new LinkedHashSet<>();
    final Set<NodeVariable> forbiddenVariables = new LinkedHashSet<>();

    for (final CfgBlockStatement block : dataPath) {

      final List<Assignment> assignments = block.getAssignments();

      for (final Assignment assignment : assignments) {

        final Collection<NodeVariable> stmtUses = assignment.getUses();
        final Set<NodeVariable> defines = assignment.getDefines();

        if (areSignalDependent(stmtUses)) {

          /* Variables that are defined here, cannot be returned. */
          forbiddenVariables.addAll(defines);

        } else {
          defined.addAll(CollectionUtils.complementSets(defines, forbiddenVariables));
        }
      }

      final Collection<NodeVariable> blockDefines = block.getDefines();
      for (final NodeVariable blockDefined : blockDefines) {
        if (!forbiddenVariables.contains(blockDefined)) {
          defined.add(blockDefined);
        }
      }
    }
    this.forbidden.addAll(forbiddenVariables);
    return defined;
  }

  private boolean areSignalDependent(final Collection<NodeVariable> stmtUses) {
    boolean dependOnSignal = false;
    for (final NodeVariable use : stmtUses) {

      final Object data = use.getUserData();

      if (data == null) {
        Log.printWarn(String.format("No data for '%s' variable", use.getName()));
      }

      if (VariableType.isInputSignal(((VariableDescriptor) use.getUserData()).getVariableType())) {
        dependOnSignal = true;
        break;
      }
    }
    return dependOnSignal;
  }

  private List<NodeVariable> getUsedVariables(final List<CfgStatement> controlPath) {
    final List<NodeVariable> used = new LinkedList<>();

    for (final CfgStatement node : controlPath) {

      if (node.getType() == CfgStatementType.SWITCH) {

        final Collection<NodeVariable> nodeUses =
            ExprUtils.getVariables(((CfgSwitchStatement)node).getCondition());

        for (final NodeVariable nodeUse : nodeUses) {

          if (!used.contains(nodeUse)) {
            used.add(nodeUse);
          }
        }
      } else if (node.getType() == CfgStatementType.CASE) {

        final EventList eventList = ((CfgCaseStatement)node).getEventList();

        for (final Event event : eventList) {

          final RangedVariable eventVariable = event.getRangedVariable();
          final Set<NodeVariable> eventUses = eventVariable.getUses();

          for (final NodeVariable eventUsed : eventUses) {

            if (!used.contains(eventUsed)) {
              used.add(eventUsed);
            }
          }
        }
      }
    }
    return used;
  }
}
