/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.GuardedAction;

import java.util.Collection;
import java.util.Iterator;

/**
 * Component that checks whether transitions are reset-like (heuristic-based algorithm).
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class EfsmRstChecker {

  private final Efsm efsm;

  /**
   * Constructs checker for the specified EFSM.
   * @param efsm The EFSM which transitions to be checked.
   */
  EfsmRstChecker(final Efsm efsm) {
    this.efsm = efsm;
  }

  /**
   * Returns the EFSM that is associated with the checker.
   * @return The EFSM that is associated with the checker.
   */
  Efsm getEfsm() {
    return efsm;
  }

  /**
   * Checks whether the specified transition is state-like.
   *
   * @param transition The transition to be checked.
   *
   * @return {@code true} when the specified transition is reset-like, {@code false} otherwise.
   */
  boolean isResetTransition(final EfsmTransition transition) {

    return hasInputDependGuard(transition) && hasConstantAction(transition);
  }

  /* Check whether guard depends on exactly one in-inout signal
  . or guard is trivial and clocks collection is not empty. */
  private boolean hasInputDependGuard(final GuardedAction guardedAction) {

    final Node guard = guardedAction.getGuardNode();
    final Collection<NodeVariable> guardUses = ExprUtils.getVariables(guard);

    boolean oneInputDepend = false;

    for (final NodeVariable guardUseVar : guardUses) {

      final String name = guardUseVar.getName();
      final MetaInfoType phaseVarName = MetaInfoType.PHASE_VAR_NAME;

      if (getEfsm().hasMetaInfo(phaseVarName) && getEfsm().getMetaInfo(phaseVarName).equals(name)) {
        continue;
      }

      final Object data = guardUseVar.getUserData();

      if (data instanceof VariableDescriptor
          && VariableType.isInputSignal(((VariableDescriptor) data).getVariableType())) {

        oneInputDepend = true;
      }
    }

    if (isTrivialGuard(guardedAction, oneInputDepend)) {
      oneInputDepend = true;
    }

    return oneInputDepend;
  }

  private boolean isTrivialGuard(final GuardedAction guardedAction, final boolean oneInputGuard) {

    final Collection<NodeVariable> guardVariables =
        ExprUtils.getVariables(guardedAction.getGuardNode());

    final MetaInfoType phaseVarNameKey = MetaInfoType.PHASE_VAR_NAME;
    final String phaseVarName = getEfsm().getMetaInfo(phaseVarNameKey);
    final Iterator<NodeVariable> iterator = guardVariables.iterator();
    final NodeVariable firstVariable = iterator.next();

    final boolean phaseOnlyDependGuard = guardVariables.size() == 1
        && firstVariable.getName().equals(phaseVarName);

    return (guardVariables.isEmpty() || phaseOnlyDependGuard)
        && !oneInputGuard
        && !guardedAction.getClocks().isEmpty();
  }

  private boolean hasConstantAction(final EfsmTransition transition) {

    /* Check whether non-empty action contains only assignments like var := constant. */

    final Action action = transition.getAction();
    boolean checkAction = false;

    for (final Assignment assignment : action.getAssignments()) {

      checkAction = assignment.getUses().isEmpty();

      if (!checkAction) {
        break;
      }
    }
    return checkAction;
  }
}
