/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmEmptyVisitor;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.util.NodeUtils;

import java.util.LinkedList;
import java.util.List;

/**
 * EFSM visitor that build a sequence of transitions that performs model resetting.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class EfsmRstVisitor extends EfsmEmptyVisitor {

  /* The phase variable for the EFSM to be visited. */
  private final NodeVariable phaseVariable;

  /* The sequence of reset-like transitions for the EFSM to be visited. */
  private final LinkedList<EfsmTransition> rstTransitions;

  /* The auxiliary checker for transitions of the EFSM to be visited. */
  private final EfsmRstChecker rstChecker;

  private Status status;

  /**
   * Constructs visitor for the specified EFSM.
   *
   * @param efsm The related EFSM.
   */
  EfsmRstVisitor(final Efsm efsm) {

    this.rstTransitions = new LinkedList<>();
    this.status = Status.OK;

    final MetaInfoType phaseVarNameKey = MetaInfoType.PHASE_VAR_NAME;
    final String phaseVarName = efsm.getMetaInfo(phaseVarNameKey);
    this.phaseVariable = efsm.getVariable(phaseVarName);

    this.rstChecker = new EfsmRstChecker(efsm);
  }

  /**
   * Returns the sequence of reset-like transitions.
   *
   * @return The sequence of reset-like transitions.
   */
  List<EfsmTransition> getRstTransitions() {
    return this.rstTransitions;
  }

  private NodeVariable getPhaseVariable() {
    return this.phaseVariable;
  }

  private EfsmRstChecker getChecker() {
    return this.rstChecker;
  }

  @Override
  public Status getStatus() {
    return this.status;
  }

  @Override
  public void setStatus(final Status status) {
    this.status = status;
  }

  @Override
  public void onStateBegin(final EfsmState state) {

    final Efsm efsm = getChecker().getEfsm();
    final Declaration phaseVarDeclaration = efsm.getDeclaration(getPhaseVariable());
    final Node initPhase = phaseVarDeclaration.getInitialValue();

    final NodeValue statePhase = NodeValue.newInteger(state.getPhase());
    final Node phase = TypeConversion.coerce(statePhase, getPhaseVariable().getDataType());

    final Node phaseEquation = NodeUtils.getEquation(initPhase, phase);

    if (!getRstTransitions().isEmpty() && ExprUtils.isSAT(phaseEquation)) {
      this.status = Status.ABORT;
    }
  }

  @Override
  public void onTransitionBegin(final EfsmTransition transition) {

    if (getChecker().isResetTransition(transition)) {

      this.rstTransitions.add(transition);

    } else {
      this.status = Status.SKIP;
    }
  }

  @Override
  public void onTransitionEnd(final EfsmTransition transition) {

    if (this.status == Status.SKIP) {

      this.status = Status.OK;

    } else if (getRstTransitions().contains(transition)) {

      final EfsmTransition last = this.rstTransitions.removeLast();
      InvariantChecks.checkTrue(last.equals(transition));
    }
  }
}
