/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.hldd;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddConditionNode;
import ru.ispras.retrascope.model.hldd.HlddEdge;
import ru.ispras.retrascope.model.hldd.HlddModel;
import ru.ispras.retrascope.model.hldd.HlddModule;
import ru.ispras.retrascope.model.hldd.HlddNode;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.model.hldd.HlddTerminalNode;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This class represents a visitor which is used to construct a HLDD model from a CGAA model.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
final class CgaaHlddVisitor extends CfgEmptyVisitor {

  // HLDD being constructed at the moment.
  private Hldd currentHldd;
  // Resulting model.
  private final HlddModel hlddModel;
  // HlddNode stack.
  private LinkedList<HlddNode> nodeStack;
  // Switch - corresponding variable mapping.
  private Map<Node, NodeVariable> switchesAsVariables;
  // Counter of guard conditions.
  private int guardCounter;
  // Current module variables.
  private Collection<NodeVariable> currentVariables;
  // Contains resulting process HLDDs.
  private HlddProcess processHldds;
  // Current node level.
  private int currentLevel;
  // HlddModule stack.
  private final LinkedList<Module> moduleStack;


  /**
   * Constructs a visitor.
   */
  public CgaaHlddVisitor() {
    super();
    currentHldd = new Hldd();
    hlddModel = new HlddModel();
    nodeStack = new LinkedList<>();
    switchesAsVariables = new HashMap<>();
    guardCounter = 0;
    currentLevel = 0;
    moduleStack = new LinkedList<>();
  }

  /**
   * Returns the constructed HLDD model.
   *
   * @return The {@link HlddModel}.
   */
  public HlddModel getModel() {
    return hlddModel;
  }

  @Override
  public void onModuleBegin(final Module module) {
    final HlddModule hlddModule = new HlddModule(module);
    moduleStack.push(hlddModule);
    currentVariables = new ArrayList<>();
    currentVariables.addAll(hlddModule.getVariables());
    // Keep only outputs and regs.
    currentVariables.removeAll(hlddModule.getInputs());
  }

  @Override
  public void onModuleEnd(final Module module) {
    final Module hlddModule = moduleStack.pop();
    if (moduleStack.isEmpty()) {
      hlddModel.addModule(hlddModule);
    } else {
      moduleStack.peek().add(hlddModule);
    }
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    currentHldd = new Hldd();
    processHldds = new HlddProcess(process);
    nodeStack = new LinkedList<>();
    moduleStack.peek().add(processHldds);
    switchesAsVariables = new HashMap<>();
    guardCounter = 0;
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {
    final Collection<NodeVariable> variables = new ArrayList<>(process.getVariables());
    variables.addAll(currentVariables);

    // Create separate HLDDs.
    for (NodeVariable variable : variables) {
      final Hldd hldd = removeWrongVariableNodes(currentHldd.deepCopy(), variable);
      if (!hldd.getTerminalNodes().isEmpty()) {
        this.processHldds.add(hldd);
      }
    }

    // Create guard variable HLDDs.
    final List<Entry<Node, NodeVariable>> entries = new ArrayList<>(switchesAsVariables.entrySet());
    sortByName(entries);
    for (Entry<Node, NodeVariable> entry : entries) {
      final Node switchNode = entry.getKey();
      final NodeVariable var = entry.getValue();
      this.processHldds.add(createHlddFromSwitch(switchNode, var));
    }

    currentLevel--;

  }

  private void sortByName(final List<Entry<Node, NodeVariable>> entries) {
    Collections.sort(entries, new Comparator<Entry<Node, NodeVariable>>() {
      @Override
      public int compare(Entry<Node, NodeVariable> o1, Entry<Node, NodeVariable> o2) {
        final String var1 = o1.getValue().getName();
        final String var2 = o2.getValue().getName();
        return var1.compareTo(var2);
      }
    });
  }

  /*
   * Removes all the nodes which are not associated with the specified variable.
   *
   * @param hldd - the specified HLDD prototype.
   *
   * @param variable - the specified variable.
   *
   * @return The {@link Hldd} containing only nodes with the specified variable.
   */
  private Hldd removeWrongVariableNodes(final Hldd newHldd, final NodeVariable variable) {
    newHldd.setVariable(variable);
    for (HlddNode node : newHldd.getNodes()) {
      if (node.isTerminal()) {
        if (!((HlddTerminalNode) node).getTargetVariable().equals(variable)) {
          newHldd.removeNode(node);
        } else {
          final HlddEdge parent = (HlddEdge) node.getParent();
          if (parent != null) {
            parent.addChild(node);
          }
        }
      }
    }
    // If a HLDD consists of only one assignment node.
    if (newHldd.getNodes().size() == 1) {
      newHldd.setRootNode(newHldd.getNodes().iterator().next());
    }
    return newHldd;
  }


  @Override
  public void onBlockBegin(final CfgBlockStatement block) {
    final HlddNode parent = nodeStack.isEmpty() ? null : nodeStack.peek();
    for (Assignment assignment : block.getAssignments()) {
      final HlddTerminalNode node =
          new HlddTerminalNode(parent, assignment.getVariable(), assignment.getValue());
      node.setLevel(currentLevel);
      currentHldd.addNode(node);
    }
  }


  @Override
  public void onCaseBegin(final CfgCaseStatement caseNode) {
    final HlddConditionNode parent = (HlddConditionNode) nodeStack.peek();
    final HlddEdge edge = new HlddEdge(parent);
    edge.setLevel(currentLevel);
    currentLevel++;
    edge.setEventList(caseNode.getEventList());
    if (caseNode.isDefault()) {
      edge.setDefault();
    } else {
      edge.addExpressions(caseNode.getValues());
    }
    currentHldd.addNode(edge);
    parent.addChild(edge);
    nodeStack.push(edge);
  }

  @Override
  public void onCaseEnd(final CfgCaseStatement caseNode) {
    nodeStack.pop();
    currentLevel--;
  }

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {
    createVariableFromSwitch(switchNode);

    final HlddNode parent = nodeStack.isEmpty() ? null : nodeStack.peek();
    final HlddConditionNode condition =
        new HlddConditionNode(parent, switchesAsVariables.get(switchNode.getCondition()));
    condition.setLevel(currentLevel);
    currentHldd.addNode(condition);
    if (parent == null) {
      currentHldd.setRootNode(condition);
    } else {
      ((HlddEdge) parent).addChild(condition);
    }

    nodeStack.push(condition);
    currentLevel++;
  }

  @Override
  public void onSwitchEnd(final CfgSwitchStatement switchNode) {
    nodeStack.pop();
    currentLevel--;
  }

  /*
   * Associates a variable with the specified switch node.
   */
  private void createVariableFromSwitch(final CfgSwitchStatement switchNode) {
    if (!switchesAsVariables.keySet().contains(switchNode.getCondition())) {
      final NodeVariable variable =
          new NodeVariable("guard" + guardCounter, switchNode.getCondition().getDataType());
      switchesAsVariables.put(switchNode.getCondition(), variable);
      guardCounter++;
    }
  }


  /*
   * Constructs a one-node HLDD from the specified switch node and associated variable.
   *
   * @param switchNode - the specified {@link CfgSwitchStatement}.
   *
   * @param variable - the specified variable.
   *
   * @return The {@link Hldd} associated with a switch.
   */
  private Hldd createHlddFromSwitch(final Node switchNode, final NodeVariable variable) {
    final Hldd result = new Hldd(variable);
    final HlddTerminalNode singleNode =
        new HlddTerminalNode(null, switchesAsVariables.get(switchNode), switchNode);
    singleNode.setLevel(0);
    result.addNode(singleNode);
    result.setRootNode(singleNode);
    result.setImmediateAssignment(true);
    return result;
  }

}
