/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.hldd;

import ru.ispras.fortress.expression.Node;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddConditionNode;
import ru.ispras.retrascope.model.hldd.HlddEdge;
import ru.ispras.retrascope.model.hldd.HlddEmptyVisitor;
import ru.ispras.retrascope.model.hldd.HlddNode;
import ru.ispras.retrascope.model.hldd.HlddTerminalNode;

import java.util.ArrayList;
import java.util.List;

/**
 * This class represents a visitor which is used to optimize a visited HLDD.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddOptimizationVisitor extends HlddEmptyVisitor {

  // The current visited HLDD.
  private Hldd currentHldd;

  public HlddOptimizationVisitor() {
    super();
  }

  @Override
  public void onHlddBegin(final Hldd hldd) {
    currentHldd = hldd;
  }

  /*
   * Adds the default assignment node to the value node.
   *
   * @param node - the specified value node.
   */
  private void addDefaultAssignment(final HlddEdge node) {
    if (node.isEdge()) {
      final HlddTerminalNode defaultNode =
          new HlddTerminalNode(node, currentHldd.getVariable(), currentHldd.getVariable());
      defaultNode.setLevel(node.getLevel() + 1);
      node.addChild(defaultNode);
    }
  }

  @Override
  public void onConditionBegin(final HlddConditionNode node) {
    if (node.getChildren().size() == 1) {
      final HlddNode onlyValue = node.getChildren().iterator().next();
      final HlddEdge secondValue = new HlddEdge(node);
      secondValue.setLevel(onlyValue.getLevel());
      secondValue.setDefault();
      currentHldd.addNode(secondValue);
    }
  }

  @Override
  public void onConditionEnd(final HlddConditionNode node) {
    final List<HlddNode> assignments = new ArrayList<>(currentHldd.getFollowingTerminalNodes(node));
    if (containIdenticalNodes(assignments)) {
      final Node sameExpr = ((HlddTerminalNode) assignments.get(0)).getAssignedValue();
      final HlddEdge parent = (HlddEdge) node.getParent();
      if (parent != null) {
        parent.removeChild(node);
        parent.addChild(new HlddTerminalNode(node, currentHldd.getVariable(), sameExpr));
      }
      final ArrayList<HlddNode> unusedNodes = new ArrayList<>();
      unusedNodes.addAll(currentHldd.getFollowingNodes(node));
      unusedNodes.add(node);
      currentHldd.removeNodes(unusedNodes);
    }
  }

  /*
   * Checks whether all nodes in the list contain equal expressions.
   *
   * @param nodes - the specified list of nodes.
   *
   * @return <code>true</code> if all nodes contain equal expressions, <code>false</code> otherwise.
   */
  private boolean containIdenticalNodes(final List<HlddNode> nodes) {
    if (!nodes.isEmpty() && nodes.size() > 1) {
      final Node sameNode = ((HlddTerminalNode) nodes.get(0)).getAssignedValue();
      for (HlddNode assignment : nodes) {
        if (!((HlddTerminalNode)assignment).getAssignedValue().equals(sameNode)) {
          return false;
        }
      }
      return true;
    } else {
      return false;
    }
  }

  @Override
  public void onEdgeEnd(final HlddEdge node) {
    if (node.getChildren().isEmpty()) {
      addDefaultAssignment(node);
    }
  }

}
