/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.solver.SolverResultBuilder;
import ru.ispras.retrascope.engine.basis.AtomicTestGenerator;
import ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction.NearestUncoveredEfsmTransitionChooser;
import ru.ispras.retrascope.engine.efsm.simulator.ModuleSimulator;
import ru.ispras.retrascope.engine.efsm.simulator.ProcessSimulator;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.Efsm.TransitionWeightBasis;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Directed test generator for single EFSM.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see RandomGenerator
 * @see EfsmTestGenerator
 */
final class DirectedGenerator extends AtomicTestGenerator {
  private static final Logger LOGGER = Log.getLogger();
  private final String logEntryHeader;

  private final Set<List<EfsmTransition>> traversablePaths;
  private final Map<EfsmTransition, Set<EfsmTransition>> testedCoreTransitions;

  private final ProcessSimulator simulator;
  private final int loopLimit;

  private final EfsmTransitionCoverage totalCoverage;

  private NearestUncoveredEfsmTransitionChooser directedChooser;
  private final EfsmAnalyser analyser;

  private final Map<EfsmState, Map<EfsmState, List<EfsmTransition>>> preferredBridges;
  private final Set<List<List<EfsmTransition>>> untraversableFragmentedPaths;
  private final Set<List<EfsmTransition>> untraversablePlainPaths;

  /**
   * Constructs a new EFSM atomic test generator for the specified EFSM.
   *
   * @param efsm the EFSM
   * @param traversablePaths a set of paths which are guaranteed to be traversable
   * @param loopLimit max number of single EFSM loop traversing in one path
   */
  DirectedGenerator(
      final Efsm efsm,
      final Set<List<EfsmTransition>> traversablePaths,
      final int loopLimit,
      final ModuleSimulator moduleSimulator) {
    this.logEntryHeader = "EFSM.TestGenerator.DirectedGenerator." + efsm.getName();

    this.loopLimit = loopLimit;

    this.simulator = moduleSimulator.getSubSimulator(efsm);

    this.totalCoverage = new EfsmTransitionCoverage(efsm);

    this.analyser = new EfsmAnalyser(efsm);

    for (List<EfsmTransition> traversablePath: traversablePaths) {
      for (EfsmTransition coveredTransition: traversablePath) {
        this.totalCoverage.markTransitionAsCovered(coveredTransition);
      }
    }

    this.traversablePaths = new HashSet<>(traversablePaths);
    this.testedCoreTransitions = new HashMap<>();

    this.directedChooser = null;

    this.preferredBridges = new HashMap<>();
    this.untraversableFragmentedPaths = new HashSet<>();
    this.untraversablePlainPaths = new HashSet<>();
  }

  /**
   * Returns an EFSM which this generator is related to.
   *
   * @return the EFSM
   */
  public Efsm getEfsm() {
    return this.simulator.getSimulatedModel();
  }

  /**
   * Returns an iterator enabling to retrieve next sequence of vectors produced by this generator.
   *
   * @return the iterator. If this generator is already finished generation,
   *         {@code null} is returned
   */
  @Override
  public Iterator<Vector> getNextSequenceIterator() {
    if (!this.totalCoverage.isFull()) {
      if (this.directedChooser == null) {
        this.directedChooser = new NearestUncoveredEfsmTransitionChooser(
            this.simulator.getSimulatedModel(), this.totalCoverage);
      }

      for (EfsmTransition transition = this.directedChooser.chooseNext(); transition != null;
          transition = this.directedChooser.chooseNext()) {
        List<Vector> vectors = this.generateSequence(transition);

        if (!vectors.isEmpty()) {
          this.directedChooser = new NearestUncoveredEfsmTransitionChooser(
              this.simulator.getSimulatedModel(), this.totalCoverage);

          return vectors.iterator();
        }
      }
    }

    return null;
  }

  private List<Vector> generateSequence(EfsmTransition transition) {
    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
        + ": generating a sequence to cover the transition: " + transition);

    List<Vector> vectors = tryDirectPath(transition);

    if (!vectors.isEmpty()) {
      return vectors;
    }

    Map<EfsmTransition, Set<List<EfsmTransition>>> definitionPaths = new HashMap<>();
    Map<EfsmTransition, Set<Set<List<EfsmTransition>>>> loopDefinitionPaths = new HashMap<>();

    for (List<EfsmTransition> definitionPath: this.analyser.getDefinitionPaths(transition)) {
      final EfsmTransition coreTransition = definitionPath.get(0);
      Set<List<EfsmTransition>> relatedPaths = definitionPaths.get(coreTransition);

      if (relatedPaths == null) {
        relatedPaths = new HashSet<>();
        definitionPaths.put(coreTransition, relatedPaths);
      }

      relatedPaths.add(definitionPath);
    }

    for (final Set<List<EfsmTransition>> definitionLoops :
        this.analyser.getLoopDefinitionPaths(transition)) {
      final EfsmTransition coreTransition = definitionLoops.iterator().next().get(0);
      Set<Set<List<EfsmTransition>>> relatedLoops = loopDefinitionPaths.get(coreTransition);

      if (relatedLoops == null) {
        relatedLoops = new HashSet<>();
        loopDefinitionPaths.put(coreTransition, relatedLoops);
      }

      relatedLoops.add(definitionLoops);
    }

    final Set<EfsmTransition> coreTransitions = new HashSet<>(definitionPaths.keySet());
    coreTransitions.addAll(loopDefinitionPaths.keySet());

    if (coreTransitions.isEmpty()) {
      LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the transition is independent of inner variables");
    }

    for (final EfsmTransition coreTransition : coreTransitions) {
      if (this.testedCoreTransitions.containsKey(transition)
          && this.testedCoreTransitions.get(transition).contains(coreTransition)) {
        continue;
      }

      if (this.totalCoverage.isCovered(coreTransition)) {
        LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
            + ": trying full definition paths from the transition: " + coreTransition);

        final List<EfsmTransition> shortestHeadPath = findShortestTraversablePath(coreTransition);

        if (shortestHeadPath.isEmpty()) {
          LOGGER.log(LogLevel.DEBUG,
              this.logEntryHeader + ": the shortest next head path is empty");
        } else {
          LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": the shortest head path: ");
          logPlainPath(shortestHeadPath, LogLevel.DEBUG);
        }

        if (loopDefinitionPaths.containsKey(coreTransition)) {
          for (final Set<List<EfsmTransition>> loopsGroup :
              loopDefinitionPaths.get(coreTransition)) {
            vectors = traverseLoop(loopsGroup, shortestHeadPath, transition);

            if (!vectors.isEmpty()) {
              break;
            }
          }
        }

        if (vectors.isEmpty() && definitionPaths.containsKey(coreTransition)) {
          for (final List<EfsmTransition> definitionPath : definitionPaths.get(coreTransition)) {
            final List<List<EfsmTransition>> fragmentedDefPath = fragmentPath(definitionPath);
            LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": trying the definition path: ");
            logFragmentedPath(fragmentedDefPath, LogLevel.DEBUG);

            final List<List<EfsmTransition>> corePath = new ArrayList<>();
            corePath.add(shortestHeadPath);
            corePath.addAll(fragmentedDefPath);
            corePath.add(Collections.singletonList(transition));
            vectors = traverseFragmentedPath(corePath);

            if (!vectors.isEmpty()) {
              break;
            }
          }
        }

        if (!vectors.isEmpty()) {
          return vectors;
        }

        Set<EfsmTransition> testedCoreTransitions = this.testedCoreTransitions.get(transition);

        if (testedCoreTransitions == null) {
          testedCoreTransitions = new HashSet<>();
          this.testedCoreTransitions.put(transition, testedCoreTransitions);
        }

        testedCoreTransitions.add(coreTransition);
      }
    }

    return new LinkedList<>();
  }

  private List<Vector> tryDirectPath(final EfsmTransition transition) {
    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": trying the shortest direct path to the target transition");

    List<EfsmTransition> directPath = null;

    for (EfsmTransition predecessor :
        EfsmUtils.getIncoming(this.simulator.getSimulatedModel(), transition.getSrc())) {
      if (this.totalCoverage.isCovered(predecessor)) {
        final List<EfsmTransition> candidatePath =
            new ArrayList<>(findShortestTraversablePath(predecessor));

        if ((directPath == null) || (candidatePath.size() + 1 < directPath.size())) {
          candidatePath.add(predecessor);
          directPath = candidatePath;
        }
      }
    }

    assert directPath != null;
    directPath.add(transition);

    return traversePlainPath(directPath);
  }

  private List<List<EfsmTransition>> fragmentPath(final List<EfsmTransition> plainPath) {
    if (LOGGER.isLoggable(LogLevel.DEBUG)) {
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": fragmenting the path: ");
      logPlainPath(plainPath, LogLevel.DEBUG);
    }

    final List<List<EfsmTransition>> fragmentedPath = new ArrayList<>();
    List<EfsmTransition> currentFragment = new ArrayList<>();
    fragmentedPath.add(currentFragment);
    EfsmState currentState = null;

    for (final EfsmTransition transition : plainPath) {
      if (!currentFragment.isEmpty()) {

        final EfsmState nextState = transition.getSrc();

        assert currentState != null;
        if (!currentState.equals(nextState)) {
          currentFragment = new ArrayList<>();
          fragmentedPath.add(currentFragment);
        }
      }

      currentFragment.add(transition);
      currentState = transition.getDst();
    }

    if (LOGGER.isLoggable(LogLevel.DEBUG)) {
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": fragmented path: ");
      logFragmentedPath(fragmentedPath, LogLevel.DEBUG);
    }

    return fragmentedPath;
  }

  private List<List<EfsmTransition>> smoothPath(final List<List<EfsmTransition>> fragmentedPath) {
    if (LOGGER.isLoggable(LogLevel.DEBUG)) {
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": smoothing the fragmented path: ");
      logFragmentedPath(fragmentedPath, LogLevel.DEBUG);
    }

    final List<List<EfsmTransition>> smoothPath = new ArrayList<>();
    EfsmState currentState = null;

    for (final List<EfsmTransition> fragment : fragmentedPath) {
      if (!fragment.isEmpty()) {
        if (currentState != null) {
          final EfsmState nextState = fragment.get(0).getSrc();

          if (!currentState.equals(nextState)) {
            List<EfsmTransition> bridge;

            if (this.preferredBridges.containsKey(currentState)
                && this.preferredBridges.get(currentState).containsKey(nextState)) {
              bridge = this.preferredBridges.get(currentState).get(nextState);
            } else {
              bridge = EfsmUtils.getShortestPath(
                  this.simulator.getSimulatedModel(),
                  currentState,
                  nextState,
                  TransitionWeightBasis.GUARD_REGISTER_AMOUNT);
            }

            smoothPath.add(bridge);
          }
        }

        smoothPath.add(fragment);
        currentState = fragment.get(fragment.size() - 1).getDst();
      }
    }

    if (LOGGER.isLoggable(LogLevel.DEBUG)) {
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": smooth path: ");
      logFragmentedPath(fragmentedPath, LogLevel.DEBUG);
    }

    return smoothPath;
  }

  private List<EfsmTransition> plainPath(final List<List<EfsmTransition>> fragmentedPath) {
    final List<EfsmTransition> plainPath = new ArrayList<>();

    for (final List<EfsmTransition> fragment: fragmentedPath) {
      plainPath.addAll(fragment);
    }

    return plainPath;
  }

  private List<Vector> traverseLoop(
      final Set<List<EfsmTransition>> plainLoopCores,
      final List<EfsmTransition> headPath,
      final EfsmTransition targetTransition) {

    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": fragmenting the loop cores");

    final Set<List<List<EfsmTransition>>> loopCores = new HashSet<>(plainLoopCores.size());

    for (final List<EfsmTransition> plainLoopCore: plainLoopCores) {
      final LinkedList<EfsmTransition> normalisedLoopCore = new LinkedList<>(plainLoopCore);
      normalisedLoopCore.addLast(normalisedLoopCore.removeFirst());
      loopCores.add(fragmentPath(normalisedLoopCore));
    }

    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": traversing the loop");
    final List<EfsmTransition> currentPath = new ArrayList<>(headPath);
    currentPath.add(plainLoopCores.iterator().next().get(0));

    return traverseMixedLoop(loopCores, fragmentPath(currentPath), targetTransition, 0);
  }

  private List<Vector> traverseMixedLoop(
      final Set<List<List<EfsmTransition>>> loopCores,
      final List<List<EfsmTransition>> currentPath,
      final EfsmTransition targetTransition,
      final int limit) {
    if (limit == this.loopLimit) {
      LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + ": loop limit is reached: " + this.loopLimit);

      return new LinkedList<>();
    }

    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": current path at iteration " + limit + ":");
    logFragmentedPath(currentPath, LogLevel.DEBUG);

    for (final List<List<EfsmTransition>> loopCore : loopCores) {
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": proceeding with the loop:");
      logFragmentedPath(loopCore, LogLevel.DEBUG);

      final List<List<EfsmTransition>> newCurrentPath = new ArrayList<>(currentPath);
      newCurrentPath.addAll(loopCore);

      if (checkFragmentedPathTraversability(newCurrentPath)) {
        LOGGER.log(LogLevel.DEBUG, this.logEntryHeader
            + ": the new loop is traversable. Trying to reach the target transition");

        final List<List<EfsmTransition>> completePath = new ArrayList<>(newCurrentPath);
        completePath.add(Collections.singletonList(targetTransition));

        List<Vector> vectors = traverseFragmentedPath(completePath);

        if (!vectors.isEmpty()) {
          return vectors;
        } else {
          LOGGER.log(LogLevel.DEBUG,
              this.logEntryHeader + ": the target transition is still unreachable");

          vectors = traverseMixedLoop(loopCores, newCurrentPath, targetTransition, limit + 1);

          if (!vectors.isEmpty()) {
            return vectors;
          }
        }
      } else {
        LOGGER.log(LogLevel.DEBUG,
            this.logEntryHeader + ": the new loop is untraversable. Rolling back");
      }
    }

    return new LinkedList<>();
  }

  private boolean checkFragmentedPathTraversability(final List<List<EfsmTransition>> corePath) {
    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": checking the fragmented path for traversability: ");
    logFragmentedPath(corePath, LogLevel.DEBUG);

    for (final List<List<EfsmTransition>> untraversablePath: this.untraversableFragmentedPaths) {
      if ((corePath.size() >= untraversablePath.size())
          && corePath.subList(0, untraversablePath.size()).equals(untraversablePath)) {

        return false;
      }
    }

    final List<List<EfsmTransition>> smoothPath = this.smoothPath(corePath);
    boolean traversability = checkPlainPathTraversability(plainPath(smoothPath));

    if (!traversability) {
      final List<List<EfsmTransition>> currentCorePath = new ArrayList<>();
      final List<List<EfsmTransition>> newSmoothPath = new ArrayList<>();
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": looking for untraversable fragment");

      for (final List<EfsmTransition> fragment : smoothPath) {
        if (corePath.contains(fragment)) {
          currentCorePath.add(fragment);
        }

        newSmoothPath.add(fragment);
        final List<EfsmTransition> plainPath = plainPath(newSmoothPath);
        traversability = checkPlainPathTraversability(plainPath);

        if (!traversability) {
          LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": the untraversable fragment: ");
          logPlainPath(fragment, LogLevel.DEBUG);

          if (corePath.contains(fragment)) {
            LOGGER.log(LogLevel.DEBUG,
                this.logEntryHeader + ": the untraversable fragment is in the core path");

            this.untraversableFragmentedPaths.add(currentCorePath);

            return false;
          } else {
            final EfsmState srcState = fragment.get(0).getSrc();
            final EfsmState dstState = fragment.get(fragment.size() - 1).getDst();
            final Set<EfsmTransition> forbiddenTransitions = new HashSet<>();

            /*for (Iterator<List<EfsmTransition>> i = corePath.listIterator(currentCorePath.size());
                i.hasNext();) {
              forbiddenTransitions.addAll(i.next());
            }*/

            final List<List<EfsmTransition>> pathCandidates =
                findAllPaths(srcState, dstState, forbiddenTransitions);
            boolean success = false;

            for (final List<EfsmTransition> pathCandidate : pathCandidates) {
              if (!pathCandidate.equals(fragment)) {
                LOGGER.log(LogLevel.DEBUG,
                    this.logEntryHeader + ": replacing the untraversable fragment with: ");
                logPlainPath(pathCandidate, LogLevel.DEBUG);

                newSmoothPath.remove(newSmoothPath.size() - 1);
                newSmoothPath.add(pathCandidate);

                if (checkPlainPathTraversability(plainPath(newSmoothPath))) {
                  success = true;

                  LOGGER.log(LogLevel.DEBUG,
                      this.logEntryHeader + ": replacement is successful");

                  Map<EfsmState, List<EfsmTransition>> bridges =
                      this.preferredBridges.get(srcState);

                  if (bridges == null) {
                    bridges = new HashMap<>();
                    this.preferredBridges.put(srcState, bridges);
                  }

                  bridges.put(dstState, pathCandidate);

                  break;
                }
              }
            }

            if (!success) {
              LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": replacement is unsuccessful");

              currentCorePath.add(corePath.get(currentCorePath.size()));
              this.untraversableFragmentedPaths.add(currentCorePath);

              return false;
            }
          }
        }
      }
    }

    return traversability;
  }

  private List<Vector> traverseFragmentedPath(final List<List<EfsmTransition>> corePath) {
    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": traversing the fragmented path: ");
    this.logFragmentedPath(corePath, LogLevel.DEBUG);

    for (final List<List<EfsmTransition>> untraversablePath: this.untraversableFragmentedPaths) {
      if ((corePath.size() >= untraversablePath.size())
          && corePath.subList(0, untraversablePath.size()).equals(untraversablePath)) {

        return new LinkedList<>();
      }
    }

    final List<List<EfsmTransition>> smoothPath = smoothPath(corePath);
    List<Vector> vectors = traversePlainPath(plainPath(smoothPath));

    if (vectors.isEmpty()) {
      final List<List<EfsmTransition>> currentCorePath = new ArrayList<>();
      final List<List<EfsmTransition>> newSmoothPath = new ArrayList<>();
      LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": looking for untraversable fragment");

      for (final List<EfsmTransition> fragment : smoothPath) {
        if (corePath.contains(fragment)) {
          currentCorePath.add(fragment);
        }

        newSmoothPath.add(fragment);
        final List<EfsmTransition> plainPath = plainPath(newSmoothPath);

        boolean success;

        if (currentCorePath.size() == corePath.size()) {
          vectors = traversePlainPath(plainPath);
          success = vectors != null;
        } else {
          success = checkPlainPathTraversability(plainPath);
        }

        if (!success) {
          LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": the untraversable fragment: ");
          logPlainPath(fragment, LogLevel.DEBUG);

          if (corePath.contains(fragment)) {
            LOGGER.log(LogLevel.DEBUG,
                this.logEntryHeader + ": the untraversable fragment is in the core path. "
                    + "Finishing traverse of this core path");

            this.untraversableFragmentedPaths.add(currentCorePath);

            return new LinkedList<>();
          } else {
            final EfsmState srcState = fragment.get(0).getSrc();
            final EfsmState dstState = fragment.get(fragment.size() - 1).getDst();
            final Set<EfsmTransition> forbiddenTransitions = new HashSet<>();

            /*for (Iterator<List<EfsmTransition>> i = corePath.listIterator(currentCorePath.size());
                i.hasNext();) {
              forbiddenTransitions.addAll(i.next());
            }*/

            final List<List<EfsmTransition>> pathCandidates =
                findAllPaths(srcState, dstState, forbiddenTransitions);
            success = false;

            for (final List<EfsmTransition> pathCandidate : pathCandidates) {
              if (!pathCandidate.equals(fragment)) {
                LOGGER.log(LogLevel.DEBUG,
                    this.logEntryHeader + ": replacing the untraversable fragment with: ");
                logPlainPath(pathCandidate, LogLevel.DEBUG);

                newSmoothPath.remove(newSmoothPath.size() - 1);
                newSmoothPath.add(pathCandidate);

                if (checkPlainPathTraversability(plainPath(newSmoothPath))) {
                  success = true;

                  LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": replacement is successful");

                  break;
                }
              }
            }

            if (!success) {
              LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": replacement is unsuccessful");

              currentCorePath.add(corePath.get(currentCorePath.size()));
              this.untraversableFragmentedPaths.add(currentCorePath);

              return new LinkedList<>();
            }
          }
        }
      }
    }

    return vectors;
  }

  private List<Vector> traversePlainPath(final List<EfsmTransition> path) {
    final SolverResult solverResult = resolvePlainPath(path);

    if (solverResult.getStatus().equals(Status.SAT)) {
      LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the path is traversable. Generating input vectors");

      final Map<Integer, Map<String, Data>> inputs =
          Substitutor.deannotateInputs(solverResult.getVariables());

      final List<Vector> vectors = new ArrayList<>(path.size());

      for (final ListIterator<EfsmTransition> i = path.listIterator(); i.hasNext();) {
        final int index = i.nextIndex();
        final EfsmTransition transition = i.next();

        final Map<String, Data> vectorInputs = inputs.get(index);
        final Transaction transaction = new Transaction();

        if (vectorInputs != null) {
          for (final Entry<String, Data> input: vectorInputs.entrySet()) {
            transaction.addVariable(input.getKey(), new NodeValue(input.getValue()));
          }
        }

        final Vector vector = new Vector();
        vector.addTransaction(transaction);
        vector.addEvents(transition.getClocks());
        vectors.add(vector);

        LOGGER.log(LogLevel.DEBUG,
            this.logEntryHeader + ": generated vector: " + vector);

        this.totalCoverage.markTransitionAsCovered(transition);
      }

      this.traversablePaths.add(path);

      return vectors;
    } else {
      LOGGER.log(LogLevel.DEBUG,
          this.logEntryHeader + ": the path is untraversable");

      this.untraversablePlainPaths.add(path);

      return new LinkedList<>();
    }
  }

  private boolean checkPlainPathTraversability(final List<EfsmTransition> path) {
    final boolean traversability = resolvePlainPath(path).getStatus().equals(Status.SAT);

    if (!traversability) {
      this.untraversablePlainPaths.add(path);
    }

    return traversability;
  }

  private SolverResult resolvePlainPath(final List<EfsmTransition> path) {
    for (final List<EfsmTransition> untraversablePath : this.untraversablePlainPaths) {
      if ((path.size() >= untraversablePath.size())
          && path.subList(0, untraversablePath.size()).equals(untraversablePath)) {

        return new SolverResultBuilder(Status.UNSAT).build();
      }
    }

    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": checking the plain path for traversability: ");
    logPlainPath(path, LogLevel.DEBUG);

    if (!this.simulator.getCurrentState().equals(path.get(0).getSrc())) {
      throw new IllegalArgumentException("The path is started not in the current state of the"
          + " simulator");
    }

    final Efsm efsm = this.simulator.getSimulatedModel();
    final Substitutor substitutor = new Substitutor(efsm);
    Node pathConstraint = substitutor.backtrack(path);

    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": the traverse constraint: " + pathConstraint);

    pathConstraint =
        this.simulator.fillExpression(pathConstraint, Collections.<String, NodeValue>emptyMap());

    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": the constraint to resolve: " + pathConstraint);

    return SolverUtils.solve(pathConstraint);
  }

  private List<List<EfsmTransition>> findAllPaths(
      final EfsmState sourceState,
      final EfsmState destinationState,
      final Set<EfsmTransition> forbiddenTransitions) {

    List<List<EfsmTransition>> foundPaths;
    final Efsm efsm = this.simulator.getSimulatedModel();

    if (EfsmUtils.isReachable(efsm, sourceState, destinationState)) {
      foundPaths = new ArrayList<>();

      for (EfsmTransition firstTransition: EfsmUtils.getOutgoing(efsm, sourceState)) {
        foundPaths.addAll(this.buildPaths(
            Collections.singletonList(firstTransition), destinationState,
            forbiddenTransitions));
      }
    } else {
      foundPaths = Collections.emptyList();
    }

    Collections.sort(foundPaths, new Comparator<List<EfsmTransition>>() {
      @Override
      public int compare(List<EfsmTransition> list1, List<EfsmTransition> list2) {
        return list1.size() - list2.size();
      }
    });

    return foundPaths;
  }

  private Set<List<EfsmTransition>> buildPaths(
      final List<EfsmTransition> currentPath,
      final EfsmState dstState,
      final Set<EfsmTransition> forbiddenTransitions) {

    final Set<List<EfsmTransition>> foundPaths = new HashSet<>();
    final EfsmState currentState = currentPath.get(currentPath.size() - 1).getDst();

    if (currentState.equals(dstState)) {
      foundPaths.add(currentPath);
    } else {
      final Efsm efsm = this.simulator.getSimulatedModel();

      for (EfsmTransition nextTransition: EfsmUtils.getOutgoing(efsm, currentState)) {
        if (!currentPath.contains(nextTransition)
            && !forbiddenTransitions.contains(nextTransition)) {
          final List<EfsmTransition> newCurrentPath = new ArrayList<>(currentPath);
          newCurrentPath.add(nextTransition);
          foundPaths.addAll(buildPaths(newCurrentPath, dstState, forbiddenTransitions));
        }
      }
    }

    return foundPaths;
  }

  private void logFragmentedPath(
      final List<List<EfsmTransition>> fragmentedPath,
      final Level logLevel) {
    for (final List<EfsmTransition> fragment : fragmentedPath) {
      LOGGER.log(logLevel, this.logEntryHeader + ": path fragment: ");
      logPlainPath(fragment, logLevel);
    }
  }

  private void logPlainPath(
      final List<EfsmTransition> plainPath,
      final Level logLevel) {
    for (final EfsmTransition transition : plainPath) {
      LOGGER.log(logLevel, this.logEntryHeader + ": " + transition.toString());
    }
  }

  private List<EfsmTransition> findShortestTraversablePath(final EfsmTransition transition) {
    List<EfsmTransition> shortestPath = null;

    for (final List<EfsmTransition> traversablePath : this.traversablePaths) {
      final int index = traversablePath.indexOf(transition);

      if (index != -1) {
        if ((shortestPath == null) || (index + 1 < shortestPath.size())) {
          shortestPath = traversablePath.subList(0, index);
        }
      }
    }

    return shortestPath;
  }

  @Override
  public void reset() {
    // TODO Auto-generated method stub
  }
}
