/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.EdgeType;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;
import ru.ispras.retrascope.util.LogMessage;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Set;
import java.util.logging.Logger;

/**
 * The class which performs the learning phase of the FATE algorithm.
 *
 * @author <a href="mailto:myllyenko@ya.ru">Igor Melnichenko</a>
 */
final class EfsmAnalyser {
  private static final Logger LOGGER = Log.getLogger(EfsmAnalyser.class);
  private final String logEntryHeader;

  private final Efsm efsm;

  private final Graph<EfsmTransition, VariableDependency> controlDependencyGraph;
  private final Graph<EfsmTransition, VariableDependency> dataDependencyGraph;
  private final Map<NodeVariable, Set<List<EfsmTransition>>> counters;
  private final Map<EfsmTransition, Set<List<EfsmTransition>>> definitionPaths;
  private final Map<EfsmTransition, Set<Set<List<EfsmTransition>>>> loopDefinitionPaths;

  EfsmAnalyser(final Efsm efsm) {
    this.logEntryHeader = "EFSM.FATE.Analyser." + efsm.getName();

    this.efsm = efsm;

    this.controlDependencyGraph = detectControlDependencies(efsm);
    this.dataDependencyGraph = detectDataDependencies(efsm);
    this.counters = new HashMap<>();
    this.definitionPaths = new HashMap<>();
    this.loopDefinitionPaths = new HashMap<>();
    this.detectCounters(efsm);
  }

  private Graph<EfsmTransition, VariableDependency> detectControlDependencies(final Efsm efsm) {

    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting control dependencies");

    final Graph<EfsmTransition, VariableDependency> ctrlDepGraph = new DirectedSparseMultigraph<>();

    for (final EfsmTransition transition : efsm.getTransitions()) {
      if (!ctrlDepGraph.containsVertex(transition)) {
        ctrlDepGraph.addVertex(transition);
      }

      for (final NodeVariable usedVariable : transition.getGuardUses()) {
        final String usedVariableName = usedVariable.getName();

        final Object varData = usedVariable.getUserData();

        if (varData instanceof VariableDescriptor) {
          if (((VariableDescriptor)varData).getVariableType() == VariableType.REG) {
            for (final EfsmTransition definingTransition :
                EfsmAnalyser.findVariableDefinition(efsm, transition, usedVariableName)) {
              if (!ctrlDepGraph.containsVertex(definingTransition)) {
                ctrlDepGraph.addVertex(definingTransition);
              }

              final VariableDependency edge =
                  new VariableDependency(definingTransition, transition, usedVariable);
              ctrlDepGraph.addEdge(edge, definingTransition, transition,
                  EdgeType.DIRECTED);
            }
          }
        } else {
          throw RetrascopeRuntimeException.makeException(
              LogMessage.UNSUPPORTED_TYPE,
              varData.getClass());
        }
      }
    }

    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": control dependencies have been detected successfully");

    return ctrlDepGraph;
  }

  private Graph<EfsmTransition, VariableDependency> detectDataDependencies(final Efsm efsm) {
    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting data dependencies");

    final Graph<EfsmTransition, VariableDependency> dataDepGraph = new DirectedSparseMultigraph<>();

    for (final EfsmTransition transition : efsm.getTransitions()) {
      if (!dataDepGraph.containsVertex(transition)) {
        dataDepGraph.addVertex(transition);
      }

      for (final NodeVariable usedVariable : transition.getActionUses()) {
        final String usedVariableName = usedVariable.getName();

        final Object varData = usedVariable.getUserData();

        if (varData instanceof VariableDescriptor) {
          if (((VariableDescriptor)varData).getVariableType().equals(VariableType.REG)) {
            for (final EfsmTransition definingTransition :
                EfsmAnalyser.findVariableDefinition(efsm, transition, usedVariableName)) {
              if (!dataDepGraph.containsVertex(definingTransition)) {
                dataDepGraph.addVertex(definingTransition);
              }

              final VariableDependency edge =
                  new VariableDependency(definingTransition, transition, usedVariable);
              dataDepGraph.addEdge(edge, definingTransition, transition, EdgeType.DIRECTED);
            }
          }
        } else {
          throw RetrascopeRuntimeException.makeException(
              LogMessage.UNSUPPORTED_TYPE,
              varData.getClass());
        }
      }
    }

    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": data dependencies have been detected successfully");

    return dataDepGraph;
  }

  private static List<EfsmTransition> findVariableDefinition(
      final Efsm efsm,
      final EfsmTransition transition,
      final String usedVariableName) {

    final Set<EfsmTransition> visitedTransitions = new HashSet<>();
    final Queue<EfsmTransition> queue = new LinkedList<>();
    queue.add(transition);
    final List<EfsmTransition> definingTransitions = new ArrayList<>();

    while (!queue.isEmpty()) {
      final EfsmTransition currentTransition = queue.poll();

      for (EfsmTransition previousTransition :
          EfsmUtils.getIncoming(efsm, currentTransition.getSrc())) {

        if (!visitedTransitions.contains(previousTransition)) {
          boolean useDefMatching = false;

          for (final NodeVariable definedVariable : previousTransition.getDefines()) {
            if (definedVariable.getName().equals(usedVariableName)) {
              useDefMatching = true;

              break;
            }
          }

          if (useDefMatching) {
            definingTransitions.add(previousTransition);
          } else {
            queue.add(previousTransition);
            visitedTransitions.add(previousTransition);
          }
        }
      }
    }

    return definingTransitions;
  }

  private void detectCounters(final Efsm efsm) {
    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting counters");

    final Map<NodeVariable, Set<EfsmTransition>> processedCandidates = new HashMap<>();

    for (final VariableDependency dependency : this.controlDependencyGraph.getEdges()) {
      final EfsmTransition definingTransition = this.controlDependencyGraph.getSource(dependency);

      if (processedCandidates.containsKey(dependency.getVariable())) {
        final Set<EfsmTransition> processedTransitions =
            processedCandidates.get(dependency.getVariable());

        if (processedTransitions.contains(definingTransition)) {
          continue;
        } else {
          processedTransitions.add(definingTransition);
        }
      } else {
        final Set<EfsmTransition> transitions = new HashSet<>();
        transitions.add(definingTransition);
        processedCandidates.put(dependency.getVariable(), transitions);
      }

      final EfsmState srcState = definingTransition.getSrc();
      final EfsmState dstState =
          this.controlDependencyGraph.getDest(dependency).getDst();
      final Set<NodeVariable> defVars =
          EfsmUtils.getResultingDependencies(definingTransition.getAction())
              .get(dependency.getVariable());

      for (final Iterator<NodeVariable> iterator = defVars.iterator(); iterator.hasNext();) {
        if (efsm.declaresInput(iterator.next().getName())) {
          iterator.remove();
        }
      }

      if (!defVars.isEmpty() && EfsmUtils.isReachable(efsm, dstState, srcState)) {
        final Set<NodeVariable> variables = new HashSet<>(1);
        variables.add(dependency.getVariable());

        for (final VariableDependency dataDependency :
            this.dataDependencyGraph.getOutEdges(definingTransition)) {
          if (dataDependency.getVariable().equals(dependency.getVariable())) {
            final EfsmTransition transition = this.dataDependencyGraph.getDest(dataDependency);
            findCounterLoop(dataDependency.getVariable(), transition,
                defVars, variables, Collections.singletonList(definingTransition));
          }
        }
      }
    }

    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": counters have been detected successfully");
  }

  //target transition is the first in the returned list
  private void findCounterLoop(
      final NodeVariable potentialCounter,
      final EfsmTransition currentTransition,
      final Set<NodeVariable> targetVariables,
      final Set<NodeVariable> currentVariables,
      final List<EfsmTransition> currentPath) {
    if (currentPath.contains(currentTransition)) {
      if (currentTransition.equals(currentPath.get(0))) {
        final Set<NodeVariable> intersection = new HashSet<>(targetVariables);
        intersection.retainAll(currentVariables);

        if (!intersection.isEmpty()) {
          Set<List<EfsmTransition>> counterData =
              this.counters.get(potentialCounter);

          if (counterData == null) {
            counterData = new HashSet<>(1);
            this.counters.put(potentialCounter, counterData);
          }

          counterData.add(currentPath);

          LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": the variable "
              + potentialCounter.getName() + " defined by the transition " + currentPath.get(0)
              + " is a counter in the loop " + currentPath);
        }
      }
    } else {
      final Set<NodeVariable> newCurrentVariables = new HashSet<>();

      for (final Entry<NodeVariable, Set<NodeVariable>> dependency :
          EfsmUtils.getResultingDependencies(currentTransition.getAction())
              .entrySet()) {
        final NodeVariable variable = dependency.getKey();
        final Set<NodeVariable> intersection = new HashSet<>(currentVariables);
        intersection.retainAll(dependency.getValue());

        if (!intersection.isEmpty()) {
          newCurrentVariables.add(variable);
        }
      }

      if (!newCurrentVariables.isEmpty()) {
        final List<EfsmTransition> newCurrentPath = new ArrayList<>(currentPath);
        newCurrentPath.add(currentTransition);

        for (final VariableDependency dataDependency:
            this.dataDependencyGraph.getOutEdges(currentTransition)) {
          if (newCurrentVariables.contains(dataDependency.getVariable())) {
            final EfsmTransition transition = this.dataDependencyGraph.getDest(dataDependency);
            this.findCounterLoop(potentialCounter, transition, targetVariables, newCurrentVariables,
                newCurrentPath);
          }
        }
      }
    }
  }

  public Set<List<EfsmTransition>> getDefinitionPaths(final EfsmTransition transition) {
    Set<List<EfsmTransition>> cachedPaths = this.definitionPaths.get(transition);

    if (cachedPaths == null) {
      cachedPaths = new HashSet<>();
      this.definitionPaths.put(transition, cachedPaths);
      final Map<EfsmTransition, Set<NodeVariable>> dependencies = new HashMap<>();

      for (final VariableDependency dependency :
          this.controlDependencyGraph.getInEdges(transition)) {

        final NodeVariable dependencyVariable = dependency.getVariable();
        final EfsmTransition definingTransition = this.controlDependencyGraph.getSource(dependency);
        final Set<NodeVariable> variables =
            EfsmUtils.getResultingDependencies(definingTransition.getAction())
                .get(dependencyVariable);

        Set<NodeVariable> storedVariables = dependencies.get(definingTransition);

        if (storedVariables == null) {
          storedVariables = new HashSet<>();
          dependencies.put(definingTransition, storedVariables);
        }

        storedVariables.addAll(variables);
      }

      for (final Entry<EfsmTransition, Set<NodeVariable>> dependency: dependencies.entrySet()) {
        cachedPaths.addAll(this.findDefinitionPaths(Collections.singletonList(dependency.getKey()),
            dependency.getValue()));
      }
    }

    return cachedPaths;
  }

  public Set<Set<List<EfsmTransition>>> getLoopDefinitionPaths(final EfsmTransition transition) {
    Set<Set<List<EfsmTransition>>> cachedPaths = this.loopDefinitionPaths.get(transition);

    if (cachedPaths == null) {
      cachedPaths = new HashSet<>();
      this.loopDefinitionPaths.put(transition, cachedPaths);

      for (final VariableDependency dependency :
          this.controlDependencyGraph.getInEdges(transition)) {
        final NodeVariable dependencyVariable = dependency.getVariable();
        final EfsmTransition definingTransition = this.controlDependencyGraph.getSource(dependency);
        final Set<List<EfsmTransition>> paths = new HashSet<>();

        if (this.counters.containsKey(dependencyVariable)) {
          for (final List<EfsmTransition> path : this.counters.get(dependencyVariable)) {
            if (path.get(0).equals(definingTransition)) {
              paths.add(path);
            }
          }

          if (!paths.isEmpty()) {
            cachedPaths.add(paths);
          }
        }
      }
    }

    return cachedPaths;
  }

  private Set<List<EfsmTransition>> findDefinitionPaths(
      final List<EfsmTransition> currentPath,
      final Set<NodeVariable> currentVariables) {

    final Set<List<EfsmTransition>> foundPaths = new HashSet<>();
    final Set<NodeVariable> innerDependencies = new HashSet<>();

    for (final NodeVariable variable : currentVariables) {
      if (this.efsm.declaresRegister(variable.getName())
          || this.efsm.getParentContainer().declaresRegister(variable.getName())) {
        innerDependencies.add(variable);
      }
    }

    if (innerDependencies.isEmpty()) {
      foundPaths.add(currentPath);
    } else {
      if (innerDependencies.size() < currentVariables.size()) {
        foundPaths.add(currentPath);
      }

      final EfsmTransition currentTransition = currentPath.get(0);
      final Map<EfsmTransition, Map<NodeVariable, Set<NodeVariable>>> deps = new HashMap<>();

      for (final VariableDependency dependency :
          this.dataDependencyGraph.getInEdges(currentTransition)) {

        final NodeVariable dependencyVariable = dependency.getVariable();

        if (innerDependencies.contains(dependencyVariable)) {
          final EfsmTransition definingTransition = this.dataDependencyGraph.getSource(dependency);
          final Set<NodeVariable> variables =
              EfsmUtils.getResultingDependencies(definingTransition.getAction())
                  .get(dependencyVariable);

          Map<NodeVariable, Set<NodeVariable>> storedVariables =
              deps.get(definingTransition);

          if (storedVariables == null) {
            storedVariables = new HashMap<>();
            deps.put(definingTransition, storedVariables);
          }

          storedVariables.put(dependencyVariable, variables);
        }
      }

      for (final Entry<EfsmTransition, Map<NodeVariable,Set<NodeVariable>>> dependency
          : deps.entrySet()) {

        if (currentPath.contains(dependency.getKey())) {
          final int limit = dependency.getKey().getAction().getAssignments().size();
          int occurrences = 0;

          for (final EfsmTransition transition : currentPath) {
            if (transition.equals(dependency.getKey())) {
              occurrences++;
            }
          }

          if (occurrences >= limit) {
            continue;
          }
        }

        final List<EfsmTransition> newPath = new ArrayList<>(currentPath);
        final Set<NodeVariable> variables = new HashSet<>(innerDependencies);
        variables.removeAll(dependency.getValue().keySet());

        for (final Set<NodeVariable> newVariables : dependency.getValue().values()) {
          variables.addAll(newVariables);
        }

        newPath.add(0, dependency.getKey());
        foundPaths.addAll(this.findDefinitionPaths(newPath, variables));
      }
    }

    return foundPaths;
  }
}
