/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.retrascope.engine.basis.TestGenerator;
import ru.ispras.retrascope.engine.efsm.simulator.ModuleSimulator;
import ru.ispras.retrascope.engine.efsm.simulator.ProcessSimulator;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestCoverage;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The test generator for EFSM-based hardware models.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
final class Generator extends TestGenerator {

  private final String logEntryHeader;

  private final ModuleSimulator simulator;

  // TODO: this field is assigned but never accessed
  private final int loopLimit;

  private final EfsmTransitionCoverage totalCoverage;

  private final Map<Efsm, Set<List<EfsmTransition>>> traversedPaths;

  // TODO: this field is assigned but never accessed
  private final boolean singleSequenceMode;

  /**
   * Constructs a generator for the specified EFSM module.
   *
   * @param module the EFSM module to be tested
   * @param loopLimit a loop iteration limit
   *
   * @throws IllegalArgumentException either if {@code module} is {@code null} or if
   *     {@code loopLimit} is lesser than zero
   */
  public Generator(EfsmModule module, int loopLimit) {
    this.simulator = new ModuleSimulator(module);

    if (loopLimit < 0) {
      throw new IllegalArgumentException("loopLimit is lesser than zero");
    }

    this.logEntryHeader = "EFSM.TestGenerator." + module.getName() + ": ";

    this.loopLimit = loopLimit;

    this.totalCoverage = new EfsmTransitionCoverage(this.simulator.getCoverage());

    this.traversedPaths = new HashMap<>();

    boolean singleSequenceModeBuffer = false;

    for (Efsm efsm: module.getNonInitialEfsms()) {

      if (efsm.singleTestable()) {
        this.traversedPaths.put(efsm, new HashSet<List<EfsmTransition>>());

        if (efsm.getResetGuardedActions().isEmpty()) {
          singleSequenceModeBuffer = true;
        }
      } else {
        Log.printWarn("The efsm can't be tested as a single one: " + efsm.getName());
      }
    }

    this.singleSequenceMode = singleSequenceModeBuffer;
  }

  //TODO process initial EFSMs
  @Override
  public Test generateTest() {
    Test test = new Test();
    List<Efsm> efsms = new ArrayList<>(this.simulator.getSimulatedModel().getNonInitialEfsms());

    if (!efsms.isEmpty()) {
      //sequences.putAll(this.generateDirectly());

      Log.printDebug(this.logEntryHeader
          + "generation has been finished successfully. Total coverage: " + this.totalCoverage);
      Log.printDebug(this.logEntryHeader
          + "total amount of sequences in the generated test before minimisation: "
          + test.getSequencesAmount());
      Log.printDebug(this.logEntryHeader
          + "total amount of vectors in the generated test before minimisation: "
          + test.getVectorsAmount());

      Map<Sequence, EfsmTransitionCoverage> sequences = this.generateRandomly();
      test = TestMinimiser.minimise(sequences);

      Log.printDebug(this.logEntryHeader
          + "total amount of sequences in the generated test after minimisation: "
          + test.getSequencesAmount());
      Log.printDebug(this.logEntryHeader
          + "total amount of vectors in the generated test after minimisation: "
          + test.getVectorsAmount());

      for (EfsmTransition transition: this.totalCoverage.getUncoveredTransitions()) {
        Log.printDebug(this.logEntryHeader + "uncovered transition: " + transition);
      }
    }

    return test;
  }

  private Map<Sequence, EfsmTransitionCoverage> generateRandomly() {
    Map<Sequence, EfsmTransitionCoverage> sequences = new HashMap<>();

    Set<Efsm> efsms = this.simulator.getSimulatedModel().getNonInitialEfsms();

    Map<Efsm, Iterator<Vector>> vectorIterators = new HashMap<>(efsms.size());

    final int noProgressLimit = this.getNoProgressLimit();
    final List<RandomGenerator> generators = new LinkedList<>();

    for (final Efsm efsm: efsms) {
      if (efsm.singleTestable()) {
        generators.add(new RandomGenerator(this.simulator.getSubSimulator(efsm)));
      }
    }

    int sequenceNoProgressCounter = 0;

    Comparator<RandomGenerator> comparator = new Comparator<RandomGenerator>() {
      @Override
      public int compare(RandomGenerator generator1, RandomGenerator generator2) {
        BigDecimal coverage1 =
            simulator.getSubSimulator(generator1.getEfsm()).getCoverage().getPercentage();
        BigDecimal coverage2 =
            simulator.getSubSimulator(generator2.getEfsm()).getCoverage().getPercentage();

        return coverage2.compareTo(coverage1);
      }
    };

    while (!this.totalCoverage.isFull() && (sequenceNoProgressCounter <= noProgressLimit)) {
      Log.printDebug(this.logEntryHeader + "starting a new sequence");

      this.simulator.reset();

      Vector resetVector = this.resetModuleSimulator();

      if (resetVector == null) {
        break;
      }

      this.simulator.simulateVector(resetVector);

      for (RandomGenerator generator: generators) {
        vectorIterators.put(generator.getEfsm(), generator.getNextSequenceIterator());
      }

      Sequence sequence = new Sequence();
      sequence.addVector(resetVector);
      Map<Efsm, List<EfsmTransition>> currentPaths = new HashMap<>(efsms.size());

      for (Efsm efsm : efsms) {
        currentPaths.put(efsm, new ArrayList<EfsmTransition>());
      }

      int vectorNoProgressCounter = 0;
      List<Vector> bufferedVectors = new ArrayList<>();
      EfsmTransitionCoverage sequenceCoverage = this.simulator.getCoverage();

      while (vectorNoProgressCounter <= noProgressLimit) {
        Collections.sort(generators, comparator);
        Vector currentVector = new Vector();
        boolean success = false;

        for (RandomGenerator generator: generators) {
          Iterator<Vector> sequenceIterator = vectorIterators.get(generator.getEfsm());
          generator.fixInputs(currentVector);
          Vector generatedVector = sequenceIterator.next();

          if (generatedVector != null) {
            success = true;
            currentVector = generatedVector;
          }
        }

        if (!success) {
          break;
        }

        TestGeneratorUtils.randomiseUnusedInputs(this.simulator.getSimulatedModel(),
            currentVector);

        this.simulator.simulateVector(currentVector);

        for (final Efsm efsm : efsms) {
          if (efsm.singleTestable()) {
            final ProcessSimulator subSimulator = this.simulator.getSubSimulator(efsm);
            final EfsmTransition lastTraversed = subSimulator.getLastTraversedTransition();
            final List<EfsmTransition> currentPath = currentPaths.get(efsm);
            currentPath.add(lastTraversed);
          }
        }

        final BigDecimal oldCoveragePercentage = sequenceCoverage.getPercentage();
        final EfsmTransitionCoverage bufferedSequenceCoverage = this.simulator.getCoverage();
        final BigDecimal newCoveragePercentage = bufferedSequenceCoverage.getPercentage();

        if (newCoveragePercentage.compareTo(oldCoveragePercentage) > 0) {
          Log.printDebug(this.logEntryHeader
              + ": the new vector has increased current sequence coverage from "
              + oldCoveragePercentage + " to " + newCoveragePercentage + "%");

          if (!bufferedVectors.isEmpty()) {
            sequence.addVectors(bufferedVectors);
            bufferedVectors.clear();
            vectorNoProgressCounter = 0;
          }

          sequence.addVector(currentVector);
          sequenceCoverage = bufferedSequenceCoverage;

          EfsmTransitionCoverage testedTotalCoverage =
              new EfsmTransitionCoverage(this.totalCoverage);
          testedTotalCoverage.include(sequenceCoverage);

          if (testedTotalCoverage.isFull()) {
            Log.printDebug(this.logEntryHeader + ": total coverage is full now");

            this.totalCoverage.include(sequenceCoverage);

            break;
          }
        } else {
          vectorNoProgressCounter++;
          bufferedVectors.add(currentVector);

          Log.printDebug(this.logEntryHeader
              + ": there is no progress in sequence coverage. Buffering the last vector. "
              + "No progress counter: " + vectorNoProgressCounter
              + ". No progress limit: " + noProgressLimit);
        }
      }

      BigDecimal oldCoveragePercentage = this.totalCoverage.getPercentage();
      EfsmTransitionCoverage bufferedTotalCoverage = new EfsmTransitionCoverage(this.totalCoverage);
      bufferedTotalCoverage.include(sequenceCoverage);
      BigDecimal newCoveragePercentage = bufferedTotalCoverage.getPercentage();

      if (newCoveragePercentage.compareTo(oldCoveragePercentage) > 0) {
        Log.printDebug(this.logEntryHeader
            + ": the new sequence has increased total coverage from "
            + oldCoveragePercentage + " to " + newCoveragePercentage + "%");

        this.totalCoverage.include(bufferedTotalCoverage);

        for (final Efsm efsm : efsms) {

          if (efsm.singleTestable()) {

            final List<EfsmTransition> path = currentPaths.get(efsm);

            for (int i = 1; i <= bufferedVectors.size(); i++) {
              path.remove(path.size() - 1);
            }

            this.traversedPaths.get(efsm).add(path);
          }
        }

        sequences.put(sequence, new EfsmTransitionCoverage(sequenceCoverage));
      } else {
        sequenceNoProgressCounter++;

        Log.printDebug(this.logEntryHeader
            + ": there is no progress in total coverage. Skipping the last sequence. "
            + "No progress counter: " + sequenceNoProgressCounter
            + ". No progress limit: " + noProgressLimit);
      }
    }

    Log.printDebug(this.logEntryHeader
        + ": the random phase of generation has been finished successfully. Total coverage: "
        + this.totalCoverage);

    return sequences;
  }


//TODO
  /*private Map<Sequence, EfsmTransitionCoverage> generateDirectly() {
    if (!this.totalCoverage.isFull()) {
      Set<Efsm> efsms = this.simulator.getSimulatedModel().getNonInitialEfsms();

      if (efsms.size() == 1) {
        Generator.LOGGER.log(LogLevel.DEBUG,
            this.logEntryHeader + ": starting the directed phase of generation");

        for (RandomGenerator randomGenerator: generators) {
          Efsm efsm = randomGenerator.getEfsm();

          if (randomGenerator.getTotalCoverage().isFull()) {
            Generator.LOGGER.log(LogLevel.DEBUG,
                this.logEntryHeader + ": the " + efsm + " EFSM is fully covered. Skipping it");
          } else {
            Generator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": the " + efsm +
                " EFSM has uncovered transitions. Processing it");

            DirectedGenerator directedGenerator =
                new DirectedGenerator(efsm, traversedPaths.get(efsm), loopLimit);

            for (Iterator<Vector> vectorIterator = directedGenerator.getNextSequenceIterator();
                vectorIterator != null;
                vectorIterator = directedGenerator.getNextSequenceIterator()) {
              Sequence sequence = new Sequence();

              while (vectorIterator.hasNext()) {
                Vector vector = vectorIterator.next();

                TestGeneratorUtils.randomiseUnusedInputs(this.simulator.getSimulatedModel(),
                    vector);

                sequence.addVector(vector);
              }

              test.addSequence(sequence);

              EfsmTransitionCoverage sequenceCoverage = directedGenerator.getSequenceCoverage();
              totalCoverage.include(sequenceCoverage);
              sequencesCoverage.put(sequence, sequenceCoverage);

              Generator.LOGGER.log(LogLevel.DEBUG,
                  this.logEntryHeader + ": the directed sequence has been generated successfully");
              Generator.LOGGER.log(LogLevel.DEBUG,
                  this.logEntryHeader + ": current coverage: " + totalCoverage);
            }
          }
        }

        Generator.LOGGER.log(LogLevel.DEBUG, this.logEntryHeader +
                ": the directed phase of generation has been finished successfully");
      } else {
        Generator.LOGGER.log(LogLevel.WARNING, this.logEntryHeader +
            ": directed generation for multi-EFSM systems isn't supported yet");

        return Collections.emptyMap();
      }
    }
  }*/

  private Vector resetModuleSimulator() {
    Log.printDebug(this.logEntryHeader + ": resetting the module");
    Vector vector = new Vector();

    for (final ProcessSimulator processSimulator : this.simulator.getSubSimulators()) {
      final Collection<GuardedAction> rstActions =
          processSimulator.getSimulatedModel().getResetGuardedActions();

      for (final Iterator<GuardedAction> iterator = rstActions.iterator(); iterator.hasNext();) {

        final GuardedAction next = iterator.next();
        final EventList eventList = vector.getEventList();
        final Map<String, NodeValue> assignments = vector.getAssignments();

        final Vector candidateVector =
            processSimulator.enableGuardedAction(next, eventList, assignments);

        if (candidateVector != null) {
          vector = candidateVector;

          break;
        } else {
          if (!iterator.hasNext()) {
            Log.printWarn(this.logEntryHeader
                + ": failed to reset the process: " + processSimulator.getName());
          }
        }
      }
    }

    Log.printDebug(this.logEntryHeader + ": the module has been reset");

    TestGeneratorUtils.randomiseUnusedInputs(this.simulator.getSimulatedModel(), vector);

    return vector;
  }

  private int getNoProgressLimit() {
    int transitionsAmount = 0;
    int statesAmount = 0;

    for (Efsm efsm : this.simulator.getSimulatedModel().getNonInitialEfsms()) {
      transitionsAmount = transitionsAmount + efsm.getTransitionNum();
      statesAmount = statesAmount + efsm.getStateNum();
    }

    return transitionsAmount / statesAmount;
  }

  @Override
  public TestCoverage getCoverage() {
    return this.totalCoverage;
  }
}
