/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 *
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class Substitutor {
  private static final String RENAMING_SUFFIX_START = "<";
  private static final String RENAMING_SUFFIX_END = ">";

  private static final Pattern NAME_PATTERN = Pattern.compile(".*(?="
      + Substitutor.RENAMING_SUFFIX_START + ".*"
      + Substitutor.RENAMING_SUFFIX_END + ")");
  private static final Pattern SUFFIX_PATTERN = Pattern.compile("(?<="
      + Substitutor.RENAMING_SUFFIX_START + ").*(?="
      + Substitutor.RENAMING_SUFFIX_END + ")");

  private final Efsm efsm;
  private final Map<String, NodeVariable> annotatedVariables;

  /**
   * Constructs a substitutor for the specified EFSM.
   *
   * @param efsm the EFSM
   */
  public Substitutor(final Efsm efsm) {
    this.efsm = efsm;
    this.annotatedVariables = new HashMap<>();
  }

  /**
   * Analyses cycle number-annotated variables and returns mappings between cycle numbers and
   * corresponding variables' values.
   *
   * @param variables the cycle number-annotated variables
   *
   * @return a map whose key is a cycle number and values is a mappings between variables' names and
   *     values
   */
  public static Map<Integer, Map<String, Data>> deannotateInputs(
      final Iterable<Variable> variables) {

    final Map<Integer, Map<String, Data>> mappedVariables = new HashMap<>();

    for (final Variable variable : variables) {

      final String annotatedName = variable.getName();
      final int number = Substitutor.getCycleNumberForAnnotatedInput(annotatedName);

      if (number != -1) {

        final String name = Substitutor.deannotateName(annotatedName);

        if (!mappedVariables.containsKey(number)) {
          mappedVariables.put(number, new HashMap<String, Data>());
        }

        mappedVariables.get(number).put(name, variable.getData());
      }
    }

    return mappedVariables;
  }

  /**
   * Extracts a cycle number from the specified cycle number-annotated name.
   *
   * @param name the cycle number-annotated name
   *
   * @return the cycle number
   */
  private static int getCycleNumberForAnnotatedInput(final String name) {
    final Matcher suffixMatcher = Substitutor.SUFFIX_PATTERN.matcher(name);

    return suffixMatcher.find() ? Integer.parseInt(suffixMatcher.group()) : -1;
  }

  /**
   * Deannotates the specified cycle number-annotated name.
   *
   * @param name the cycle number-annotated name
   *
   * @return a deannotated name
   */
  public static String deannotateName(final String name) {
    final Matcher nameMatcher = Substitutor.NAME_PATTERN.matcher(name);

    return nameMatcher.find() ? nameMatcher.group() : name;
  }

  /**
   * Backtracks the specified path performing symbolic substitutions and returns an expression with
   * cycle number-annotated variables that must be satisfied in order to traverse the path.
   *
   * @return the expression with that must be satisfied in order to traverse the path
   */
  public Node backtrack(final List<EfsmTransition> path) {
    Node substitutedExpression = null;

    for (final ListIterator<EfsmTransition> i = path.listIterator(path.size()); i.hasPrevious();) {
      final int index = i.previousIndex();
      substitutedExpression = simulateAction(substitutedExpression, i.previous(), index);
    }

    return substitutedExpression;
  }

  private Node simulateAction(
      Node targetExpression,
      final EfsmTransition transition,
      final int number) {
    final Action action = transition.getAction();

    if (targetExpression == null) {
      targetExpression = transition.getGuardNode();
    } else {
      final Map<NodeVariable, Node> substitutions = EfsmUtils.getResultingSubstitutions(action);
      final List<BoundVariable> bindingList = new ArrayList<>(substitutions.size());

      for (final Map.Entry<NodeVariable, Node> entry : substitutions.entrySet()) {
        bindingList.add(NodeBinding.bindVariable(entry.getKey(), entry.getValue()));
      }

      targetExpression =
          Transformer.substituteAllBindings(new NodeBinding(targetExpression, bindingList));

      targetExpression = Node.AND(targetExpression, transition.getGuardNode());
    }

    if (targetExpression.getKind() == Node.Kind.OPERATION) {
      targetExpression = annotateInputsWithCycleNumbers((NodeOperation) targetExpression, number);
    } else {
      if (targetExpression.getKind() == Node.Kind.VARIABLE) {
        targetExpression = annotateInputWithCycleNumber((NodeVariable) targetExpression, number);
      }
    }

    return targetExpression;
  }

  private NodeOperation annotateInputsWithCycleNumbers(
      final NodeOperation expression,
      final int number) {

    NodeOperation annotatedExpression = expression;

    for (final NodeVariable variable : ExprUtils.getVariables(expression)) {
      final String originalName = variable.getName();
      final NodeVariable annotatedVariable = annotateInputWithCycleNumber(variable, number);

      if (!annotatedVariable.equals(variable)) {
        annotatedExpression =
            (NodeOperation) Transformer.substitute(annotatedExpression, originalName,
                annotatedVariable);
      }
    }

    return annotatedExpression;
  }

  private NodeVariable annotateInputWithCycleNumber(
      final NodeVariable variable,
      final int number) {

    final String originalName = variable.getName();

    if (this.efsm.declaresInput(originalName)
        || this.efsm.getParentContainer().declaresInput(originalName)) {
      final String newName = originalName + Substitutor.RENAMING_SUFFIX_START
          + number + Substitutor.RENAMING_SUFFIX_END;

      if (this.annotatedVariables.containsKey(newName)) {

        return this.annotatedVariables.get(newName);
      } else {
        final NodeVariable annotatedVariable =
            new NodeVariable(new Variable(newName, variable.getData()));
        this.annotatedVariables.put(newName, annotatedVariable);

        return annotatedVariable;
      }
    } else {
      return variable;
    }
  }
}
