/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test;

import ru.ispras.retrascope.engine.efsm.simulator.EfsmSimulatorUtils;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestModel;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 * The class which performs test minimisation.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class TestMinimiser {

  private TestMinimiser() {
    /* Empty. */
  }

  /**
   * Minimises the specified set of sequences by excluding redundant (in terms of coverage)
   * sequences.
   *
   * @param sequencesCoverage a mappings between source sequences
   *                          and sets of transitions covered by them
   */
  static Test minimise(final Map<Sequence, EfsmTransitionCoverage> sequencesCoverage) {

    final Test test = new Test();
    final Map<Sequence, Set<EfsmTransition>> covered =
        new LinkedHashMap<>(sequencesCoverage.size());
    final Set<EfsmTransition> targets = new LinkedHashSet<>();
    final Map<EfsmTransition, Sequence> unique = new LinkedHashMap<>();

    for (final Entry<Sequence, EfsmTransitionCoverage> entry : sequencesCoverage.entrySet()) {

      final Sequence sequence = entry.getKey();
      final Set<EfsmTransition> transitions = entry.getValue().getCoveredTransitions();

      if (covered.containsKey(sequence)) {
        covered.get(sequence).addAll(transitions);
      } else {
        covered.put(sequence, transitions);
      }

      for (final EfsmTransition transition : transitions) {

        if (unique.containsKey(transition)) {
          unique.remove(transition);
        } else {
          if (!targets.contains(transition)) {
            unique.put(transition, sequence);
          }
        }
      }

      targets.addAll(transitions);
    }

    for (final Sequence sequence : new LinkedHashSet<>(unique.values())) {

      test.addSequence(sequence);
      targets.removeAll(covered.remove(sequence));
    }

    while (!targets.isEmpty()) {

      int maxCoverage = Integer.MIN_VALUE;
      Sequence currentSequence = null;

      for (Iterator<Entry<Sequence, Set<EfsmTransition>>> i =
          covered.entrySet().iterator(); i.hasNext();) {

        Entry<Sequence, Set<EfsmTransition>> entry = i.next();
        Set<EfsmTransition> effectiveCoverage = new LinkedHashSet<>(entry.getValue());
        effectiveCoverage.retainAll(targets);

        if (effectiveCoverage.isEmpty()) {
          i.remove();
        } else {
          int coverage = effectiveCoverage.size();
          Sequence sequence = entry.getKey();

          if ((coverage > maxCoverage)
              || ((coverage == maxCoverage)
              && (sequence.getLength() < currentSequence.getLength()))) {

            maxCoverage = coverage;
            currentSequence = sequence;
          }

          entry.setValue(effectiveCoverage);
        }
      }

      test.addSequence(currentSequence);
      targets.removeAll(covered.remove(currentSequence));
    }

    return test;
  }

  /**
   * Minimizes test according to the coverage it reaches on transitions of the specified EFSM model.
   * <p>Input data are not modified by this method.</p>
   * @param testModel Test to be minimized.
   * @param efsmModel EFSM model which transition coverage is used as minimization criterion.
   * @return Test that reaches the same EFSM transition coverage level than input one does
   *         but with less or equal number of test sequences.
   */
  public static TestModel minimise(final TestModel testModel, final EfsmModel efsmModel) {

    final Map<Sequence, EfsmTransitionCoverage> covMap =
        EfsmSimulatorUtils.collectCoverage(testModel, efsmModel);

    final Test minTest = TestMinimiser.minimise(covMap);
    EfsmSimulatorUtils.printEfsmTransitionCoverage(covMap, minTest);

    return new TestModel(minTest);
  }
}
