/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test.fate;

import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.EdgeType;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.Set;
import java.util.logging.Logger;

/**
 * The class which performs the learning phase of the FATE algorithm.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
final class FateEfsmAnalyser {
  private static final Logger LOGGER = Log.getLogger(FateEfsmAnalyser.class);
  private final String logEntryHeader;

  private final Graph<EfsmTransition, VariableDependency> controlDependencyGraph;
  private final Graph<EfsmTransition, VariableDependency> dataDependencyGraph;
  private final Map<NodeVariable, Set<List<EfsmTransition>>> counters;

  private FateEfsmAnalyser(final Efsm efsm) {
    this.logEntryHeader = "EFSM.FATE.Analyser." + efsm.getName();

    this.controlDependencyGraph = detectControlDependencies(efsm);
    this.dataDependencyGraph = detectDataDependencies(efsm);
    this.counters = new HashMap<>();
    detectCounters(efsm);
  }

  private Graph<EfsmTransition, VariableDependency> detectControlDependencies(final Efsm efsm) {

    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting control dependencies");

    final Graph<EfsmTransition, VariableDependency> ctrlDepGraph = new DirectedSparseMultigraph<>();

    for (final EfsmTransition transition : efsm.getTransitions()) {
      if (!ctrlDepGraph.containsVertex(transition)) {
        ctrlDepGraph.addVertex(transition);
      }

      for (final NodeVariable used : transition.getGuardUses()) {
        final String usedVariableName = used.getName();

        final VariableDescriptor varData = (VariableDescriptor) used.getUserData();

        if (varData.getVariableType() == VariableType.REG) {
          for (final EfsmTransition defTransition :
              FateEfsmAnalyser.findVariableDefinition(efsm, transition, usedVariableName)) {
            if (!ctrlDepGraph.containsVertex(defTransition)) {
              ctrlDepGraph.addVertex(defTransition);
            }

            final VariableDependency edge = new VariableDependency(transition, defTransition, used);
            ctrlDepGraph.addEdge(edge, defTransition, transition, EdgeType.DIRECTED);
          }
        }
      }
    }

    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": control dependencies have been detected successfully");

    return ctrlDepGraph;
  }

  private Graph<EfsmTransition, VariableDependency> detectDataDependencies(final Efsm efsm) {
    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting data dependencies");

    final Graph<EfsmTransition, VariableDependency> dataDepGraph = new DirectedSparseMultigraph<>();

    for (final EfsmTransition transition : efsm.getTransitions()) {
      if (!dataDepGraph.containsVertex(transition)) {
        dataDepGraph.addVertex(transition);
      }

      for (final NodeVariable used : transition.getActionUses()) {

        final String usedVariableName = used.getName();

        final VariableDescriptor varData = (VariableDescriptor) used.getUserData();

        if ((varData.getVariableType() == VariableType.REG)) {
          for (final EfsmTransition defTransition :
              FateEfsmAnalyser.findVariableDefinition(efsm, transition, usedVariableName)) {
            if (!dataDepGraph.containsVertex(defTransition)) {
              dataDepGraph.addVertex(defTransition);
            }

            final VariableDependency edge = new VariableDependency(transition, defTransition, used);
            dataDepGraph.addEdge(edge, defTransition, transition, EdgeType.DIRECTED);
          }
        }
      }
    }

    LOGGER.log(LogLevel.DEBUG,
        this.logEntryHeader + ": data dependencies have been detected successfully");

    return dataDepGraph;
  }

  private static List<EfsmTransition> findVariableDefinition(
      final Efsm efsm,
      final EfsmTransition transition,
      final String usedVariableName) {

    final Set<EfsmTransition> visitedTransitions = new HashSet<>();
    final Queue<EfsmTransition> queue = new LinkedList<>();
    queue.add(transition);

    final List<EfsmTransition> definingTransitions = new ArrayList<>();

    while (!queue.isEmpty()) {
      final EfsmTransition current = queue.poll();

      for (final EfsmTransition previous : EfsmUtils.getIncoming(efsm, current.getSrc())) {

        if (!visitedTransitions.contains(previous)) {
          boolean useDefMatching = false;

          for (final NodeVariable definedVariable : previous.getDefines()) {
            if (definedVariable.getName().equals(usedVariableName)) {
              useDefMatching = true;

              break;
            }
          }

          if (useDefMatching) {
            definingTransitions.add(previous);
          } else {
            queue.add(previous);
            visitedTransitions.add(previous);
          }
        }
      }
    }

    return definingTransitions;
  }

  private void detectCounters(final Efsm efsm) {
    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": detecting counters");

    final Map<NodeVariable, Set<EfsmTransition>> processedCandidates = new HashMap<>();

    for (final VariableDependency dependency : this.controlDependencyGraph.getEdges()) {
      final EfsmTransition defTransition = this.controlDependencyGraph.getSource(dependency);

      if (processedCandidates.containsKey(dependency.getVariable())) {
        final Set<EfsmTransition> processedTransitions =
            processedCandidates.get(dependency.getVariable());

        if (processedTransitions.contains(defTransition)) {
          continue;
        } else {
          processedTransitions.add(defTransition);
        }
      } else {
        final Set<EfsmTransition> transitions = new HashSet<>();
        transitions.add(defTransition);
        processedCandidates.put(dependency.getVariable(), transitions);
      }

      final EfsmState sourceState = defTransition.getSrc();
      final EfsmState destinationState =
          this.controlDependencyGraph.getDest(dependency).getDst();
      final Set<NodeVariable> defVariables = EfsmUtils
          .getResultingDependencies(defTransition.getAction())
              .get(dependency.getVariable());

      for (final Iterator<NodeVariable> iterator = defVariables.iterator(); iterator.hasNext();) {
        final String variableName = iterator.next().getName();
        if (efsm.declaresInput(variableName)
            || efsm.getParentContainer().declaresInput(variableName)) {
          iterator.remove();
        }
      }

      if (!defVariables.isEmpty() && EfsmUtils.isReachable(efsm, destinationState, sourceState)) {

        final Set<NodeVariable> variables = new HashSet<>(1);
        variables.add(dependency.getVariable());

        for (final VariableDependency dataDependency :
            this.dataDependencyGraph.getOutEdges(defTransition)) {
          if (dataDependency.getVariable().equals(dependency.getVariable())) {
            EfsmTransition transition = this.dataDependencyGraph.getDest(dataDependency);
            this.findCounterLoop(dataDependency.getVariable(), transition,
                defVariables, variables, Collections.singletonList(defTransition));
          }
        }
      }
    }

    LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": counters have been detected successfully");
  }

  //target transition is the first in the returned list
  private void findCounterLoop(
      final NodeVariable potentialCounter,
      final EfsmTransition current,
      final Set<NodeVariable> targetVariables,
      final Set<NodeVariable> currentVars,
      final List<EfsmTransition> currentPath) {

    if (currentPath.contains(current) && current.equals(currentPath.get(0))) {

      final Set<NodeVariable> intersect = new HashSet<>(targetVariables);
      intersect.retainAll(currentVars);

      if (!intersect.isEmpty()) {
        Set<List<EfsmTransition>> counterData = this.counters.get(potentialCounter);

        if (counterData == null) {
          counterData = new HashSet<>(1);
          this.counters.put(potentialCounter, counterData);
        }

        counterData.add(currentPath);

        LOGGER.log(LogLevel.DEBUG, this.logEntryHeader + ": the variable "
            + potentialCounter.getName() + " defined by the transition " + currentPath.get(0)
            + " is a counter in the loop " + currentPath);
      }
    } else {
      final Set<NodeVariable> newCurrentVariables = new HashSet<>();

      for (final Entry<NodeVariable, Set<NodeVariable>> dependency:
          EfsmUtils.getResultingDependencies(current.getAction()).entrySet()) {

        final NodeVariable variable = dependency.getKey();
        final Set<NodeVariable> intersect = new HashSet<>(currentVars);
        intersect.retainAll(dependency.getValue());

        if (!intersect.isEmpty()) {
          newCurrentVariables.add(variable);
        }
      }

      if (!newCurrentVariables.isEmpty()) {
        final List<EfsmTransition> newCurrentPath = new ArrayList<>(currentPath);
        newCurrentPath.add(current);

        for (final VariableDependency dataDependency :
            this.dataDependencyGraph.getOutEdges(current)) {
          if (newCurrentVariables.contains(dataDependency.getVariable())) {
            final EfsmTransition transition = this.dataDependencyGraph.getDest(dataDependency);
            findCounterLoop(potentialCounter, transition, targetVariables, newCurrentVariables,
                newCurrentPath);
          }
        }
      }
    }
  }
}
