/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.generator.test.heuristic.direction;

import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * The class representing a random-based heuristic for choosing a transitions to traverse.
 * <br>
 * Each instance of {@code RandomEfsmTransitionChooser} is bounded to the specified EFSM and its
 * state. Only transitions outgoing from this state may be chosen.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public final class RandomEfsmTransitionChooser extends EfsmTransitionChooser {
  private final List<EfsmTransition> possibleTransitions;

  private final Random randomNumberGenerator;

  /**
   * Creates a new RandomEfsmTransitionChooser for the specified EFSM and its state.
   * <br>
   * Note that list of choosable transitions is constructed only during a
   * RandomEfsmTransitionChooser instance initialisation, so any future changes
   * in the specified EFSM and state instances will not affect a result of
   * choosing by that RandomEfsmTransitionChooser instance.
   *
   * @param efsm the EFSM
   * @param state the state of the EFSM
   * @param seed an initial seed for the random number generator
   */
  public RandomEfsmTransitionChooser(
      final Efsm efsm,
      final EfsmState state,
      final long seed) {
    this.possibleTransitions = new ArrayList<>(EfsmUtils.getOutgoing(efsm, state));
    this.randomNumberGenerator = new Random(seed);
  }

  /**
   * Chooses and returns a transition for traversing.
   * <br>
   * The transition is chosen at random among previously unchosen transitions outgoing from the
   * {@link EfsmState} which this RandomEfsmTransitionChooser is bounded to.
   *
   * @return the transition proposed for traversing. {@code null} is returned if all choosable
   *         transitions were chosen or there are no choosable transitions at all
   */
  @Override
  public EfsmTransition chooseNext() {
    if (this.possibleTransitions.isEmpty()) {
      return null;
    } else {
      final int chosenIndex = this.randomNumberGenerator.nextInt(this.possibleTransitions.size());
      final EfsmTransition chosenTransition = this.possibleTransitions.get(chosenIndex);
      this.possibleTransitions.remove(chosenIndex);

      return chosenTransition;
    }
  }

  @Override
  public boolean hasNext() {
    return !this.possibleTransitions.isEmpty();
  }
}
