/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.printer.graphml;

import org.apache.commons.cli.ParseException;

import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.engine.efsm.EfsmEngine;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.efsm.EfsmDfsWalker;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.util.Log;

import java.util.Map;

/**
 * EFSM printer to GraphML format.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */

public class EfsmGraphMlPrinter extends EfsmEngine {

  /**
   * The engine identifier.
   */
  public static final String ENGINE_ID = "efsm-graphml-printer";

  /**
   *  The engine output entity type.
   */
  public static final EntityType OUT_TYPE = EntityType.get("efsm-graphml");

  /**
   * The output GraphML-file-name command line parameter.
   */
  public static final Parameter OUT_FILE_NAME =
      new Parameter("efsm-graphml", true, "Output GraphML file");

  /* The default file name. */
  private static final String DEFAULT_FILE_NAME = "efsm-model.graphml";

  /* The printer result. */
  private FileNames fileNames;

  /**
   * The default constructor.
   */
  public EfsmGraphMlPrinter() {
    super(EfsmGraphMlPrinter.ENGINE_ID, new EfsmGraphMlVisitor(), EfsmGraphMlPrinter.OUT_TYPE);
    addParameter(OUT_FILE_NAME);
  }

  @Override
  public void initialise(final Map<EntityType, Entity> inputs) {

    final Arguments arguments = (Arguments)inputs.get(Arguments.TYPE);

    try {

      /* Get arguments from command line. */
      final Arguments engineArgs = parseCommandLine(arguments.getCommandLine());

      /* Create GraphML file. */
      fileNames = new FileNames("graphml");
      fileNames.add(engineArgs.getValue(OUT_FILE_NAME, DEFAULT_FILE_NAME));

      /* Add GraphML file to visitor. */
      initGraphMlVisitor(fileNames.getList().get(0));
    } catch (ParseException e) {
      Log.printError(e.toString());
    }
  }

  /**
   * Initializes visitor with the GraphML file name.
   *
   * @param fileName the output GraphML file name.
   */
  private void initGraphMlVisitor(final String fileName) {
    ((EfsmGraphMlVisitor)this.visitor).initFile(fileName);
  }

  @Override
  public Entity getOutput() {
    return this.fileNames;
  }

  /**
   * Prints the specified {@link ru.ispras.retrascope.model.efsm.EfsmModel} model to GraphML file.
   *
   * @param fileName the output file name
   * @param efsm extended finite state machine model to be printed
   */
  public static void print(final String fileName, final EfsmModel efsm) {

    final EfsmGraphMlVisitor efsmGraphMlVisitor = new EfsmGraphMlVisitor();
    efsmGraphMlVisitor.initFile(fileName);

    final ModelWalker efsmWalker = new EfsmDfsWalker(efsmGraphMlVisitor);
    efsmWalker.start(efsm);
  }
}
