/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.simulator;

import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.Simulator;
import ru.ispras.retrascope.engine.efsm.simulator.ProcessSimulator.ProcessSimulationSnapshot;
import ru.ispras.retrascope.model.basis.State;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmModule;
import ru.ispras.retrascope.model.efsm.EfsmModuleState;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;

/**
 * A simulator for {@code EfsmModule}.
 *
 * @author <a href="mailto:myllyenko@ya.ru">Igor Melnichenko</a>
 *
 * @see EfsmModule
 */
public class ModuleSimulator extends Simulator<EfsmModule> {

  private final Set<ProcessSimulator> efsmSimulators;
  private final Map<String, Map<ProcessSimulator, NodeValue>> currentVariableModifications;

  private final String logEntryHeader;

  private final Random rndGen;

  /**
   * Constructs a simulator for the specified EFSM module.
   *
   * @param efsmModule the EFSM module which the simulator is constructed for
   *
   * @throws IllegalArgumentException if {@code efsmModule} is {@code null}
   */
  public ModuleSimulator(final EfsmModule efsmModule) {
    super(efsmModule);

    this.efsmSimulators = new HashSet<>();

    final String name = this.simulatedModel.getName();
    this.logEntryHeader = "EFSM.Simulator." + name + ": ";

    for (final Efsm efsm : this.simulatedModel.getNonInitialEfsms()) {
      if (efsm.singleTestable()) {
        this.efsmSimulators.add(new ProcessSimulator(this, efsm));
      } else {
        Log.printWarn(
            String.format("The efsm can't be tested as a single one: %s", efsm.getName()));
      }
    }

    final Collection<String> registerNames = this.simulatedModel.getRegisterNames();
    final Collection<String> outputNames = this.simulatedModel.getOutputNames();

    this.currentVariableModifications = new HashMap<>(registerNames.size() + outputNames.size());
    resetModifications();

    this.reset();

    this.rndGen = new Random(0L);
  }

  private void resetModifications() {
    final Collection<String> registerNames = this.simulatedModel.getRegisterNames();
    final Collection<String> outputNames = this.simulatedModel.getOutputNames();

    for (final String variable : registerNames) {
      this.currentVariableModifications.put(variable, new HashMap<ProcessSimulator, NodeValue>());
    }

    for (final String variable: outputNames) {
      this.currentVariableModifications.put(variable, new HashMap<ProcessSimulator, NodeValue>());
    }
  }

  @Override
  public EfsmModule getSimulatedModel() {
    return this.simulatedModel;
  }

  @Override
  public Vector simulateVector(final Vector inputVector) {
    super.readVector(inputVector);

    Log.printDebug(this.logEntryHeader + "processing vector: " + inputVector);

    final Transaction outTransaction = new Transaction();

    for (final ProcessSimulator efsmSimulator : this.efsmSimulators) {
      efsmSimulator.simulateVector(inputVector);
    }

    updateMemory(outTransaction);

    final Vector outputVector = new Vector(outTransaction);

    Log.printDebug(this.logEntryHeader
        + "vector has been processed successfully. Output vector: " + outputVector);

    return outputVector;
  }

  private void updateMemory(final Transaction outTransaction) {

    for (final Entry<String, NodeValue> modification: resolveAssignmentConflicts().entrySet()) {

      final String variableName = modification.getKey();
      final NodeValue variableValue = modification.getValue();

      if (this.innerVariablesMemory.containsVariable(variableName)) {

        this.innerVariablesMemory.setValue(variableName, variableValue);

      } else {

        if (this.outputsMemory.containsVariable(variableName)) {

          this.outputsMemory.setValue(variableName, variableValue);
          outTransaction.addVariable(variableName, variableValue);
        } else {
          throw new RetrascopeRuntimeException("The variable " + variableName
              + " isn't found neither among inner variables nor among outputs");
        }
      }
    }
  }

  @Override
  public Sequence simulateSequence(final Sequence inputSequence) {
    InvariantChecks.checkNotNull(inputSequence);

    Log.printDebug(this.logEntryHeader + "processing sequence: " + inputSequence);

    final Sequence outputSequence = new Sequence();

    inputSequence.initIterator();
    while (inputSequence.hasNext()) {
      outputSequence.addVector(this.simulateVector(inputSequence.next()));
    }

    Log.printDebug(this.logEntryHeader
        + "sequence has been processed successfully. Output sequence: " + outputSequence);

    return outputSequence;
  }

  @Override
  public void reset() {
    Log.printDebug(this.logEntryHeader + "resetting");

    super.reset();

    for (final ProcessSimulator efsmSimulator : this.efsmSimulators) {
      efsmSimulator.reset();
    }

    resetModifications();

    Log.printDebug(this.logEntryHeader + "reset has been finished successfully");
  }

  @Override
  public EfsmTransitionCoverage getCoverage() {
    final EfsmTransitionCoverage coverage = new EfsmTransitionCoverage();

    for (final ProcessSimulator efsmSimulator: this.efsmSimulators) {
      coverage.include(efsmSimulator.getCoverage());
    }

    return coverage;
  }

  /**
   * Returns a coverage of the specified EFSM.
   *
   * @param efsm the target EFSM
   *
   * @return the coverage of the specified EFSM. If there is no this EFSM in a simulated module,
   * {@code null} is returned
   */
  public EfsmTransitionCoverage getCoverage(final Efsm efsm) {

    for (final ProcessSimulator efsmSimulator: this.efsmSimulators) {
      if (efsmSimulator.getSimulatedModel().equals(efsm)) {
        return efsmSimulator.getCoverage();
      }
    }

    return null;
  }

  /**
   * Returns a simulator for the specified EFSM. The returned simulator represents a current state
   * of EFSM simulation.
   *
   * @param efsm the target EFSM
   *
   * @return the simulator for the specified EFSM. If there is no this EFSM in a simulated module,
   * {@code null} is returned
   */
  public ProcessSimulator getSubSimulator(final Efsm efsm) {

    for (final ProcessSimulator efsmSimulator: this.efsmSimulators) {
      if (efsmSimulator.getSimulatedModel().equals(efsm)) {
        return efsmSimulator;
      }
    }

    return null;
  }

  /**
   * Returns all process simulators of this simulator. The returned simulators represent a current
   * state of simulation of the processes.
   *
   * @return the simulator for the specified EFSM. If there is no this EFSM in a simulated module,
   * {@code null} is returned
   */
  public Set<ProcessSimulator> getSubSimulators() {
    return Collections.unmodifiableSet(this.efsmSimulators);
  }

  /**
   * Submits a variable modification request to this simulator. If there is no inner/output variable
   * with the specified name in a simulated module, this method has no any effects.
   * <br>
   * The request is valid only for the current simulation cycle. At the end of the cycle the
   * simulator performs all requested modifications.
   *
   * @param simulator a process simulator which caused variable modification
   * @param variableName the name of a variable to be modified
   * @param value a new requested value of the variable
   *
   * @throws IllegalArgumentException if {@code efsmModule} is {@code null}
   */
  void modifyVariable(
      final ProcessSimulator simulator,
      final String variableName,
      final NodeValue value) {
    final Map<ProcessSimulator, NodeValue> modifications =
        this.currentVariableModifications.get(variableName);

    if (modifications == null) {
      Log.printWarn(this.logEntryHeader
          + "a request to modify unknown variable " + variableName + " has been received");
    } else {
      modifications.put(simulator, value);
    }
  }

  private Map<String, NodeValue> resolveAssignmentConflicts() {
    final Map<String, NodeValue> effectiveModifications =
        new HashMap<>(this.currentVariableModifications.size());

    for (final Entry<String, Map<ProcessSimulator, NodeValue>> modifications :
        this.currentVariableModifications.entrySet()) {
      final String variableName = modifications.getKey();
      final Map<ProcessSimulator, NodeValue> assignedValues = modifications.getValue();

      switch (assignedValues.size()) {
        case 0: {
          break;
        }

        case 1: {
          effectiveModifications.put(variableName, assignedValues.values().iterator().next());

          break;
        }

        default: {
          Log.printWarn(this.logEntryHeader
              + "there are multiple simultaneous assignments to the variable " + variableName
              + ": " + assignedValues);

          final int chosenAssignment = this.rndGen.nextInt(assignedValues.size());
          final NodeValue chosenValue =
              new ArrayList<>(assignedValues.values()).get(chosenAssignment);

          Log.printDebug(this.logEntryHeader + "randomly chosen assigned value: " + chosenValue);

          effectiveModifications.put(variableName, chosenValue);
        }
      }
    }

    return effectiveModifications;
  }

  @Override
  public EfsmModuleState getCurrentState() {
    final Collection<State> subStates = new ArrayList<>();

    for (final ProcessSimulator subSimulator: this.efsmSimulators) {
      subStates.add(subSimulator.getCurrentState());
    }

    return new EfsmModuleState(subStates);
  }

  @Override
  public void resetCoverage() {
    for (final ProcessSimulator subSimulator: this.efsmSimulators) {
      subSimulator.resetCoverage();
    }
  }

  public class ModuleSimulationSnapshot extends SimulationSnapshot {
    private final Map<ProcessSimulator, ProcessSimulationSnapshot> snapshots;

    ModuleSimulationSnapshot() {
      super();

      this.snapshots = new HashMap<>(ModuleSimulator.this.efsmSimulators.size());

      for (final ProcessSimulator subSimulator: ModuleSimulator.this.efsmSimulators) {
        this.snapshots.put(subSimulator, subSimulator.makeSimulationSnapshot());
      }
    }
  }

  @Override
  public ModuleSimulationSnapshot makeSimulationSnapshot() {
    return new ModuleSimulationSnapshot();
  }

  /**
   * Loads the specified simulation snapshot discarding current simulation state and history.
   *
   * @param snapshot the snapshot
   */
  public void loadSnapshot(final ModuleSimulationSnapshot snapshot) {

    Log.printDebug(this.logEntryHeader + ": loading a simulation snapshot");

    super.loadSnapshot(snapshot);

    for (final ProcessSimulator subSimulator: ModuleSimulator.this.efsmSimulators) {
      subSimulator.loadSnapshot(snapshot.snapshots.get(subSimulator));
    }
  }
}
