/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.efsm.simulator;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeBinding.BoundVariable;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.SolverResult.Status;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.basis.Simulator;
import ru.ispras.retrascope.engine.efsm.generator.test.Substitutor;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.efsm.Action;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmState;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.model.efsm.EfsmUtils;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.result.test.efsm.EfsmTransitionCoverage;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.Set;

/**
 * The EFSM-based representation of a hardware model which encapsulates a runtime state along with a
 * static structure of an original device.
 *
 * @author <a href="mailto:myllyenko@ya.ru">Igor Melnichenko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class ProcessSimulator extends Simulator<Efsm> {
  private final String logEntryHeader;

  private final ModuleSimulator parentModuleSimulator;

  private final String name;

  private EfsmState currentState;

  private EventList currentEvents;

  private final EfsmTransitionCoverage coverage;

  private EfsmTransition lastTraversedTransition;

  /**
   * Constructs a simulator above the specified EFSM.
   *
   * @param parentModuleSimulator a simulator of a module which the specified EFSM belongs to
   * @param efsm an EFSM which the simulator is based on
   *
   * @throws IllegalArgumentException if at least one argument is {@code null}
   *
   * @see Efsm#getInitialState()
   * @see Efsm#getResetGuardedActions()
   */
  ProcessSimulator(ModuleSimulator parentModuleSimulator, Efsm efsm) {
    super(efsm);
    this.parentModuleSimulator = parentModuleSimulator;
    InvariantChecks.checkNotNull(parentModuleSimulator);

    this.logEntryHeader = "EFSM.Simulator." + this.simulatedModel.getName() + ": ";

    this.name = this.simulatedModel.getName();

    this.coverage = new EfsmTransitionCoverage(this.simulatedModel);
  }

  /**
   * Returns a name of this simulator.
   *
   * @return the name
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns a state of an underlying EFSM which this simulator is currently in.
   *
   * @return the current state of this simulator. Returns the same state as
   *         {@link Efsm#getInitialState()} do if called immediately
   *         after construction/reset of this simulator
   *
   * @see ProcessSimulator#reset()
   */
  @Override
  public EfsmState getCurrentState() {
    return this.currentState;
  }

  /**
   * Returns a transition which was traversed by this simulator on last cycle.
   *
   * @return the last traversed transition
   */
  public EfsmTransition getLastTraversedTransition() {
    return this.lastTraversedTransition;
  }

  /**
   * Returns a test coverage reached during simulation since last reset.
   *
   * @return the test coverage
   */
  @Override
  public EfsmTransitionCoverage getCoverage() {
    return new EfsmTransitionCoverage(this.coverage);
  }

  /**
   * Resets this simulator, that is, returns it to an uninitialised state.
   */
  @Override
  public void reset() {

    Log.printDebug(this.logEntryHeader + "resetting");

    super.reset();

    this.coverage.reset();

    this.currentState = this.simulatedModel.getInitialState();

    if (this.currentState.equals(Efsm.UNINIT_STATE)) {
      final EfsmTransition initTransition =
          EfsmUtils.getOutgoing(getSimulatedModel(), this.currentState).iterator().next();

      if (initTransition == null) {
        throw new RetrascopeRuntimeException("No init transition has been found.");
      }
      executeAction(initTransition.getAction());
      this.currentState = initTransition.getDst();
      this.lastTraversedTransition = initTransition;
    }

    this.lastTraversedTransition = null;

  }

  private void traverse(final EfsmTransition transition) {

    Log.printDebug(this.logEntryHeader + "traversing the transition " + transition);

    executeAction(transition.getAction());
    this.currentState = transition.getDst();
    this.coverage.markTransitionAsCovered(transition);
    this.lastTraversedTransition = transition;

    Log.printDebug(this.logEntryHeader + "traversing has been finished successfully");
  }

  private List<EfsmTransition> getTraversableTransitions() {

    Log.printDebug(this.logEntryHeader + "performing transitions traversability test");

    final List<EfsmTransition> possibleTransitions =
        new LinkedList<>(EfsmUtils.getOutgoing(this.simulatedModel, this.currentState));

    for (final Iterator<EfsmTransition> possibleTransIterator = possibleTransitions.iterator();
        possibleTransIterator.hasNext();) {

      final EfsmTransition testedTransition = possibleTransIterator.next();

      Log.printDebug(this.logEntryHeader + "testing the transition " + testedTransition);

      if (this.currentEvents == null && testedTransition.clockFree()
          || this.currentEvents != null && testedTransition.isSensible(this.currentEvents)) {

        final Node testedConstraint = guardToConstraint(testedTransition.getGuardNode());

        if (SolverUtils.isSat(testedConstraint)) {

          Log.printDebug(this.logEntryHeader + "the transition is traversable");

        } else {

          Log.printDebug(this.logEntryHeader + "the transition isn't traversable now");

          possibleTransIterator.remove();
        }

      } else {

        Log.printDebug(
            this.logEntryHeader + "the transition isn't sensible to the occurred events");

        possibleTransIterator.remove();
      }
    }

    return possibleTransitions;
  }

  private EfsmTransition chooseTransitionToTraverse(final List<EfsmTransition> transitions) {

    EfsmTransition traversedTransition;

    if (transitions.size() == 1) {

      traversedTransition = transitions.get(0);

    } else {

      Log.printWarn(this.logEntryHeader
          + "there are multiple traversable transitions now. Choosing one of them at random");

      for (final EfsmTransition transition: transitions) {

        Log.printWarn(this.logEntryHeader + "possible transition: " + transition);
      }

      Random randomGenerator = new Random(transitions.size());

      traversedTransition =
          transitions.get(randomGenerator.nextInt(transitions.size()));
    }

    return traversedTransition;
  }

  private Node guardToConstraint(final Node guardNode) {

    return substituteVariables(guardNode, Collections.<String, NodeValue>emptyMap());
  }

  private void executeAction(Action action) {
    InvariantChecks.checkNotNull(action);

    Log.printDebug(this.logEntryHeader + "executing the action: " + action);

    final List<Assignment> assignments = action.getAssignments();

    Map<String, NodeValue> bufferedVariables = new HashMap<>(assignments.size());

    for (final Assignment assignment : assignments) {

      bufferedVariables = this.executeAssignment(assignment, bufferedVariables);
    }

    for (Entry<String, NodeValue> bufferedVariable: bufferedVariables.entrySet()) {

      final String variableName = bufferedVariable.getKey();
      final NodeValue variableValue = bufferedVariable.getValue();

      if (this.simulatedModel.getRegisterNames().contains(variableName)) {

        this.innerVariablesMemory.setValue(variableName, variableValue);

      } else {

        if (this.simulatedModel.getOutputNames().contains(variableName)) {

          this.outputsMemory.setValue(variableName, variableValue);

        } else {
          this.parentModuleSimulator.modifyVariable(this, variableName, variableValue);
        }
      }
    }

    Log.printDebug(this.logEntryHeader + "the action has been executed successfully: " + action);
  }

  private Map<String, NodeValue> executeAssignment(
      final Assignment assignment,
      final Map<String, NodeValue> bufferedVars) {

    Log.printDebug(this.logEntryHeader + "executing the assignment: " + assignment);

    final RangedVariable rangedVariable = assignment.getTarget();
    Node expression = assignment.getValue();

    if ((rangedVariable == null) && (expression == null)) {
      return Collections.emptyMap();
    }

    InvariantChecks.checkNotNull(rangedVariable);
    InvariantChecks.checkNotNull(expression);

    final NodeVariable nodeVariable = rangedVariable.getVariable();
    final DataType dataType = nodeVariable.getDataType();

    if (rangedVariable.isRanged()) {

      if (!dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)
          && !dataType.getTypeId().equals(DataTypeId.MAP)) {

        throw new IllegalArgumentException("Unsupported data type of ranged variable "
            + rangedVariable.getVariableName() +  ": " + dataType);
      }

      final Range range = rangedVariable.getRange();

      final Node leftBoundExpression = this.substituteVariables(range.getYoung(), bufferedVars);
      final Node rightBoundExpression = this.substituteVariables(range.getOld(), bufferedVars);

      final NodeValue leftBound = this.computeExpression(leftBoundExpression, DataType.INTEGER);
      final NodeValue rightBound = this.computeExpression(rightBoundExpression, DataType.INTEGER);

      final int leftBoundInteger = leftBound.getInteger().intValue();
      final int rightBoundInteger = rightBound.getInteger().intValue();

      final int variableSize = dataType.getSize();

      if (leftBoundInteger < rightBoundInteger) {

        throw new IllegalArgumentException("This type of subvectors is currently unsupported");
      }

      NodeOperation mostSignificantPart = null;
      NodeOperation lessSignificantPart = null;

      if (leftBoundInteger < variableSize - 1) {

        final NodeValue leftIndex =
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(variableSize - 1), 10));

        final NodeValue rightIndex =
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(leftBoundInteger + 1), 10));

        mostSignificantPart = new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex,
            nodeVariable);
      }

      if (rightBoundInteger > 0) {
        NodeValue leftIndex =
            new NodeValue(DataType.INTEGER.valueOf(String.valueOf(rightBoundInteger - 1), 10));
        NodeValue rightIndex = new NodeValue(DataType.INTEGER.valueOf("0", 10));

        lessSignificantPart = new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex,
            nodeVariable);
      }

      if ((mostSignificantPart != null) && (lessSignificantPart != null)) {
        expression = new NodeOperation(StandardOperation.BVCONCAT, mostSignificantPart, expression,
            lessSignificantPart);
      } else {
        if (mostSignificantPart != null) {
          expression =
              new NodeOperation(StandardOperation.BVCONCAT, mostSignificantPart, expression);
        } else {
          if (lessSignificantPart != null) {
            expression =
                new NodeOperation(StandardOperation.BVCONCAT, expression, lessSignificantPart);
          }
        }
      }
    }

    expression = this.substituteVariables(expression, bufferedVars);
    NodeValue assignedValue = this.computeExpression(expression, dataType);
    bufferedVars.put(rangedVariable.getVariableName(), assignedValue);

    Log.printDebug(this.logEntryHeader + "the assigned value: " + assignedValue);

    return bufferedVars;
  }

  private NodeValue computeExpression(Node expression, final DataType resultDataType) {

    if (expression instanceof NodeValue) {
      NodeValue value = (NodeValue) expression;

      if (!value.getDataType().equals(resultDataType)) {
        throw new IllegalArgumentException("Incompatible data types: " + value.getDataType()
            + " vs " + resultDataType);
      }

      return value;
    } else {
      if (!ExprUtils.getVariables(expression).isEmpty()) {
        throw new IllegalArgumentException("The expression to be computed (" + expression
            + ") contains unevaluated variables: " + ExprUtils.getVariables(expression));
      }

      Log.printDebug(this.logEntryHeader + "computing the expression " + expression);

      expression =
          NodeUtils.getEquation(new NodeVariable("assignmentResult", resultDataType), expression);

      final SolverResult result = SolverUtils.solve(expression);

      if (result.getStatus().equals(Status.SAT)) {

        final NodeValue resultValue =
            new NodeValue(result.getVariables().iterator().next().getData());

        Log.printDebug(this.logEntryHeader + "the computation result: " + resultValue);

        return resultValue;
      } else {
        throw new RetrascopeRuntimeException(
            "Failed to resolve the assignment constraint due to errors: " + result.getErrors());
      }
    }
  }

  private void logInnerDump() {

    Log.printDebug(this.logEntryHeader + "current state: " + this.currentState);
    Log.printDebug(this.logEntryHeader + "inner variables' dump: " + this.innerVariablesMemory);
    Log.printDebug(this.logEntryHeader + "output dump: " + this.outputsMemory);
    Log.printDebug(this.logEntryHeader + "current coverage: " + this.coverage);
  }

  @Override
  public Vector simulateVector(final Vector inputVector) {

    Log.printDebug(this.logEntryHeader + "processing vector: " + inputVector);

    super.readVector(inputVector);

    if (inputVector.hasEvents()) {
      this.currentEvents = inputVector.getEventList();
    }

    final List<EfsmTransition> possibleTransitions = getTraversableTransitions();

    if (possibleTransitions.isEmpty()) {

      Log.printDebug(this.logEntryHeader
          + "there are no transitions to traverse. Finishing event handling");

      this.lastTraversedTransition = null;

    } else {
      this.traverse(this.chooseTransitionToTraverse(possibleTransitions));
    }

    final List<EfsmTransition> enabled = getEnabledTransitions();

    while (!enabled.isEmpty()) {

      this.traverse(this.chooseTransitionToTraverse(enabled));
      enabled.clear();
      enabled.addAll(getEnabledTransitions());
    }

    this.logInnerDump();

    Log.printDebug(this.logEntryHeader + "model has finished to process the events");

    return super.getOutputVector();
  }

  private List<EfsmTransition> getEnabledTransitions() {

    final Integer currentPhase = this.currentState.getPhase();
    if (currentPhase == 0) {
      this.currentEvents = null;
    }

    final List<EfsmTransition> enabled = new LinkedList<>();

    final Collection<EfsmTransition> possibleTransitions =
        EfsmUtils.getOutgoing(this.simulatedModel, this.currentState);

    for (final EfsmTransition transition : possibleTransitions) {

      Log.printDebug(this.logEntryHeader + "testing the transition " + transition);

      boolean isTriggeredByEvent = false;

      if (!transition.clockFree()) {

        if (this.currentEvents == null) {
          break;
        }

        if (transition.isSensible(this.currentEvents)) {
          isTriggeredByEvent = true;
        }
      }

      if (transition.clockFree() || (isTriggeredByEvent && this.currentEvents != null)) {

        final Node testedConstraint = guardToConstraint(transition.getGuardNode());

        if (SolverUtils.isSat(testedConstraint)) {

          Log.printDebug(this.logEntryHeader + "the transition is traversable");
          enabled.add(transition);

        } else {

          Log.printDebug(this.logEntryHeader + "the transition isn't traversable now");
        }
      }
    }

    return enabled;
  }

  @Override
  public Sequence simulateSequence(final Sequence inputSequence) {

    throw new UnsupportedOperationException("simulateSequence(Sequence) isn't implemented yet");
  }

  @Override
  public Efsm getSimulatedModel() {
    return this.simulatedModel;
  }

  @Override
  public Set<String> getInnerVariableNames() {

    final Set<String> variableNames = new HashSet<>(this.innerVariablesMemory.getValues().keySet());
    variableNames.addAll(this.parentModuleSimulator.getInnerVariableNames());

    return variableNames;
  }

  @Override
  public Set<String> getOutputVariableNames() {

    final Set<String> variableNames = new HashSet<>(this.outputsMemory.getValues().keySet());
    variableNames.addAll(this.parentModuleSimulator.getOutputVariableNames());

    return variableNames;
  }

  @Override
  protected boolean hasInput(final String variableName) {
    return EfsmUtils.isInputDeclared(this.simulatedModel, variableName);
  }

  @Override
  protected boolean hasInnerVariable(final String variableName) {
    return EfsmUtils.isRegisterDeclared(this.simulatedModel, variableName);
  }

  @Override
  public boolean hasOutput(final String variableName) {
    return EfsmUtils.isOutputDeclared(this.simulatedModel, variableName);
  }

  @Override
  public NodeValue getVariableValue(final String variableName) {

    NodeValue value = super.getVariableValue(variableName);

    if (value == null) {
      value = this.parentModuleSimulator.getVariableValue(variableName);
    }

    return value;
  }

  /**
   * Generates an input vector to enable the specified guarded action considering the specified
   * fixed events and inputs.
   *
   * @param guardedAction the guarded action to be enabled
   * @param fixedEvents the fixed events
   * @param fixedInputs the fixed inputs
   *
   * @return the generated vector; if one failed to enable the guarded action, {@code null} is
   *     returned
   */
  public Vector enableGuardedAction(
      final GuardedAction guardedAction,
      final EventList fixedEvents,
      final Map<String, NodeValue> fixedInputs) {

    if (!fixedEvents.isEmpty()) {

      final EventList gaEventList = new EventList();
      gaEventList.addEvents(guardedAction.getClocks());

      if (!gaEventList.compatibleWith(fixedEvents)) {
        return null;
      }
    }

    final Node guardExpression = this.fillExpression(guardedAction.getGuardNode(), fixedInputs);

    final SolverResult solverResult = SolverUtils.solve(guardExpression);

    if (solverResult.getStatus().equals(Status.SAT)) {

      final Vector inputVector = new Vector();
      final Transaction transaction = new Transaction();

      for (final Variable variable : solverResult.getVariables()) {

        final String variableName = variable.getName();

        if (this.hasInput(variableName)) {

          transaction.addVariable(variableName, new NodeValue(variable.getData()));
        }
      }

      inputVector.addEvents(guardedAction.getClocks());

      if (!transaction.isEmpty()) {
        inputVector.addTransaction(transaction);
      }

      inputVector.addEvents(fixedEvents);

      if (!fixedInputs.isEmpty()) {
        inputVector.addTransaction(new Transaction(fixedInputs));
      }

      return inputVector;
    } else {
      return null;
    }
  }

  /**
   * Fills the specified expression with the specified inputs and current values of inner variables.
   *
   * @param expression the expression to be filled
   * @param fixedInputs the fixed inputs
   *
   * @return an expression where variables replaced by their values
   */
  public Node fillExpression(Node expression, final Map<String, NodeValue> fixedInputs) {

    final Set<NodeVariable> usedVariables = new HashSet<>(ExprUtils.getVariables(expression));
    final List<BoundVariable> bindingList = new LinkedList<>();
    final Set<Node> inputConstraints = new HashSet<>();

    for (final NodeVariable usedVar: usedVariables) {

      final String varName = Substitutor.deannotateName(usedVar.getName());

      if (this.hasInput(varName)) {

        final Node variableConstraint = this.simulatedModel.getInvariant(varName);

        if (variableConstraint != null) {

          inputConstraints.add(Transformer.substitute(variableConstraint, varName,
              usedVar));
        }

        if (fixedInputs.containsKey(varName)) {

          bindingList.add(NodeBinding.bindVariable(usedVar,
              fixedInputs.get(varName)));
        }
      } else if (this.hasInnerVariable(varName)) {

        bindingList.add(NodeBinding.bindVariable(usedVar, this.getVariableValue(varName)));
      }
    }

    expression = Transformer.substituteAllBindings(new NodeBinding(expression, bindingList));

    if (!inputConstraints.isEmpty()) {
      expression = Node.AND(expression, NodeUtils.getConjunction(inputConstraints));
    }

    return expression;
  }

  @Override
  public void resetCoverage() {
    this.coverage.reset();
  }

  /**
   * A process simulation snapshot.
   */
  class ProcessSimulationSnapshot extends SimulationSnapshot {

    private final EfsmState currentState;
    private final EfsmTransition lastTraversedTransition;
    private final EfsmTransitionCoverage coverage;

    ProcessSimulationSnapshot() {
      super();

      this.currentState = ProcessSimulator.this.currentState;
      this.lastTraversedTransition = ProcessSimulator.this.lastTraversedTransition;
      this.coverage = new EfsmTransitionCoverage(ProcessSimulator.this.coverage);
    }
  }

  @Override
  public ProcessSimulationSnapshot makeSimulationSnapshot() {
    return new ProcessSimulationSnapshot();
  }

  /**
   * Loads the specified simulation snapshot discarding current simulation state and history.
   *
   * @param snapshot the snapshot
   */
  void loadSnapshot(ProcessSimulationSnapshot snapshot) {

    Log.printDebug(this.logEntryHeader + "loading a simulation snapshot");

    this.currentState = snapshot.currentState;
    this.lastTraversedTransition = snapshot.lastTraversedTransition;
    this.coverage.reset();
    this.coverage.include(snapshot.coverage);

    super.loadSnapshot(snapshot);
  }
}
