/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.retrascope.basis.FileCreator;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.MapUtils;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvExprPrinter;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvPrinterUtils;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.hldd.Hldd;
import ru.ispras.retrascope.model.hldd.HlddConditionNode;
import ru.ispras.retrascope.model.hldd.HlddEdge;
import ru.ispras.retrascope.model.hldd.HlddEmptyVisitor;
import ru.ispras.retrascope.model.hldd.HlddNode;
import ru.ispras.retrascope.model.hldd.HlddProcess;
import ru.ispras.retrascope.model.hldd.HlddTerminalNode;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.File;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * Basic HLDD visitor that provides SMV representation.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class HlddSmvVisitor extends HlddEmptyVisitor {

  // Prints expressions into file.
  private final PrintWriter printWriter;

  // Module being visited.
  private Module currentModule;

  // Process being visited.
  private HlddProcess currentProcess;

  // SMV expression printer.
  private final SmvExprPrinter exprPrinter;

  // Indentation.
  private int indentationLevel;

  // Current HLDD variable.
  private NodeVariable hlddVariable;

  // Enables printing.
  private boolean printingEnabled;

  // The node which children are not printed.
  private HlddNode disablingNode;

  private Collection<NodeVariable> currentInputs;

  protected Map<NodeVariable, Declaration> currentPhaseVars;

  private static final int BASIC_INDENT = 3;

  /**
   * Constructs a new visitor and creates a new *.smv file.
   *
   * @param fileName - the specified *.smv file name.
   */
  public HlddSmvVisitor(final String fileName) {
    super();
    final File smvFile = FileCreator.newFile(fileName);
    exprPrinter = new SmvExprPrinter();
    printingEnabled = true;
    disablingNode = null;
    printWriter = ReadWriteUtils.newPrintWriter(smvFile);
    currentInputs = new ArrayList<>();
    indentationLevel = BASIC_INDENT;
  }

  /**
   * Returns the current SMV expression printer.
   *
   * @return The current SMV expression printer.
   */
  protected SmvExprPrinter getExprPrinter() {
    return this.exprPrinter;
  }

  /**
   * Returns the current control flow graph model module.
   *
   * @return the current control flow graph model module.
   */
  protected Module getCurrentModule() {
    return this.currentModule;
  }

  /**
   * Sets the specified module as a current one.
   *
   * @param module The module to be set as the current one.
   */
  private void setCurrentModule(final Module module) {
    this.currentModule = module;
  }

  /**
   * Sets the specified process as a current one.
   *
   * @param process The process to be set as a current one.
   */
  private void setCurrentProcess(final HlddProcess process) {
    this.currentProcess = process;
  }

  /**
   * Appends the specified string to the output file.
   *
   * @param string The string to be added to the output file.
   */
  protected void append(final String string) {
    if (printingEnabled) {
      printWriter.print(string);
    }
  }

  /**
   * Appends the specified line to the output file.
   *
   * @param line The line to be added to the output file.
   */
  protected void appendLine(final String line) {
    if (printingEnabled) {
      printWriter.println(line);
    }
  }

  /**
   * Appends an empty line to the output file.
   */
  protected void appendLine() {
    if (printingEnabled) {
      printWriter.println();
    }
  }

  /**
   * Returns the expression description string.
   *
   * @param expression The expression to be represented in string format.
   * @return The SMV string representation for the specified expression.
   */
  protected String getExprDescription(final Node expression) {
    return exprPrinter.toString(expression).replaceAll("true", "TRUE").replaceAll("false", "FALSE");
  }

  @Override
  public void onModelBegin(final Model model) {

    appendLine();
    // Print "main".
    appendLine(SmvPrinterUtils.getModuleDeclaration("main"));

    final Map<NodeVariable, Declaration> inputVariables = new HashMap<>();

    // Declare nested modules.
    for (final Module hlddModule : model.getModules()) {
      final String moduleName = SmvPrinterUtils.getFormattedName(hlddModule.getName());
      final String instanceName = "m_" + moduleName;
      for (final NodeVariable input : hlddModule.getInputs()) {
        inputVariables.put(input, hlddModule.getDeclaration(input));
      }

      appendLine(
          SmvPrinterUtils.getModuleInstantiation(instanceName, moduleName, hlddModule.getInputs()));
    }

    // Declare input variables.
    appendLine(SmvPrinterUtils.getVarDeclareSection(inputVariables, exprPrinter));
  }

  @Override
  public void onModelEnd(final Model model) {
    appendLine();
    printWriter.close();
  }

  @Override
  public void onModuleBegin(final Module hlddModule) {
    currentInputs = hlddModule.getInputs();
    setCurrentModule(hlddModule);
    final String moduleName = " " + SmvPrinterUtils.getFormattedName(hlddModule.getName());
    appendLine();
    // Print module.
    appendLine(SmvPrinterUtils.getModuleDeclaration(moduleName, currentInputs));
    final List<NodeVariable> regs = SmvPrinterUtils.getRegs(hlddModule);
    if (!regs.isEmpty()) {
      final Map<NodeVariable, Declaration> regDeclarations = new HashMap<>();
      for (NodeVariable reg : regs) {
        regDeclarations.put(reg, hlddModule.getDeclaration(reg));
      }
      appendLine(SmvPrinterUtils.getVarDeclareSection(regDeclarations, exprPrinter));
      currentInputs.addAll(regs);
    }
    currentPhaseVars = getPhaseVariables(hlddModule.getProcesses());
    if (!currentPhaseVars.isEmpty()) {
      currentInputs.addAll(currentPhaseVars.keySet());
      appendLine(SmvPrinterUtils.getVarDeclareSection(currentPhaseVars, exprPrinter));
    }
    // Declare nested processes.
    appendLine(SmvPrinterUtils.getProcessInstantiation(hlddModule.getProcesses(), currentInputs));
  }

  private Map<NodeVariable, Declaration> getPhaseVariables(
      final Collection<ModuleProcess> processes) {
    final Map<NodeVariable, Declaration> result = new HashMap<>();
    for (ModuleProcess process : processes) {
      if (process.hasMetaInfo(MetaInfoType.PHASE_VAR_NAME)) {
        final String phaseVarName = process.getMetaInfo(MetaInfoType.PHASE_VAR_NAME);
        final NodeVariable phaseVar = process.getVariable(phaseVarName);
        final Declaration declaration = process.getDeclaration(phaseVar);
        result.put(phaseVar, declaration);
      }
    }
    return result;
  }

  @Override
  public void onModuleEnd(final Module module) {
    appendLine();
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    setCurrentProcess((HlddProcess) process);
    final String moduleName = SmvPrinterUtils.getFormattedName(process.getName());
    appendLine();
    // Print process.
    appendLine(SmvPrinterUtils.getModuleDeclaration(moduleName, currentInputs));
    appendLine();
    // Declare variables.
    appendLine(
        SmvPrinterUtils.getVarDeclareSection(getUsedVariables((HlddProcess) process), exprPrinter));
    appendLine(SmvExprPrinter.ASSIGNMENT_DECLARATION);
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {
    appendLine(MapUtils.printMaps(exprPrinter));
    MapUtils.resetMaps();
  }

  @Override
  public void onHlddBegin(final Hldd hldd) {
    if (!hldd.getVariable().isType(DataTypeId.MAP)) {
      final String name = SmvPrinterUtils.getName(hldd.getVariable());
      if (hldd.isImmediateAssignment()) {
        append(String.format("  %s := ", name));
      } else {
        append(String.format("  next(%s) := ", name));
      }
    }
    hlddVariable = hldd.getVariable();
  }

  @Override
  public void onHlddEnd(final Hldd hldd) {
    appendLine();
  }

  @Override
  public void onConditionBegin(final HlddConditionNode node) {
    if (!isConstantGuard(node)) {
      indentationLevel++;
      appendLine(String.format("%n%scase", SmvPrinterUtils.getWhitespace(indentationLevel)));
    }
  }

  /*
   * Checks whether the specified condition node represents a constant expression.
   *
   * @param node - the specified node.
   * @return {@code true} if the specified node contains a NodeValue, {@code false} otherwise.
   */
  private boolean isConstantGuard(final HlddConditionNode node) {
    final Node guard = node.getExpression();
    for (final Hldd hldd : currentProcess.getHldds()) {
      if (hldd.getVariable().equals(guard)) {
        final Node expr =
            ((HlddTerminalNode) hldd.getTerminalNodes().iterator().next()).getAssignedValue();
        if (expr instanceof NodeValue) {
          return true;
        }
      }
    }
    return false;
  }

  @Override
  public void onConditionEnd(final HlddConditionNode node) {
    if (!node.hasDefaultEdge()) {
      appendLine(String.format("%s  TRUE : %s;", SmvPrinterUtils.getWhitespace(indentationLevel),
          SmvPrinterUtils.getName(hlddVariable)));
    }

    if (!isConstantGuard(node)) {
      appendLine(String.format("%sesac;", SmvPrinterUtils.getWhitespace(indentationLevel)));
      indentationLevel--;
    }
  }

  @Override
  public void onEdgeBegin(final HlddEdge node) {

    final HlddConditionNode parent = (HlddConditionNode) node.getParent();

    if (!isConstantGuard(parent)) {
      // If parent condition is not a constant.
      indentationLevel++;
      if (node.isDefault()) {
        append(String.format("%sTRUE : ", SmvPrinterUtils.getWhitespace(indentationLevel)));
      } else {
        final Node finalExpression = SmvPrinterUtils.makeSubstitutions(node.getConditionEquation());
        append(String.format("%s%s : ", SmvPrinterUtils.getWhitespace(indentationLevel),
            getExprDescription(finalExpression)));
      }
    } else {
      // If parent condition is a constant and the edge contains "false".
      if (!((NodeValue) node.getExpressions().iterator().next()).getBoolean()) {
        printingEnabled = false;
        disablingNode = parent;
        Log.printWarn("Unreachable path detected. Variable: " + parent.getExpression().toString());
      }
    }
  }

  @Override
  public void onEdgeEnd(final HlddEdge node) {
    final HlddConditionNode parent = (HlddConditionNode) node.getParent();
    if (!isConstantGuard(parent)) {
      indentationLevel--;
    } else {
      if (parent.equals(disablingNode)) {
        printingEnabled = true;
      }
    }
  }

  @Override
  public void onAssignment(final HlddTerminalNode node) {
    final Node expression = node.getAssignedValue().deepCopy();
    if (expression.getKind() == Node.Kind.VALUE) {
      appendLine(getExprDescription(coerceTypes(expression)) + ";");
    } else {
      final Node finalExpression = SmvPrinterUtils.makeSubstitutions(expression);
      appendExpression(getExpressionType(finalExpression).getTypeId(),
          getExprDescription(finalExpression));
    }
  }

  /*
   * Coerces expression type to the target variable type.
   */
  private Node coerceTypes(final Node expression) {
    final DataType targetDataType = hlddVariable.getDataType();
    if (!expression.getDataType().equals(targetDataType)
        && expression.getDataTypeId() != DataTypeId.MAP) {
      final Node coercedExpr = TypeConversion.coerce(expression, targetDataType);
      if (coercedExpr != null) {
        return coercedExpr;
      }
    }
    return expression;
  }

  /*
   * Returns the specified expressions data type if possible.
   */
  private DataType getExpressionType(final Node expression) {
    DataType exprType = DataType.UNKNOWN;
    try {
      exprType = expression.getDataType();
    } catch (IllegalStateException | NullPointerException e) {
      exprType = DataType.UNKNOWN;
    } finally {
      if (exprType == null) {
        exprType = DataType.UNKNOWN;
      }
    }
    return exprType;
  }

  /*
   * Prints the specified expression description and adds the type conversion expression if needed.
   */
  private void appendExpression(final DataTypeId exprType, final String exprDescription) {
    if (conversionNeeded(exprType)) {
      appendLine(String.format("%s(%s);",
          SmvPrinterUtils.getTypeConversionPrefix(exprType, hlddVariable.getDataTypeId()),
          exprDescription));
    } else {
      appendLine(exprDescription + ";");
    }
  }

  /*
   * Shows whether the specified type needs conversion to target variable type.
   */
  private boolean conversionNeeded(final DataTypeId exprType) {
    return exprType != hlddVariable.getDataTypeId() && exprType != DataTypeId.UNKNOWN;
  }

  /*
   * Searches for variable declarations.
   */
  private Map<NodeVariable, Declaration> getUsedVariables(final HlddProcess process) {
    final Map<NodeVariable, Declaration> variables = new HashMap<>();
    for (final NodeVariable var : process.getAssignedVariables()) {
      if (process.declares(var)) {
        if (!currentPhaseVars.containsKey(var)) {
          variables.put(var, process.getDeclaration(var));
        }
      } else if (currentModule.declaresOutput(var.getName())) {
        variables.put(var, currentModule.getDeclaration(var));
      }
    }
    return variables;
  }


}
