/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.assertion;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.hldd.printer.smv.HlddSmvVisitor;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvExprPrinter;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvPrinterUtils;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.efsm.GuardedAction;
import ru.ispras.retrascope.result.assertion.AssertionContainer;
import ru.ispras.retrascope.result.assertion.AssertionGroup;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

/**
 * This class represents a visitor which contains SMV printing methods.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddAssertSmvVisitor extends HlddSmvVisitor {

  // Model assertions.
  private Map<String, AssertionContainer> assertions;
  // Current assertion container (module).
  private AssertionContainer curAssertContainer;
  // Current assertion group (process).
  private AssertionGroup curAssertGroup;

  private Map<NodeVariable, Node> inputs;

  private boolean printCtl;

  private NodeVariable currentPhaseVar;

  /**
   * Constructs a new visitor and creates a new *.smv file.
   *
   * @param fileName - the specified *.smv file name.
   */
  public HlddAssertSmvVisitor(final String fileName) {
    super(fileName);
    assertions = new HashMap<>();
    printCtl = false;
    inputs = new HashMap<>();
  }

  /**
   * Adds model assertions to the visitor.
   *
   * @param assertions - the specified assertions.
   */
  public void addAssertions(final Map<String, AssertionContainer> assertions) {
    this.assertions = assertions;
  }

  /**
   * Sets the type of specifications to be checked.
   *
   * @param method The specification method.
   */
  public void setSpecType(final String method) {
    switch (method) {
      case "bdd":
        printCtl = true;
        break;
      case "bmc":
        printCtl = false;
        break;
      default:
        throw new RetrascopeRuntimeException("Unknown model checking method: " + method);
    }
  }

  @Override
  public void onModuleBegin(final Module hlddModule) {
    super.onModuleBegin(hlddModule);

    setCurAssertContainer(hlddModule.getName());

    if (curAssertContainer != null && curAssertContainer.getClock() != null) {
      final NodeVariable clock = curAssertContainer.getClock();
      append(SmvPrinterUtils.getClockAssignment(clock, currentPhaseVars.keySet()));
      appendLine();

      final Collection<NodeVariable> inputs = hlddModule.getInputs();
      inputs.remove(clock);
      append(SmvPrinterUtils.getInputConstraints(inputs, clock));
      appendLine();
    }

    for (final NodeVariable input : hlddModule.getInputs()) {
      inputs.put(input, hlddModule.getDeclaration(input).getInitialValue());
    }
  }

  /*
   * Searches for AssertionContainer using the specified key name. If found, assigns the value to
   * the corresponding variable, assigns null otherwise.
   */
  private void setCurAssertContainer(final String name) {
    if (assertions.containsKey(name)) {
      curAssertContainer = assertions.get(name);
    } else {
      curAssertContainer = null;
    }
  }

  @Override
  public void onProcessBegin(final ModuleProcess process) {
    super.onProcessBegin(process);
    setCurAssertGroup(process.getName());
    if (process.hasMetaInfo(MetaInfoType.PHASE_VAR_NAME)) {
      final String phaseName = process.getMetaInfo(MetaInfoType.PHASE_VAR_NAME);
      currentPhaseVar = process.getVariable(phaseName);
    } else {
      currentPhaseVar = null;
    }
  }

  /*
   * Sets the current assertion group.
   */
  private void setCurAssertGroup(final String name) {
    if (curAssertContainer != null) {
      final Map<String, AssertionGroup> groups = curAssertContainer.getGroupMap();
      if (groups.containsKey(name)) {
        curAssertGroup = groups.get(name);
      } else {
        curAssertGroup = null;
      }
    }
  }

  @Override
  public void onProcessEnd(final ModuleProcess process) {
    if (curAssertGroup != null) {
      printAssertions();
    }
    super.onProcessEnd(process);
  }

  /*
   * Prints the current assertion group.
   */
  private void printAssertions() {
    final String assertionDecl =
        printCtl ? SmvExprPrinter.CTL_SPEC_DECLARATION : SmvExprPrinter.LTL_SPEC_DECLARATION;
    final String futureOperator = printCtl ? "EF" : "F";
    for (final Node node : curAssertGroup.getAssertions()) {
      append(assertionDecl);
      final String reachCond = String.format("(%s %s)", futureOperator,
          getExprDescription(SmvPrinterUtils.makeSubstitutions(node)));
      final List<GuardedAction> rst = curAssertGroup.getRstGuardedAction();


      if (rst == null) {
        appendLine(String.format("  ! %s;", reachCond));
      } else {
        final List<Node> guards = new ArrayList<>();
        for (GuardedAction guardedAction : rst) {
          final Node guard = guardedAction.getGuardNode();
          final Set<NodeVariable> uses = guardedAction.getGuardUses();
          if (currentPhaseVar == null || !uses.contains(currentPhaseVar) || (
              uses.contains(currentPhaseVar) && uses.size() > 1)) {
            guards.add(guard);
          }
        }
        final Node resetGuardNode =
            ExprUtils.getConjunction(guards.toArray(new Node[guards.size()]));
        Node initStateNode = curAssertGroup.getInitValueState();
        final Node resetNode;
        if (initStateNode.getKind() == Node.Kind.VALUE && initStateNode.isType(DataType.BOOLEAN)) {
          resetNode = resetGuardNode;
        } else {
          resetNode = ExprUtils.getConjunction(initStateNode, resetGuardNode);
        }

        final String rstCond =
            String.format("(%s)", getExprDescription(SmvPrinterUtils.makeSubstitutions(resetNode)));

        if (printCtl) {
          final String inputsCond =
              getExprDescription(SmvPrinterUtils.makeSubstitutions(getInitialInputsCondition()));
          appendLine(String.format(" ! E [(%s) U (%s & %s)];", inputsCond, rstCond, reachCond));
        } else {
          appendLine(String.format(" ! ( X %s & %s);", rstCond, reachCond));

        }
      }
    }
    appendLine();
  }

  private Node getInitialInputsCondition() {
    final Collection<Node> equations = new ArrayList<>();
    for (final Entry<NodeVariable, Node> var : inputs.entrySet()) {
      if (var.getValue() != null) {
        equations.add(new NodeOperation(StandardOperation.EQ, var.getKey(), var.getValue()));
      }
    }
    return ExprUtils.getConjunction(equations.toArray(new Node[equations.size()]));
  }

}
