/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.hldd.printer.smv.utils;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.Node.Kind;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.DeclarationUtils;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This class contains static methods for *.smv file printing.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvPrinterUtils {

  /**
   * SMV language reserved words.
   */
  private static final String[] RESERVED_WORDS =
      {"A", "ABF", "ABG", "AF", "AG", "array", "ASSIGN", "AX", "bool", "boolean", "BU", "case",
          "COMPASSION", "COMPID", "COMPUTE", "COMPWFF", "CONSTANTS", "CONSTRAINT", "count",
          "CTLSPEC", "CTLWFF", "DEFINE", "E", "EBF", "EBG", "EF", "EG", "esac", "EX", "extend", "F",
          "FAIRNESS", "FALSE", "floor", "FROZENVAR", "FUN", "G", "H", "IN", "in", "INIT", "init",
          "Integer", "integer", "INVAR", "INVARSPEC", "ISA", "ITYPE", "IVAR", "JUSTICE", "LTLSPEC",
          "LTLWFF", "MAX", "max", "MDEFINE", "MIN", "MIRROR", "mod", "MODULE", "NAME", "next",
          "NEXTWFF", "O", "of", "PRED", "PREDICATES", "PSLSPEC", "READ", "Real", "real", "resize",
          "S", "self", "signed", "SIMPWFF", "sizeof", "SPEC", "swconst", "T", "toint", "TRANS",
          "TRUE", "U", "union", "unsigned", "uwconst", "V", "VAR", "Word", "word", "word1", "WRITE",
          "X", "xnor", "xor", "Y", "Z"};

  /* Statement end with double new line. */
  private static final String ENDLINE = ";\n\n";
  private static final String BASE = "  ";

  private SmvPrinterUtils() {
    /* Empty. */
  }

  /**
   * Generates a module declaration in SMV format.
   *
   * @param name   - the specified module name.
   * @param inputs - the specified collection of parameters.
   *
   * @return The string containing the module declaration.
   */
  public static String getModuleDeclaration(final String name,
      final Collection<NodeVariable> inputs) {

    final StringBuilder result = new StringBuilder();
    result.append(String.format("%s %s", SmvExprPrinter.MODULE_DECLARATION, name));
    if (!inputs.isEmpty()) {
      result.append(String.format("(%s)", getModuleParametersDescription(inputs)));
    }
    result.append("\n");
    return result.toString();
  }

  /**
   * Generates a module declaration in SMV format with empty parameter list.
   *
   * @param name - the specified module name.
   *
   * @return The string containing the module declaration.
   */
  public static String getModuleDeclaration(final String name) {
    return getModuleDeclaration(name, new ArrayList<NodeVariable>());
  }

  /**
   * Generates a module instantiation in SMV format.
   *
   * @param instanceName - the specified module instance name.
   * @param moduleName   - the specified module name.
   * @param parameters   - the specified collection of parameters.
   *
   * @return The string containing the module instantiation.
   */
  public static String getModuleInstantiation(final String instanceName, final String moduleName,
      final Collection<NodeVariable> parameters) {
    return String
        .format("%s%n%s : %s(%s);%n", SmvExprPrinter.VARIABLE_DECLARATION, instanceName, moduleName,
            getModuleParametersDescription(parameters));
  }

  /**
   * Generates process instantiations in SMV format.
   *
   * @param processes      - the specified process collection.
   * @param inputVariables - the specified input variable collection.
   *
   * @return The string containing the process instantiations.
   */
  public static String getProcessInstantiation(final Collection<ModuleProcess> processes,
      final Collection<NodeVariable> inputVariables) {
    final StringBuilder builder = new StringBuilder();
    for (final ModuleProcess process : processes) {
      final String processName = SmvPrinterUtils.getFormattedName(process.getName());
      final String instanceName = "  m_" + processName;
      builder.append(
          SmvPrinterUtils.getModuleInstantiation(instanceName, processName, inputVariables));
    }
    return builder.toString();
  }

  /*
   * Generates a list of module parameters from a collection of variables.
   *
   * @param parameters - the specified collection of variables.
   *
   * @return The string containing the list of module parameters.
   */
  private static String getModuleParametersDescription(final Collection<NodeVariable> parameters) {
    final StringBuilder result = new StringBuilder();
    final ArrayList<NodeVariable> variables = new ArrayList<>(parameters);
    for (int j = 0; j < variables.size(); j++) {
      if (j != 0) {
        result.append(", ");
      }
      result.append(getName(variables.get(j)));
    }
    return result.toString();
  }

  /**
   * Checks whether the variable name is acceptable by SMV syntax.
   *
   * @param variable - the specified variable.
   *
   * @return The modified variable name if it is forbidden by SMV, lower case variable name
   *     otherwise.
   */
  public static String getName(final NodeVariable variable) {
    final String varName = variable.getName();
    if (Arrays.asList(RESERVED_WORDS).contains(varName)) {
      return varName + SmvExprPrinter.SUB_ENDING;
    }
    return varName;
  }

  /**
   * Generates a VAR declaration section in SMV format from a collection of variables.
   *
   * @param variables - the variable-declaration mapping
   * @param printer   - the specified {@link SmvExprPrinter}.
   *
   * @return The string containing the VAR section.
   */
  public static String getVarDeclareSection(final Map<NodeVariable, Declaration> variables,
      final SmvExprPrinter printer) {
    return getVarsDescription(variables, SmvExprPrinter.VARIABLE_DECLARATION, printer);
  }

  /*
   * Generates a variable declaration in SMV format.
   *
   * @param variable - the specified variable.
   *
   * @param declaration - the variable declaration.
   *
   * @param printer - the specified {@link SmvExprPrinter}.
   *
   * @return The string containing the variable declaration.
   */
  private static String getVariableDeclaration(final NodeVariable variable,
      final Declaration declaration, final SmvExprPrinter printer, final String keyword) {
    final StringBuilder result = new StringBuilder();
    final String varName = getName(variable);
    result.append(String.format("%s%n", keyword));
    result.append(
        String.format("  %s : %s%s", varName, printer.getType(variable.getDataType()), ENDLINE));

    if (declaration != null && declaration.containsInvariant()) {
      result.append("INVAR\n");
      final Node invariant =
          makeSubstitutions(DeclarationUtils.getInvariant(declaration, variable));
      final String exprDescription =
          printer.toString(invariant).replaceAll("true", "TRUE").replaceAll("false", "FALSE");
      result.append(String.format("  %s%s", exprDescription, ENDLINE));
    }

    if (declaration != null && declaration.containsInitialValue()) {
      result.append(SmvExprPrinter.ASSIGNMENT_DECLARATION + "\n");
      result.append(String
          .format("  init(%s) := %s%s", varName, printer.toString(declaration.getInitialValue()),
              ENDLINE));
    }


    return result.toString();
  }

  /*
   * Generates a declaration section in SMV format from variable collection.
   *
   * @param variables
   *
   * @param keyword - the specified declaration section keyword.
   *
   * @param printer
   *
   * @return The string containing the declaration section.
   */
  private static String getVarsDescription(final Map<NodeVariable, Declaration> variables,
      final String keyword, final SmvExprPrinter printer) {
    final StringBuilder result = new StringBuilder();
    result.append("\n");
    for (Entry<NodeVariable, Declaration> entry : variables.entrySet()) {
      result.append(getVariableDeclaration(entry.getKey(), entry.getValue(), printer, keyword));
    }
    return result.toString();
  }

  /**
   * Returns the indentation space according to the specified indentation level.
   *
   * @param level - the specified indentation level.
   *
   * @return The whitespace string.
   */
  public static String getWhitespace(final int level) {
    StringBuilder builder = new StringBuilder();
    for (int i = 0; i < level; i++) {
      builder.append(BASE);
    }
    return builder.toString();
  }

  /*
   * Searches for power operations in the node and replaces them by arithmetic left shift operation.
   *
   * @param node - the specified {@link Node}.
   *
   * @return The node with substitution performed.
   */
  private static Node substitutePower(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();

    final TransformerRule powerRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node arg0) {
        return ExprUtils.isOperation(arg0, StandardOperation.POWER);
      }

      @Override
      public Node apply(final Node arg0) {
        return new NodeOperation(StandardOperation.BVASHL, new NodeValue(Data.newInteger(1)),
            ((NodeOperation) arg0).getOperand(1));
      }
    };

    transformer.addRule(StandardOperation.POWER, powerRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }

  private static Node substituteBvrepeat(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();

    final TransformerRule bvrepeatRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node arg0) {
        return ExprUtils.isOperation(arg0, StandardOperation.BVREPEAT);
      }

      @Override
      public Node apply(final Node arg0) {
        final NodeOperation operation = (NodeOperation) arg0;
        final Node pattern = operation.getOperand(1);
        Node result = pattern;
        if (operation.getOperand(0).getKind() == Kind.VALUE) {
          final NodeValue repeats = (NodeValue) TransformerUtils.toInt(operation.getOperand(0));
          assert repeats != null;
          for (int i = 0; i < repeats.getInteger().intValue() - 1; i++) {
            result = new NodeOperation(StandardOperation.BVCONCAT, pattern, result);
          }
        }
        return result;
      }
    };

    transformer.addRule(StandardOperation.BVREPEAT, bvrepeatRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }

  private static Node substituteNor(final Node node) {
    final NodeTransformer transformer = new NodeTransformer();

    final TransformerRule norRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node arg0) {
        return ExprUtils.isOperation(arg0, StandardOperation.BVNORR) || ExprUtils
            .isOperation(arg0, StandardOperation.BVNOR);
      }

      @Override
      public Node apply(final Node arg0) {
        final NodeOperation nor = (NodeOperation) arg0;
        final List<Node> operands = nor.getOperands();
        final StandardOperation op = ExprUtils.isOperation(arg0, StandardOperation.BVNORR)
            ? StandardOperation.BVORR
            : ExprUtils.isOperation(arg0, StandardOperation.BVNOR) ? StandardOperation.BVOR : null;
        final NodeOperation or = new NodeOperation(op, operands);
        return new NodeOperation(StandardOperation.BVNOT, or);
      }
    };

    transformer.addRule(StandardOperation.BVNORR, norRule);
    transformer.addRule(StandardOperation.BVNOR, norRule);
    transformer.walk(node);
    return transformer.getResult().iterator().next();
  }


  /**
   * Performs the following substitutions on the specified Node:
   * <ul>
   * <li>Substitutes a power operation by an arithmetic left shift operation.
   * <li>If a map value is found, a corresponding variable is created in the declarations section
   * and the value is replaced by this variable.
   * <li>If a BVEXTRACT operation extracts only one bit it is replaced by a SELECT operation.
   * </ul>
   *
   * @param node - the specified node.
   *
   * @return The node with the substitutions made.
   */
  public static Node makeSubstitutions(final Node node) {
    Node subNode = substitutePower(node);
    subNode = MapUtils.substituteMaps(subNode);
    subNode = substituteBvrepeat(subNode);
    subNode = substituteNor(subNode);
    return subNode;
  }

  /**
   * Returns the type conversion string from the source type into the target type if possible.
   *
   * @param sourceType - the specified source type.
   * @param targetType - the specified target type.
   *
   * @return Type conversion expression in SMV format if conversion is possible, empty string
   *     otherwise.
   */
  public static String getTypeConversionPrefix(final DataTypeId sourceType,
      final DataTypeId targetType) {
    String result = "";
    if (targetType == DataTypeId.BIT_VECTOR && sourceType == DataTypeId.LOGIC_BOOLEAN) {
      result = "word1";
    }
    if (targetType == DataTypeId.LOGIC_BOOLEAN && sourceType == DataTypeId.BIT_VECTOR) {
      result = "bool";
    }
    if (targetType == DataTypeId.LOGIC_INTEGER && (sourceType == DataTypeId.LOGIC_BOOLEAN
        || sourceType == DataTypeId.BIT_VECTOR)) {
      result = "toint";
    }
    return result;
  }

  /**
   * Generates clock assignment.
   *
   * @param clock - the specified clock variable.
   *
   * @return Clock assignment in SMV format.
   */
  public static String getClockAssignment(final NodeVariable clock,
      final Collection<NodeVariable> phaseVars) {
    final StringBuilder result = new StringBuilder();
    final String clockName = getName(clock);
    result.append(SmvExprPrinter.ASSIGNMENT_DECLARATION + "\n");
    if (phaseVars.size() == 1) {
      final NodeVariable phaseVar = phaseVars.iterator().next();
      final String phaseName = phaseVar.getName();
      final int size = phaseVar.getDataType().getSize();
      result.append(String.format("  next(%s) :=%n", clockName));
      result.append("    case\n");
      result.append(String.format("      (%s = 0d%d_0) : !%s;%n", phaseName, size, clockName));
      //result.append(String.format("      (next(%s) = 0d%d_0) : 0d1_0;%n", phaseName, size));
      result.append(String.format("      TRUE : %s;%n", clockName));
      result.append("    esac;\n");
    } else {
      result.append(String.format("  next(%s) := ! %s;%n", clockName, clockName));
    }
    return result.toString();
  }

  /**
   * Converts a string of process identifiers into a single identifier.
   *
   * @param name The string of process identifiers.
   */
  public static String getFormattedName(final String name) {
    String result = name.replaceAll("[\\s.,(){}:=\\-\\[\\]@]", "_");
    result = result.replaceAll("_+", "_");
    return result;
  }

  /**
   * Returns a list of registers of the specified module.
   *
   * @param module - the specified {@link Module}.
   *
   * @return The list of register variables.
   */
  public static List<NodeVariable> getRegs(final Module module) {
    final List<NodeVariable> result = module.getVariables();
    result.removeAll(module.getInputs());
    result.removeAll(module.getOutputs());
    return result;
  }

  /**
   * Generates input variable constraints that prohibit value changes while clock = 1.
   *
   * @param inputs - the collection of input variables.
   * @param clock  - the clock variable.
   *
   * @return The string containing a set of constraints.
   */
  public static String getInputConstraints(final Collection<NodeVariable> inputs,
      final NodeVariable clock) {
    final StringBuilder result = new StringBuilder();
    final String clockName = clock.getName();
    for (NodeVariable variable : inputs) {
      final String inputName = getName(variable);
      result.append("TRANS\n");
      result.append(String
          .format("  (%s = 0d1_1 & next(%s) != 0d1_0) -> (next(%s) = %s)%n", clockName, clockName,
              inputName, inputName));
      result.append("TRANS\n");
      result.append(String
          .format("  (%s = 0d1_0 & next(%s) != 0d1_1) -> (next(%s) = %s)%n", clockName, clockName,
              inputName, inputName));
      result.append("\n");
    }
    return result.toString();
  }

}
