/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.smv.launcher;

import org.apache.commons.cli.ParseException;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.FileNames;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.hldd.printer.smv.assertion.HlddAssertSmvPrinter;
import ru.ispras.retrascope.engine.smv.test.parser.parsers.SmvSpecificationParser;
import ru.ispras.retrascope.engine.smv.test.parser.parsers.SmvStatisticsParser;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.SmvChecker;
import ru.ispras.retrascope.util.SmvChecker.Result;

import java.io.File;
import java.util.Map;

/**
 * This engine checks the generated *.smv files with the model checker.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvModelCheckerLauncher extends Engine {

  /** The engine ID. */
  public static final String ENGINE_ID = "smv-modelchecker-launcher";

  /** The engine output entity type. */
  public static final EntityType OUT_TYPE = EntityType.get("smv-launch");

  /** NuXMV log file name. */
  private static final String NUXMV_LOG_FILE = "nuxmv_traces.log";

  // Command file for NuXMV.
  private static final String BDD_COMMANDS = "nuxmv/nuxmv_commands_bdd.txt";
  private static final String BMC_COMMANDS = "nuxmv/nuxmv_commands_bmc.txt";

  private static final Parameter CHECK_METHOD = new Parameter("check-method", new String[] {"bdd",
      "bmc"}, "Model checking method");

  private int counterexamples;

  /**
   * Constructs a new engine object.
   */
  public SmvModelCheckerLauncher() {
    super(ENGINE_ID);
    addParameter(CHECK_METHOD);
    addInputType(HlddAssertSmvPrinter.OUT_TYPE);
    setOutputType(OUT_TYPE);
    counterexamples = 0;
  }

  @Override
  public Entity start(Map<EntityType, Entity> inputs) {

    final FileNames result = new FileNames(OUT_TYPE.getId());
    final SmvChecker checker = SmvChecker.getChecker();
    checker.setNuXmv();

    try {
      final FileNames smvFiles = (FileNames) inputs.get(HlddAssertSmvPrinter.OUT_TYPE);
      final Arguments arguments =
          this.parseCommandLine(((Arguments) inputs.get(Arguments.TYPE)).getCommandLine());
      final String checkType = arguments.getValue(CHECK_METHOD, "bmc").toLowerCase();
      String commands;

      switch (checkType) {
        case "bdd":
          commands = BDD_COMMANDS;
          break;
        case "bmc":
          commands = BMC_COMMANDS;
          break;
        default:
          throw new RetrascopeRuntimeException("Unknown model checking method: " + checkType);
      }

      final String commandFile =
          new File(this.getClass().getClassLoader().getResource(commands).getFile())
              .getAbsolutePath();

      for (String file : smvFiles.getArray()) {
        checker.runChecker(file, "-source", commandFile);
        if (checker.getResult() == Result.OK && checker.getLogFile() != null) {
          result.add(NUXMV_LOG_FILE);
        }
        Log.printInfo("Checker result: " + checker.getResult().name());
        if (checker.getResult() == Result.OK) {
          printLaunchInformation(file);
        }
      }

    } catch (final ParseException e) {
      throw new RetrascopeRuntimeException(e);
    }

    return result;
  }

  public int getCounterexampleCount() {
    return counterexamples;
  }

  /*
   * Prints model checking information.
   */
  private void printLaunchInformation(final String fileName) {
    final SmvSpecificationParser specParser = new SmvSpecificationParser();
    final SmvStatisticsParser statParser = new SmvStatisticsParser();
    statParser.parseStats(fileName + "log");
    specParser.parseSpecifications(fileName);
    Log.printInfo("Specifications total: " + specParser.getOverallSpecs());
    Log.printInfo("Counterexamples generated: " + statParser.getCounterexamples());
    Log.printInfo("Specifications proven TRUE: " + statParser.getTrueSpecs());
    counterexamples = statParser.getCounterexamples();
  }

}
