/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.smv.test.parser.parsers;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.engine.hldd.printer.smv.utils.SmvExprPrinter;
import ru.ispras.retrascope.model.basis.DeclarationContainer;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.Transaction;
import ru.ispras.retrascope.result.test.Vector;

import java.math.BigInteger;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;

/**
 * This class is used to parse a SMV log file and construct a {@link Test} object.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class SmvCounterexampleParser extends LogParser {

  private Test test;
  private Sequence sequence;
  private Vector vector;
  private Vector previousVector;
  private Map<String, DataType> inputs;
  private EventList events;
  private final Set<NodeVariable> sensitivityVars;


  /**
   * Constructs a new parser object.
   */
  public SmvCounterexampleParser() {
    super();
    inputs = new HashMap<>();
    sensitivityVars = new HashSet<>();
  }

  // TODO: Extract inputs from the hldd model.

  /**
   * Sets the interface which would be used during parsing.
   *
   * @param iface - the specified interface.
   */
  public void setSignalInterface(final CfgInterface iface) {
    InvariantChecks.checkNotNull(iface);
    inputs = new HashMap<>();
    for (Map.Entry<String, DeclarationContainer> entry : iface.getMapping().entrySet()) {
      for (NodeVariable input : entry.getValue().getInputs()) {
        inputs.put(input.getName(), input.getDataType());
      }
    }
  }

  /**
   * Parses a file for counterexamples.
   *
   * @param fileName - the specified log file name.
   * @return The {@link Test} object extracted from the file.
   */
  public Test parseCounterexamples(final String fileName) {
    test = new Test();
    parse(fileName);
    addVector();
    addSequence();
    return test;
  }

  @Override
  protected void parseLine(final String line) {
    final Matcher commentMatcher = comment.matcher(line);
    final Matcher specMatcher = counterexample.matcher(line);
    final Matcher stateMatcher = state.matcher(line);
    final Matcher valueMatcher = value.matcher(line);

    if (!line.isEmpty() && !commentMatcher.find()) {
      if (specMatcher.find()) {
        addVector();
        addSequence();
        sequence = new Sequence();
        vector = null;
        previousVector = null;
      }

      if (stateMatcher.find()) {
        addVector();
        previousVector = vector;
        vector = new Vector();
      }

      if (valueMatcher.find()) {
        addTransaction(line);
      }
    }
  }

  /*
   * Adds a new transaction to the test.
   */
  private void addTransaction(final String line) {
    final Transaction transaction = createTransaction(line);
    if (!transaction.isEmpty()) {
      vector.addTransaction(transaction);
    }
  }

  /*
   * Adds a new vector to the test.
   */
  private void addVector() {
    if (vector != null && !vector.isEmpty()) {
      final EventList eventList = extractEvents();
      if (!eventList.isEmpty()) {
        vector.addEvents(eventList);
      }
      final Collection<Transaction> prevTransactions =
          previousVector != null ? previousVector.getTransactions() : null;
      if (!(vector.getTransactions()).equals(prevTransactions)) {
        sequence.addVector(vector);
      }
    }
  }

  /*
   * Adds a new sequence to the test.
   */
  private void addSequence() {
    if (sequence != null && sequence.getLength() != 0 && !test.getSequences().contains(sequence)) {
      test.addSequence(sequence);
    }
  }

  /*
   * Creates a new transaction from the specified string.
   */
  private Transaction createTransaction(final String line) {
    final String[] elements = line.split(" = ");
    String varName = elements[0];
    final String value = elements[1];
    final Transaction result = new Transaction();
    varName = parseName(varName);
    if (varName != null) {
      result.addVariable(varName, parseValue(value, inputs.get(varName)));
    }
    return result;
  }

  /*
   * Parses a variable name.
   */
  private String parseName(final String varName) {
    String result;
    String[] values = varName.split("\\.");
    if (values.length == 1) {
      result = values[0];
      result = result.replaceAll("\\s", "");
      if (result.endsWith(SmvExprPrinter.SUB_ENDING)) {
        result = result.substring(0, result.length() - SmvExprPrinter.SUB_ENDING.length());
      }
      if (!inputs.containsKey(result)) {
        return null;
      }
    } else {
      return null;
    }
    return result;
  }

  /*
   * Parses a variable value.
   */
  private NodeValue parseValue(final String string, final DataType dataType) {
    NodeValue result = null;
    String tmp = string;
    String sign = "";
    if (tmp.matches("TRUE|FALSE")) {
      result = new NodeValue(Data.newBoolean(Boolean.valueOf(tmp)));
    }
    if (tmp.matches("\\d+")) {
      result = new NodeValue(Data.newInteger(Integer.parseInt(tmp)));
    }
    if (tmp.matches("-\\w+")) {
      tmp = tmp.substring(1);
      sign = "-";
    }
    if (tmp.matches("0ud\\d+_\\d+") || tmp.matches("0sd\\d+_\\d+")) {
      final String[] values = tmp.split("_");
      final String size = values[0].substring(3, values[0].length());
      final String value = sign + values[1];
      final int intSize = Integer.parseInt(size);
      final long intValue = Long.parseLong(value);
      if (dataType.getTypeId() == DataTypeId.LOGIC_INTEGER) {
        result = new NodeValue(Data.newInteger(intValue));
      } else {
        result = new NodeValue(Data.newBitVector(intValue, intSize));
      }
    }
    return result;
  }

  /*
   * Transforms values of the variables from the sensitivity list into events.
   */
  private EventList extractEvents() {
    InvariantChecks.checkNotEmpty(sensitivityVars);
    final EventList result = new EventList();

    for (NodeVariable var : sensitivityVars) {
      final NodeValue value = (NodeValue) vector.getValue(var.getName());
      final NodeValue previousValue =
          previousVector == null ? null : (NodeValue) previousVector.getValue(var.getName());
      final BigInteger intValue = getIntValue(value);
      final BigInteger prevIntValue = getIntValue(previousValue);

      if (value != null && intValue != null) {
        if (intValue.intValue() == 1) {

          // First "1" value is considered as posedge.
          if (prevIntValue == null || prevIntValue.intValue() == 0) {
            final Event event = new Event(new RangedVariable(var), EventType.POSITIVE_EDGE);
            result.add(event);
          }
        } else if (intValue.intValue() == 0) {

          // First "0" value is not considered as edge.
          if (prevIntValue != null && prevIntValue.intValue() == 1) {
            final Event event = new Event(new RangedVariable(var), EventType.NEGATIVE_EDGE);
            result.add(event);
          }
        }
      }
    }
    return result;
  }

  private BigInteger getIntValue(final NodeValue value) {
    if (value == null) {
      return null;
    }
    switch (value.getDataTypeId()) {
      case BIT_VECTOR:
        return value.getBitVector().bigIntegerValue(false);
      case LOGIC_INTEGER:
        return value.getInteger();
      default:
        return null;
    }
  }

  /**
   * Returns the {@link EventList} associated with the parser.
   *
   * @return The associated {@link EventList}.
   */
  public EventList getEventList() {
    return events;
  }

  /**
   * Associates the specified {@link EventList} with the parser.
   *
   * @param events - the specified {@link EventList}.
   */
  public void setEventList(final EventList events) {
    this.events = events;
    for (Event event : events) {
      sensitivityVars.add(event.getVariable());
    }
  }
}
