/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.test.printer.testbench;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Vector;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.ReadWriteUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

/**
 * The class for printing tests to files in the DECIDER-like manner.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
final class DeciderLikeTestPatternPrinter {
  private static final String PATTERN_HEADER = "#Pattern: ";
  private static final String SEQUENCE_HEADER = "#New sequence";
  private static final String SIGNAL_DELIMETER = " ";
  private static final String VALUE_FOR_UNSPECIFIED_SIGNALS = "X";

  private DeciderLikeTestPatternPrinter() {
    /* Empty. */
  }

  /**
   * Prints the specified sequence to the specified file.
   *
   * @param sequence the test sequence to be printed
   * @param file the file where the test will be printed to
   * @param inputNames a set of all input signals of a tested module
   * @param overwrite indicates whether or not the target file should be overwritten if it already
   *        exists
   *
   * @throws IOException if the specified file already exists or if an I/O error occurred
   */
  public static void print(
      final Sequence sequence,
      final Path file,
      final Collection<String> inputNames,
      final boolean overwrite)
      throws IOException {

    InvariantChecks.checkTrue(overwrite || !Files.exists(file));

    final BufferedWriter testWriter = ReadWriteUtils.newBufferedWriter(file);
    InvariantChecks.checkNotNull(testWriter);
    testWriter.append(PATTERN_HEADER);

    for (final Iterator<String> iterator = inputNames.iterator(); iterator.hasNext();) {
      testWriter.append(iterator.next());

      if (iterator.hasNext()) {
        testWriter.append(SIGNAL_DELIMETER);
      }
    }

    testWriter.newLine();
    testWriter.append(SEQUENCE_HEADER);
    testWriter.newLine();

    sequence.initIterator();
    while (sequence.hasNext()) {

      final Vector vector = sequence.next();
      final Map<String, NodeValue> eventValues = getEventValues(vector);
      final Map<String, NodeValue> invertedEventValues = invertEventValues(eventValues);

      final StringBuilder firstHalfBuilder = new StringBuilder();
      final StringBuilder secondHalfBuilder = new StringBuilder();

      for (final Iterator<String> inputIter = inputNames.iterator(); inputIter.hasNext();) {
        final String inputName = inputIter.next();

        if (eventValues.containsKey(inputName)) {
          firstHalfBuilder.append(invertedEventValues.get(inputName).getData().getValue());
          secondHalfBuilder.append(eventValues.get(inputName).getData().getValue());
        } else {
          if (vector.containsVariable(inputName)) {
            final Node value = vector.getValue(inputName);

            if (ExprUtils.isKind(Node.Kind.VALUE, value)) {
              firstHalfBuilder.append(((NodeValue) value).getData().getValue());
              secondHalfBuilder.append(((NodeValue) value).getData().getValue());
            } else {

              throw RetrascopeRuntimeException.makeException(
                  LogMessage.UNSUPPORTED_TYPE,
                  value.getClass());
            }
          } else {
            firstHalfBuilder.append(VALUE_FOR_UNSPECIFIED_SIGNALS);
            secondHalfBuilder.append(VALUE_FOR_UNSPECIFIED_SIGNALS);
          }
        }

        if (inputIter.hasNext()) {
          firstHalfBuilder.append(SIGNAL_DELIMETER);
          secondHalfBuilder.append(SIGNAL_DELIMETER);
        }
      }

      testWriter.append(firstHalfBuilder);
      testWriter.newLine();
      testWriter.append(secondHalfBuilder);
      testWriter.newLine();
    }

    testWriter.flush();
  }

  private static Map<String, NodeValue> getEventValues(Vector vector) {
    Map<String, NodeValue> eventValues = new HashMap<>();

    for (Event event: vector.getEventList()) {
      NodeVariable variable = event.getVariable();
      String variableName = variable.getName();
      DataType dataType = variable.getDataType();

      switch (event.getType()) {
        case NEGATIVE_EDGE:
          if (dataType.getTypeId().equals(DataTypeId.LOGIC_BOOLEAN)) {
            eventValues.put(variableName, SolverUtils.FALSE);
          } else {
            eventValues.put(variableName, new NodeValue(dataType.valueOf("0", 10)));
          }

          break;
        case POSITIVE_EDGE:
          if (dataType.getTypeId().equals(DataTypeId.LOGIC_BOOLEAN)) {
            eventValues.put(variableName, SolverUtils.TRUE);
          } else {
            eventValues.put(variableName, new NodeValue(dataType.valueOf("1", 10)));
          }

          break;
        case ANY_EDGE:
          if (vector.containsVariable(variableName)) {
            eventValues.put(variableName, (NodeValue) vector.getValue(variableName));
          } else {
            if (dataType.getTypeId().equals(DataTypeId.LOGIC_BOOLEAN)) {
              eventValues.put(variableName, SolverUtils.TRUE);
            } else {
              eventValues.put(variableName, new NodeValue(dataType.valueOf("1", 10)));
            }
          }

          break;
        default:
          throw new IllegalArgumentException("Unsupported event type: " + event.getType());
      }
    }

    return eventValues;
  }

  private static Map<String, NodeValue> invertEventValues(Map<String, NodeValue> eventValues) {
    Map<String, NodeValue> invertedEventValues = new HashMap<>();

    for (Entry<String, NodeValue> entry: eventValues.entrySet()) {

      final String varName = entry.getKey();
      final NodeValue value = entry.getValue();
      final Data data = value.getData();
      final DataType dataType = value.getDataType();

      if (ExprUtils.isType(DataType.BOOLEAN, value)) {

        if (data.equals(Data.newBoolean(false))) {

          invertedEventValues.put(varName, SolverUtils.TRUE);

        } else {

          if (data.equals(Data.newBoolean(true))) {

            invertedEventValues.put(varName, SolverUtils.FALSE);

          } else {
            throw new IllegalArgumentException("Unexpected boolean event value: " + value);
          }
        }
      } else if (ExprUtils.isType(DataTypeId.BIT_VECTOR, value)) {

        final int size = dataType.getSize();
        final NodeValue zeroVector = NodeValue.newBitVector(BitVector.valueOf(0, size));
        invertedEventValues.put(varName, zeroVector);

      } else {

        final Data zero = dataType.valueOf("0", 10);
        final Data one = dataType.valueOf("1", 10);

        if (data.equals(zero)) {

          invertedEventValues.put(varName, new NodeValue(one));

        } else {

          if (data.equals(one)) {

            invertedEventValues.put(varName, new NodeValue(zero));

          } else {
            throw new IllegalArgumentException("Unexpected numeric event value: " + value);
          }
        }
      }
    }

    return invertedEventValues;
  }
}
