/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.test.printer.testbench;

import org.apache.commons.cli.ParseException;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Arguments;
import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.Parameter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.DeclarationContainer;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.result.cfginterface.CfgInterface;
import ru.ispras.retrascope.result.test.Sequence;
import ru.ispras.retrascope.result.test.Test;
import ru.ispras.retrascope.result.test.TestModel;
import ru.ispras.retrascope.result.testbench.Testbench;
import ru.ispras.retrascope.result.testbench.TestbenchModel;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.FileAlreadyExistsException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Testbench printer abstract engine.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class TestbenchPrinterEngine extends Engine {

  /**
   * Path to directory for testbenches to be generated.
   */
  private static final Parameter TESTBENCH_DIRECTORY =
      new Parameter(
          "testbench-dir",
          true,
          "A directory where testbenches are saved");

  /**
   * When specified, allows the engine to overwrite testbench files.
   */
  public static final Parameter OVERWRITE_EXISTING =
      new Parameter(
          "overwrite-existing",
          false,
          "Overwrite target files if they already exist");

  private static final String DEFAULT_CLOCK_PERIOD_VALUE = "10";
  private static final String DEFAULT_CLOCK_PERIOD_UNIT = "ns";

  private static final String DEFAULT_DIRECTORY = "testbenches/" + PlaceHolder.MODULE_NAME;

  private static final String FILE_EXISTS = "The specified file already exists: ";

  /**
   * Creates a new {@code TestVhdlTestbenchPrinter} instance.
   */
  protected TestbenchPrinterEngine(final String engineId) {
    super(engineId, TestModel.TYPE, TestbenchModel.TYPE);

    addInputType(CfgInterface.TYPE);
    addParameter(TESTBENCH_DIRECTORY);
    addParameter(OVERWRITE_EXISTING);
  }

  /**
   * Returns an extension of the testbench files.
   *
   * @return Testbench files extension.
   */
  protected abstract String getTestbenchExt();

  /**
   * Returns string of the path to template resource file.
   *
   * @return Path to template resource file.
   */
  protected abstract String getTemplateResource();

  /**
   * Returns string of the path to TST parser resource file.
   *
   * @return Path to TST parser resource file.
   */
  protected abstract String getTstParserResource();

  /**
   * Returns string description of input port.
   *
   * @return Input port string description.
   */
  protected abstract String getInputPortKind();

  /**
   * Returns string description of output port.
   *
   * @return Output port string description.
   */
  protected abstract String getOutputPortKind();

  /**
   * Returns bit vector declaration format.
   *
   * @see PlaceHolder
   *
   * @return Bit vector declaration format.
   */
  protected abstract String getBitVectorFmt();

  /**
   * Returns port declaration format.
   *
   * @see PlaceHolder
   *
   * @return Port declaration format.
   */
  protected abstract String getPortDeclarationFmt();

  /**
   * Returns input signal declaration format.
   *
   * @see PlaceHolder
   *
   * @return Input signal declaration format.
   */
  protected abstract String getInputDeclarationFmt();

  /**
   * Returns output signal declaration format.
   *
   * @see PlaceHolder
   *
   * @return Output signal declaration format.
   */
  protected abstract String getOutputDeclarationFmt();

  /**
   * Returns buffer variable declaration format.
   *
   * @see PlaceHolder
   *
   * @return Buffer variable declaration format.
   */
  protected abstract String getBufferDeclarationFmt();

  /**
   * Returns port mapping declaration format.
   *
   * @see PlaceHolder
   *
   * @return Port mapping declaration format.
   */
  protected abstract String getPortMapDeclarationFmt();

  /**
   * Returns signal assignment format.
   *
   * @see PlaceHolder
   *
   * @return Signal assignment format.
   */
  protected abstract String getSignalAssignmentFmt();

  /**
   * Returns buffer assignment format.
   *
   * @see PlaceHolder
   *
   * @return Buffer assignment format.
   */
  protected abstract String getBufferAssignmentFmt();

  /**
   * Returns mapping from data types to their string representations.
   *
   * @return Map from data types to their string representations.
   */
  protected abstract Map<DataType, String> getDataTypeMap();

  /**
   * Returns mapping from data type string description to "string-to" conversion functions names.
   *
   * @return Map from data types to function names extracting them from strings.
   */
  protected abstract Map<String, String> getConvertFunctionMap();

  @Override
  public Entity start(final Map<EntityType, Entity> inputs) {

    final boolean overwrite = isOverwrite(inputs);

    final TestModel testModel = getInputEntity(inputs, TestModel.class, TestModel.TYPE);
    final CfgInterface cfgInterface = getInputEntity(inputs, CfgInterface.class, CfgInterface.TYPE);
    final List<Test> tests = testModel.getTests();
    final Map<String, DeclarationContainer> interfaceMapping = cfgInterface.getMapping();

    InvariantChecks.checkTrue(tests.size() == interfaceMapping.size());

    final Iterator<Test> testIterator = tests.iterator();
    final Iterator<Map.Entry<String, DeclarationContainer>> entryIterator =
        interfaceMapping.entrySet().iterator();
    final String directory = getTestbenchDirectory(inputs);
    final List<Set<Testbench>> testbenches = new ArrayList<>(tests.size());

    while (testIterator.hasNext() && entryIterator.hasNext()) {
      final Map.Entry<String, DeclarationContainer> interfaceEntry = entryIterator.next();
      final Test test = testIterator.next();

      final Testbench testbench = printTestbench(test, interfaceEntry, directory, overwrite);
      testbenches.add(new HashSet<>(Collections.singletonList(testbench)));
    }

    return new TestbenchModel(testbenches);
  }

  private boolean isOverwrite(final Map<EntityType, Entity> inputs) {

    final Arguments args = Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);

    boolean overwrite = false;

    if (args != null) {
      try {
        final Arguments parsedArgs = parseCommandLine(args.getCommandLine());
        overwrite = parsedArgs.contains(OVERWRITE_EXISTING);
      } catch (final ParseException e) {
        throwParseException(args, e);
      }
    }
    return overwrite;
  }

  private String getTestbenchDirectory(final Map<EntityType, Entity> inputs) {

    final Arguments args = Engine.getOptionalInputEntity(inputs, Arguments.class, Arguments.TYPE);

    String directory = DEFAULT_DIRECTORY;

    if (args != null) {
      try {
        final Arguments parsedArgs = parseCommandLine(args.getCommandLine());
        directory = parsedArgs.getValue(TESTBENCH_DIRECTORY, directory);
      } catch (final ParseException e) {
        throwParseException(args, e);
      }
    }
    return directory;
  }

  private Testbench printTestbench(
      final Test test,
      final Map.Entry<String, DeclarationContainer> entry,
      final String testbenchDir,
      final boolean overwrite) {

    final String moduleName = normaliseModuleName(entry.getKey());
    final Path dirPath = Paths.get(processFileName(testbenchDir, moduleName));

    try {
      Files.createDirectories(dirPath);
    } catch (final IOException exception) {
      throwCreateException(dirPath.toString(), exception);
    }

    final DeclarationContainer signals = entry.getValue();
    final Collection<String> inputSignalNames = NodeUtils.getVariableNames(signals.getInputs());

    final Set<Path> additionalFiles = new HashSet<>();

    int valueLength = 0;
    int lineLength = 0;
    int sequenceNumber = 0;

    for (final Sequence sequence : test.getSequences()) {
      final Path tstFile = dirPath.resolve(getTestFileName(moduleName, sequenceNumber));
      try {
        DeciderLikeTestPatternPrinter.print(sequence, tstFile, inputSignalNames, overwrite);
      } catch (final IOException e) {

        final String msg = LogMessage.makeMessage(LogMessage.CANT_CREATE_FILE, tstFile);
        throw new RetrascopeRuntimeException(msg, e);
      }
      sequenceNumber++;

      additionalFiles.add(tstFile);

      try (final BufferedReader testReader = ReadWriteUtils.newBufferedReader(tstFile)) {

        int maxLineLength = 0;
        int maxValueLength = 0;

        for (String line = testReader.readLine(); line != null; line = testReader.readLine()) {
          if (line.length() > maxLineLength) {
            maxLineLength = line.length();
          }

          final String[] values = line.split(" ");

          for (final String value : values) {
            if (value.length() > maxValueLength) {
              maxValueLength = value.length();
            }
          }

          if (maxLineLength > lineLength) {
            lineLength = maxLineLength;
          }
          if (maxValueLength > valueLength) {
            valueLength = maxValueLength;
          }
        }
      } catch (final IOException exception) {
        throw new RetrascopeRuntimeException(exception);
      }
    }

    final Path testPatternFile = dirPath.resolve(getTestFileName(moduleName, 0));

    try {

      if (getTstParserResource() != null) {
        final Path tstParserFile = processTstParser(dirPath, moduleName, valueLength, 0, overwrite);
        additionalFiles.add(tstParserFile);
      }

      final Path testbenchFile =
          processTestbenchFile(dirPath, moduleName, testPatternFile,
              lineLength, valueLength, signals, inputSignalNames, 0, overwrite);

      return new Testbench(testbenchFile, additionalFiles);
    } catch (final IOException e) {
      throw new RetrascopeRuntimeException("The exception has occurred while printing a file", e);
    }
  }

  private static void throwParseException(
      final Arguments arguments,
      final ParseException exception) {
    throwException(arguments.toString(), exception, LogMessage.CANT_PARSE);
  }

  private static void throwCreateException(
      final String file,
      final IOException exception) {
    throwException(file, exception, LogMessage.CANT_CREATE_FILE);
  }

  private static void throwException(
      final String description,
      final Exception exception,
      final LogMessage type) {

    final String msg = LogMessage.makeMessage(type, description);
    throw new RetrascopeRuntimeException(msg, exception);
  }

  private Path processTstParser(
      final Path testbenchDirectory,
      final String moduleName,
      final int maxValueLength,
      final int testNumber,
      final boolean overwrite) throws IOException {

    final String fileName = getTstParserFileName(moduleName, testNumber);
    final Path tstParserFile = testbenchDirectory.resolve(fileName);

    if (!overwrite && Files.exists(tstParserFile)) {
      throw new FileAlreadyExistsException(FILE_EXISTS + tstParserFile);
    }

    final ClassLoader classLoader = getClass().getClassLoader();
    final InputStream stream = classLoader.getResourceAsStream(getTstParserResource());

    final BufferedReader tstParserReader = ReadWriteUtils.newBufferedReader(stream);
    final BufferedWriter tstParserWriter = ReadWriteUtils.newBufferedWriter(tstParserFile);

    for (String line = tstParserReader.readLine(); line != null; line =
        tstParserReader.readLine()) {
      line = line.replace(PlaceHolder.MAX_VALUE_LENGTH, String.valueOf(maxValueLength));
      tstParserWriter.write(line);
      tstParserWriter.newLine();
    }

    tstParserWriter.flush();

    return tstParserFile;
  }

  private Path processTestbenchFile(
      final Path testbenchDirectory,
      final String moduleName,
      final Path testPatternFile,
      final int maxLineLength,
      final int maxValueLength,
      final DeclarationContainer externalSignals,
      final Collection<String> inputSignalNames,
      final int testNumber,
      final boolean overwrite) throws IOException {
    InvariantChecks.checkNotNull(testPatternFile);
    InvariantChecks.checkNotNull(externalSignals);

    final String fileName = getTestbenchFileName(moduleName, testNumber);
    final Path testbenchFile = testbenchDirectory.resolve(fileName);

    if (!overwrite && Files.exists(testbenchFile)) {
      throw new FileAlreadyExistsException(FILE_EXISTS + testbenchFile);
    }

    final InputStream templateStream =
        getClass().getClassLoader().getResourceAsStream(getTemplateResource());

    final BufferedReader templateReader = ReadWriteUtils.newBufferedReader(templateStream);
    final BufferedWriter testbenchWriter = ReadWriteUtils.newBufferedWriter(testbenchFile);

    final Map<String, VhdlSignal> signalMap = mapExternalSignals(externalSignals);

    for (String templateLine = templateReader.readLine(); templateLine != null; templateLine =
        templateReader.readLine()) {

      final String testbenchLine =
          templateLine.replace(PlaceHolder.MODULE_NAME, moduleName)
              .replace(PlaceHolder.TEST_FILE_NAME, testPatternFile.getFileName().toString())
              .replace(PlaceHolder.CLOCK_PERIOD_UNIT, DEFAULT_CLOCK_PERIOD_UNIT)
              .replace(PlaceHolder.CLOCK_PERIOD_VALUE, DEFAULT_CLOCK_PERIOD_VALUE)
              .replace(PlaceHolder.MAX_LINE_LENGTH, String.valueOf(maxLineLength))
              .replace(PlaceHolder.MAX_VALUE_LENGTH, String.valueOf(maxValueLength))
              .replace(PlaceHolder.VECTOR_FMT, getInputFormat(externalSignals.getInputs()))
              .replace(PlaceHolder.VECTOR_SIGNALS, getSignals(inputSignalNames));

      if (testbenchLine.contains(PlaceHolder.PORT_DECLARATION)) {
        final String tab = getTab(testbenchLine, PlaceHolder.PORT_DECLARATION);

        for (final String portDeclaration : buildPortDeclaration(tab, externalSignals)) {
          testbenchWriter.append(portDeclaration);
          testbenchWriter.newLine();
        }

        continue;
      }

      if (testbenchLine.contains(PlaceHolder.SIGNAL_DECLARATION)) {
        final String tab = getTab(testbenchLine, PlaceHolder.SIGNAL_DECLARATION);

        for (String signalDeclaration : buildSignalDeclaration(tab, signalMap.values())) {
          testbenchWriter.append(signalDeclaration);
          testbenchWriter.newLine();
        }

        continue;
      }

      if (testbenchLine.contains(PlaceHolder.BUFFER_VARIABLE_DECLARATION)) {
        final String tab = getTab(testbenchLine, PlaceHolder.BUFFER_VARIABLE_DECLARATION);

        for (String signalDeclaration : buildBufferVariableDeclaration(tab, signalMap.values())) {
          testbenchWriter.append(signalDeclaration);
          testbenchWriter.newLine();
        }

        continue;
      }

      if (testbenchLine.contains(PlaceHolder.PORT_MAP_DECLARATION)) {
        final String tab = getTab(testbenchLine, PlaceHolder.PORT_MAP_DECLARATION);

        for (final String signalDeclaration : buildPortMapDeclaration(tab, signalMap)) {
          testbenchWriter.append(signalDeclaration);
          testbenchWriter.newLine();
        }

        continue;
      }

      if (testbenchLine.contains(PlaceHolder.SIGNAL_ASSIGNMENT)) {
        final String tab = getTab(testbenchLine, PlaceHolder.SIGNAL_ASSIGNMENT);

        for (final String inputSignalName : inputSignalNames) {
          testbenchWriter.append(buildSignalAssignment(tab, signalMap.get(inputSignalName)));
          testbenchWriter.newLine();
        }

        continue;
      }

      if (testbenchLine.contains(PlaceHolder.BUFFER_VARIABLE_ASSIGNMENT)) {
        final String tab = getTab(testbenchLine, PlaceHolder.BUFFER_VARIABLE_ASSIGNMENT);

        for (final String inName : inputSignalNames) {
          testbenchWriter.append(buildBufferVariableAssignment(tab, signalMap.get(inName)));
          testbenchWriter.newLine();
        }

        continue;
      }

      testbenchWriter.append(testbenchLine);
      testbenchWriter.newLine();
    }

    testbenchWriter.flush();

    return testbenchFile;
  }

  private CharSequence getSignals(final Collection<String> signalNames) {

    final Iterator<String> iterator = signalNames.iterator();
    final StringBuilder builder = new StringBuilder();

    while (iterator.hasNext()) {
      builder.append(iterator.next());

      if (iterator.hasNext()) {
        builder.append(", ");
      }
    }

    return builder.toString();
  }

  private CharSequence getInputFormat(final Collection<NodeVariable> inputs) {

    final Iterator<NodeVariable> iterator = inputs.iterator();
    final StringBuilder builder = new StringBuilder();

    while (iterator.hasNext()) {
      builder.append(getSpecifier(iterator.next().getDataType()));

      if (iterator.hasNext()) {
        builder.append(" ");
      }
    }

    return builder.toString();
  }

  private String getSpecifier(final DataType dataType) {

    final DataTypeId typeId = dataType.getTypeId();

    switch (typeId) {
      case LOGIC_BOOLEAN:
        return "%b";
      case LOGIC_INTEGER:
        return "%d";
      case BIT_VECTOR:
        return "%" + dataType.getSize() + "b";
      case LOGIC_REAL:
        return "%f";
      default:
        throw RetrascopeRuntimeException.makeException(LogMessage.UNSUPPORTED_TYPE, typeId);
    }
  }

  private String getTestFileName(
      final String moduleName,
      final int testNumber) {

    return getTestbenchFile(moduleName, "test", testNumber, "tst");
  }

  private String getTestbenchFileName(
      final String moduleName,
      final int testNumber) {

    return getTestbenchFile(moduleName, "testbench", testNumber, getTestbenchExt());
  }

  private String getTstParserFileName(
      final String moduleName,
      final int testNumber) {

    return getTestbenchFile(moduleName, "tst_parser", testNumber, getTestbenchExt());
  }

  private static String getTestbenchFile(
      final String prefix,
      final String meaning,
      final int fileNum,
      final String ext) {
    return String.format("%s_%s%s.%s", prefix, meaning, (fileNum > 0) ? "_" + fileNum : "", ext);
  }

  private static String getTab(
      final String line,
      final String placeholder) {
    return line.substring(0, line.indexOf(placeholder));
  }

  private List<String> buildPortDeclaration(
      final String tab,
      final DeclarationContainer externalSignals) {

    final List<String> portDeclaration = new ArrayList<>();

    for (final NodeVariable input : externalSignals.getInputs()) {
      if (!portDeclaration.isEmpty()) {
        final int lastIndex = portDeclaration.size() - 1;
        portDeclaration.set(lastIndex, portDeclaration.get(lastIndex) + ";");
      }

      portDeclaration.add(getPortDeclarationFmt().replace(PlaceHolder.TAB, tab)
          .replace(PlaceHolder.PORT_NAME, input.getName())
          .replace(PlaceHolder.PORT_KIND, getInputPortKind())
          .replace(PlaceHolder.PORT_TYPE, mapDataType(input)));
    }

    for (final NodeVariable output : externalSignals.getOutputs()) {
      if (!portDeclaration.isEmpty()) {
        final int lastIndex = portDeclaration.size() - 1;
        portDeclaration.set(lastIndex, portDeclaration.get(lastIndex) + ";");
      }

      portDeclaration.add(getPortDeclarationFmt().replace(PlaceHolder.TAB, tab)
          .replace(PlaceHolder.PORT_NAME, output.getName())
          .replace(PlaceHolder.PORT_KIND, getOutputPortKind())
          .replace(PlaceHolder.PORT_TYPE, mapDataType(output)));
    }

    return portDeclaration;
  }

  private List<String> buildSignalDeclaration(
      final String tab,
      final Collection<VhdlSignal> signals) {

    final List<String> signalDeclaration = new ArrayList<>();

    for (final VhdlSignal signal : signals) {

      final String declarationFmt =
          signal.isInput ? getInputDeclarationFmt() : getOutputDeclarationFmt();

      signalDeclaration.add(declarationFmt.replace(PlaceHolder.TAB, tab)
          .replace(PlaceHolder.SIGNAL_NAME, signal.name)
          .replace(PlaceHolder.SIGNAL_TYPE, signal.type));
    }

    return signalDeclaration;
  }

  private List<String> buildBufferVariableDeclaration(
      final String tab,
      final Collection<VhdlSignal> signals) {

    final List<String> signalDeclaration = new ArrayList<>();

    for (final VhdlSignal signal : signals) {
      signalDeclaration.add(getBufferDeclarationFmt().replace(PlaceHolder.TAB, tab)
          .replace(PlaceHolder.SIGNAL_NAME, signal.name)
          .replace(PlaceHolder.SIGNAL_TYPE, signal.type));
    }

    return signalDeclaration;
  }

  private List<String> buildPortMapDeclaration(
      final String tab,
      final Map<String, VhdlSignal> portMap) {

    final List<String> portMapDeclaration = new ArrayList<>();

    for (final Map.Entry<String, VhdlSignal> entry : portMap.entrySet()) {
      if (!portMapDeclaration.isEmpty()) {
        int lastIndex = portMapDeclaration.size() - 1;
        portMapDeclaration.set(lastIndex, portMapDeclaration.get(lastIndex) + ",");
      }

      portMapDeclaration.add(getPortMapDeclarationFmt().replace(PlaceHolder.TAB, tab)
          .replace(PlaceHolder.PORT_NAME, entry.getKey())
          .replace(PlaceHolder.SIGNAL_NAME, entry.getValue().name));
    }

    return portMapDeclaration;
  }

  private String mapDataType(final NodeVariable variable) {

    final VariableDescriptor descriptor = (VariableDescriptor) variable.getUserData();

    if (descriptor != null) {
      final String vhdlDataType = descriptor.getMetaInfo(MetaInfoType.HDL_TYPE_STR.name());

      if (vhdlDataType != null) {
        return vhdlDataType;
      }
    }

    final DataType dataType = variable.getDataType();

    if (getDataTypeMap().containsKey(dataType)) {
      return getDataTypeMap().get(dataType);
    } else {
      if (dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {
        return getBitVectorFmt().replace(PlaceHolder.UPPER_BOUND,
            String.valueOf(dataType.getSize() - 1)).replace(PlaceHolder.LOWER_BOUND, "0");
      } else {
        throw RetrascopeRuntimeException.makeException(LogMessage.UNSUPPORTED_TYPE, dataType);
      }
    }
  }

  private Map<String, VhdlSignal> mapExternalSignals(final DeclarationContainer externalSignals) {

    final Collection<NodeVariable> inputs = externalSignals.getInputs();
    final Collection<NodeVariable> outputs = externalSignals.getOutputs();

    final Map<String, VhdlSignal> mapping = new HashMap<>(inputs.size() + outputs.size());

    for (final NodeVariable input : inputs) {
      final VhdlSignal inSignal =
          new VhdlSignal(input.getName(), mapDataType(input), input.getDataType(), true);
      mapping.put(input.getName(), inSignal);
    }

    for (final NodeVariable output : outputs) {
      final VhdlSignal outSignal =
          new VhdlSignal(output.getName(), mapDataType(output), output.getDataType(), false);
      mapping.put(output.getName(), outSignal);
    }

    return mapping;
  }

  private String buildBufferVariableAssignment(
      final String tab,
      final VhdlSignal signal) {
    String assignment =
        getBufferAssignmentFmt().replace(PlaceHolder.TAB, tab)
            .replace(PlaceHolder.SIGNAL_NAME, signal.name);

    final String signalCoreType = signal.type.split(" |\\(")[0];
    final DataType signalDataType = signal.dataType;

    if (getConvertFunctionMap().containsKey(signalCoreType)) {
      assignment =
          assignment.replace(PlaceHolder.CONVERSION_FUNCTION,
            getConvertFunctionMap().get(signalCoreType));
    } else if (getDataTypeMap().containsKey(signalDataType)) {
      assignment =
          assignment.replace(PlaceHolder.CONVERSION_FUNCTION,
              getDataTypeMap().get(signalDataType));
    } else {
      throw RetrascopeRuntimeException.makeException(LogMessage.UNSUPPORTED_TYPE, signal.type);
    }

    return assignment;
  }

  private String buildSignalAssignment(
      final String tab,
      final VhdlSignal signal) {
    return getSignalAssignmentFmt().replace(PlaceHolder.TAB, tab).replace(
        PlaceHolder.SIGNAL_NAME, signal.name);
  }

  private static class VhdlSignal {

    final String name;
    final String type;
    final DataType dataType;
    final boolean isInput;

    VhdlSignal(
        final String name,
        final String type,
        final DataType dataType,
        final boolean isInput) {

      this.name = name;
      this.type = type;
      this.dataType = dataType;
      this.isInput = isInput;
    }
  }

  private static String normaliseModuleName(final String moduleName) {
    InvariantChecks.checkNotNull(moduleName);

    String[] splittedName = moduleName.split("\\.");

    if (splittedName.length == 1) {
      return moduleName;
    } else {
      return splittedName[splittedName.length - 1].replaceAll("\\(.*\\)", "");
    }
  }

  private static String processFileName(final String name, final String moduleName) {
    return name.replace(PlaceHolder.MODULE_NAME, moduleName);
  }
}
