/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.engine.test.printer.testbench.vhdl;

import ru.ispras.fortress.data.DataType;
import ru.ispras.retrascope.engine.test.printer.testbench.PlaceHolder;
import ru.ispras.retrascope.engine.test.printer.testbench.TestbenchPrinterEngine;

import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;

/**
 * The engine that makes VHDL testbenches from tests.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class TestVhdlTestbenchPrinter extends TestbenchPrinterEngine {

  /**
   * The engine ID of all {@code TestVhdlTestbenchPrinter} instances.
   */
  private static final String ENGINE_ID = "test-vhdl-testbench-printer";

  private static final String BUFFER_VARIABLE_NAME_POSTFIX = "_buffer";

  private static final Map<DataType, String> DATA_TYPES_MAP = new HashMap<>();

  private static final Map<String, String> CONVERSION_FUNCTIONS_MAP = new TreeMap<>(
      String.CASE_INSENSITIVE_ORDER);

  /**
   * Creates a new {@code TestVhdlTestbenchPrinter} instance.
   */
  public TestVhdlTestbenchPrinter() {
    super(ENGINE_ID);
  }

  static {
    DATA_TYPES_MAP.put(DataType.BIT_VECTOR(1), "std_logic");
    DATA_TYPES_MAP.put(DataType.INTEGER, "integer");
    DATA_TYPES_MAP.put(DataType.BOOLEAN, "bit");
    DATA_TYPES_MAP.put(DataType.REAL, "real");

    CONVERSION_FUNCTIONS_MAP.put("std_logic", "string_to_std_logic");
    CONVERSION_FUNCTIONS_MAP.put("std_logic_vector", "string_to_std_logic_vector");
    CONVERSION_FUNCTIONS_MAP.put("bit", "string_to_bit");
    CONVERSION_FUNCTIONS_MAP.put("bit_vector", "string_to_bit_vector");
    CONVERSION_FUNCTIONS_MAP.put("integer", "string_to_integer");
    CONVERSION_FUNCTIONS_MAP.put("real", "string_to_real");
  }

  @Override
  public String getTestbenchExt() {
    return "vhd";
  }

  @Override
  public String getTemplateResource() {
    return "templates/vhdl/testbench";
  }

  @Override
  public String getTstParserResource() {
    return "templates/vhdl/tst_parser";
  }

  @Override
  public String getInputPortKind() {
    return "in";
  }

  @Override
  public String getOutputPortKind() {
    return "out";
  }

  @Override
  public String getBitVectorFmt() {
    return String.format(
        "std_logic_vector (%s to %s)",
        PlaceHolder.LOWER_BOUND,
        PlaceHolder.UPPER_BOUND);
  }

  @Override
  public String getPortDeclarationFmt() {
    return String.format(
        "%s%s: %s %s",
        PlaceHolder.TAB,
        PlaceHolder.PORT_NAME,
        PlaceHolder.PORT_KIND,
        PlaceHolder.PORT_TYPE);
  }

  @Override
  public String getInputDeclarationFmt() {
    return getSignalDeclarationFormat();
  }

  @Override
  public String getOutputDeclarationFmt() {
    return getSignalDeclarationFormat();
  }

  @Override
  public String getBufferDeclarationFmt() {
    return  String.format(
        "%svariable %s%s: %s;",
        PlaceHolder.TAB,
        PlaceHolder.SIGNAL_NAME,
        BUFFER_VARIABLE_NAME_POSTFIX,
        PlaceHolder.SIGNAL_TYPE);
  }

  @Override
  public String getPortMapDeclarationFmt() {
    return  String.format(
        "%s%s => %s",
        PlaceHolder.TAB,
        PlaceHolder.PORT_NAME,
        PlaceHolder.SIGNAL_NAME);
  }

  @Override
  public String getSignalAssignmentFmt() {
    return  String.format(
        "%s%s <= %s%s;",
        PlaceHolder.TAB,
        PlaceHolder.SIGNAL_NAME,
        PlaceHolder.SIGNAL_NAME,
        BUFFER_VARIABLE_NAME_POSTFIX);
  }

  @Override
  public String getBufferAssignmentFmt() {
    return  String.format(
        "%s%s(test_line_string, %s%s);",
        PlaceHolder.TAB,
        PlaceHolder.CONVERSION_FUNCTION,
        PlaceHolder.SIGNAL_NAME,
        BUFFER_VARIABLE_NAME_POSTFIX);
  }

  @Override
  public Map<DataType, String> getDataTypeMap() {
    return DATA_TYPES_MAP;
  }

  @Override
  public Map<String, String> getConvertFunctionMap() {
    return CONVERSION_FUNCTIONS_MAP;
  }

  private static String getSignalDeclarationFormat() {
    return String.format(
        "%ssignal %s: %s;",
        PlaceHolder.TAB,
        PlaceHolder.SIGNAL_NAME,
        PlaceHolder.SIGNAL_TYPE);
  }
}
