/*
 * Copyright 2013-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.printer.ExprPrinter;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.MetaInfo;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Assignment statement.
 *
 * <p>Consists of defined {@link RangedVariable} target
 * variable and used value expression.</p>
 *
 * @see RangedVariable
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 *
 */
public class Assignment extends MetaInfo implements Identifiable,UseDef {

  private RangedVariable target;

  private Node value;

  /**
   * Default constructor.
   */
  public Assignment() {
    super();
  }

  /**
   * Constructs object with the specified target variable and value expression.
   *
   * @param target Defined target variable.
   * @param value Used expression.
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public Assignment(final RangedVariable target, final Node value) {
    super();
    InvariantChecks.checkNotNull(target);
    InvariantChecks.checkNotNull(value);

    this.target = target;
    this.value = value;
  }

  /**
   * Constructs object with the specified target non-ranged variable and value expression.
   *
   * @param var Defined target variable.
   * @param expr Used expression.
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public Assignment(final NodeVariable var, final Node expr) {
    this(new RangedVariable(var), expr);
  }

  /**
   * Returns a deep copy of the object.
   * @return A ddep copy of the object.
   */
  public Assignment deepCopy() {
    return new Assignment(this);
  }

  private Assignment(final Assignment assignment) {
    this(assignment.getTarget().deepCopy(), assignment.getValue().deepCopy());
  }

  /**
   * Returns defined ranged variable.
   *
   * @return Target ranged variable.
   */
  public RangedVariable getTarget() {
    return this.target;
  }

  /**
   * Returns defined variable.
   *
   * @return Target variable.
   */
  public NodeVariable getVariable() {
    return this.target.getVariable();
  }

  /**
   * Checks whether this object contains a defining to ranged variable.
   *
   * @return {@code true} if this object defines ranged variable,
   *         {@code false} otherwise.
   */
  public boolean ranged() {
    return this.target.isRanged();
  }

  /**
   * Sets the specified variable as objects' target.
   *
   * @param variable New target variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setVariable(final RangedVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.target = variable;
  }

  /**
   * Returns value expression.
   *
   * @return Used expression.
   */
  public Node getValue() {
    return this.value;
  }

  /**
   * Sets new value expression.
   *
   * @param value New value expression.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void setValue(final Node value) {
    InvariantChecks.checkNotNull(value);

    this.value = value;
  }

  /**
   * Returns names of variables that are used in this object.
   *
   * @return Names of used variables.
   */
  public Set<String> getUsedVariableNames() {

    final Collection<NodeVariable> usedVariables = getUses();
    final Set<String> usedVariableNames = new LinkedHashSet<>();

    for (final NodeVariable variable: usedVariables) {
      usedVariableNames.add(variable.getName());
    }
    return usedVariableNames;
  }

  /**
   * Returns this assignment description that is presented in format of the specified printer
   * with the specified separator.
   *
   * @param printer Expression printer.
   * @param separator Assignment separator between defined and used variables.
   *
   * @return {@code String} description of assignment with the specified separator.
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public String getSeparatedDescription(final ExprTreePrinter printer, final String separator) {
    InvariantChecks.checkNotNull(printer);
    InvariantChecks.checkNotNull(separator);

    return this.target.toString() + separator + printer.toString(this.value);
  }

  /**
   * Returns assignment separator for the specified printer and basic block concurrent type.
   * <p>An assignment separator is a symbol that stands between target and value expressions
   * of assignment.</p>
   * @param printer Expression printer.
   * @param concurrent Concurrent kind of block statement above.
   * @return Assignment separator for the block statement of the specified concurrent type.
   */
  public String getSeparator(final ExprTreePrinter printer, final boolean concurrent) {

    String kind = null;

    if (printer instanceof ExprPrinter) {

      if (((ExprPrinter) printer).name().equals(ExprPrinter.VERILOG.name())) {

        kind = concurrent ? "<=" : "=";

      } else if (((ExprPrinter) printer).name().equals(ExprPrinter.VHDL.name())) {

        kind = concurrent ? "<=" : ":=";
      }
    }
    if (kind == null) {
      throw RetrascopeRuntimeException.makeException(
          LogMessage.UNSUPPORTED_TYPE,
          printer.getClass());
    }
    return kind;
  }

  @Override
  public Set<NodeVariable> getUses() {

    return getValue() == null
        ? new LinkedHashSet<NodeVariable>()
        : new LinkedHashSet<>(ExprUtils.getVariables(getValue()));
  }

  @Override
  public Set<NodeVariable> getDefines() {

    final Set<NodeVariable> defines = new LinkedHashSet<>();

    if (getTarget() != null) {
      defines.add(getVariable());
    }

    return defines;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((this.value == null) ? 0 : this.value.hashCode());
    result = prime * result + this.target.hashCode();

    return result;
  }

  @Override
  public boolean equals(final Object object) {
    if (this == object) {
      return true;
    }

    if (!(object instanceof Assignment)) {
      return false;
    }

    final Assignment other = (Assignment) object;

    if (this.value == null) {
      if (other.value != null) {
        return false;
      }
    } else if (!this.value.equals(other.value)) {
      return false;
    }

    return this.target.equals(other.target);
  }

  @Override
  public String toString() {
    return String.format("%s := %s", getTarget(), getValue());
  }

  @Override
  public String getId() {
    return getDescription();
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    return getSeparatedDescription(printer, "=");
  }

  @Override
  public String getDescription() {
    return getDescription(Retrascope.getConfiguration().getStylePrinter());
  }
}
