/*
 * Copyright 2016-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import org.apache.commons.collections15.Closure;
import org.apache.commons.collections15.CollectionUtils;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfo;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * The variables and their declarations container.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class DeclarationContainer extends MetaInfo {

  private final Map<NodeVariable, Declaration> variables;

  /**
   * Default constructor.
   */
  public DeclarationContainer() {
    this.variables = new LinkedHashMap<>();
  }

  /**
   * Constructs the object with the specified data.
   * @param map The mapping from variables to their declarations.
   */
  public DeclarationContainer(final Map<NodeVariable, Declaration> map) {
    this.variables = map;
  }

  /**
   * Returns the declaration for the specified variable.
   * @param variable The variable that declaration should be returned.
   * @return The declaration for the specified variable if it is declared here,
   *         {@code null} otherwise.
   */
  public Declaration getDeclaration(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    return this.variables.get(variable);
  }

  /**
   * Returns the mapping from declared variables to their declarations.
   *
   * @return The mapping from declared variables to their declarations.
   */
  public Map<NodeVariable, Declaration> getDeclarationMap() {
    return Collections.unmodifiableMap(this.variables);
  }

  /**
   * Returns variables declarations.
   *
   * @return declarations of node variables
   */
  public Collection<Declaration> getDeclarations() {
    return this.variables.values();
  }

  /**
   * Returns declared variable with specified name.
   *
   * @param name name of variable to be returned
   *
   * @return declared variable with the specified name,
   *         or {@code null} if there is no such variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public NodeVariable getVariable(final String name) {
    InvariantChecks.checkNotNull(name);

    NodeVariable result = null;

    for (final NodeVariable variable : this.variables.keySet()) {

      if (variable.getName().equals(name)) {

        result = variable;
        break;
      }
    }

    return result;
  }

  /**
   * Returns declared variables.
   *
   * @return variables collection
   */
  public List<NodeVariable> getVariables() {
    return new LinkedList<>(this.variables.keySet());
  }

  /**
   * Returns the collection of names for already declared variables.
   *
   * @return The collection of names for already declared variables.
   */
  public Collection<String> getVariableNames() {

    final Collection<String> names = new LinkedHashSet<>();

    final Closure<NodeVariable> varClosure = new Closure<NodeVariable>() {
      @Override
      public void execute(final NodeVariable variable) {
        names.add(variable.getName());
      }
    };

    CollectionUtils.forAllDo(this.variables.keySet(), varClosure);
    return names;
  }

  /**
   * Returns the collection of names of already declared input variables.
   * @return The collection of names of already declared input variables.
   */
  public Collection<String> getInputNames() {

    final Collection<String> names = new LinkedHashSet<>();

    final Closure<NodeVariable> inputClosure = new Closure<NodeVariable>() {
      @Override
      public void execute(final NodeVariable variable) {

        final Object userData = variable.getUserData();
        if (userData instanceof VariableDescriptor
            && VariableType.isInputSignal(((VariableDescriptor) userData).getVariableType())) {
          names.add(variable.getName());
        }
      }
    };

    CollectionUtils.forAllDo(this.variables.keySet(), inputClosure);

    return names;
  }

  /**
   * Returns the collection of names of already declared output variables.
   * @return The collection of names of already declared output variables.
   */
  public Collection<String> getOutputNames() {

    final Collection<String> names = new LinkedHashSet<>();

    final Closure<NodeVariable> outputClosure = new Closure<NodeVariable>() {
      @Override
      public void execute(final NodeVariable variable) {

        final Object userData = variable.getUserData();
        if (userData instanceof VariableDescriptor
            && VariableType.isOutputSignal(((VariableDescriptor) userData).getVariableType())) {
          names.add(variable.getName());
        }
      }
    };

    CollectionUtils.forAllDo(this.variables.keySet(), outputClosure);

    return names;
  }

  /**
   * Returns the collection of names of already declared internal (register) variables.
   * @return The collection of names of already declared internal (register) variables.
   */
  public Collection<String> getRegisterNames() {

    final Collection<String> names = new LinkedHashSet<>();

    final Closure<NodeVariable> regClosure = new Closure<NodeVariable>() {
      @Override
      public void execute(final NodeVariable variable) {

        final Object userData = variable.getUserData();
        if (userData instanceof VariableDescriptor
            && VariableType.isRegister(((VariableDescriptor) userData).getVariableType())) {
          names.add(variable.getName());
        }
      }
    };

    CollectionUtils.forAllDo(this.variables.keySet(), regClosure);
    return names;
  }

  /**
   * Returns the number of declared variables.
   * @return The number of declared variables.
   */
  public int getVariableNum() {
    return this.variables.size();
  }

  /**
   * Returns a collection of input signal variables that are declared into this object.
   *
   * @return Input signal variables.
   */
  public Collection<NodeVariable> getInputs() {
    final Collection<NodeVariable> inputs = new LinkedHashSet<>();

    final Closure<NodeVariable> inputClosure = new Closure<NodeVariable>() {
      @Override
      public void execute(final NodeVariable variable) {

        final Object userData = variable.getUserData();
        if (userData instanceof VariableDescriptor
            && VariableType.isInputSignal(((VariableDescriptor) userData).getVariableType())) {
          inputs.add(variable);
        }
      }
    };

    CollectionUtils.forAllDo(this.variables.keySet(), inputClosure);
    return inputs;
  }

  /**
   * Returns a collection of output signal variables that are declared into this object.
   *
   * @return Output signal variables.
   */
  public Collection<NodeVariable> getOutputs() {
    final Collection<NodeVariable> outputs = new LinkedHashSet<>();

    final Closure<NodeVariable> outputClosure = new Closure<NodeVariable>() {
      @Override
      public void execute(final NodeVariable variable) {

        final Object userData = variable.getUserData();
        if (userData instanceof VariableDescriptor
            && VariableType.isOutputSignal(((VariableDescriptor) userData).getVariableType())) {
          outputs.add(variable);
        }
      }
    };

    CollectionUtils.forAllDo(this.variables.keySet(), outputClosure);
    return outputs;
  }

  /**
   * Returns data type of the declared variable with the specified name.
   *
   * @param name variable name
   *
   * @return data type if exists the declared variable with the specified name,
   *         {@code null} otherwise.
   */
  public DataType getDataType(final String name) {
    DataType dataType = null;

    for (final NodeVariable variable : this.variables.keySet()) {

      if (variable.getName().equals(name)) {
        dataType = variable.getDataType();
        break;
      }
    }

    return dataType;
  }

  /**
   * Returns invariant of the declared variable with the specified name.
   *
   * @param name variable name
   *
   * @return invariant if it exist for the declared variable with the specified name,
   *         {@code null} otherwise.
   */
  public Node getInvariant(final String name) {
    Node invariant = null;

    for (final Map.Entry<NodeVariable, Declaration> entry : this.variables.entrySet()) {

      final NodeVariable var = entry.getKey();

      if (var.getName().equals(name)) {

        final Declaration declaration = entry.getValue();
        invariant = DeclarationUtils.getInvariant(declaration, var);
        break;
      }
    }

    return invariant;
  }

  /**
   * Returns the initial value for declared variable with the specified name.
   * @param name The name for declared variable.
   * @return The initial value for variable if it is already declared, {@code null} otherwise.
   */
  public Node getInitialValue(final String name) {
    Node initial = null;

    for (final Map.Entry<NodeVariable, Declaration> entry : this.variables.entrySet()) {

      if (entry.getKey().getName().equals(name)) {

        initial = entry.getValue().getInitialValue();
        break;
      }
    }
    return initial;
  }

  /**
   * Checks does this object contain the specified variable.
   *
   * @param variable variable to be checked
   *
   * @return {@code true} if module already has such variable declared, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean containsVariable(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    return this.variables.containsKey(variable);
  }

  /**
   * Checks whether object contains a variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if object contains a variable with the specified name,
   *         {@code false} otherwise.
   */
  public boolean containsVariable(final String name) {
    boolean result = false;

    for (final NodeVariable variable : this.variables.keySet()) {

      if (variable.getName().equals(name)) {

        result = true;
        break;
      }
    }

    return result;
  }

  /**
   * Checks whether this object contain input signal variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if this object contains an input signal variable with the specified name,
   *         {@code false} otherwise.
   */
  public boolean containsInput(final String name) {
    boolean result = false;

    for (final NodeVariable variable : this.variables.keySet()) {

      final Object userData = variable.getUserData();

      if (variable.getName().equals(name)
          && userData instanceof VariableDescriptor
          && VariableType.isInputSignal(((VariableDescriptor) userData).getVariableType())) {

        result = true;
        break;
      }
    }

    return result;
  }

  /**
   * Checks whether this object contains output signal variable with the specified name.
   *
   * @param name variable name
   *
   * @return {@code true} if this object contains an output signal variable with the specified name,
   *         {@code false} otherwise.
   */
  public boolean containsOutput(final String name) {
    boolean result = false;

    for (final NodeVariable variable : this.variables.keySet()) {

      final Object userData = variable.getUserData();

      if (variable.getName().equals(name)
          && userData instanceof VariableDescriptor
          && VariableType.isOutputSignal(((VariableDescriptor) userData).getVariableType())) {

        result = true;
        break;
      }
    }

    return result;
  }

  /**
   * Checks whether this object contains internal variable with the specified name.
   *
   * @param name The name of declared variable.
   *
   * @return {@code true} if this object contains an internal variable with the specified name,
   *         {@code false} otherwise.
   */
  public boolean containsRegister(final String name) {
    boolean result = false;

    for (final NodeVariable variable : this.variables.keySet()) {

      final Object userData = variable.getUserData();

      if (variable.getName().equals(name)
          && userData instanceof VariableDescriptor
          && VariableType.isRegister(((VariableDescriptor) userData).getVariableType())) {

        result = true;
        break;
      }
    }
    return result;
  }

  /**
   * Declares non-existing variable.
   *
   * <p>Object MUST NOT contain the specified variable.
   * Check this fact by corresponding check-method first.</p>
   *
   * @param variable variable to be added
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalStateException when this object
   *         already contains such variable.
   */
  public void declareVariable(final NodeVariable variable) {
    declareVariable(variable, null, null);
  }

  /**
   * Declares the specified variable with the specified initial value.
   * @param variable The variable to be declared.
   * @param initValue The initial value of the specified variable.
   * @throws IllegalArgumentException when initial value argument is {@code null}.
   */
  public void declareVariable(final NodeVariable variable, final Node initValue) {
    InvariantChecks.checkNotNull(initValue);

    declareVariable(variable, initValue, null);
  }

  /**
   * Declares the specified variable with the specified invariant.
   *
   * <p>mObject MUST NOT contain the specified variable.
   * Check this fact by corresponding method first.</p>
   *
   * @param variable variable to be added
   * @param invariant variable invariant (can be {@code null})
   *
   * @throws IllegalArgumentException when variable argument is {@code null}.
   * @throws IllegalStateException when this module object
   *         already contains such variable.
   */
  public void declareVariable(final NodeVariable variable, final Range invariant) {

    if (invariant == null) {
      declareVariable(variable);
    } else {
      declareVariable(variable, null, invariant);
    }
  }

  /**
   * Declares the specified variable with the specified initial value and the specified invariant.
   *
   * <p>Invariant keeps a constraint on variable value.</p>
   *
   * @param variable variable
   * @param initialValue initial value of variable
   * @param invariant variable invariant
   * @throws IllegalArgumentException when variable argument is {@code null}.
   */
  public void declareVariable(
      final NodeVariable variable,
      final Node initialValue,
      final Range invariant) {
    InvariantChecks.checkNotNull(variable);

    if (this.variables.keySet().contains(variable)) {
      this.variables.remove(variable);
    }
    this.variables.put(variable, new Declaration(initialValue, invariant));
  }

  /**
   * Declares all the variables from the specified mapping.
   * @param map The mapping from variables to their declaration descriptors
   */
  public void declareVariables(final Map<NodeVariable, Declaration> map) {

    for (final Map.Entry<NodeVariable, Declaration> entry : map.entrySet()) {

      final Declaration declaration = entry.getValue();
      declareVariable(entry.getKey(), declaration.getInitialValue(), declaration.getInvariant());
    }
  }

  /**
   * Removes the declaration for the variable with the specified name.
   * @param name The name for variable to be removed.
   */
  public void removeDeclaration(final String name) {
    NodeVariable toBeRemoved = null;

    for (final NodeVariable variable : this.variables.keySet()) {

      if (variable.getName().equals(name)) {

        toBeRemoved = variable;
        break;
      }
    }

    if (toBeRemoved != null) {
      this.variables.remove(toBeRemoved);
    }
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final DeclarationContainer that = (DeclarationContainer) obj;

    return getVariables().equals(that.getVariables());
  }

  @Override
  public int hashCode() {
    return getVariables().hashCode();
  }
}
