/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

/**
 * Variable declarations related utility methods.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class DeclarationUtils {

  /**
   * Prefix for new variables that are created by the tool.
   */
  public static final String NEW_VAR_PREFIX = "NEW_VAR_";

  private DeclarationUtils() {
    /* Empty. */
  }

  /**
   * Returns the constraint representation of the specified variable.
   * @param declaration The declaration that contains invariant.
   * @param var The variable for which an invariant is built.
   * @return The constraint representation of the invariant from the specified declaration
   *         for the specified variable.
   */
  public static Node getInvariant(final Declaration declaration, final NodeVariable var) {
    InvariantChecks.checkNotNull(declaration);
    InvariantChecks.checkNotNull(var);

    if (!declaration.containsInvariant()) {
      return null;
    }

    final Range invariant = declaration.getInvariant();
    final Node old = invariant.getOld();
    final Node young = invariant.getYoung();
    final DataType oldType = old.getDataType();
    final DataType youngType = young.getDataType();

    InvariantChecks.checkTrue(oldType.equals(youngType) && oldType.equals(var.getDataType()));

    final DataTypeId varTypeId = var.getDataTypeId();

    final Enum<?> gtEq = getId(varTypeId, StandardOperation.GREATEREQ, StandardOperation.BVUGE);
    final Enum<?> lsEq = getId(varTypeId, StandardOperation.LESSEQ, StandardOperation.BVULE);

    return ExprUtils.getConjunction(
        new NodeOperation(gtEq, var, young),
        new NodeOperation(lsEq, var, old));
  }

  private static Enum<?> getId(
      final DataTypeId id,
      final Enum<?> logicId,
      final Enum<?> nonLogicId) {

    return id.isLogic() ? logicId : nonLogicId;
  }
}
