/*
 * Copyright 2013-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.jaxb.JaxbEventListAdapter;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import javax.xml.bind.annotation.XmlType;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

/**
 * Collection of events.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
@XmlType(name = "eventList")
@XmlJavaTypeAdapter(JaxbEventListAdapter.class)
public class EventList implements Iterable<Event> {

  private final Set<Event> events;

  /**
   * Default constructor.
   */
  public EventList() {
    this.events = new LinkedHashSet<>();
  }

  /**
   * Constructs object from the specified event object.
   *
   * @param event Object from which event list is constructed.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public EventList(final Event event) {
    InvariantChecks.checkNotNull(event);

    this.events = new LinkedHashSet<>();
    this.events.add(event);
  }

  /**
   * Constructs object with the specified events collection.
   *
   * @param events Events.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public EventList(final Collection<Event> events) {
    InvariantChecks.checkNotNull(events);

    this.events = new LinkedHashSet<>(events);
  }

  /**
   * Checks whether this object contains the specified event.
   *
   * @param event Event object.
   *
   * @return {@code true} if event is here, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public boolean contains(final Event event) {
    InvariantChecks.checkNotNull(event);

    return this.events.contains(event);
  }

  /**
   * Checks whether this object contains an event on the specified variable.
   *
   * @param variable Variable of an event of the object to be checked.
   *
   * @return {@code true} if object contains an event with the specified variable,
   * {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  private boolean contains(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    boolean contains = false;

    for (final Event event : this.events) {

      if (event.getVariable().equals(variable)) {
        contains = true;
        break;
      }
    }

    return contains;
  }

  /**
   * Adds event.
   *
   * <p>If this object already contains an event with the same
   * {@link RangedVariable} as the specified {@link Event} has,
   * but their {@link EventType} types are different,
   * the {@link Event} object with the same {@link RangedVariable}
   * and {@code EventType.ANY_EDGE} event type.</p>
   *
   * @param event Event object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void add(final Event event) {
    InvariantChecks.checkNotNull(event);

    final RangedVariable eVariable = event.getRangedVariable();
    final EventType eType = event.getType();
    Event sameVarEvent = null;

    for (final Event thisEvent : this.events) {

      if (thisEvent.getRangedVariable().equals(eVariable) && !thisEvent.getType().equals(eType)) {

        sameVarEvent = thisEvent;
        break;
      }
    }

    if (sameVarEvent == null) {

      this.events.add(event);

    } else {

      this.events.remove(sameVarEvent);
      final Event combination = new Event(event.getRangedVariable(), EventType.ANY_EDGE);
      this.events.add(combination);
    }
  }

  /**
   * Removes an event on the specified ranged variable.
   *
   * @param variable The event's ranged variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void remove(final RangedVariable variable) {
    InvariantChecks.checkNotNull(variable);

    final Collection<Event> removed = new LinkedHashSet<>();

    for (final Event event : this.events) {

      if (event.getRangedVariable().equals(variable)) {
        removed.add(event);
      }
    }

    for (final Event rmEvent : removed) {
      this.events.remove(rmEvent);
    }
  }

  /**
   * Removes an event on the specified variable.
   *
   * @param variable The event's variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void remove(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    final Collection<Event> removed = new LinkedHashSet<>();

    for (final Event event : this.events) {

      if (event.getVariable().equals(variable)) {
        removed.add(event);
      }
    }

    for (final Event rmEvent : removed) {
      this.events.remove(rmEvent);
    }
  }

  /**
   * Removes the specified event from object's sensitivity list.
   *
   * @param event The event to be removed.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void remove(final Event event) {
    InvariantChecks.checkNotNull(event);

    this.events.remove(event);
  }

  /**
   * Adds a collection of events.
   *
   * @param events Event collection.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  private void addEvents(final Collection<? extends Event> events) {
    InvariantChecks.checkNotNull(events);

    for (final Event event : events) {
      add(event);
    }
  }

  /**
   * Adds events from the specified collection to this one.
   *
   * @param events The container of events to be added to this object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addEvents(final EventList events) {
    InvariantChecks.checkNotNull(events);

    addEvents(events.events);
  }

  /**
   * Checks if this object is empty.
   *
   * @return {@code true} if this event list doesn't contain any events, {@code false} otherwise.
   */
  public boolean isEmpty() {
    return this.events.isEmpty();
  }

  /**
   * Applies expressions to bounded event variables.
   *
   * @param bindings Mapping of event variables names to expressions that are to be applied.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void applyToEventList(final Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(bindings);

    final EventList eventList = new EventList();
    boolean isBounded = false;

    for (final Event event : this.events) {

      for (final Map.Entry<String, Node> entry : bindings.entrySet()) {

        if (event.getVariable().getName().equals(entry.getKey())) {

          isBounded = true;


          final Collection<NodeVariable> vars = ExprUtils.getVariables(entry.getValue());

          for (final NodeVariable variable : vars) {

            if (eventList.isEmpty() || !eventList.contains(variable)) {

              /* Add any-edge event. */
              eventList.add(new Event(new RangedVariable(variable)));
            }
          }
        }
      }

      if (isBounded) {

        isBounded = false;

      } else {

        eventList.add(new Event(event.getRangedVariable(), event.getType()));
      }
    }
  }

  /**
   * Checks if this object is compatible with the specified one.
   *
   * <p>Checks if this object and the specified object are not mutually exclusive.</p>
   *
   * @param eventList Event list to be tested for compatibility with this one.
   *
   * @return {@code true} if the lists are compatible, otherwise returns {@code false}
   *
   * @see Event#incompatibleWith(Event)
   */
  public boolean compatibleWith(final EventList eventList) {

    for (final Event event: this) {

      for (final Event testedEvent: eventList) {

        if (event.incompatibleWith(testedEvent)) {
          return false;
        }
      }
    }
    return true;
  }

  /**
   * Checks if this object is always triggered along with the specified argument.
   *
   * <p>Checks if this object is triggered by the specified {@link Event} object, for example,
   * if this event is any edge of some variable while other is positive edge of the same variable.
   * To trigger this object it is necessary and sufficient to trigger at least one of its
   * events.</p>
   *
   * @param event Event to be tested.
   *
   * @return {@code true} if this event is always triggered along with the specified one,
   *         {@code false} otherwise.
   */
  public boolean isTriggered(final Event event) {

    for (final Event expectedEvent: this) {

      if (expectedEvent.isTriggered(event)) {
        return true;
      }
    }

    return false;
  }

  /**
   * Returns a deep copy of this object.
   *
   * @return Object's deep copy.
   */
  public EventList deepCopy() {
    final EventList copy = new EventList();

    for (final Event event : this.events) {
      final Event eventCopy = event.deepCopy();

      copy.add(eventCopy);
    }

    return copy;
  }

  /**
   * Removes all the internal events.
   */
  public void clear() {
    this.events.clear();
  }

  @Override
  public String toString() {

    final StringBuilder text = new StringBuilder("");
    final Iterator<Event> iterator = this.events.iterator();

    while (iterator.hasNext()) {

      text.append(iterator.next().toString());

      if (iterator.hasNext()) {
        text.append(',');
      }
    }
    return text.toString();
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }

    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final EventList that = (EventList) obj;

    return this.events.equals(that.events);
  }

  @Override
  public int hashCode() {
    return this.events.hashCode();
  }

  @Override
  public Iterator<Event> iterator() {
    return this.events.iterator();
  }
}
