/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;

/**
 * The abstract representation of a device model.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class Model extends Entity {

  private final Collection<Module> modules;

  /**
   * Constructs an entity with a given type.
   *
   * @param type the entity type.
   */
  protected Model(EntityType type) {
    super(type);

    this.modules = new LinkedList<>();
  }

  /**
   * Adds the specified module as a child one for this object.
   * @param module The module to be added as a child one.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addModule(final Module module) {
    InvariantChecks.checkNotNull(module);

    this.modules.add(module);
  }

  /**
   * Adds objects from the specified collection as children for this object.
   * @param modules The objects to be added as child modules.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addModules(final Collection<? extends Module> modules) {
    InvariantChecks.checkNotNull(modules);

    this.modules.addAll(modules);
  }

  /**
   * Returns a collection of child modules.
   * @return A collection of child modules.
   */
  public Collection<Module> getModules() {
    return Collections.unmodifiableCollection(this.modules);
  }

  /**
   * Adds the specified entity.
   *
   * <p>Able to add entity of the appropriate type.</p>
   *
   * @param entity entity to be added
   *
   * @throws IllegalArgumentException when argument is {@code null};
   * @throws RetrascopeRuntimeException when the specified entity
   *         is not of {@link ru.ispras.retrascope.model.cfg.CfgModel} type.
   */
  @Override
  public void add(final Entity entity) {
    InvariantChecks.checkNotNull(entity);

    if (entity instanceof Model && entity.getEntityType() == getEntityType()) {

      final Collection<Module> entityModules = ((Model)entity).getModules();
      addModules(entityModules);

    } else {
      throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, entity.getEntityType());
    }
  }
}
