/*
 * Copyright 2016-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.util.NodeUtils;

import java.util.LinkedList;
import java.util.List;

/**
 * Utility methods for models and their components.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class ModelUtils {

  /**
   * Transforms the specified events container to the node.
   *
   * @param events Container of events.
   * @return Node that is a conjunction of "signal == signal_level" equations that are extracted
   *         from events.
   */
  public static Node eventsToNode(final EventList events) {

    final List<Node> nodes = new LinkedList<>();

    for (final Event event : events) {

      final RangedVariable rangedVariable = event.getRangedVariable();
      final NodeVariable variable = rangedVariable.getVariable();

      Node value = null;
      switch (event.getType()) {

        case NEGATIVE_EDGE:
          value = newValue(variable.getDataType(), 0);
          break;

        case POSITIVE_EDGE:
          value = newValue(variable.getDataType(), 1);
          break;

        default:
          /* Do nothing. */
          break;
      }

      if (value != null) {

        nodes.add(NodeUtils.getEquation(
            variable,
            rangedVariable.isRanged()
                ? RangedVariableUtils.transformNodeWithRanged(rangedVariable, value) : value));
      }
    }
    return nodes.isEmpty() ? null : NodeUtils.getConjunction(nodes);
  }

  private static Node newValue(final DataType dataType, final int num) {

    switch (dataType.getTypeId()) {

      case BIT_VECTOR:
        return NodeValue.newBitVector(BitVector.valueOf(num, dataType.getSize()));

      case LOGIC_BOOLEAN:
        return NodeValue.newBoolean(num != 0);

      case LOGIC_INTEGER:
        return NodeValue.newInteger(num);

      default:
        return null;
    }
  }
}
