/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * The model component with it's own name, instance name and variable declarations.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class Module extends DecStatement {

  private final String name;
  private String instanceName;

  /**
   * Default constructor.
   */
  protected Module() {
    super();

    this.name = null;
  }

  /**
   * Constructs object with the specified name.
   * @param name The object name.
   */
  protected Module(final String name) {
    super();

    this.name = name;
  }

  /**
   * Constructs a copy of the specified module object.
   * @param module The module object to be copied.
   */
  protected Module(final Module module) {
    super(module);

    this.name = module.getName();
    this.instanceName = module.getInstanceName();
  }

  /**
   * Returns the module name.
   * @return The module name.
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns a full name.
   * <p>A full name is a combination of name and instance name.</p>
   * @return full name
   */
  public String getFullName() {

    final StringBuilder builder = new StringBuilder(getName());

    if (getInstanceName() != null) {

      builder.append('.');
      builder.append(getInstanceName());
    }
    return builder.toString();
  }

  /**
   * Sets instance name.
   *
   * @param instanceName The name of current instance.
   */
  protected void setInstanceName(final String instanceName) {
    this.instanceName = instanceName;
  }

  /**
   * Returns instance name.
   *
   * @return instance name
   */
  public String getInstanceName() {
    return this.instanceName;
  }

  /**
   * Returns processes.
   *
   * @return processes
   */
  public Collection<ModuleProcess> getProcesses() {

    final Set<ModuleProcess> processes = new LinkedHashSet<>();

    for (int i = 0; i < getStatementNum(); i++) {

      final Statement statement = getStatement(i);

      if (statement.getType() == StatementType.PROCESS) {
        processes.add((ModuleProcess) statement);
      }
    }
    return processes;
  }

  /**
   * Returns sub-modules.
   *
   * @return sub-modules
   */
  public Collection<Module> getModules() {

    final Set<Module> modules = new LinkedHashSet<>();

    for (int i = 0; i < getStatementNum(); i++) {

      final Statement statement = getStatement(i);

      if (statement.getType() == StatementType.MODULE) {
        modules.add((Module) statement);
      }
    }
    return modules;
  }

  /**
   * Remove internal processes.
   */
  public void removeProcesses() {

    final List<ModuleProcess> processes = new LinkedList<>();

    for (int i = 0; i < getStatementNum(); i++) {

      final Statement statement = getStatement(i);

      if (statement.isType(StatementType.PROCESS)) {
        processes.add((ModuleProcess) statement);
      }
    }

    for (final ModuleProcess process : processes) {
      remove(process);
    }
  }

  @Override
  public void add(final Statement statement) {

    if (!(statement instanceof DecStatement)) {
      throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, statement);
    }
    super.add(statement);
  }

  @Override
  public void setParent(final Statement statement) {

    final Enum<?> type = statement.getType();

    if (!type.equals(StatementType.MODULE)) {
      throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, statement);
    }
    super.setParent(statement);
  }

  @Override
  public boolean isSequential() {
    return false;
  }

  @Override
  public StatementType getType() {
    return StatementType.MODULE;
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    return String.format("%s %s", getType().name(), getFullName());
  }
}
