/*
 * Copyright 2013-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorAlgorithm;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Variable or it's sub-part restricted by the specified range.
 *
 * <p>When the range is not {@code null} it means that not all variable is in use
 * but only it's sub-array.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class RangedVariable implements UseDef {

  private final NodeVariable variable;

  private Range range;

  private static final String WRONG_RANGE = "Wrong range: ";

  /**
   * Constructs object with the specified variable.
   *
   * @param variable Variable.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public RangedVariable(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    this.variable = variable;
    this.range = null;
  }

  /**
   * Constructs object with the specified variable and it's range.
   *
   * <p>The ranged variable MUST have either bit vector data type
   * or map data type.</p>
   *
   * @param variable Variable.
   * @param range Range.
   *
   * @throws IllegalArgumentException in the following situations:
   *         (1) when any of the arguments is {@code null};
   *         (2) when variable argument is neither of
   *         {@link ru.ispras.fortress.data.types.bitvector.BitVector} data type
   *         nor of {@link ru.ispras.fortress.data.types.datamap.DataMap} data type;
   *         (3) when variable is of {@link ru.ispras.fortress.data.types.bitvector.BitVector} type
   *         but range old/young bit numbers are out of
   *         {@link ru.ispras.fortress.data.types.bitvector.BitVector} size;
   *         (4) when variable is of {@link ru.ispras.fortress.data.types.datamap.DataMap} type
   *         but range old/young bit numbers are not equal.
   */
  public RangedVariable(final NodeVariable variable, final Range range) {
    InvariantChecks.checkNotNull(variable);
    InvariantChecks.checkNotNull(range);

    checkParams(variable, range);

    this.variable = variable;
    this.range = range;
  }

  /**
   * Creates a deep copy of the current object.
   *
   * @return Deep copy of this object.
   */
  public RangedVariable deepCopy() {

    final RangedVariable deepCopy = new RangedVariable((NodeVariable) this.variable.deepCopy());

    if (this.isRanged()) {
      deepCopy.setRange(this.range.deepCopy());
    }
    return deepCopy;
  }

  /**
   * Returns variable.
   *
   * @return Variable.
   */
  public NodeVariable getVariable() {
    return this.variable;
  }

  /**
   * Returns variable name.
   *
   * @return Variable name.
   */
  public String getVariableName() {
    return this.getVariable().getName();
  }

  /**
   * Returns range.
   *
   * @return Range.
   */
  public Range getRange() {
    return this.range;
  }

  /**
   * Sets new range.
   *
   * @param range Range to be set.
   *
   *  @throws IllegalArgumentException in the following situations:
   *         (1) when any argument is {@code null};
   *         (2) when this object's variable is neither of
   *         {@link ru.ispras.fortress.data.types.bitvector.BitVector}
   *         nor of {@link ru.ispras.fortress.data.types.datamap.DataMap} data types;
   *         (3) when this object's variable
   *         is of {@link ru.ispras.fortress.data.types.bitvector.BitVector} type
   *         but range old/young bit numbers are out of
   *         {@link ru.ispras.fortress.data.types.bitvector.BitVector} size;
   *         (4) when variable is of {@link ru.ispras.fortress.data.types.datamap.DataMap} type
   *         but range old/young bit numbers are not equal.
   */
  public void setRange(final Range range) {
    InvariantChecks.checkNotNull(range);
    checkParams(getVariable(), range);

    this.range = range;
  }

  /**
   * Checks whether this is ranged.
   *
   * @return {@code true} if this contains non-{@code null} range, {@code false} otherwise.
   */
  public boolean isRanged() {
    return this.range != null;
  }

  /**
   * Returns the data type of this object.
   * @return The data type of this object or DataType.UNKNOWN if it fails.
   */
  public DataType getDataType() {

    final DataType varType = getVariable().getDataType();
    DataType rangedType = varType;

    if (isRanged()) {

      final DataTypeId typeId = varType.getTypeId();
      final Node old = getRange().getOld();
      final DataType oldType = old.getDataType();
      final Node young = getRange().getYoung();
      final DataType youngType = young.getDataType();

      if (ExprUtils.isKind(Node.Kind.VALUE, old, young) && oldType.equals(youngType)) {

        switch (typeId) {

          case BIT_VECTOR:

            final boolean areInt = ExprUtils.isType(DataType.INTEGER, old, young);

            final NodeOperation size =
                new NodeOperation(areInt ? StandardOperation.ADD : StandardOperation.BVADD,
                    new NodeOperation(
                        areInt
                            ? StandardOperation.SUB
                            : StandardOperation.BVSUB, old, young),
                    areInt
                        ? NodeValue.newInteger(1)
                        : NodeValue.newBitVector(
                            BitVector.valueOf(1, oldType.getSize())));
            rangedType =
                DataType.BIT_VECTOR(
                    ((NodeValue)TransformerUtils.toInt(size)).getInteger().intValue());
            break;

          case MAP:

            if (old.equals(young)) {
              rangedType = (DataType) varType.getParameters()[1];
            } else {
              rangedType = DataType.UNKNOWN;
            }
            break;

          default:
            rangedType = DataType.UNKNOWN;
            break;
        }
      } else {
        rangedType = DataType.UNKNOWN;
      }
    }

    return rangedType;
  }

  private static void checkParams(final NodeVariable variable, final Range range) {

    if (!variable.isType(DataTypeId.BIT_VECTOR)
        && !variable.isType(DataTypeId.MAP)) {

      throw RetrascopeRuntimeException.makeException(
          LogMessage.WRONG_TYPE,
          variable.getDataTypeId());
    }

    final Node young = range.getYoung();
    final Node old = range.getOld();

    if (variable.isType(DataTypeId.BIT_VECTOR)) {

      /* Check whether range does not spread outside the 0..size-1 bit numbers. */

      if (ExprUtils.isType(DataTypeId.BIT_VECTOR, old, young)) {

        final DataType dataType = old.getDataType();
        final int size = dataType.getSize();

        final BitVector zero = BitVector.newEmpty(size);
        BitVectorAlgorithm.fill(zero, (byte) 0);
        final NodeValue zeroValue = NodeValue.newBitVector(zero);

        final BitVector max = BitVector.newEmpty(size);
        BitVectorAlgorithm.fill(max, (byte) 1);
        final NodeValue maxValue = NodeValue.newBitVector(max);

        final Node inRange =
            Node.AND(
                new NodeOperation(StandardOperation.BVUGE, young, zeroValue),
                new NodeOperation(StandardOperation.BVULE, old, maxValue));

        if (!SolverUtils.isSat(inRange)) {

          throw new RetrascopeRuntimeException(
              String.format("%s%s < %s or %s > %s.", WRONG_RANGE, young, zero, old, max));
        }
      } else if (ExprUtils.isType(DataType.INTEGER, old, young)) {

        final NodeValue zero = NodeValue.newInteger(0);
        final NodeValue max = NodeValue.newInteger(variable.getDataType().getSize());

        final Node inRange =
            Node.AND(
                new NodeOperation(StandardOperation.GREATEREQ, young, zero),
                new NodeOperation(StandardOperation.LESSEQ, old, max));

        if (!SolverUtils.isSat(inRange)) {

          throw new RetrascopeRuntimeException(
            String.format("%s%s < %s or %s > %s.", WRONG_RANGE, young, zero, old, max));
        }
      }
    }

    if (variable.isType(DataTypeId.MAP)
        && !SolverUtils.isSat(NodeUtils.getEquation(young, old))) {

      throw new RetrascopeRuntimeException(
          String.format("%s%s != %s for map", WRONG_RANGE, old, young));
    }
  }

  @Override
  public String toString() {

    final StringBuilder builder = new StringBuilder();
    builder.append(this.variable.toString());

    if (this.range != null) {

      builder.append('[');
      builder.append(this.range.toString());
      builder.append(']');
    }
    return builder.toString();
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    boolean result;
    final RangedVariable that = (RangedVariable) obj;

    if (this.range == null || that.getRange() == null) {

      result = this.variable.equals(that.variable);

    } else {
      result = this.variable.equals(that.variable) && this.range.equals(that.range);
    }
    return result;
  }

  @Override
  public int hashCode() {

    int result = this.variable.hashCode();

    if (this.range != null) {

      result = 31 * result + this.range.hashCode();
    }
    return result;
  }

  @Override
  public Set<NodeVariable> getUses() {

    final Set<NodeVariable> uses = new LinkedHashSet<>();
    uses.add(getVariable());

    if (isRanged()) {

      uses.addAll(ExprUtils.getVariables(getRange().getOld()));
      uses.addAll(ExprUtils.getVariables(getRange().getYoung()));
    }

    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {
    return Collections.emptySet();
  }
}
