/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;
import ru.ispras.retrascope.util.TransformerUtils;

/**
 * Utility methods for ranged variable class.
 *
 * @see RangedVariable
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class RangedVariableUtils {

  private RangedVariableUtils() {
    /* Empty. */
  }

  /**
   * Transforms the specified node with assistance of the specified ranged variable.
   * <p>For example, we have a pair '(a[x], b)' where 'a[x]' is ranged variable and 'b'
   * is node. Such pair may come from {@link Assignment} node constructor parameters
   * that should be transformed to (NodeVariable, Node) pair. This method transforms 'b'
   * expression in such way that the result can be assigned to RangedVariable's NodeVariable field
   * without any Range.</p>
   * @param rangedVariable Ranged variable that is used in transformation.
   * @param node Node to be transformed.
   * @return Result of the specified node's transformation with using of the specified
   *         ranged variable.
   */
  public static Node transformNodeWithRanged(
      final RangedVariable rangedVariable,
      final Node node) {

    final NodeVariable variable = rangedVariable.getVariable();
    final Range range = rangedVariable.getRange();
    final DataType rangeType =
        range.getOld().getDataType() == null
        ? range.getOld().getDataType()
        : range.getYoung().getDataType();
    final int bitSize = rangeType.getSize();
    final NodeValue zero = createValue(rangeType, bitSize, 0);
    final NodeValue one = createValue(rangeType, bitSize, 1);

    if (variable.isType(DataTypeId.BIT_VECTOR)) {

      final int dataTypeSize = variable.getDataType().getSize();
      final NodeValue size = NodeValue.newInteger(dataTypeSize);
      final Node rangeTypeSize = TypeConversion.coerce(size, rangeType);
      final StandardOperation subOp =
          rangeType.equals(DataType.INTEGER) ? StandardOperation.SUB : StandardOperation.BVSUB;
      final Node maxNum = new NodeOperation(subOp, rangeTypeSize, one);

      assert dataTypeSize != 0 : "Can't elaborate non-bitvector ranged variable: " + rangedVariable;

      final Node old = range.getOld();
      final Node young = range.getYoung();

      if (ExprUtils.isKind(Node.Kind.VALUE, old, young)) {
        return TransformerUtils.standardize(getNodeValueRanged(rangedVariable, node));
      } else {

        /* young == 0 && old == size -1 */
        final Node nonRanged = getNonRangedCondition(zero, maxNum, range);

        /* young > 0 && old == size -1 */
        final Node oldBorder = getOldBorderCondition(zero, maxNum, range);
        final Node oldBorderNode = getOldBorderExpr(node, rangedVariable, zero, one);

        /* young == 0 && old < size -1 */
        final Node youngBorder = getYoungBorderCondition(zero, maxNum, range);
        final Node youngBorderNode = getYoungBorderExpr(node, rangedVariable, one, maxNum);

        final Node generalNode = getGeneralCaseExpr(node, rangedVariable, zero, one, maxNum);

        return new NodeOperation(
            StandardOperation.ITE,
            nonRanged,
            node,
            new NodeOperation(
                StandardOperation.ITE,
                oldBorder, oldBorderNode,
                new NodeOperation(
                    StandardOperation.ITE, youngBorder, youngBorderNode, generalNode)));
      }
    } else if (variable.isType(DataTypeId.MAP)) {

      assert range.getOld().equals(range.getYoung())
          : "Can't elaborate multi-index MAP entry: " + range;

      return new NodeOperation(StandardOperation.STORE, variable, range.getOld(), node);
    }
    return null;
  }

  private static NodeOperation getGeneralCaseExpr(
      final Node node,
      final RangedVariable rangedVariable,
      final NodeValue zero,
      final NodeValue one,
      final Node maxNum) {

    final Range range = rangedVariable.getRange();
    final DataType dataType =
        range.getOld().getDataType() == null
        ? range.getOld().getDataType()
        : range.getYoung().getDataType();

    final Node maxNumExpr = maxNum instanceof NodeValue ? TransformerUtils.toInt(maxNum) : maxNum;

    final Node oldPlusOneExpr = new NodeOperation(
        dataType.equals(DataType.INTEGER)
            ? StandardOperation.ADD
            : StandardOperation.BVADD,
            rangedVariable.getRange().getOld(),
        one);
    final Node oldPlusOneInt = TransformerUtils.toInt(oldPlusOneExpr);
    final Node oldPlusOne = oldPlusOneInt == null ? oldPlusOneExpr : oldPlusOneInt;

    final Node youngMinusOneExpr = new NodeOperation(
        dataType.equals(DataType.INTEGER)
            ? StandardOperation.SUB
            : StandardOperation.BVSUB,
        rangedVariable.getRange().getYoung(),
        one);
    final Node youngMinusOneInt = TransformerUtils.toInt(youngMinusOneExpr);
    final Node youngMinusOne = youngMinusOneInt == null ? youngMinusOneExpr : youngMinusOneInt;

    return new NodeOperation(
        StandardOperation.BVCONCAT,
        new NodeOperation(
            StandardOperation.BVEXTRACT,
            maxNumExpr,
            oldPlusOne,
            rangedVariable.getVariable()),
        node,
        new NodeOperation(
            StandardOperation.BVEXTRACT,
            youngMinusOne,
            TransformerUtils.toInt(zero),
            rangedVariable.getVariable()));
  }

  private static Node getYoungBorderCondition(
      final NodeValue zero,
      final Node maxNum,
      final Range range) {

    return Node.AND(
        NodeUtils.getEquation(range.getYoung(), zero),
        new NodeOperation(
            range.getOld().isType(DataType.INTEGER)
                ? StandardOperation.LESS : StandardOperation.BVULT,
            range.getOld(),
            maxNum));
  }

  private static Node getOldBorderCondition(
      final NodeValue zero,
      final Node maxNum,
      final Range range) {

    return Node.AND(
        new NodeOperation(
            range.getYoung().isType(DataType.INTEGER)
                ? StandardOperation.GREATER : StandardOperation.BVUGT,
            range.getYoung(),
            zero),
        NodeUtils.getEquation(range.getOld(), maxNum));
  }

  private static Node getNonRangedCondition(
      final NodeValue zero,
      final Node maxNum,
      final Range range) {

    return Node.AND(
        NodeUtils.getEquation(range.getYoung(), zero),
        NodeUtils.getEquation(range.getOld(), maxNum));
  }

  private static NodeOperation getYoungBorderExpr(
      final Node node,
      final RangedVariable rangedVar,
      final NodeValue one,
      final Node maxNum) {

    final Node old = rangedVar.getRange().getOld();
    final Node oldPlusOneExpr =
        new NodeOperation(
            old.isType(DataType.INTEGER) ? StandardOperation.ADD : StandardOperation.BVADD,
            old,
            one);
    final Node oldPlusOne = TransformerUtils.toInt(oldPlusOneExpr);
    final Node maxInt = TransformerUtils.toInt(maxNum);

    return new NodeOperation(
        StandardOperation.BVCONCAT,
        new NodeOperation(
            StandardOperation.BVEXTRACT,
            maxInt,
            oldPlusOne == null ? oldPlusOneExpr : oldPlusOne,
            rangedVar.getVariable()),
        node);
  }

  private static NodeOperation getOldBorderExpr(
      final Node node,
      final RangedVariable rangedVar,
      final NodeValue zero,
      final NodeValue one) {

    final Node young = rangedVar.getRange().getYoung();
    final Node youngMinusOneExpr =
        new NodeOperation(
            young.isType(DataType.INTEGER) ? StandardOperation.SUB : StandardOperation.BVSUB,
            young,
            one);

    final Node youngMinusOne = TransformerUtils.toInt(youngMinusOneExpr);
    final Node zeroInt = TransformerUtils.toInt(zero);

    return new NodeOperation(
        StandardOperation.BVCONCAT,
        node,
        new NodeOperation(
            StandardOperation.BVEXTRACT,
            youngMinusOne == null ? youngMinusOneExpr : youngMinusOne,
            zeroInt,
            rangedVar.getVariable()));
  }

  private static Node getNodeValueRanged(
      final RangedVariable rangedVariable,
      final Node node) {

    final Range range = rangedVariable.getRange();
    final NodeVariable variable = rangedVariable.getVariable();

    /* Calculate 'one' and 'zero' values depending on range data type. */
    final DataType dataType =
                range.getOld().getDataType() == null
                    ? range.getYoung().getDataType() : range.getOld().getDataType();
    final int bitSize =
                dataType.equals(DataType.INTEGER) ? DataType.INTEGER.getSize() : dataType.getSize();

    final NodeValue zero = createValue(dataType, bitSize, 0);
    final NodeValue one = createValue(dataType, bitSize, 1);

    final int dataTypeSize = variable.getDataType().getSize();
    final NodeValue size = createValue(dataType, bitSize, dataTypeSize);
    final Node maxNum = new NodeOperation(
        dataType.equals(DataType.INTEGER) ? StandardOperation.SUB : StandardOperation.BVSUB,
        size,
        TypeConversion.coerce(one, dataType));

    if (SolverUtils.isSat(getNonRangedCondition(zero, maxNum, range))) {
      return node;
    }
    if (SolverUtils.isSat(getOldBorderCondition(zero, maxNum, range))) {

      return getOldBorderExpr(node, rangedVariable, zero, one);
    }
    if (SolverUtils.isSat(getYoungBorderCondition(zero, maxNum, range))) {

      return getYoungBorderExpr(node, rangedVariable, one, maxNum);
    }

    return getGeneralCaseExpr(node, rangedVariable, zero, one, maxNum);
  }

  private static NodeValue createValue(
      final DataType dataType,
      final int bitSize,
      final int value) {

    return dataType.equals(DataType.INTEGER)
        ? NodeValue.newInteger(value)
        : NodeValue.newBitVector(BitVector.valueOf(value, bitSize));
  }
}
