/*
 * Copyright 2014-2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.basis.jaxb;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.jaxb.JaxbDataType;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.util.LogMessage;

import javax.xml.bind.annotation.adapters.XmlAdapter;

/**
 * The adapter class for JAXB marshalling/unmarshalling of {@link Event} objects.
 *
 * <p>Performs conversion between {@link Event} and {@link JaxbEvent} instances.</p>
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 *
 * @see Event
 * @see JaxbEvent
 */
public class JaxbEventAdapter extends XmlAdapter<JaxbEvent, Event> {

  @Override
  public JaxbEvent marshal(final Event event) throws Exception {
    if (event == null) {
      return null;
    }

    final JaxbEvent jaxbEvent = new JaxbEvent();
    jaxbEvent.type = event.getType();
    final JaxbEventSignal jaxbEventSignal = new JaxbEventSignal();
    final RangedVariable rangedVariable = event.getRangedVariable();
    final Range range = rangedVariable.getRange();

    if (range != null) {
      jaxbEventSignal.leftIndex = range.getYoung();
      jaxbEventSignal.rightIndex = range.getOld();
    } else {
      jaxbEventSignal.leftIndex = null;
      jaxbEventSignal.rightIndex = null;
    }

    final NodeVariable variable = rangedVariable.getVariable();
    jaxbEventSignal.name = variable.getName();
    final Data data = variable.getData();
    final DataType dataType = data.getType();
    jaxbEventSignal.dataType = JaxbDataType.valueOf(dataType.getTypeId().name());
    jaxbEventSignal.size = dataType.getSize();

    if (variable.getUserData() instanceof VariableDescriptor) {
      jaxbEventSignal.signalType = ((VariableDescriptor) variable.getUserData()).getVariableType();
    } else {
      throw new IllegalArgumentException(
          String.format("%s%s.", LogMessage.UNSUPPORTED_TYPE, variable.getUserData().getClass()));
    }

    final Object value = data.getValue();

    if (value != null) {
      jaxbEventSignal.value = String.valueOf(value);
    }

    jaxbEvent.signal = jaxbEventSignal;

    return jaxbEvent;
  }

  @Override
  public Event unmarshal(final JaxbEvent jaxbEvent) throws Exception {
    if (jaxbEvent == null) {
      return null;
    }

    final JaxbEventSignal jaxbEventSignal = jaxbEvent.signal;
    final DataTypeId typeId = DataTypeId.valueOf(jaxbEventSignal.dataType.name());
    final DataType type = DataType.newDataType(typeId, jaxbEventSignal.size);
    final Data data =
        (jaxbEventSignal.value == null)
            ? type.valueUninitialized() : type.valueOf(jaxbEventSignal.value, type.getTypeRadix());

    final NodeVariable variable = new NodeVariable(new Variable(jaxbEventSignal.name, data));
    variable.setUserData(new VariableDescriptor(jaxbEventSignal.signalType));

    if ((jaxbEventSignal.leftIndex == null) && (jaxbEventSignal.rightIndex == null)) {
      return new Event(new RangedVariable(variable), jaxbEvent.type);
    } else {
      final Range range = new Range(jaxbEventSignal.leftIndex, jaxbEventSignal.rightIndex);

      return new Event(new RangedVariable(variable, range), jaxbEvent.type);
    }
  }
}
