/*
 * Copyright 2016-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis.walker;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;

import java.util.Collection;

/**
 * Depth-first search model walker.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class DfsWalker extends ModelWalker {

  /**
   * Constructs walker object with the specified parameters.
   *
   * @param visitor visitor to be applied to control flow graph model
   * @param mode The walking mode.
   * @throws IllegalArgumentException when any argument is {@code null}.
   * @throws IllegalArgumentException when {@code mode} argument equals neither
   * {@link WalkerMode#DFS} nor {@link WalkerMode#DFS_NO_RPT}
   */
  public DfsWalker(
      final ModelVisitor visitor,
      final WalkerMode mode) {
    super(visitor, mode);
    InvariantChecks.checkTrue(mode == WalkerMode.DFS || mode == WalkerMode.DFS_NO_RPT);
  }

  @Override
  protected void visitModelStatement(final Statement statement) {
    /* Empty. */
  }

  @Override
  public void goToNextStatement() {
    /* Empty. */
  }

  @Override
  public void processChildren(final Statement statement) {

    final Collection<Statement> successors = StatementUtils.getSuccessors(statement);

    for (final Statement successor : successors) {

      final String id = successor.getId();

      if (getMode().equals(WalkerMode.DFS_NO_RPT)) {
        if (getVisited().isEmpty() || !getVisited().contains(id)) {

          addVisited(id);
          visitNode(successor);
          if (isStatus(ModelVisitor.Status.ABORT)) {
            return;
          }
        }
      } else {
        visitNode(successor);

        if (isStatus(ModelVisitor.Status.ABORT)) {
          return;
        }
      }
    }
  }
}
