/*
 * Copyright 2016-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis.walker;

import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;

/**
 * Basic interface for model visitors.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public interface ModelVisitor {

  /*
   * The enumeration describes possible statuses of the visitor. Status serves as a directive
   * for the walker to alter its behavior depending on events that may occur in the visitor.
   */
  enum Status {
    /** Continue traversing. */
    OK,

    /** Skip child nodes. */
    SKIP,

    /** Stop traversing. */
    ABORT
  }

  /**
   * Returns the current status of the visitor. The status guides further actions of the walker.
   *
   * @return Current visitor status.
   */
  Status getStatus();

  /**
   * Notifies that processing has been started.
   */
  void onRootBegin();

  /**
   * Notifies that processing has been finished.
   */
  void onRootEnd();

  /**
   * Starts visiting {@link Model} object.
   */
  void onModelBegin(final Model model);

  /**
   * Finishes visiting {@link Model} object.
   */
  void onModelEnd(final Model model);

  /**
   * Starts visiting {@link Statement} statement.
   *
   * @param statement Statement to be visited.
   */
  void onStatementBegin(final Statement statement);

  /**
   * Finishes visiting {@link Statement} statement.
   *
   * @param statement Statement to be visited.
   */
  void onStatementEnd(final Statement statement);

  /**
   * Starts visiting {@link Module} node.
   *
   * @param module Module node to be visited.
   */
  void onModuleBegin(final Module module);

  /**
   * Finishes visiting {@link Module} node.
   *
   * @param module Module node to be visited.
   */
  void onModuleEnd(final Module module);

  /**
   * Starts visiting {@link ModuleProcess} node.
   *
   * @param process Process node to be visited.
   */
  void onProcessBegin(final ModuleProcess process);

  /**
   * Finishes visiting {@link ModuleProcess} node.
   *
   * @param process Process node to be visited.
   */
  void onProcessEnd(final ModuleProcess process);
}
