/*
 * Copyright 2016-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis.walker;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementType;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Set;

/**
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public abstract class ModelWalker {

  private final ModelVisitor visitor;
  private final WalkerMode mode;

  /* Identifiers of visited statements. */
  private final Set<String> visited = new LinkedHashSet<>();

  /**
   * Constructs walker with the specified parameters.
   *
   * @param visitor The model visitor.
   * @param mode The walking mode.
   *
   * @throws IllegalArgumentException when any argument is {@code null}.
   */
  ModelWalker(final ModelVisitor visitor, final WalkerMode mode) {
    InvariantChecks.checkNotNull(visitor);
    InvariantChecks.checkNotNull(mode);

    this.visitor = visitor;
    this.mode = mode;
  }

  /**
   * Returns the model visitor.
   *
   * @return The model visitor.
   */
  public ModelVisitor getVisitor() {
    return this.visitor;
  }

  /**
   * Returns the walking mode.
   * @return The walking mode.
   */
  WalkerMode getMode() {
    return this.mode;
  }

  /**
   * Returns a collection of identifiers for statements that are already visited.
   *
   * @return identifiers of already visited statements
   */
  Collection<String> getVisited() {
    return this.visited;
  }

  /**
   * Add the specified statement identifier to the storage of identifiers
   * for statements that are already visited by this walker.
   *
   * @param visitedId The statement identifier that should be treated
   *                  by this walker as already visited.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  void addVisited(final String visitedId) {
    InvariantChecks.checkNotNull(visitedId);

    this.visited.add(visitedId);
  }

  /**
   * Add statement identifiers from the specified collection to the storage of identifiers
   * for statements that are already visited by this walker.
   *
   * @param visitedIds Statement identifiers that should be treated as already visited.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addVisited(final Collection<String> visitedIds) {
    InvariantChecks.checkNotNull(visitedIds);

    for (final String visitedId : visitedIds) {
      addVisited(visitedId);
    }
  }

  /**
   * Starts walking the specified model.
   *
   * @param model The model to be walked.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void start(final Model model) {
    InvariantChecks.checkNotNull(model);

    getVisitor().onRootBegin();

    if (isStatus(ModelVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(ModelVisitor.Status.OK)) {

      visitModel(model);

      if (isStatus(ModelVisitor.Status.ABORT)) {
        return;
      }
    }
    getVisitor().onRootEnd();
  }

  /**
   * Starts walking the specified statement.
   *
   * @param statement The statement to be walked.
   */
  public void start(final Statement statement) {
    InvariantChecks.checkNotNull(statement);

    getVisitor().onRootBegin();

    if (isStatus(ModelVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(ModelVisitor.Status.OK)) {

      visitNode(statement);

      if (isStatus(ModelVisitor.Status.ABORT)) {
        return;
      }
    }

    getVisitor().onRootEnd();
  }

  /**
   * Visits the specified node.
   *
   * @param statement node to be visited
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void visitNode(final Statement statement) {
    InvariantChecks.checkNotNull(statement);

    getVisitor().onStatementBegin(statement);

    if (statement.getType().equals(StatementType.MODULE)) {

      visitModule((Module) statement);

    } else if (statement.getType().equals(StatementType.PROCESS)) {

      visitProcess((ModuleProcess) statement);

    } else {
      visitModelStatement(statement);
    }
    getVisitor().onStatementEnd(statement);

    goToNextStatement();
  }

  /**
   * Visits the specified model statement.
   * @param statement The statement to be visited,
   */
  protected abstract void visitModelStatement(final Statement statement);

  /**
   * Process child statements of the specified one.
   *
   * @param statement Statement those children are to be processed.
   */
  protected abstract void processChildren(final Statement statement);

  /**
   * Next statement for walking choosing method.
   */
  protected abstract void goToNextStatement();

  /**
   * Checks rather this walker object has the specified status value.
   *
   * @param status Status value
   * @return {@code true} whether this object has the specified status value,
   *         {@code false} otherwise.
   */
  boolean isStatus(final ModelVisitor.Status status) {
    return getVisitor().getStatus().equals(status);
  }

  /**
   * Visits the specified model object.
   * @param model The model to be visited.
   */
  private void visitModel(final Model model) {

    getVisitor().onModelBegin(model);

    if (isStatus(ModelVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(ModelVisitor.Status.OK)) {

      final Collection<Module> modules = model.getModules();

      for (final Module module : modules) {

        visitModule(module);

        if (isStatus(ModelVisitor.Status.ABORT)) {
          return;
        }
      }
    }
    getVisitor().onModelEnd(model);
  }

  /**
   * Visits the specified module.
   * @param module The module to be visited.
   */
  private void visitModule(final Module module) {

    getVisitor().onModuleBegin(module);

    if (isStatus(ModelVisitor.Status.ABORT)) {
      return;
    }

    if (isStatus(ModelVisitor.Status.OK)) {

      final Collection<ModuleProcess> processes = module.getProcesses();

      for (final ModuleProcess process : processes) {

        visitProcess(process);

        if (isStatus(ModelVisitor.Status.ABORT)) {
          return;
        }
      }

      final Collection<Module> modules = module.getModules();

      for (final Module subModule : modules) {

        visitModule(subModule);

        if (isStatus(ModelVisitor.Status.ABORT)) {
          return;
        }
      }
    }

    getVisitor().onModuleEnd(module);
  }

  /**
   * Visits the specified process.
   * @param node The process to be visited.
   */
  protected void visitProcess(final ModuleProcess node) {

    getVisitor().onProcessBegin(node);

    if (processStatement(node)) {
      return;
    }

    getVisitor().onProcessEnd(node);
  }

  /**
   * Processes the specified model statement.
   * @param node The model statement to be processed.
   * @return {@code true} if argument processing has been successfulle finished,
   *         {@code false} otherwise.
   */
  protected boolean processStatement(final Statement node) {

    if (isStatus(ModelVisitor.Status.ABORT)) {
      return true;
    }

    if (isStatus(ModelVisitor.Status.OK)) {

      processChildren(node);

      if (isStatus(ModelVisitor.Status.ABORT)) {
        return true;
      }
    }
    return false;
  }
}
