/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Statement;

/**
 * Assertion statement of control flow graph model.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgAssertStatement extends CfgAtomicStatement {

  /**
   * Assertion statement severity levels.
   */
  public enum Severity {
    NOTE,
    WARNING,
    ERROR,
    FAILURE,
  }

  private final Severity severityLevel;
  private final Node condition;
  private final String report;

  /**
   * Constructs assertion statement with the specified parameters.
   * @param condition condition
   * @param severityLevel severity level
   * @param report error report
   *
   * @throws IllegalArgumentException when condition argument is {@code null}.
   */
  public CfgAssertStatement(
      final Node condition,
      final Severity severityLevel,
      final String report) {
    super();
    InvariantChecks.checkNotNull(condition);

    this.condition = condition;
    this.severityLevel = severityLevel == null ? Severity.ERROR : severityLevel;
    this.report = report == null ? "Assertion violation." : report;
  }

  private CfgAssertStatement(final CfgAssertStatement statement) {
    super(statement);
    this.severityLevel = statement.getSeverity();
    this.condition = statement.getCondition();
    this.report = statement.getReport();
  }

  /**
   * Returns severity level.
   * @return severity level
   */
  public Severity getSeverity() {
    return this.severityLevel;
  }

  /**
   * Returns condition.
   * @return condition
   */
  public Node getCondition() {
    return this.condition;
  }

  /**
   * Returns error report.
   * @return error report
   */
  public String getReport() {
    return this.report;
  }

  @Override
  public Enum<?> getType() {
    return CfgStatementType.ASSERT;
  }

  @Override
  public Statement deepCopy() {
    return new CfgAssertStatement(this);
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    return String.format(
        "%s %s%s%s",
        getType().name(),
        printer.toString(this.condition),
        getReport() == null ? "" : "report " + getReport(),
        getSeverity() == null ? "" : "severity " + getSeverity());
  }

  @Override
  public boolean equals(final Object object) {

    if (this == object) {
      return true;
    }
    if (object == null || getClass() != object.getClass()) {
      return false;
    }
    if (!super.equals(object)) {
      return false;
    }

    final CfgAssertStatement that = (CfgAssertStatement) object;

    return severityLevel == that.severityLevel
        && getCondition().equals(that.getCondition())
        && (getReport() != null ? getReport().equals(that.getReport()) : that.getReport() == null);
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (severityLevel != null ? severityLevel.hashCode() : 0);
    result = 31 * result + getCondition().hashCode();
    result = 31 * result + (getReport() != null ? getReport().hashCode() : 0);
    return result;
  }
}
