/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.Set;

/**
 * Case statement of control flow graph model.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgCaseStatement extends CfgStatement {

  private final CfgCaseData cfgCaseData;

  /**
   * Constructs the case statement with the specified parameters.
   * @param events The events this statement is sensible to.
   * @param value The case value.
   *
   * @throws IllegalArgumentException when both arguments are {@code null}.
   */
  public CfgCaseStatement(final EventList events, final NodeValue value) {
    InvariantChecks.checkTrue(
        (events == null || events.isEmpty()) && value != null
            || events != null && !events.isEmpty());

    this.cfgCaseData = new CfgCaseData(events, value);
  }

  /**
   * Constructs the case statement with the specified parameters.
   * @param events The events this statement is sensible to.
   * @param values The case values.
   *
   * @throws IllegalArgumentException when value argument is {@code null}.
   */
  public CfgCaseStatement(final EventList events, final Collection<NodeValue> values) {
    InvariantChecks.checkNotNull(values);
    InvariantChecks.checkTrue(
        (events == null || events.isEmpty()) && !values.isEmpty()
            || events != null && !events.isEmpty());

    this.cfgCaseData = new CfgCaseData(events, values);
  }

  /**
   * Constructs case statement of the specified default type.
   *
   * @param defaultNode default type
   */
  public CfgCaseStatement(final boolean defaultNode) {

    this.cfgCaseData = new CfgCaseData(defaultNode);
  }

  /**
   * Constructs case statement with the specified value.
   *
   * @param value case value
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public CfgCaseStatement(final NodeValue value) {
    InvariantChecks.checkNotNull(value);

    this.cfgCaseData = new CfgCaseData(value);
  }

  /**
   * Constructs case statement with the specified values.
   *
   * @param values case values
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeRuntimeException when values are not of the same data type.
   */
  public CfgCaseStatement(final Collection<NodeValue> values) {
    this(null, values);
  }

  private CfgCaseStatement(final CfgCaseStatement statement) {
    super(statement);
    this.cfgCaseData = new CfgCaseData(
        statement.getEventList(), statement.getValues(), statement.isDefault());
  }

  /**
   * Returns the data descriptor for this statement.
   * @return The data descriptor for this statement.
   */
  public CfgCaseData getCfgCaseData() {
    return this.cfgCaseData;
  }

  /**
   * Returns case statement values.
   *
   * @return values
   */
  public Set<NodeValue> getValues() {
    return getCfgCaseData().getValues();
  }

  /**
   * Returns case statement default type.
   *
   * @return default type
   */
  public boolean isDefault() {
    return getCfgCaseData().isValueFree();
  }

  /**
   * Sets default type of this case statement object.
   * @param isDefault default type
   *
   * @throws RetrascopeRuntimeException when argument is {@code true}
   *         and this statement already contains related {@link NodeValue} values
   */
  public void setDefault(final boolean isDefault) {

    if (!getCfgCaseData().setValueFree(isDefault)) {
      throw new RetrascopeRuntimeException(
          String.format("Cannot be both default and have values: %s.", toString()));
    }
  }

  /**
   * Adds value to this case statement object.
   *
   * @param value case value
   *
   * @throws RetrascopeRuntimeException when this object is of default kind.
   */
  public void add(final NodeValue value) {
    InvariantChecks.checkNotNull(value);

    if (isDefault()) {
      throw new RetrascopeRuntimeException(
          String.format("Cannot be both default and have values: %s.", getType()));
    }
    getCfgCaseData().add(value);
  }

  /**
   * Removes the specified value from this case statement.
   *
   * @param value case value
   */
  public void remove(final NodeValue value) {
    InvariantChecks.checkNotNull(value);

    getCfgCaseData().remove(value);
  }

  /**
   * Removes all values of this case statement.
   */
  public void removeValues() {
    getCfgCaseData().removeValues();
  }

  /**
   * Returns event list.
   *
   * @return event list
   */
  public EventList getEventList() {
    return getCfgCaseData().getEventList();
  }

  /**
   * Checks whether this object contains an event.
   *
   * @return {@code true} when this object containt an event,
   *         {@code false} otherwise
   */
  public boolean hasEvents() {
    return getCfgCaseData().hasEvents();
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {
    return getCfgCaseData().getDescription(printer);
  }

  @Override
  public CfgStatementType getType() {
    return CfgStatementType.CASE;
  }

  @Override
  public Statement deepCopy() {
    return new CfgCaseStatement(this);
  }

  @Override
  public void setParent(final Statement parent) {
    checkClass(parent);
    super.setParent(parent);
  }

  @Override
  public void removeParent(final Statement parent) {
    checkClass(parent);
    super.removeParent(parent);
  }

  private static void checkClass(final Statement parent) {

    if (!parent.isType(CfgStatementType.SWITCH)) {
      throw RetrascopeRuntimeException.makeException(LogMessage.WRONG_TYPE, parent);
    }
  }

  @Override
  public boolean equals(final Object object) {

    if (this == object) {
      return true;
    }
    if (object == null || getClass() != object.getClass()) {
      return false;
    }
    if (!super.equals(object)) {
      return false;
    }

    final CfgCaseStatement that = (CfgCaseStatement) object;

    return this.cfgCaseData.equals(that.cfgCaseData);
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    return 31 * result + this.cfgCaseData.hashCode();
  }
}
