/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.util.NodeUtils;

import java.util.Collection;
import java.util.Collections;
import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * Auxiliary class for symbolic execution of block statements with the specified pre-condition.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgContext {

  private final Node pre;
  private final CfgBlockStatement[] blocks;
  private Map<NodeVariable, Deque<NodeVariable>> versionMap;
  private Map<NodeVariable, Node> context;
  private Node performedContext;

  /**
   * Constructs non-elaborated context object from the specified input data.
   * @param pre The pre-condition.
   * @param blocks The sequence of block statements.
   */
  CfgContext(final Node pre, final CfgBlockStatement... blocks) {

    this.pre = pre;
    this.blocks = blocks;
  }

  private static void updateContext(
      final List<Assignment> list,
      final Map<NodeVariable, Deque<NodeVariable>> versionMap,
      final Map<NodeVariable, Node> context) {

    for (final Assignment assignment : list) {

      final Node performed = processAssignment(versionMap, context, assignment);

      final NodeVariable targetVersion = getTargetVersion(versionMap, assignment);

      context.put(targetVersion, performed);
    }
  }

  private static NodeVariable getTargetVersion(
      final Map<NodeVariable, Deque<NodeVariable>> versionMap,
      final Assignment assignment) {

    final NodeVariable target = assignment.getVariable();
    final Deque<NodeVariable> targetVersions = versionMap.get(target);
    return targetVersions.peek();
  }

  private static Node processAssignment(
      final Map<NodeVariable, Deque<NodeVariable>> versionMap,
      final Map<NodeVariable, Node> context,
      final Assignment assignment) {

    final Node performed = performByContext(assignment.getValue(), versionMap, context);

    putToDeque(versionMap, assignment.getVariable());
    return performed;
  }

  private static Node performByContext(
      final Node node,
      final Map<NodeVariable, Deque<NodeVariable>> versionMap,
      final Map<NodeVariable, Node> context) {

    final Collection<NodeVariable> nodeVariables = ExprUtils.getVariables(node);

    final List<NodeBinding.BoundVariable> boundVariables = new LinkedList<>();

    for (final NodeVariable variable : nodeVariables) {

      if (versionMap.containsKey(variable)) {

        final NodeVariable currentVersion = versionMap.get(variable).peek();

        boundVariables.add(
            NodeBinding.bindVariable(
                variable,
                context.containsKey(currentVersion)
                    ? context.get(currentVersion)
                    : currentVersion));
      } else {
        putToDeque(versionMap, variable);
      }
    }
    return NodeUtils.substituteAllBindings(node, boundVariables);
  }

  private static List<Assignment> filterSameVarAssignments(final List<Assignment> assignments) {

    final Map<NodeVariable, Node> map = new LinkedHashMap<>();

    for (final Assignment assignment : assignments) {

      final NodeVariable variable = assignment.getVariable();

      if (map.containsKey(variable)) {
        map.remove(variable);
      }
      map.put(variable, assignment.getValue());
    }

    final List<Assignment> filtered = new LinkedList<>();

    for (final Map.Entry<NodeVariable, Node> entry : map.entrySet()) {
      filtered.add(new Assignment(entry.getKey(), entry.getValue()));
    }
    return filtered;
  }

  private static Node performContext(Map<NodeVariable, Node> context) {

    final Node[] contextEquations = new Node[context.size()];

    int entryNum = 0;
    for (final Map.Entry<NodeVariable, Node> entry : context.entrySet()) {

      final NodeVariable target = entry.getKey();
      final Node value = entry.getValue();
      final DataType targetType = target.getDataType();
      final DataType valueType = value.getDataType();

        /* Modify value when it's type differs from target's one. */
      contextEquations[entryNum] =
          NodeUtils.getEquation(
              target,
              targetType.equals(valueType) ? value : TypeConversion.coerce(value, targetType));
      entryNum++;
    }
    return ExprUtils.getConjunction(contextEquations);
  }

  Map<NodeVariable, Deque<NodeVariable>> getVersionMap() {
    return versionMap;
  }

  public Map<NodeVariable, Node> getContext() {
    return context;
  }

  Node getPerformedContext() {
    return performedContext;
  }

  /**
   * Invokes the context object elaboration.
   * @return The elaborated object itself.
   */
  CfgContext invoke() {

    /* Separate basic blocks to blocking & concurrent. */
    final List<Assignment> blocking = new LinkedList<>();
    final List<Assignment> concurrent = new LinkedList<>();

    for (final CfgBlockStatement blockStatement : blocks) {

      final List<Assignment> blockAssignments = blockStatement.getAssignments();

      if (blockStatement.isConcurrent()) {

        concurrent.addAll(blockAssignments);

      } else {
        blocking.addAll(blockAssignments);
      }
    }

    versionMap = new LinkedHashMap<>();
    context = new LinkedHashMap<>();

    /* Initialize map for variable versions by pre-condition processing. */
    final Collection<NodeVariable> preCondVariables =
        pre == null ? Collections.<NodeVariable>emptyList() : ExprUtils.getVariables(pre);

    for (final NodeVariable variable : preCondVariables) {
      putToDeque(versionMap, variable);
    }

    /* Blocking assignments should be processed first. */
    updateContext(blocking, versionMap, context);

    /* Filter concurrent assignments to the same variable. */
    final List<Assignment> filteredConcurrent = filterSameVarAssignments(concurrent);

    /* Concurrent assignments processing. */
    final Map<NodeVariable, Node> concurrentContext = new LinkedHashMap<>();

    updateContext(filteredConcurrent, versionMap, concurrentContext);

    context.putAll(concurrentContext);

    performedContext = context.isEmpty() ? null : performContext(context);
    return this;
  }

  private static void putToDeque(
      final Map<NodeVariable, Deque<NodeVariable>> versionMap,
      final NodeVariable var) {

    if (!versionMap.containsKey(var)) {
      versionMap.put(var, new LinkedList<NodeVariable>());
    }

    final Deque<NodeVariable> varDeque = versionMap.get(var);

    varDeque.push(varDeque.isEmpty() ? var : getVersion(var, varDeque));
  }

  private static NodeVariable getVersion(final NodeVariable var, final Deque<NodeVariable> deque) {

    final String versionName = getVersionName(var.getName(), deque.size());
    final NodeVariable version = new NodeVariable(versionName, var.getDataType());
    version.setUserData(var.getUserData());

    return version;
  }

  private static String getVersionName(String name, int postfix) {

    return String.format("%s_VERSION_%s", name, postfix);
  }
}
