/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.walker.DfsWalker;
import ru.ispras.retrascope.model.basis.walker.ModelVisitor;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.util.LogMessage;

/**
 * Control flow graph model walker running in a depth-first search mode.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgDfsWalker extends DfsWalker {

  /**
   * Constructs walker object with the specified parameters.
   *
   * @param visitor visitor to be applied to control flow graph model
   * @throws IllegalArgumentException when any argument is {@code null}.
   * @throws IllegalArgumentException when {@code mode} argument equals neither
   * {@link WalkerMode#DFS} nor {@link WalkerMode#DFS_NO_RPT}
   */
  public CfgDfsWalker(
      final ModelVisitor visitor,
      final WalkerMode mode) {
    super(visitor, mode);
    InvariantChecks.checkTrue(mode == WalkerMode.DFS || mode == WalkerMode.DFS_NO_RPT);
  }

  /**
   * Returns the internal visitor.
   * @return The internal visitor.
   */
  private CfgVisitor getCfgVisitor() {
    return (CfgVisitor) getVisitor();
  }

  @Override
  public void goToNextStatement() {
    /* Do nothing. */
  }

  @Override
  protected void visitModelStatement(final Statement statement) {

    final Enum<?> type = statement.getType();

    if (type.equals(CfgStatementType.ASSERT)) {

      visitAssert((CfgAssertStatement) statement);

    } else if (type.equals(CfgStatementType.BASIC_BLOCK)) {

      visitBlock((CfgBlockStatement) statement);

    } else if (type.equals(CfgStatementType.SWITCH)) {

      visitSwitch((CfgSwitchStatement) statement);

    } else if (type.equals(CfgStatementType.CASE)) {

      visitCase((CfgCaseStatement) statement);

    } else if (type.equals(CfgStatementType.WAIT)) {

      visitWait((CfgWaitStatement) statement);

    } else if (type.equals(CfgStatementType.LOOP)) {

      visitLoop((CfgLoopStatement) statement);

    } else {
      throw RetrascopeRuntimeException.makeException(LogMessage.UNSUPPORTED_TYPE, type);
    }
  }

  private void visitLoop(final CfgLoopStatement node) {

    getCfgVisitor().onLoopBegin(node);

    if (processStatement(node)) {
      return;
    }

    getCfgVisitor().onLoopEnd(node);
  }

  private void visitAssert(final CfgAssertStatement assertStatement) {

    getCfgVisitor().onAssertBegin(assertStatement);

    if (processStatement(assertStatement)) {
      return;
    }

    getCfgVisitor().onAssertEnd(assertStatement);
  }

  private void visitSwitch(final CfgSwitchStatement node) {

    getCfgVisitor().onSwitchBegin(node);

    if (processStatement(node)) {
      return;
    }
    getCfgVisitor().onSwitchEnd(node);
  }

  private void visitCase(final CfgCaseStatement node) {

    getCfgVisitor().onCaseBegin(node);

    if (processStatement(node)) {
      return;
    }
    getCfgVisitor().onCaseEnd(node);
  }

  private void visitBlock(final CfgBlockStatement node) {

    getCfgVisitor().onBlockBegin(node);

    if (processStatement(node)) {
      return;
    }
    getCfgVisitor().onBlockEnd(node);
  }

  private void visitWait(final CfgWaitStatement node) {

    getCfgVisitor().onWaitBegin(node);

    if (processStatement(node)) {
      return;
    }
    getCfgVisitor().onWaitEnd(node);
  }
}
