/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Statement;

/**
 * Loop statement for control flow graph node.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgLoopStatement extends CfgStatement {

  /* Loop type. */
  private final Type type;

  /* Loop counter initialization (only for {@code FOR} loops). */
  private final Assignment init;

  /* Loop condition (or iteration number). */
  private final Node condition;

  /* Loop counter increment (only for {@code FOR} loops). */
  private final Assignment increment;

  /**
   * Loop statement types.
   */
  public enum Type {
    /* Forever loop: {@code forever ...}. */
    FOREVER,
    /* Repeat loop: {@code repeat(n) ...}. */
    REPEAT,
    /* While loop: {@code while(c) ...}. */
    WHILE,
    /* For loop: {@code for(a; b; c) ...}. */
    FOR,
  }

  /**
   * Constructs loop statement with the specified parameters.
   * <p>Constructs loop statement of the specified type, with the specified condition,
   * initialization and incrementation assignments on loop counter</p>
   * @param type loop type
   * @param condition loop counter condition
   * @param init loop counter initialization assignment
   * @param increment loop counter incrementation assignment
   *
   * @throws IllegalArgumentException when type argument is {@code null}.
   */
  public CfgLoopStatement(
      final Type type,
      final Node condition,
      final Assignment init,
      final Assignment increment) {
    super();
    InvariantChecks.checkNotNull(type);

    if (type.equals(Type.FOR) || type.equals(Type.REPEAT)) {
      InvariantChecks.checkNotNull(increment);
      InvariantChecks.checkNotNull(init);
    }

    if (!type.equals(Type.FOREVER)) {
      InvariantChecks.checkNotNull(condition);
    }

    this.type = type;
    this.condition = condition;
    this.init = init;
    this.increment = increment;
  }

  private CfgLoopStatement(final CfgLoopStatement statement) {
    super(statement);

    this.type = statement.getLoopType();
    this.init = statement.getInit();
    this.increment = statement.getIncrement();
    this.condition = statement.getCondition();
  }

  /**
   * Returns loop type.
   * @return loop type
   */
  public Type getLoopType() {
    return type;
  }

  /**
   * Returns loop counter initialization assignment.
   * @return loop counter initialization assignment
   */
  public Assignment getInit() {
    return this.init;
  }

  /**
   * Returns loop counter condition.
   * @return loop counter condition
   */
  public Node getCondition() {
    return this.condition;
  }

  /**
   * Returns loop counter incrementation assignment.
   * @return loop counter incrementation assignment
   */
  public Assignment getIncrement() {
    return increment;
  }

  @Override
  public CfgStatementType getType() {
    return CfgStatementType.LOOP;
  }

  @Override
  public Statement deepCopy() {
    return new CfgLoopStatement(this);
  }

  @Override
  public String getDescription(final ExprTreePrinter printer) {

    final String description =
        getCondition() == null
            ? ""
            : String.format(
                "(%s%s%s)",
                getInit() == null ? "" : getInit().getDescription(printer) + ';',
                printer.toString(getCondition()),
                getIncrement() == null ? "" : ';' + getIncrement().getDescription(printer));

    return String.format("%s %s", getLoopType().name(), description);
  }

  @Override
  public boolean equals(final Object object) {

    if (this == object) {
      return true;
    }
    if (object == null || getClass() != object.getClass()) {
      return false;
    }
    if (!super.equals(object)) {
      return false;
    }

    final CfgLoopStatement that = (CfgLoopStatement) object;

    return getType() == that.getType()
        && (getInit() != null
            ? getInit().equals(that.getInit())
            : that.getInit() == null
                && (getCondition() != null
                    ? getCondition().equals(that.getCondition())
                    : that.getCondition() == null
                        && (getIncrement() != null
                            ? getIncrement().equals(that.getIncrement())
                            : that.getIncrement() == null)));
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (getType() != null ? getType().hashCode() : 0);
    result = 31 * result + (getInit() != null ? getInit().hashCode() : 0);
    result = 31 * result + (getCondition() != null ? getCondition().hashCode() : 0);
    result = 31 * result + (getIncrement() != null ? getIncrement().hashCode() : 0);
    return result;
  }
}
