/*
 * Copyright 2013-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.Statement;

import java.util.Map;

/**
 * Module of control flow graph model.
 *
 * <p>Object contains child {@link CfgProcess} processes and {@link CfgModule} sub-modules.
 * When object is an instance of another module, it MUST have the instance name
 * with module name.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgModule extends Module {

  /**
   * Constructs object with the specified name.
   *
   * @param name module name
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public CfgModule(final String name) {
    super(name);
  }

  /**
   * Constructs the object that copies data from thw specified one.
   * <p>The internal statements are omitted in the copy.</p>
   * @param module The object which data is to be copied.
   */
  public CfgModule(final Module module) {
    this(module.getName());
    setInstanceName(module.getInstanceName());
    declareVariables(module.getDeclarationMap());
  }

  private CfgModule(final CfgModule module) {
    super(module);
  }

  /**
   * Instantiates this object with the specified name and bindings.
   *
   * @param instanceName instance name
   * @param bindings instance bindings
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public void instantiate(final String instanceName, final Map<String, Node> bindings) {
    InvariantChecks.checkNotNull(instanceName);
    InvariantChecks.checkNotNull(bindings);

    setInstanceName(instanceName);
    apply(bindings);
  }

  @Override
  public Statement deepCopy() {
    return new CfgModule(this);
  }

  @Override
  public void apply(final Map<String, Node> bindings) {

    /* Change declarations. */
    for (final String varName : bindings.keySet()) {
      if (declares(varName)) {

        /* Remove old declaration for the variable that is substituted by bindings. */
        removeDeclaration(varName);
      }
    }
    super.apply(bindings);
  }
}
