/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.Set;

/**
 * Switch statement of control flow graph model.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgSwitchStatement extends CfgConditionStatement {

  /**
   * Constructs an object with the specified condition.
   *
   * @param condition Switch condition.
   */
  public CfgSwitchStatement(final Node condition) {
    super(condition);
  }

  private CfgSwitchStatement(final CfgSwitchStatement statement) {
    super(statement);
  }

  @Override
  public boolean isSequential() {
    return false;
  }

  /**
   * Sets new {@link Node} condition.
   *
   * @param expr expression
   *
   * @throws IllegalArgumentException when argument is either {@code null}
   *         or not of the same {@link DataType} data type
   *         that its children {@link CfgCaseStatement} nodes have.
   */
  @Override
  public void setCondition(final Node expr) {
    InvariantChecks.checkNotNull(expr);

    super.setCondition(expr);
  }

  @Override
  public CfgStatementType getType() {
    return CfgStatementType.SWITCH;
  }

  @Override
  public Statement deepCopy() {
    return new CfgSwitchStatement(this);
  }

  @Override
  public void add(final Statement statement) {
    checkClass(statement);
    checkValues(this, (CfgCaseStatement)statement);
    super.add(statement);
  }

  @Override
  public void remove(final Statement statement) {
    checkClass(statement);
    super.remove(statement);
  }

  @Override
  public void addAll(final Collection<Statement> statements) {
    InvariantChecks.checkNotNull(statements);

    for (final Statement statement : statements) {
      add(statement);
    }
  }

  private static void checkClass(final Statement statement) {

    if (statement.getType() != CfgStatementType.CASE) {
      throw new RetrascopeRuntimeException(
          LogMessage.makeMessage(LogMessage.WRONG_TYPE, statement));
    }
  }

  private void checkValues(
      final CfgSwitchStatement switchStatement,
      final CfgCaseStatement caseStatement) {

    final Set<NodeValue> caseValues = caseStatement.getValues();
    final boolean isDefault = caseStatement.isDefault();

    final Collection<Statement> otherCases = StatementUtils.getSuccessors(switchStatement);

    for (final Statement otherSuccessor : otherCases) {

      final CfgCaseStatement otherCase = (CfgCaseStatement) otherSuccessor;

      if (isDefault && otherCase.isDefault()) {
        throw new RetrascopeRuntimeException(
            switchStatement.getDescription()
                + " can't include more than one 'default' node.");
      }

      if (CollectionUtils.areIntersectedSets(caseValues, otherCase.getValues())) {
        throw new RetrascopeRuntimeException(
            switchStatement.getDescription()
                + " can't include 2 or more cases with coinciding values.");
      }
    }
  }
}
