/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.basis.walker.ModelWalker;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.ModelTransformUtils;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.util.Collection;
import java.util.Collections;
import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;

/**
 * Utility methods for control flow graph model and it's components.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgUtils {

  private CfgUtils() {
    /* Empty. */
  }

  /**
   * Returns sub-statement with the specified identifier.
   *
   * @param id Statement identifier.
   * @return Sub-statement that has the specified identifier.
   *
   * @throws NoSuchElementException when there is no such sub-statement with the specified
   *         identifier.
   */
  public static Statement getStatement(final Statement statement, final String id) {

    class CfgGetByIdVisitor extends CfgEmptyVisitor {

      private String id;
      private Statement result;

      private CfgGetByIdVisitor(final String id) {
        super();
        this.id = id;
        this.result = null;
      }

      private Statement getStatementWithId() {
        return this.result;
      }

      @Override
      public void onStatementBegin(final Statement statement) {
        if (statement.getId().equals(this.id)) {
          this.result = statement;
        }
      }
    }

    final CfgGetByIdVisitor visitor = new CfgGetByIdVisitor(id);
    final ModelWalker walker = new CfgDfsWalker(visitor, WalkerMode.DFS_NO_RPT);
    walker.visitNode(statement);

    final Statement statementWithId = visitor.getStatementWithId();

    if (statementWithId == null) {
      throw new NoSuchElementException(
          String.format("%s node with '%s' id.", LogMessage.CANT_FIND, id));
    }
    return statementWithId;
  }

  /**
   * Substitutes usages of variable with the specified name
   * by the specified expression in all the elements of the specified collection.
   * <p>
   * Substitutes usages of variable with the specified {@link java.lang.String} name
   * by the specified {@link ru.ispras.fortress.expression.Node} expression
   * in all the elements of the specified collection. The resulting collection
   * consists of node copies of the specified collection. In which node the every
   * usage of the variable with the specified name will be substituted by the
   * specified expression.
   * </p>
   *
   * @param list collection which nodes will be copied and transformed
   * @param name variable name
   * @param node expression
   *
   * @return collection of modified nodes
   */
  public static List<Statement> substitute(
      final List<Statement> list,
      final String name,
      final Node node) {

    final int listSize = list.size();
    final List<Statement> result = new LinkedList<>();
    final Map<Statement, Statement> map = new LinkedHashMap<>(listSize);

    /* Substitute variable with the specified name by the specified node. */
    for (final Statement elem : list) {

      final Statement copy = substituteStatement(elem, name, node);

      map.put(elem, copy);
      result.add(copy);
    }

    /* Establish parent-child links between nodes if they have them. */
    for (int i = 0; i < listSize; i++) {

      final Statement one = list.get(i);

      for (final Statement another : list) {

        if (one.contains(another)) {
          map.get(one).add(map.get(another));
        }
      }
    }

    return result;
  }

  private static Statement substituteStatement(
      final Statement elem,
      final String name,
      final Node node) {

    final Enum<?> elemType = elem.getType();

    if (elemType.equals(CfgStatementType.SWITCH)) {

      final Node switchCondition = ((CfgSwitchStatement) elem).getCondition();
      return new CfgSwitchStatement(Transformer.substitute(switchCondition, name, node));

    } else {

      if (elemType.equals(CfgStatementType.BASIC_BLOCK)) {

        final CfgBlockStatement blockElem = (CfgBlockStatement) elem;
        final List<Assignment> elemAssignments = blockElem.getAssignments();
        final List<Assignment> substAssignments = new LinkedList<>();

        for (final Assignment assign : elemAssignments) {

          Assignment substituted;
          final Node copyValue = Transformer.substitute(assign.getValue(), name, node);

          final RangedVariable target = assign.getTarget();

          if (target.isRanged()) {

            final Range range = target.getRange();
            final Node old = range.getOld();
            final Node young = range.getYoung();

            final Node newOld = TransformerUtils.substitute(old, name, node);
            final Node newYoung = TransformerUtils.substitute(young, name, node);
            final Range newRange = new Range(newYoung, newOld);
            final RangedVariable newRangeVar = new RangedVariable(target.getVariable(), newRange);
            substituted = new Assignment(newRangeVar, copyValue);

          } else {
            substituted = new Assignment(target.getVariable(), copyValue);
          }
          substAssignments.add(substituted);
        }

        return new CfgBlockStatement(substAssignments, blockElem.isConcurrent());

      } else if (elemType.equals(CfgStatementType.ASSERT)) {

        final CfgAssertStatement assertElem = (CfgAssertStatement) elem;
        final Node newCond = Transformer.substitute(assertElem.getCondition(), name, node);

        return new CfgAssertStatement(newCond, assertElem.getSeverity(), assertElem.getReport());

      } else if (elemType.equals(CfgStatementType.WAIT)) {

        final CfgWaitStatement waitElem = (CfgWaitStatement) elem;
        return new CfgWaitStatement(
            waitElem.getEventList(), Transformer.substitute(waitElem.getCondition(), name, node));

      } else {
        throw new RetrascopeRuntimeException(
            LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, elem.getType()));
      }
    }
  }

  /**
   * Constructs an expression from the expression of the specified switch
   * {@link CfgSwitchStatement} node and it's possible values
   * in the specified case {@link CfgCaseStatement} node.
   *
   * @param caseNode case node
   * @param switchNode switch node
   *
   * @return expression
   */
  public static Node getExpression(
      final CfgCaseStatement caseNode,
      final CfgSwitchStatement switchNode) {

    Node expr;

    if (caseNode.isDefault()) {

      /* The 'default' case processing. */
      final Collection<Statement> children = StatementUtils.getSuccessors(switchNode);
      final List<NodeOperation> list = new LinkedList<>();

      for (final Statement child : children) {

        if (!child.equals(caseNode)) {
          list.add(constructExpr((CfgCaseStatement)child, switchNode));
        }
      }
      expr = ExprUtils.getComplement(list.toArray(new Node[list.size()]));
    } else {

      /* The 'normal' case processing. */
      expr = constructExpr(caseNode, switchNode);
    }
    return TransformerUtils.standardize(expr);
  }

  private static NodeOperation constructExpr(
      final CfgCaseStatement child,
      final CfgSwitchStatement parent) {

    Node expr = null;

    for (final NodeValue value: child.getValues()) {

      if (expr == null) {

        expr = NodeUtils.getEquation(parent.getCondition(), value);

      } else {

        expr = Node.OR(expr, NodeUtils.getEquation(parent.getCondition(), value));
      }
    }
    return (NodeOperation) expr;
  }

  /**
   * Creates weakest pre-condition for the specified post-condition and the specified
   * sequence of basic blocks.
   *
   * @param postCondition Expression that is formed after all the specified blocks.
   * @param blocks Sequence of basic blocks.
   * @return Weakest pre-condition.
   */
  private static Node getWeakestPreCondition(
      final Node postCondition,
      final List<CfgBlockStatement> blocks) {

    return ModelTransformUtils.substituteBackward(blocks, postCondition);
  }

  /**
   * Creates the weakest pre-condition for the specified post-condition and the specified
   * basic block.
   *
   * @param postCond The expression that is formed after the specified block.
   * @param block The basic block.
   * @return Weakest pre-condition.
   */
  public static Node getWeakestPreCondition(final Node postCond, final CfgBlockStatement block) {

    return getWeakestPreCondition(postCond, Collections.singletonList(block));
  }

  /**
   * Returns the post-condition for the specified sequence of block statements.
   * @param blocks The sequence of block statements.
   * @return The post-condition for the specified sequence of block statements.
   */
  public static Node getPostCondition(final CfgBlockStatement ... blocks) {

    final CfgContext cfgContext = new CfgContext(null, blocks).invoke();
    return cfgContext.getPerformedContext();
  }

  /**
   * Creates strongest post-condition for the specified conditions and the specified list
   * of block nodes.
   * <p>IMPORTANT NOTE: all the block assignments should be range-free.</p>
   *
   * @param pre Expression that is formed before all the specified blocks (pre-condition).
   * @param post Expression that is formed after all the specified blocks (post-condition).
   * @param blocks The sequence of basic blocks.
   *
   * @return Strongest post-condition.
   */
  public static Node getStrongestPostCondition(
      final Node pre,
      final Node post,
      final CfgBlockStatement ... blocks) {

    final CfgContext cfgContext = new CfgContext(pre, blocks).invoke();

    final Map<NodeVariable, Deque<NodeVariable>> versionMap = cfgContext.getVersionMap();
    final Map<NodeVariable, Node> context = cfgContext.getContext();
    final Node performedContext = cfgContext.getPerformedContext();
    final List<NodeBinding.BoundVariable> boundVariables = new LinkedList<>();

    for (final Map.Entry<NodeVariable, Deque<NodeVariable>> entry : versionMap.entrySet()) {

      final NodeVariable key = entry.getKey();
      final NodeVariable value = entry.getValue().peek();

      if (!key.equals(value)) {
        boundVariables.add(NodeBinding.bindVariable(entry.getKey(), entry.getValue().peek()));
      }
    }
    return context.isEmpty()
        ? post == null ? pre : Node.AND(pre, NodeUtils.substituteAllBindings(post, boundVariables))
        : (post == null
            ? Node.AND(pre, performedContext)
            : ExprUtils.getConjunction(
                  pre,
                  performedContext,
                  NodeUtils.substituteAllBindings(post, boundVariables)));
  }

  /**
   * Returns all the variables and their declarations that are defined at the specified statement
   * and their parents (if they're exist).
   * @param statement The statement for which all the variables that can be used in it are returned.
   * @return The collection of variables and their declarations that can be used or defined
   *         at the specified statement.
   */
  static Map<NodeVariable, Declaration> getDeclared(final Statement statement) {
    InvariantChecks.checkNotNull(statement);

    final Map<NodeVariable, Declaration> declared = new LinkedHashMap<>();

    Statement nearDecStatement = statement;

    while (!(nearDecStatement instanceof DecStatement)) {
      nearDecStatement = statement.getParent();
    }

    addDeclared(declared, (DecStatement) nearDecStatement);

    return declared;
  }

  private static void addDeclared(
        final Map<NodeVariable, Declaration> declared,
        final DecStatement statement) {

    declared.putAll(statement.getDeclarationMap());

    Statement parent = statement.getParent();
    if (parent != null && parent instanceof DecStatement) {
      addDeclared(declared, (DecStatement) parent);
    }
  }
}
