/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cgaa;

import edu.uci.ics.jung.graph.Graph;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.engine.cfg.transformer.cgaa.PhaseSeqEdge;
import ru.ispras.retrascope.model.basis.Model;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.cfg.CfgModel;

import java.util.Collections;
import java.util.Map;

/**
 * Clocked Guarded Atomic Actions (i.e. CGAA) model.
 *
 * <p>The model is control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel} model
 * that is stored in the form of the clocked guarded actions decision diagram.
 * It means that internal model is an acyclic graph, each path of here contains a
 * continuous sub-sequence of switch/case nodes (may be empty) and after that -
 * a continuous sub-sequence of concurrent (non-blocking) basic blocks.
 * The basic block sequence is called "action", the switch/case sequence is called "guard".</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CgaaModel extends Model {

  /**
   * The entity type.
   */
  public static final EntityType TYPE = EntityType.get("cgaa");

  /* The internal model. */
  private final CfgModel model;

  /* Map from model processes to their phase dependency graphs. */
  private final Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> processPhaseMap;

  /**
   * Constructs an object with the specified control flow graph internal model.
   *
   * @param model the internal model
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public CgaaModel(final CfgModel model, final Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> map) {
    super(CgaaModel.TYPE);
    InvariantChecks.checkNotNull(model);

    this.model = model;
    this.processPhaseMap = map;
  }

  /**
   * Adds the specified entity.
   *
   * @param entity the entity to be added
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeRuntimeException when argument is not of {@code CgaaModel} type.
   */
  @Override
  public void add(final Entity entity) {
    InvariantChecks.checkNotNull(entity);
    this.model.add(((CgaaModel) entity).getModel());
  }

  /**
   * Returns the internal {@link CfgModel} model.
   *
   * @return model
   */
  public CfgModel getModel() {
    return this.model;
  }

  /**
   * Returns the mapping from model processes to their phase sequence graphs.
   * @return The mapping from model processes to their phase sequence graphs.
   */
  public Map<ModuleProcess, Graph<Phase, PhaseSeqEdge>> getProcessPhaseMap() {
    return Collections.unmodifiableMap(this.processPhaseMap);
  }
}
