/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.AssignmentContainer;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;
import ru.ispras.retrascope.model.basis.UseDef;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.util.Log;

import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Sequence of {@link Assignment} statements.
 *
 * <p>When non-concurrent, statements MUST be executed in the same order
 * as they are presented in this class. When concurrent, statements are executed in a
 * "parallel" mode.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class Action implements AssignmentContainer,UseDef {

  /* String representation constants. */
  private static final String LEFT_BOUND = "{";
  private static final String ASSIGN_DEL = "; ";
  private static final String RIGHT_BOUND = "}";

  private final Map<NodeVariable, Node> assignments = new LinkedHashMap<>();
  private final boolean isConcurrent;

  /**
   * Constructs object with the specified assignment.
   *
   * @param assignment Assignment to be kept in this object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Action(
      final Assignment assignment,
      final boolean isConcurrent) {
    InvariantChecks.checkNotNull(assignment);

    add(assignment);
    this.isConcurrent = isConcurrent;
  }

  /**
   * Constructs non-concurrent object with the specified assignment.
   *
   * @param assignment Assignment to be kept in this object.
   */
  public Action(final Assignment assignment) {
    this(assignment, false);
  }

  /**
   * Constructs object with the specified list of assignments of the specified concurrency mode.
   *
   * @param list Assignments to be kept in this object.
   * @param isConcurrent Concurrent mode.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public Action(
      final List<Assignment> list,
      final boolean isConcurrent) {
    InvariantChecks.checkNotNull(list);

    for (final Assignment assignment : list) {
      add(assignment);
    }
    this.isConcurrent = isConcurrent;
  }

  /**
   * Creates a completely independent copy of this object.
   * @return The completely independent copy of this object.
   */
  public Action deepCopy() {

    final List<Assignment> assignments = new LinkedList<>();

    for (final Assignment assignment : getAssignments()) {
      assignments.add(assignment.deepCopy());
    }
    return new Action(assignments, isConcurrent());
  }

  /**
   * Constructs non-concurrent object with the specified collection of assignments.
   *
   * @param list Assignments to be kept in this object.
   */
  public Action(final List<Assignment> list) {
    this(list, false);
  }

  @Override
  public void addAssignment(final Assignment stmt) {
    InvariantChecks.checkNotNull(stmt);

    add(stmt);
  }

  @Override
  public void addAssignments(final List<Assignment> statements) {
    InvariantChecks.checkNotNull(statements);

    for (final Assignment stmt : statements) {
      add(stmt);
    }
  }

  private void add(final Assignment assignment) {

    final RangedVariable target = assignment.getTarget();
    final Node value = assignment.getValue();

    if (target == null || value == null) {
      Log.printWarn("Suspicious assignment is added to action block: " + assignment.toString());
      return;
    }

    this.assignments.put(
        assignment.getVariable(),
        target.isRanged() ? RangedVariableUtils.transformNodeWithRanged(target, value) : value);
  }

  @Override
  public List<Assignment> getAssignments() {

    final List<Assignment> list = new LinkedList<>();

    for (final Map.Entry<NodeVariable, Node> entry : this.assignments.entrySet()) {
      list.add(new Assignment(entry.getKey(), entry.getValue()));
    }
    return list;
  }

  @Override
  public void removeAssignments() {
    this.assignments.clear();
  }

  @Override
  public boolean isConcurrent() {
    return this.isConcurrent;
  }

  /**
   * Transforms to concurrent basic block node.
   *
   * @return control flow graph node
   */
  public CfgBlockStatement toBlock() {
    return new CfgBlockStatement(getAssignments(), true);
  }

  /**
   * Returns unique string identifier of this object.
   *
   * @return identifier
   */
  public String getId() {

    final StringBuilder stringBuilder = new StringBuilder(Action.LEFT_BOUND);
    final int length = stringBuilder.length();

    for (final Map.Entry<NodeVariable, Node> entry : this.assignments.entrySet()) {

      if (stringBuilder.length() > length) {
        stringBuilder.append(Action.ASSIGN_DEL);
      }

      stringBuilder.append(new Assignment(entry.getKey(), entry.getValue()));
    }

    stringBuilder.append(Action.RIGHT_BOUND);

    return stringBuilder.toString();
  }

  @Override
  public Set<NodeVariable> getUses() {

    Set<NodeVariable> uses = new LinkedHashSet<>();

    for (final Node node : this.assignments.values()) {

      final Set<NodeVariable> stmtUses = new LinkedHashSet<>(ExprUtils.getVariables(node));

      if (!stmtUses.isEmpty()) {
        uses = CollectionUtils.uniteSets(uses, stmtUses);
      }
    }
    return uses;
  }

  @Override
  public Set<NodeVariable> getDefines() {

    final Set<NodeVariable> defines = new LinkedHashSet<>();
    defines.addAll(this.assignments.keySet());
    return defines;
  }

  @Override
  public int hashCode() {

    final int prime = 31;
    int result = 1;
    result = prime * result + this.assignments.hashCode();

    return result;
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }

    if (obj == null) {
      return false;
    }

    if (!(obj instanceof Action)) {
      return false;
    }

    final Action actionToCompare = (Action) obj;

    return this.assignments.equals(actionToCompare.assignments);
  }

  @Override
  public String toString() {
    return getId();
  }
}
