/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.walker.DfsWalker;
import ru.ispras.retrascope.model.basis.walker.ModelVisitor;
import ru.ispras.retrascope.model.basis.walker.WalkerMode;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;

/**
 * The EFSM walker.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class EfsmDfsWalker extends DfsWalker {

  private final EfsmVisitor visitor;
  private final Collection<EfsmTransition> visitedTransitions;
  private final LinkedList<EfsmState> statePath;

  /**
   * Constructs an EFSM walker for the specified visitor.
   *
   * @param visitor the visitor
   *
   * @throws IllegalArgumentException if {@code visitor} is {@code null}
   */
  public EfsmDfsWalker(final EfsmVisitor visitor) {
    super(visitor, WalkerMode.DFS);
    InvariantChecks.checkNotNull(visitor);

    this.visitor = visitor;
    this.visitedTransitions = new LinkedHashSet<>();
    this.statePath = new LinkedList<>();
  }

  @Override
  protected void visitProcess(final ModuleProcess process) {

    if (process instanceof Efsm) {
      this.visitor.onProcessBegin(process);
      visitEfsm((Efsm) process);
      this.visitor.onProcessEnd(process);
    }
  }

  /**
   * Performs walking across the specified EFSM.
   *
   * @param efsm the EFSM to be walked across
   *
   * @throws IllegalArgumentException if {@code efsm} is {@code null}
   */
  private void visitEfsm(final Efsm efsm) {
    InvariantChecks.checkNotNull(efsm);

    this.statePath.clear();
    this.visitedTransitions.clear();

    /* Take initial state as a first one if exists, otherwise chose random one. */
    EfsmState currentState =
        efsm.hasInitState()
            ? efsm.getInitialState()
            : efsm.getStates().iterator().next();

    visitState(efsm, currentState);
  }

  private void visitState(final Efsm efsm, final EfsmState state) {

    if (this.statePath.contains(state)) {
      return;
    } else {
      this.statePath.add(state);
    }
    this.visitor.onStateBegin(state);

    if (isStatus(ModelVisitor.Status.ABORT)) {
      return;
    }

    final Collection<EfsmTransition> transitionsToVisit = EfsmUtils.getOutgoing(efsm, state);

    visitTransitions(efsm, transitionsToVisit);

    if (isStatus(ModelVisitor.Status.ABORT)) {
      return;
    }

    this.visitor.onStateEnd(state);
    this.statePath.remove(state);
  }

  private void visitTransitions(final Efsm efsm, final Collection<EfsmTransition> transitions) {

    for (final EfsmTransition transition : transitions) {

      if (isStatus(ModelVisitor.Status.ABORT)) {
        return;
      }

      if (this.visitedTransitions.contains(transition)) {
        continue;
      } else {
        this.visitedTransitions.add(transition);
      }

      this.visitor.onTransitionBegin(transition);

      if (isStatus(ModelVisitor.Status.SKIP)) {
        setStatus(ModelVisitor.Status.OK);

      } else {
        final EfsmState dst = transition.getDst();
        visitState(efsm, dst);

        if (isStatus(ModelVisitor.Status.ABORT)) {
          return;
        }
      }
      this.visitor.onTransitionEnd(transition);
    }
  }

  private boolean isStatus(final ModelVisitor.Status status) {
    return this.visitor.getStatus().equals(status);
  }

  private void setStatus(final ModelVisitor.Status status) {
    this.visitor.setStatus(status);
  }
}
