/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.State;

/**
 * A state of an extended finite state machine. The state representation is based on a logical
 * expression in form of a {@link ru.ispras.fortress.expression.Node} instance.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class EfsmState extends State {

  private static final Integer DEFAULT_PHASE = 0;

  private final Node node;
  private final Integer phase;

  /**
   * Constructs a state represented by the specified expression.
   *
   * @param node the expression representing the state
   *
   * @throws IllegalArgumentException if {@code expression} is {@code null}
   */
  public EfsmState(final Node node) {
    this(node, DEFAULT_PHASE);
  }

  /**
   * Constructs the state with the specified expression and phase value.
   * @param node The expression that represents the state.
   * @param phase The phase value.
   */
  public EfsmState(final Node node, final Integer phase) {
    InvariantChecks.checkNotNull(node);

    this.node = node;
    this.phase = phase;
  }

  /**
   * Returns a deep copy of the state.
   * @return A deep copy of the state.
   */
  public EfsmState deepCopy() {
    return new EfsmState(this);
  }

  private EfsmState(final EfsmState state) {
    this.node = state.getNode().deepCopy();
    this.phase = state.getPhase();
  }

  /**
   * Returns an expression which this state is represented by.
   *
   * @return the expression representing this state
   */
  public Node getNode() {
    return this.node;
  }

  /**
   * Returns the associated value of phase variable.
   * @return The associated value of phase variable.
   */
  public Integer getPhase() {
    return this.phase;
  }

  /**
   * Checks whether phase field is equal to {@code null}.
   * @return {@code true} when phase field equals to {@code null}, {@code false} otherwise.
   */
  private boolean phaseFree() {
    return this.phase == null;
  }

  @Override
  public String getId() {
    return String.format("node=%s phase=%s", getNode().toString(), getPhase());
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final EfsmState state = (EfsmState) obj;

    return (phaseFree() ? state.phaseFree() : getPhase().equals(state.getPhase()))
        && getNode().equals(state.getNode());
  }

  @Override
  public int hashCode() {
    int result = getNode() == null ? 0 : getNode().hashCode();
    result = 31 * result + (getPhase() == null ? 0 : getPhase().hashCode());
    return result;
  }

  @Override
  public String toString() {
    return getId();
  }
}
