/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * The guarded action-based representation of a transition of an extended finite state machine.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class EfsmTransition extends GuardedAction {
  private static final String LEFT_STRING_BOUND = "{";
  private static final String SOURCE_STATE_STRING_HEADER = "source state: ";
  private static final String DESTINATION_STATE_STRING_HEADER = "destination state: ";
  private static final String GUARDED_ACTION_STRING_HEADER = "guarded action: ";
  private static final String STRING_DELIMITER = "; ";
  private static final String RIGHT_STRING_BOUND = "}";

  private final EfsmState src;
  private final EfsmState dst;

  /**
   * Constructs a transition between the specified states with the specified guarded action.
   *
   * @param sourceState a source state
   * @param destinationState a destination state
   * @param guardedAction the guarded action
   *
   * @throws IllegalArgumentException if at least one argument is {@code null}
   */
  public EfsmTransition(
      final EfsmState sourceState,
      final EfsmState destinationState,
      final GuardedAction guardedAction) {
    super(guardedAction);
    InvariantChecks.checkNotNull(sourceState);
    InvariantChecks.checkNotNull(destinationState);
    InvariantChecks.checkNotNull(guardedAction);

    this.src = sourceState;
    this.dst = destinationState;
  }

  /**
   * Returns a deep copy of the transition object.
   * @return A deep copy of the transition object.
   */
  public EfsmTransition deepCopy() {
    return new EfsmTransition(this);
  }

  private EfsmTransition(final EfsmTransition transition) {
    super(transition);
    this.src = transition.getSrc().deepCopy();
    this.dst = transition.getDst().deepCopy();
  }

  /**
   * Returns a state which is the start point of this transition.
   *
   * @return the source state of this transition
   */
  public EfsmState getSrc() {
    return this.src;
  }

  /**
   * Returns a state which is the end point of this transition.
   *
   * @return the destination state of this transition
   */
  public EfsmState getDst() {
    return this.dst;
  }

  @Override
  public int hashCode() {

    final int prime = 31;
    int result = 1;
    result = prime * result + super.hashCode();
    result = prime * result + this.dst.hashCode();
    result = prime * result + this.src.hashCode();

    return result;
  }

  /**
   * Compares the specified object with this transition for equality. The object is considered as
   * equal to this transition if and only if it is also an EFSM transition, it connects the same
   * states with the same direction, and it has the same guarded action as this transition does, or
   * more formally:
   * <ul>
   *    <li>it is an instance of the {@code EfsmTransition} class or its subclass;
   *    <li>its source and destination states are equal to this transition ones;
   *    <li>its guarded action is equal to this transition's one.
   * </ul>
   *
   * @param object the object to be compared for equality with this EFSM state
   *
   * @return {@code true} if the specified object is equal to this EFSM state. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(final Object object) {
    if (this == object) {
      return true;
    }

    if (!(object instanceof EfsmTransition) || (this.hashCode() != object.hashCode())) {
      return false;
    }

    final EfsmTransition transitionForComparison = (EfsmTransition) object;

    return (this.src.equals(transitionForComparison.src)
        && this.dst.equals(transitionForComparison.dst)
        && super.equals(transitionForComparison));
  }

  @Override
  public String toString() {
    return EfsmTransition.LEFT_STRING_BOUND
        + EfsmTransition.SOURCE_STATE_STRING_HEADER
        + this.src
        + EfsmTransition.STRING_DELIMITER
        + EfsmTransition.DESTINATION_STATE_STRING_HEADER
        + this.dst
        + EfsmTransition.STRING_DELIMITER
        + EfsmTransition.GUARDED_ACTION_STRING_HEADER
        + super.toString()
        + EfsmTransition.RIGHT_STRING_BOUND;
  }
}
