/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Utility methods for extended finite state machine model and it's components.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class EfsmUtils {

  private EfsmUtils() {
    /* Empty. */
  }

  /**
   * Checks if the specified input is declared either in the EFSM or in it's parent module.
   *
   * @param efsm The EFSM to be checked.
   * @param variableName The name of input variable.
   *
   * @return {@code true} if the input is declared. Otherwise returns {@code false}.
   */
  public static boolean isInputDeclared(final Efsm efsm, final String variableName) {

    return efsm.declaresInput(variableName)
        || efsm.getParentContainer().declaresInput(variableName);
  }

  /**
   * Checks if the specified output is declared either in the EFSM or in it's parent module.
   *
   * @param efsm The EFSM to be checked.
   * @param variableName The name of output variable.
   *
   * @return {@code true} if the output is declared. Otherwise returns {@code false}.
   */
  public static boolean isOutputDeclared(final Efsm efsm, final String variableName) {

    return efsm.declaresOutput(variableName)
        || efsm.getParentContainer().declaresOutput(variableName);
  }

  /**
   * Checks if the specified register declared either in the EFSM or in it's parent module.
   *
   * @param efsm The EFSM to be checked.
   * @param variableName The name of register variable.
   *
   * @return {@code true} if the register is declared. Otherwise returns {@code false}.
   */
  public static boolean isRegisterDeclared(final Efsm efsm, final String variableName) {

    return efsm.declaresRegister(variableName)
      || efsm.getParentContainer().declaresRegister(variableName);
  }

  /**
   * Checks if there is a path from the specified source state to the specified destination state
   * at the specified EFSM..
   *
   * @param efsm The EFSM to be checked.
   * @param sourceState The source state.
   * @param destinationState The destination state.
   *
   * @return {@code true} if there is a path from the source state to the destination state.
   *         Otherwise returns {@code false}
   *
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   *         in this EFSM
   */
  public static boolean isReachable(
      final Efsm efsm,
      final EfsmState sourceState,
      final EfsmState destinationState) {

    if (!efsm.containsState(sourceState)) {
      throw new IllegalArgumentException("The specified source state is not found in this EFSM");
    }

    if (!efsm.containsState(destinationState)) {
      throw new IllegalArgumentException(
        "The specified destination state is not found in this EFSM");
    }

    if (sourceState.equals(destinationState)) {
      return true;
    }

    final List<EfsmState> visitedStates = new LinkedList<>();
    final LinkedList<EfsmState> queue = new LinkedList<>();
    visitedStates.add(sourceState);
    queue.add(sourceState);

    while (!queue.isEmpty()) {

      for (final EfsmState nextState : getDestinationStates(efsm, queue.poll())) {

        if (nextState.equals(destinationState)) {
          return true;
        }

        if (!visitedStates.contains(nextState)) {
          queue.add(nextState);
          visitedStates.add(nextState);
        }
      }
    }

    return false;
  }

  /**
   * Returns transitions of the specified EFSM that end at the specified state.
   *
   * @param efsm The EFSM that contains the specified state.
   * @param state The state of the specified EFSM.
   *
   * @return Collection containing incoming transitions for the specified state.
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public static Collection<EfsmTransition> getIncoming(final Efsm efsm, final EfsmState state) {

    if (!efsm.containsState(state)) {
      throw new IllegalArgumentException(Efsm.STATE_NOT_FOUND + state);
    }

    final List<EfsmTransition> incomingTransitions = new LinkedList<>();

    for (final EfsmTransition transition: efsm.getTransitions()) {

      if (transition.getDst().equals(state)) {
        incomingTransitions.add(transition);
      }
    }

    return incomingTransitions;
  }

  /**
   * Returns transitions of the specified EFSM that start at the specified state.
   *
   * @param efsm The EFSM that contains the specified state.
   * @param state The state of the specified EFSM.
   *
   * @return A list containing outgoing transitions for the specified state.
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  public static Collection<EfsmTransition> getOutgoing(final Efsm efsm, final EfsmState state) {

    if (!efsm.containsState(state)) {
      throw new IllegalArgumentException(Efsm.STATE_NOT_FOUND + state);
    }

    final List<EfsmTransition> outgoingTransitions = new LinkedList<>();

    for (final EfsmTransition transition: efsm.getTransitions()) {

      if (transition.getSrc().equals(state)) {
        outgoingTransitions.add(transition);
      }
    }

    return outgoingTransitions;
  }

  /**
   * Returns states which are end points of outgoing transitions of the specified state.
   *
   * @param efsm The EFSM that contains the specified state.
   * @param state The state of the specified EFSM.
   *
   * @return A list containing the destination states for the specified state.
   *
   * @throws IllegalArgumentException if the specified state is not found in this EFSM
   */
  private static List<EfsmState> getDestinationStates(final Efsm efsm, final EfsmState state) {

    if (!efsm.containsState(state)) {
      throw new IllegalArgumentException(Efsm.STATE_NOT_FOUND + state);
    }

    final Collection<EfsmTransition> outgoingTransitions = getOutgoing(efsm, state);

    if (outgoingTransitions.isEmpty()) {
      return Collections.emptyList();
    }

    final List<EfsmState> destinationStates = new LinkedList<>();

    for (final EfsmTransition transition : outgoingTransitions) {
      destinationStates.add(transition.getDst());
    }

    return destinationStates;
  }

  /**
   * Returns state-like variables for the specified EFSM.
   * @param efsm The EFSM that state-like variables are returned.
   * @return State-like variables for the specified EFSM.
   */
  public static Collection<NodeVariable> getStateVariables(final Efsm efsm) {

    final Collection<NodeVariable> stateVars = new LinkedHashSet<>();

    for (final EfsmState state : efsm.getStates()) {
      stateVars.addAll(ExprUtils.getVariables(state.getNode()));
    }

    return stateVars;
  }

  /**
   * Returns mapping between modified variables and affecting variables.
   * <p>
   * Returns mappings between modified variables and variables which affects corresponding
   * modifications.
   * </p>
   * <br>
   * <p>
   * For example, if this action consists of two subsequent assignments {@code x = y} and
   * {@code z = x + i}, the resulting map contains mappings from {@code x} to {@code y} and from
   * {@code z} to {@code y} and {@code i}.
   * </p>
   *
   * @return a map containing modified variables and variables which affect corresponding
   *         modifications
   */
  public static Map<NodeVariable, Set<NodeVariable>> getResultingDependencies(final Action action) {

    final List<Assignment> assignments = action.getAssignments();
    final Map<NodeVariable, Set<NodeVariable>> substitutions = new HashMap<>(assignments.size());

    for (final Assignment assignment : assignments) {

      final Collection<NodeVariable> usedVariables = ExprUtils.getVariables(assignment.getValue());

      Set<NodeVariable> affecting;

      if (usedVariables.isEmpty()) {

        affecting = Collections.emptySet();

      } else {

        affecting = new HashSet<>(usedVariables.size());

        for (final NodeVariable variable : usedVariables) {

          if (substitutions.containsKey(variable)) {

            affecting.addAll(substitutions.get(variable));

          } else {
            affecting.add(variable);
          }
        }
      }
      substitutions.put(assignment.getVariable(), affecting);
    }
    return substitutions;
  }

  /**
   * Returns mapping between variables and corresponding substitutions.
   * <p>
   * Returns mappings between variables and corresponding substitutions representing total variable
   * modifications during execution of this action. That is, execution of this action is equal to
   * parallel evaluation of the substitutions and subsequent assignment of derived values to the
   * corresponding variables.
   * </p>
   *
   * @return a map containing modified variables and the substitutions representing these
   *         modifications
   */
  public static Map<NodeVariable, Node> getResultingSubstitutions(final Action action) {

    final List<Assignment> assignments = action.getAssignments();
    final Map<NodeVariable, Node> substitutions = new LinkedHashMap<>();

    for (final Assignment assignment : assignments) {
      Node expression = assignment.getValue();

      for (final NodeVariable variable : ExprUtils.getVariables(expression)) {

        if (substitutions.containsKey(variable)) {
          expression =
              Transformer.substitute(expression, variable.getName(), substitutions.get(variable));
        }
      }

      final RangedVariable rangedVariable = assignment.getTarget();
      final NodeVariable nodeVariable = rangedVariable.getVariable();

      if (rangedVariable.isRanged()) {

        final DataType dataType = nodeVariable.getDataType();

        if (!dataType.getTypeId().equals(DataTypeId.BIT_VECTOR)) {

          throw new IllegalArgumentException(
              LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, dataType));
        }

        final Range range = rangedVariable.getRange();

        if (!(range.getOld() instanceof NodeValue) || !(range.getYoung() instanceof NodeValue)) {

          throw new IllegalArgumentException(LogMessage.UNSUPPORTED_TYPE + "varying range bounds");
        }

        final NodeValue leftValue = (NodeValue) range.getYoung();
        final NodeValue rightValue = (NodeValue) range.getOld();

        final int leftBound = leftValue.getInteger().intValue();
        final int rightBound = rightValue.getInteger().intValue();

        if (leftBound < rightBound) {
          throw new IllegalArgumentException(LogMessage.UNSUPPORTED_TYPE.toString());
        }

        NodeOperation mostSignPart = null;
        NodeOperation lessSignPart = null;
        final int variableSize = dataType.getSize();
        final int decRadix = 10;

        if (leftBound < variableSize - 1) {

          final NodeValue leftIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(variableSize - 1), decRadix));
          final NodeValue rightIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(leftBound + 1), decRadix));

          mostSignPart =
              new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex, nodeVariable);
        }

        if (rightBound > 0) {

          final NodeValue leftIndex =
              new NodeValue(DataType.INTEGER.valueOf(String.valueOf(rightBound - 1), decRadix));
          final NodeValue rightIndex = new NodeValue(DataType.INTEGER.valueOf("0", decRadix));

          lessSignPart =
              new NodeOperation(StandardOperation.BVEXTRACT, leftIndex, rightIndex, nodeVariable);
        }

        if (mostSignPart == null || lessSignPart == null) {

          if (mostSignPart == null && lessSignPart != null) {

            expression = new NodeOperation(StandardOperation.BVCONCAT, expression, lessSignPart);
          } else {
            expression = new NodeOperation(StandardOperation.BVCONCAT, mostSignPart, expression);
          }
        } else {
          expression =
              new NodeOperation(StandardOperation.BVCONCAT, mostSignPart, expression, lessSignPart);
        }
      }

      if (substitutions.containsKey(nodeVariable)) {
        substitutions.remove(nodeVariable);
      }

      substitutions.put(nodeVariable, expression);
    }

    return substitutions;
  }

  /**
   * Constructs a synchronous composition from a list of EFSMs.
   *
   * @param efsms the list of EFSMs
   * @return the synchronous composition of the EFSMs
   * @throws IllegalArgumentException if the specified list of EFSMs is {@code null}
   *         or contains less than two models
   */
  public static Efsm product(final Efsm ... efsms) {

    if ((efsms == null) || (efsms.length < 2)) {
      throw new IllegalArgumentException("The number of EFSMs should be greater than 1!");
    }

    // This data structure represents a state with its outgoing transitions.

    class StateWithTransitions {
      private EfsmState state;
      private List<EfsmTransition> transitions;

      StateWithTransitions(EfsmState state, List<EfsmTransition> transitions) {
        this.state = state;
        this.transitions = transitions;
      }
    }

    // Constructing a new EFSM name.

    final StringBuilder builder = new StringBuilder(efsms[0].getName());

    for (int i = 1; i < efsms.length; i++) {
      builder.append(" || ").append(efsms[i].getName());
    }
    final Efsm product = new Efsm(new EfsmModule("product"), builder.toString());

    // Each EFSM is represented as a list of states with transitions
    final List<List<StateWithTransitions>> oldStates = new ArrayList<>(efsms.length);
    final List<EfsmState> initialStates = new ArrayList<>();

    for (Efsm efsm : efsms) {

      // Collecting initial states.
      initialStates.add(efsm.getInitialState());

      // Collecting input, output and inner signals.
      final List<EfsmModule> parentModules = getAllParentModules(efsm.getParentContainer());

      for (EfsmModule module : parentModules) {

        for (NodeVariable variable : module.getVariables()) {

          if (!product.declares(variable)) {
            product.declareVariable((NodeVariable) variable.deepCopy());
          }
        }
      }
      for (NodeVariable variable : efsm.getVariables()) {

        if (!product.declares(variable)) {
          product.declareVariable((NodeVariable) variable.deepCopy());
        }
      }

      final List<StateWithTransitions> swt = new ArrayList<>(efsm.getStateNum());

      for (EfsmState state : efsm.getStates()) {

        final Collection<EfsmTransition> trans = getOutgoing(efsm, state);

        if (trans.isEmpty()) {

          // If a state has no outgoing transitions, a new transition is constructed
          // (guard = true, action = null). This is needed by the construction algorithm.

          trans.add(
              new EfsmTransition(
                  state,
                  state,
                  new GuardedAction(
                      SolverUtils.TRUE,
                      new Action(new Assignment(), true))));
        } else {

          final List<Node> guards = new ArrayList<>();

          // If the guards of the outgoing transitions are not complete (a situation is
          // possible when no transition is enabled), a negation of disjunction of guards
          // is constructed.

          for (final EfsmTransition item : trans) {
            guards.add(item.getGuardNode());
          }

          if (!areComplete(trans)) {
            trans.add(
                new EfsmTransition(
                    state,
                    state,
                    new GuardedAction(guards, new Action(new Assignment(), true))));
          }
        }
        swt.add(new StateWithTransitions(state, new ArrayList<>(trans)));
      }
      oldStates.add(swt);
    }

    // Constructing a new initial state.

    Node initialStateNode = null;

    for (final EfsmState initState : initialStates) {

      if (initialStateNode == null) {

        initialStateNode = initState.getNode();
      } else {
        initialStateNode = Node.AND(initialStateNode, initState.getNode());
      }
    }

    final EfsmState newInitialState = new EfsmState(initialStateNode);

    // Combining old EFSM states to construct new EFSM states.

    final List<List<StateWithTransitions>> stateCombinations = combineLists(oldStates);

    final Map<EfsmState, List<List<EfsmTransition>>> transitionsMapping = new LinkedHashMap<>();

    // Combining transitions of new state combinations.

    for (final List<StateWithTransitions> combinedStateList : stateCombinations) {

      Node stateNode = null;
      final List<List<EfsmTransition>> newStateTransitions = new ArrayList<>();
      final List<List<EfsmTransition>> combinedTransitions;

      for (final StateWithTransitions item : combinedStateList) {

        if (stateNode == null) {
          stateNode = item.state.getNode();
        } else {
          stateNode = Node.AND(stateNode, item.state.getNode());
        }
        newStateTransitions.add(item.transitions);
      }

      combinedTransitions = combineLists(newStateTransitions);

      // Creating a new state.

      final EfsmState newState = product.addState(stateNode);
      transitionsMapping.put(newState, combinedTransitions);

      if (newState.equals(newInitialState)) {
        product.setInitialState(newState);
      }

    }

    // Constructing new transitions.

    for (final Map.Entry<EfsmState, List<List<EfsmTransition>>> entry :
        transitionsMapping.entrySet()) {

      for (final List<EfsmTransition> item : entry.getValue()) {

        final EfsmState destState = getNewDestinationState(item);

        if (transitionsMapping.keySet().contains(destState)) {

          final GuardedAction ga = constructGuardedAction(item);

          if (ga != null) {
            product.addTransition(entry.getKey(), destState, ga);
          }
        }
      }
    }

    return product;
  }

  private static List<EfsmModule> getAllParentModules(final EfsmModule module) {
    final List<EfsmModule> modules = new ArrayList<>();

    modules.add(module);

    final EfsmModule parentModule = (EfsmModule) module.getParent();

    if (parentModule != null) {

      final List<EfsmModule> parentModules = getAllParentModules(parentModule);

      if (!parentModules.isEmpty()) {
        modules.addAll(parentModules);
      }
    }
    return modules;
  }

  /**
   * Checks whether the guards of the specified transitions are complete.
   *
   * @param transitions the specified {@link Collection} of {@link EfsmTransition}
   * @return {@code true} if the guards are complete, {@code false} otherwise
   */
  private static boolean areComplete(final Collection<EfsmTransition> transitions) {

    Node constraint = null;
    final List<Node> guards = new ArrayList<>();

    for (final EfsmTransition item : transitions) {

      guards.add(item.getGuardNode());

      if (constraint == null) {
        constraint = item.getGuardNode();
      } else {
        constraint = Node.AND(constraint, item.getGuardNode());
      }
    }

    return SolverUtils.areComplete(guards.toArray(new Node[guards.size()]));
  }

  /**
   * Performs a Cartesian product of several lists (sets) of elements.
   *
   * @param lists {@link List} containing the lists to product
   * @return the new {@link List} containing a Cartesian product of the input lists
   * @throws IllegalArgumentException when argument is {@code null}
   */
  private static <T> List<List<T>> combineLists(final List<List<T>> lists) {
    InvariantChecks.checkNotNull(lists);

    int combinations = 1;

    for (final List<T> item : lists) {
      combinations = combinations * item.size();
    }

    final List<List<T>> product = new ArrayList<>();

    for (int i = 0; i < combinations; i++) {

      final List<T> combination = new ArrayList<>();
      int index = 1;

      for (final List<T> item : lists) {
        combination.add(item.get((i / index) % item.size()));
        index = index * item.size();
      }
      product.add(combination);
    }
    return product;
  }

  /**
   * Constructs a new EfsmState according to the destinations of transitions, contained in the
   * specified collection.
   *
   * @param transitions the specified {@link Collection} of {@link EfsmTransition}
   * @return a new {@link EfsmState}
   */
  private static EfsmState getNewDestinationState(final Collection<EfsmTransition> transitions) {

    final List<EfsmState> result = new ArrayList<>();
    Node constraint = null;

    for (final EfsmTransition item : transitions) {
      result.add(item.getDst());
    }

    for (final EfsmState item : result) {
      if (constraint == null) {
        constraint = item.getNode();
      } else {
        constraint = Node.AND(constraint, item.getNode());
      }
    }

    return new EfsmState(constraint);

  }

  /**
   * Constructs a single {@link GuardedAction} from the guarded actions of the specified
   * transitions.
   *
   * @param transitions a collection of the transitions
   * @return the new guarded action
   */
  private static GuardedAction constructGuardedAction(
      final Collection<EfsmTransition> transitions) {

    final List<Node> guardPredicates = new ArrayList<>();
    final List<Assignment> assignments = new ArrayList<>();
    final EventList events = new EventList();

    for (final EfsmTransition item : transitions) {

      guardPredicates.add(item.getGuardNode());
      events.addEvents(item.getClocks());

      if (item.getAction() != null) {

        final Action action = item.getAction();

        for (final Assignment atomic : action.getAssignments()) {

          if (atomic.getValue() != null) {
            assignments.add(atomic);
          }
        }
      }
    }

    if (!SolverUtils.areCompatible(guardPredicates.toArray(new Node[guardPredicates.size()]))) {
      return null;
    }

    final GuardedAction ga = new GuardedAction(guardPredicates, new Action(assignments, true));
    ga.addClocks(events);

    return ga;
  }

  /**
   * Checks whether the specified EFSM (or it's ancestor) declares the specified variable.
   *
   * @param efsm EFSM to be checked if it (or it's ancestor) declares variable
   * @param variable variable to be checked for declaration
   * @return {@code true} if the specified EFSM (or some it's ancestor) declares the specified
   *         variable, {@code false} otherwise
   */
  public static boolean hasDeclaringAncestor(final Efsm efsm, final NodeVariable variable) {

    if (efsm.declares(variable)) {

      return true;
    } else {

      EfsmModule module = efsm.getParentContainer();

      while (module != null) {

        if (module.declares(variable)) {
          return true;
        }
        module = (EfsmModule) module.getParent();
      }
      return false;
    }
  }

  /**
   * Looks for the shortest path at the specified EFSM from the specified source state
   * to the specified destination state with the specified principle of transition
   * weight defining. This method is based on the Dijkstra's algorithm.
   * <br>
   * Currently CONST and GUARD_REGISTER_AMOUNT TransitionWeightBasis types are supported.
   *
   * @param efsm the EFSM
   * @param sourceState the source state
   * @param destinationState the destination state
   * @param weightBasis the principle used for transition weight computing
   *
   * @return a list containing the transitions of found path.
   *         It's empty if the source state matches
   *         to the destination state. If there is no path
   *         from the source state to the destination
   *         state {@code null} is returned
   *
   * @throws IllegalArgumentException if the specified source and/or destination state is not found
   *         in this EFSM or if unsupported
   *         {@link ru.ispras.retrascope.model.efsm.Efsm.TransitionWeightBasis} is passed.
   */
  public static List<EfsmTransition> getShortestPath(
      final Efsm efsm,
      final EfsmState sourceState,
      final EfsmState destinationState,
      final Efsm.TransitionWeightBasis weightBasis) {

    if (!efsm.containsState(sourceState)) {
      throw new IllegalArgumentException("The specified source state is not found in this EFSM");
    }

    if (!efsm.containsState(destinationState)) {
      throw new IllegalArgumentException(
        "The specified destination state is not found in this EFSM");
    }

    if (sourceState.equals(destinationState)) {
      return Collections.emptyList();
    }

    if (!isReachable(efsm, sourceState, destinationState)) {
      return null;
    }

    final LinkedHashMap<EfsmState, Integer> distances = new LinkedHashMap<>(efsm.getStateNum());
    final LinkedHashMap<EfsmState, EfsmTransition> previousNodes =
        new LinkedHashMap<>(efsm.getStateNum());

    for (final EfsmState state : efsm.getStates()) {
      distances.put(state, Integer.MAX_VALUE);
      previousNodes.put(state, null);
    }

    distances.put(sourceState, 0);
    final LinkedHashMap<EfsmTransition, Integer> transitionWeights = new LinkedHashMap<>();

    for (final EfsmTransition transition : efsm.getTransitions()) {

      switch (weightBasis) {

        case CONSTANT:
          transitionWeights.put(transition, 1);
          break;

        case GUARD_REGISTER_AMOUNT:
          int weight = 0;

          for (final NodeVariable variable : ExprUtils.getVariables(transition.getGuardNode())) {
            final String variableName = variable.getName();
            if (efsm.declaresRegister(variableName)
                || efsm.getParentContainer().declaresRegister(variableName)) {
              weight++;
            }
          }

          transitionWeights.put(transition, weight);
          break;

        default:
          throw new IllegalArgumentException("Unsupported weight defining principle: "
            + weightBasis);
      }
    }

    final Comparator<EfsmTransition> transitionComparator = new Comparator<EfsmTransition>() {
      @Override
      public int compare(final EfsmTransition transition1, final EfsmTransition transition2) {
        return transitionWeights.get(transition1) - transitionWeights.get(transition2);
      }
    };

    final Comparator<EfsmState> stateComparator = new Comparator<EfsmState>() {
      @Override
      public int compare(final EfsmState state1, final EfsmState state2) {
        return distances.get(state1) - distances.get(state2);
      }
    };
    final List<EfsmState> visitedStates = new LinkedList<>();
    visitedStates.add(sourceState);
    final LinkedList<EfsmState> queue = new LinkedList<>();
    queue.add(sourceState);

    while (!queue.isEmpty()) {

      final EfsmState currentState = queue.poll();
      final int currentDistance = distances.get(currentState);
      final LinkedList<EfsmTransition> nextTransitions =
          new LinkedList<>(getOutgoing(efsm, currentState));
      Collections.sort(nextTransitions, transitionComparator);

      for (final EfsmTransition nextTransition : nextTransitions) {

        final EfsmState nextState = nextTransition.getDst();
        final int newDistance = currentDistance + transitionWeights.get(nextTransition);

        if (newDistance < distances.get(nextState)) {

          previousNodes.put(nextState, nextTransition);
          distances.put(nextState, newDistance);

          if (nextState.equals(destinationState)) {
            break;
          }

          if (!visitedStates.contains(nextState)) {
            visitedStates.add(nextState);
            queue.add(nextState);
          }
        }
      }

      Collections.sort(queue, stateComparator);
    }

    final LinkedList<EfsmTransition> result = new LinkedList<>();
    EfsmState currentState = destinationState;

    while (previousNodes.get(currentState) != null) {

      result.addFirst(previousNodes.get(currentState));
      currentState = previousNodes.get(currentState).getSrc();
    }

    return result;
  }
}
