/*
 * Copyright 2013-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.efsm;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfo;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.ModelUtils;
import ru.ispras.retrascope.model.basis.UseDef;

import java.util.Collection;
import java.util.Set;

/**
 * Clocked guarded action.
 *
 * <p>Clocked guarded action is a triple: a collection of clock
 * {@link ru.ispras.retrascope.model.basis.Event} events, a guard
 * {@link ru.ispras.retrascope.model.efsm.Guard} condition,
 * and an {@link ru.ispras.retrascope.model.efsm.Action} object.</p>
 * <br>
 * <p>When some clock event happens and guard is {@code true}
 * then the action proceeds.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 *
 */
public class GuardedAction extends MetaInfo implements UseDef {

  private final Guard guard;
  private final Action action;
  private final EventList clocks;

  /**
   * Constructs object with the specified guard, action and clocks.
   * @param guard Guard predicate.
   * @param action Action block.
   * @param clocks Clock events.
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  private GuardedAction(final Guard guard, final Action action, final EventList clocks) {
    super();
    InvariantChecks.checkNotNull(guard);
    InvariantChecks.checkNotNull(action);
    InvariantChecks.checkNotNull(clocks);

    this.guard = guard;
    this.action = action;
    this.clocks = new EventList();

    for (final Event clock : clocks) {
      addClock(clock);
    }
  }

  /**
   * Constructs object with the specified node guard, action and clock events.
   * @param guard Guard predicate.
   * @param action Action block.
   * @param clocks Clock events.
   */
  public GuardedAction(final Node guard, final Action action, final EventList clocks) {

    this(new Guard(guard), action, clocks);
  }

  /**
   * Constructs object with the specified guard condition and action.
   *
   * @param guard guard
   * @param action action
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public GuardedAction(final Guard guard, final Action action) {

    this(guard, action, new EventList());
  }

  /**
   * Constructs object with the specified guard expression node and action.
   *
   * @param guardNode Guard expression.
   * @param action Action.
   */
  public GuardedAction(final Node guardNode, final Action action) {

    this(new Guard(guardNode), action);
  }

  /**
   * Constructs object from the specified guard predicates and the specified action.
   *
   * @param predicates Guard predicates.
   * @param action Action.
   */
  public GuardedAction(final Collection<Node> predicates, final Action action) {

    this(new Guard(predicates), action);
  }

  /**
   * Returns a deep copy of the object.
   * @return A deep copy of the object.
   */
  public GuardedAction deepCopy() {
    return new GuardedAction(this);
  }

  /**
   * Creates a copy of the specified object.
   * @param guardedAction The object this is created from.
   */
  public GuardedAction(final GuardedAction guardedAction) {
    this.guard = guardedAction.getGuard().deepCopy();
    this.action = guardedAction.getAction().deepCopy();
    this.clocks = guardedAction.getEventList().deepCopy();
  }

  /**
   * Returns action.
   *
   * @return action
   */
  public Action getAction() {
    return this.action;
  }

  /**
   * Returns guard's node.
   * @return The guard's node.
   */
  public Node getGuardNode() {
    return getGuard().toNode();
  }

  /**
   * Returns a conjunction of guard's node and a node formed by clock events.
   * @return The conjunction of guard's node and a node formed by clock events.
   */
  public Node getClockGuardNode() {

    final Node clockNode = getClockNode();
    final Node guardNode = getGuardNode();
    return clockNode == null ? guardNode : Node.AND(clockNode, guardNode);
  }

  /**
   * Returns a node that is formed by clock events.
   * @return A node that is formed by clock events or {@code null} if there is no such.
   */
  public Node getClockNode() {
    return getClocks().isEmpty() ? null : ModelUtils.eventsToNode(getClocks());
  }

  /**
   * Returns guard.
   *
   * @return guard
   */
  private Guard getGuard() {
    return this.guard;
  }

  /**
   * Returns list of object's clock events.
   *
   * @return clocks list
   */
  private EventList getEventList() {
    return this.clocks;
  }

  /**
   * Returns clock events.
   *
   * @return clock events collection
   */
  public EventList getClocks() {
    return this.clocks;
  }

  /**
   * Checks whether this object does not contain clocks.
   *
   * @return {@code true} if this object does not contain clocks, {@code false} otherwise.
   */
  public boolean clockFree() {

    return getClocks().isEmpty();
  }

  /**
   * Adds the specified clock event to this object's collection.
   *
   * @param clock clock event
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addClock(final Event clock) {
    InvariantChecks.checkNotNull(clock);

    this.clocks.add(clock);
  }

  /**
   * Adds the specified clock events collection to the object's collection.
   *
   * @param clocks clocks collection
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addClocks(final EventList clocks) {
    InvariantChecks.checkNotNull(clocks);

    for (final Event clock : clocks) {
      addClock(clock);
    }
  }

  /**
   * Returns variables that are used in {@link Guard} condition.
   *
   * @return sequence of variables
   */
  public Set<NodeVariable> getGuardUses() {
    return this.getGuard().getUses();
  }

  /**
   * Returns variables that are used in {@link Action} action.
   *
   * @return sequence of variables
   */
  public Set<NodeVariable> getActionUses() {
    return this.getAction().getUses();
  }

  /**
   * Checks that the specified event is in the object's clock collection.
   *
   * @param event event
   *
   * @return {@code true} if the specified event is in the object's clock collection
   *         or the list is empty, {@code false} otherwise.
   */
  private boolean isSensible(final Event event) {
    return this.clocks.isTriggered(event);
  }

  /**
   * Checks if this guarded action is sensible to the specified events.
   *
   * @param eventList an event list to be tested
   *
   * @return {@code true} if at least one of the specified events
   *         is in the object's clock collection or it is empty, {@code false} otherwise.
   */
  public boolean isSensible(final EventList eventList) {

    for (final Event event : eventList) {
      if (this.isSensible(event)) {
        return true;
      }
    }

    return false;
  }

  @Override
  public Set<NodeVariable> getUses() {
    return CollectionUtils.uniteSets(this.getGuard().getUses(), this.getAction().getUses());
  }

  @Override
  public Set<NodeVariable> getDefines() {
    return this.getAction().getDefines();
  }

  @Override
  public String toString() {

    final StringBuilder builder = new StringBuilder();
    builder.append(String.format("{%s: %s -> {", this.clocks.toString(), getGuard()));

    if (this.getAction() != null) {
      builder.append(this.getAction());
    }
    builder.append("}}");
    return builder.toString();
  }

  @Override
  public boolean equals(final Object object) {

    if (this == object) {
      return true;
    }
    if (object == null || getClass() != object.getClass()) {
      return false;
    }
    if (!super.equals(object)) {
      return false;
    }

    final GuardedAction that = (GuardedAction) object;

    return getGuard() == null
        ? that.getGuard() == null
            && (getAction() == null
                ? that.getAction() == null
                    && (getClocks() == null
                        ? that.getClocks() == null
                        : getClocks().equals(that.getClocks()))
                : getAction().equals(that.getAction()))
        : getGuard().equals(that.getGuard());
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (getGuard() != null ? getGuard().hashCode() : 0);
    result = 31 * result + (getAction() != null ? getAction().hashCode() : 0);
    result = 31 * result + (getClocks() != null ? getClocks().hashCode() : 0);
    return result;
  }
}
