/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.Collection;

/**
 * This class represents the non-terminal condition node of a HLDD.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class HlddConditionNode extends HlddNode implements HlddNonTerminalNode {

  // Branch condition.
  private Node expression;
  // Node edges.
  private Collection<HlddEdge> children;

  /**
   * Constructs a new condition node.
   *
   * @param parent - the specified parent.
   * @param expression - the specified branch condition.
   */
  public HlddConditionNode(final HlddNode parent, final Node expression) {
    super(parent, NodeType.CONDITION);
    setExpression(expression);
    children = new ArrayList<>();
  }

  /**
   * Constructs a new condition node.
   *
   * @param parent - the specified parent.
   */
  public HlddConditionNode(final HlddNode parent) {
    super(parent, NodeType.CONDITION);
    children = new ArrayList<>();
  }

  /**
   * Constructs a new condition node.
   */
  public HlddConditionNode() {
    super(NodeType.CONDITION);
    children = new ArrayList<>();
  }

  private HlddConditionNode(final HlddConditionNode node) {
    super(node);
    setExpression(node.getExpression());
    children = new ArrayList<>();
  }

  /**
   * Returns the branch condition.
   *
   * @return The branch condition.
   */
  public Node getExpression() {
    return expression;
  }

  /**
   * Sets the branch expression.
   *
   * @param expression - the specified condition.
   */
  private void setExpression(final Node expression) {
    InvariantChecks.checkNotNull(expression);
    this.expression = expression;
  }

  @Override
  public HlddNode deepCopy() {
    return new HlddConditionNode(this);
  }

  @Override
  public Collection<HlddNode> getChildren() {
    return new ArrayList<HlddNode>(children);
  }

  @Override
  public void addChild(final HlddNode child) {
    InvariantChecks.checkNotNull(child);
    children.add((HlddEdge) child);
  }

  /**
   * Adds the specified collection of {@link HlddNode} to the children of the node.
   *
   * @param children - the specified collection of child nodes.
   */
  public void addChildren(final Collection<HlddNode> children) {
    for (HlddNode child : children) {
      addChild(child);
    }
  }

  @Override
  public void removeChild(final HlddNode node) {
    InvariantChecks.checkNotNull(node);
    children.remove(node);
  }

  @Override
  public boolean isNonTerminal() {
    return true;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + ((expression == null) ? 0 : expression.hashCode());
    return result;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (!super.equals(obj)) {
      return false;
    }
    if (getClass() != obj.getClass()) {
      return false;
    }
    final HlddConditionNode other = (HlddConditionNode) obj;
    if (expression == null) {
      if (other.expression != null) {
        return false;
      }
    } else if (!expression.equals(other.expression)) {
      return false;
    }
    return true;
  }

  @Override
  public String toString() {
    return "HlddConditionNode [expression=" + expression + "]";
  }

  /**
   * Shows whether the node has a default child edge.
   * @return {@code true} if the node has a default children edge, {@code false} otherwise.
   */
  public boolean hasDefaultEdge() {
    return getDefaultEdge() != null;
  }

  /**
   * Returns the default child edge.
   * @return The default child edge, if exists.
   */
  public HlddEdge getDefaultEdge() {
    for (HlddEdge child : children) {
      if (child.isDefault()) {
        return child;
      }
    }
    return null;
  }

}
