/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.util.NodeUtils;

import java.util.ArrayList;
import java.util.Collection;

/**
 * This class represents the edge node of a HLDD.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public class HlddEdge extends HlddNode implements HlddNonTerminalNode {

  // Edge values.
  private Collection<Node> expressions;
  // Indicates that this node represents a default value.
  private boolean isDefault;
  // Child node.
  private HlddNode child;

  private EventList eventList;

  /**
   * Constructs a new edge node.
   *
   * @param parent - the specified parent.
   */
  public HlddEdge(final HlddNode parent) {
    super(parent, NodeType.EDGE);
    expressions = new ArrayList<>();
    isDefault = false;
  }

  /**
   * Constructs a new edge node.
   */
  public HlddEdge() {
    super(NodeType.EDGE);
    expressions = new ArrayList<>();
    isDefault = false;
  }

  private HlddEdge(final HlddEdge node) {
    super(node);
    expressions = new ArrayList<>();
    addExpressions(node.getExpressions());
    if (node.isDefault()) {
      setDefault();
    }
  }

  /**
   * Returns the edge expressions.
   *
   * @return The edge expressions.
   */
  public Collection<Node> getExpressions() {
    return expressions;
  }

  /**
   * Adds expressions to the edge.
   *
   * @param expressions - the specified expressions.
   */
  public void addExpressions(final Collection<? extends Node> expressions) {
    InvariantChecks.checkNotNull(expressions);
    for (Node expr : expressions) {
      addExpression(expr);
    }
  }

  /**
   * Adds an expression to the edge.
   *
   * @param expression - the specified expression.
   */
  private void addExpression(final Node expression) {
    InvariantChecks.checkNotNull(expression);
    this.expressions.add(expression);
  }

  @Override
  public HlddNode deepCopy() {
    return new HlddEdge(this);
  }

  /**
   * Returns the child node.
   *
   * @return The child node.
   */
  public HlddNode getChild() {
    return child;
  }

  @Override
  public void addChild(final HlddNode child) {
    this.child = child;
  }

  @Override
  public boolean isNonTerminal() {
    return true;
  }

  /**
   * Sets the node as a default node.
   */
  public void setDefault() {
    isDefault = true;
  }

  /**
   * Shows whether the node is a default node.
   *
   * @return <code>true</code> if the node is a default node, <code>false</code> otherwise.
   */
  public boolean isDefault() {
    return isDefault;
  }

  /**
   * Returns this edge's branch equation using parent condition node.
   *
   * @return The branch equation node.
   */
  public Node getConditionEquation() {
    if (this.getParent() != null) {
      final ArrayList<Node> equations = new ArrayList<>();
      for (final Node value : this.expressions) {
        equations.add(NodeUtils.getEquation(((HlddConditionNode) this.getParent()).getExpression(),
            value));
      }
      if (eventList != null) {
        return NodeUtils.getConjunction(transformEvents(), NodeUtils.getDisjunction(equations));
      } else {
        return NodeUtils.getDisjunction(equations);
      }
    } else {
      return null;
    }
  }

  private Node transformEvents() {
    Node condition = null;
    for (Event event : eventList) {
      final Node eventCondition = getEventCondition(event);
      if (eventCondition != null && condition != null) {
        condition = new NodeOperation(StandardOperation.AND, condition, eventCondition);
      }
      if (condition == null) {
        condition = eventCondition;
      }
    }
    return condition;
  }

  /*
   * Constructs the corresponding node from the event.
   */
  private Node getEventCondition(final Event event) {
    final NodeVariable variable = event.getVariable();
    final NodeValue posEdgeValue =
        new NodeValue(Data.newBitVector(1, variable.getDataType().getSize()));
    final NodeValue negEdgeValue =
        new NodeValue(Data.newBitVector(0, variable.getDataType().getSize()));
    switch (event.getType()) {
      case NEGATIVE_EDGE:
        return NodeUtils.getEquation(variable, negEdgeValue);
      case POSITIVE_EDGE:
        return NodeUtils.getEquation(variable, posEdgeValue);
      default:
        break;
    }
    return null;
  }


  @Override
  public Collection<HlddNode> getChildren() {
    final Collection<HlddNode> result = new ArrayList<>();
    if (child != null) {
      result.add(child);
    }
    return result;
  }

  @Override
  public void removeChild(final HlddNode child) {
    if (this.child != null && this.child.equals(child)) {
      this.child = null;
    }
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = super.hashCode();
    result = prime * result + ((expressions == null) ? 0 : expressions.hashCode());
    result = prime * result + (isDefault ? 1231 : 1237);
    return result;
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (!super.equals(obj)) {
      return false;
    }
    if (getClass() != obj.getClass()) {
      return false;
    }
    final HlddEdge other = (HlddEdge) obj;
    if (expressions == null) {
      if (other.expressions != null) {
        return false;
      }
    } else if (!expressions.equals(other.expressions)) {
      return false;
    }
    return isDefault == other.isDefault;
  }

  @Override
  public String toString() {
    return "HlddEdge [expressions=" + expressions + "]";
  }

  public EventList getEventList() {
    return eventList;
  }

  public void setEventList(final EventList eventList) {
    this.eventList = eventList;
  }

}
