/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.model.hldd;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Statement;

import java.util.ArrayList;
import java.util.Collection;

/**
 * This class represents a process of a HLDD model. Contains a collection of HLDDs and the
 * corresponding variables.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public final class HlddProcess extends ModuleProcess {

  // HLDDs of the process.
  private final Collection<Hldd> hldds;

  /**
   * Constructs a new process with the specified name.
   *
   * @param name - the specified name.
   */
  public HlddProcess(final String name) {
    super(name);
    hldds = new ArrayList<>();
  }

  /**
   * Constructs a new process using the specified process.
   *
   * @param process - the specified process.
   */
  public HlddProcess(final ModuleProcess process) {
    super(process.getId());
    hldds = new ArrayList<>();
    declareVariables(process.getDeclarationMap());
    final Enum<?> infoType = MetaInfoType.PHASE_VAR_NAME;
    if (process.hasMetaInfo(infoType)) {
      this.addMetaInfo(infoType, process.getMetaInfo(infoType));
    }
  }

  private HlddProcess(final HlddProcess process) {
    super(process.getName());
    this.hldds = process.getHldds();
    declareVariables(process.getDeclarationMap());
    add(process.getHldds());
  }

  /**
   * Adds a collection of HLDDs to this process.
   *
   * @param hldds - the specified collection.
   */
  private void add(Collection<Hldd> hldds) {
    for (Hldd hldd : hldds) {
      add(hldd);
    }
  }

  /**
   * Adds a HLDD to this process.
   *
   * @param hldd - the specified HLDD.
   */
  public void add(final Hldd hldd) {
    InvariantChecks.checkNotNull(hldd);
    hldds.add(hldd);
    if (hldd.getVariable() != null && !checkDeclaration(hldd.getVariable(), this)) {
      declareVariable(hldd.getVariable());
    }
  }

  private boolean checkDeclaration(final NodeVariable variable, final DecStatement statement) {
    return statement.declares(variable)
        || statement.getParent() != null
            && checkDeclaration(variable, (DecStatement) statement.getParent());
  }

  /**
   * Returns the variables that are assigned in this process.
   *
   * @return The collection of assigned variables.
   */
  public Collection<NodeVariable> getAssignedVariables() {
    final Collection<NodeVariable> result = new ArrayList<>();
    for (Hldd hldd : hldds) {
      result.add(hldd.getVariable());
    }
    return result;
  }

  /**
   * Returns the contents of the container.
   *
   * @return The collection of objects.
   */
  public Collection<Hldd> getHldds() {
    return new ArrayList<>(hldds);
  }

  @Override
  public Statement deepCopy() {
    return new HlddProcess(this);
  }
}
