/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Module;
import ru.ispras.retrascope.model.basis.ModuleProcess;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Control flow graph model visitor that detects one-basic-block processes and transforms them.
 *
 * <p>This backend merges {@link CfgProcess} processes, which contain only one
 * {@link CfgBlockStatement} basic block node and equal
 * {@link ru.ispras.retrascope.model.basis.EventList} event lists.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class CfgOneBlockProcessBackend extends CfgEmptyVisitor {

  /**
   * Default constructor.
   */
  public CfgOneBlockProcessBackend() {
    super();
  }

  @Override
  public void onModuleBegin(final Module module) {

    final Set<CfgProcess> merged = new LinkedHashSet<>();
    final Collection<ModuleProcess> processes = module.getProcesses();

    for (final ModuleProcess process : processes) {

      if (isOneBlock((CfgProcess) process)) {

        if (merged.isEmpty()) {
          merged.add((CfgProcess) process);

        } else {

          boolean isMerged = false;

          for (final CfgProcess mergedProcess : merged) {

            /* If the target process can be merged into "merged" process. */

            final boolean mergedBlock = isOneBlock(mergedProcess);
            final EventList processSensList = process.getEventList();
            final boolean sameSensitive =
                mergedProcess.getEventList().equals(processSensList);

            if (mergedBlock
                && compatBlocking((CfgProcess) process, mergedProcess) && sameSensitive) {

              mergeProcesses(mergedProcess, (CfgProcess) process);
              isMerged = true;
              break;
            }
          }
          if (!isMerged) {
            merged.add((CfgProcess) process);
          }
        }
      } else {
        merged.add((CfgProcess) process);
      }
    }

    /* Remove "old" processes. */
    module.removeProcesses();

    /* Add new "merged" processes. */
    for (final CfgProcess process : merged) {
      module.add(process);
    }
  }

  private static boolean compatBlocking(
      final CfgProcess one,
      final CfgProcess another) {

    final CfgBlockStatement oneBlock = ((CfgBlockStatement) one.getStatement(0));
    final CfgBlockStatement anotherBlock = ((CfgBlockStatement) another.getStatement(0));

    return !CollectionUtils.areIntersectedSets(oneBlock.getUses(), anotherBlock.getDefines())
        && !CollectionUtils.areIntersectedSets(oneBlock.getDefines(), anotherBlock.getUses())
        && compatibleTargets(oneBlock, anotherBlock);
  }

  private static boolean compatibleTargets(
      final CfgBlockStatement oneBlock,
      final CfgBlockStatement anotherBlock) {

    final Map<NodeVariable, List<Range>> varRangeMap = new LinkedHashMap<>();

    for (final Assignment assignment : oneBlock.getAssignments()) {

      final RangedVariable target = assignment.getTarget();
      final NodeVariable targetVar = target.getVariable();

      if (!varRangeMap.containsKey(targetVar)) {
        varRangeMap.put(targetVar, new LinkedList<Range>());
      }
      varRangeMap.get(targetVar).add(target.getRange());
    }

    for (final Assignment assignment : anotherBlock.getAssignments()) {

      final RangedVariable target = assignment.getTarget();
      final NodeVariable targetVar = target.getVariable();

      if (varRangeMap.containsKey(targetVar)
          && compatibleWithRanges(target.getRange(), varRangeMap.get(targetVar))) {
        return false;
      }
    }

    return true;
  }

  private static boolean compatibleWithRanges(
      final Range range,
      final List<Range> ranges) {

    if (range == null || ranges.isEmpty()) {
      return true;
    } else {
      for (final Range listRange : ranges) {
        if (!compatible(range, listRange)) {
          return false;
        }
      }
    }

    return true;
  }

  private static boolean compatible(
      final Range range,
      final Range listRange) {

    if (!sameDataTypes(range, listRange)) {
      return false;
    }
    final NodeVariable rangeVar = new NodeVariable("RANGE_VAR", range.getOld().getDataType());
    return SolverUtils.areCompatible(makeNode(rangeVar, range), makeNode(rangeVar, listRange));
  }

  private static boolean sameDataTypes(
      final Range oneRange,
      final Range anotherRange) {

    return oneRange.getYoung().getDataType().equals(oneRange.getOld().getDataType())
        && anotherRange.getYoung().getDataType().equals(anotherRange.getOld().getDataType())
        && oneRange.getOld().getDataType().equals(anotherRange.getYoung().getDataType());
  }

  private static Node makeNode(
      final NodeVariable rangeVar,
      final Range range) {

    return Node.AND(
        new NodeOperation(
            range.getYoung().isType(DataType.INTEGER)
                ? StandardOperation.GREATEREQ
                : StandardOperation.BVUGE,
            rangeVar,
            range.getYoung()),
        new NodeOperation(
            range.getOld().isType(DataType.INTEGER)
                ? StandardOperation.LESSEQ
                : StandardOperation.BVULE,
            rangeVar,
            range.getOld()));
  }

  private void mergeProcesses(
      final CfgProcess target,
      final CfgProcess value) {

    /* Merge sensitivity lists. */
    for (final Event event : value.getEventList()) {
      target.addEvent(event);
    }

    /* Merge control flow graphs. */

    /* Get basic bock node at the result process. */
    final CfgBlockStatement resultBlock = (CfgBlockStatement) target.getStatement(0);

    /* Get basic block at the target process. */
    final CfgBlockStatement targetBlock = (CfgBlockStatement) value.getStatement(0);

    resultBlock.addAssignments(targetBlock.getAssignments());
  }

  private static boolean isOneBlock(final CfgProcess process) {

    return process.getStatementNum() == 1
        && process.getStatement(0).isType(CfgStatementType.BASIC_BLOCK);
  }
}
