/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.basis.backend;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.expression.printer.ExprTreePrinter;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.Configuration;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.IntegerRange;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.StatementUtils;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgEmptyVisitor;
import ru.ispras.retrascope.model.cfg.CfgStatementType;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;

/**
 * Parser backend that transforms sequences of switch statements with ranged variables in
 * conditions.
 *
 * <p>The backend optimizes sequences of control flow graph
 * {@link ru.ispras.retrascope.model.cfg.CfgSwitchStatement} switch statements. Statements to be
 * optimized should have conditions of the {@code x[i:j]} form, where {@code x} is a common
 * variable for sequence statements; {@code i} and {@code j} indices are constants; ranges
 * of subsequent statements are non-intersecting but can be joined.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgRangedSwitchBackend extends CfgEmptyVisitor {

  private CfgSwitchStatement next;
  private NodeOperation currentCond;
  private NodeOperation nextCond;
  private IntegerRange currentRange;
  private IntegerRange nextRange;

  @Override
  public void onSwitchBegin(final CfgSwitchStatement switchNode) {
    while (canCombineWithNext(switchNode)) {
      combineWithNext(switchNode);
    }
  }

  private boolean canCombineWithNext(final CfgSwitchStatement switchStmt) {
    final Statement parent = switchStmt.getParent();

    final int switchIndex = parent.indexOf(switchStmt);

    if (switchIndex == parent.getStatementNum() - 1) {

      /* Switch statement is the last, no statements to be combined with. */
      return false;
    }

    final Statement next = parent.getStatement(switchIndex + 1);

    if (!next.isType(CfgStatementType.SWITCH)) {

      /* Next statement is not a switch, no statements to be combined with. */
      return false;
    }

    this.next = (CfgSwitchStatement) next;

    final Node currentCond = switchStmt.getCondition();
    final Node nextCond = this.next.getCondition();

    if (!ExprUtils.isKind(Node.Kind.OPERATION, currentCond, nextCond)) {
      return false;
    }
    this.currentCond = (NodeOperation) currentCond;
    this.nextCond = (NodeOperation) nextCond;

    return sameRanged() && noDefaultCases(switchStmt, this.next);
  }

  private boolean noDefaultCases(final CfgSwitchStatement ... statements) {

    for (final CfgSwitchStatement statement : statements) {

      for (int i = 0; i < statement.getStatementNum(); i++) {

        final CfgCaseStatement caseStatement = (CfgCaseStatement) statement.getStatement(i);

        if (caseStatement.isDefault()) {
          return false;
        }
      }
    }

    return true;
  }

  private boolean sameRanged() {

    if (!bvExtract(this.currentCond, this.nextCond)) {
      return false;
    }

    this.currentRange = getRange(this.currentCond);
    this.nextRange = getRange(this.nextCond);

    return this.currentRange.border(this.nextRange);
  }

  private IntegerRange getRange(final NodeOperation operation) {

    final Enum<?> opId = operation.getOperationId();
    final int paramCount = StandardOperation.getParameterCount(opId);

    assert paramCount == 2 : "BVEXTRACT operation has wrong integer param number: " + paramCount;

    final NodeValue minNode = (NodeValue) operation.getOperand(1);
    final NodeValue maxNode = (NodeValue) operation.getOperand(0);

    return new IntegerRange(minNode.getInteger(), maxNode.getInteger());
  }

  private static boolean bvExtract(final Node ... nodes) {

    for (final Node node : nodes) {

      if (!ExprUtils.isOperation(node, StandardOperation.BVEXTRACT)) {
        return false;
      }
    }
    return true;
  }

  private void combineWithNext(final CfgSwitchStatement current) {

    final Statement parent = current.getParent();

    final int compareRanges = this.currentRange.compareTo(this.nextRange);

    final Collection<Statement> cases;

    if (compareRanges > 0) {
      cases = compose(this.next, current, true);

    } else if (compareRanges < 0) {
      cases = compose(current, this.next, false);

    } else {
      final Configuration config = Retrascope.getConfiguration();
      final ExprTreePrinter printer = config.getStylePrinter();

      throw RetrascopeRuntimeException.makeException(
          LogMessage.WRONG_ARG,
          printer.toString(this.currentCond));
    }

    parent.remove(this.next);

    current.setCondition(
        compareRanges > 0
            ? composeConditions(this.nextCond, this.currentCond)
            : composeConditions(this.currentCond, this.nextCond));

    current.removeAll();
    current.addAll(cases);
  }

  private Node composeConditions(
      final NodeOperation young,
      final NodeOperation old) {

    final Node expr = young.getOperand(2);
    final Node min = young.getOperand(0);
    final Node max = old.getOperand(1);

    return new NodeOperation(StandardOperation.BVEXTRACT, max, min, expr);
  }

  private Collection<Statement> compose(
      final CfgSwitchStatement top,
      final CfgSwitchStatement bottom,
      final boolean revertedOrder) {

    final Collection<Statement> composedCases = new LinkedList<>();

    for (int i = 0; i < top.getStatementNum(); i++) {

      final CfgCaseStatement topCase = (CfgCaseStatement) top.getStatement(i);

      for (int j = 0; j < bottom.getStatementNum(); j++) {

        final CfgCaseStatement bottomCase = (CfgCaseStatement) bottom.getStatement(j);

        final EventList eventList = new EventList();
        eventList.addEvents(topCase.getEventList());
        eventList.addEvents(bottomCase.getEventList());

        final Collection<NodeValue> composedValues =
            composeValues(topCase.getValues(), bottomCase.getValues(), revertedOrder);

        final CfgCaseStatement composition =
            eventList.isEmpty()
                ? new CfgCaseStatement(composedValues)
                : new CfgCaseStatement(eventList, composedValues);

        final Collection<Statement> topStatements = copyTopStatements(topCase);
        final Collection<Statement> bottomStatements = copyTopStatements(bottomCase);

        if (revertedOrder) {
          composition.addAll(bottomStatements);
          composition.addAll(topStatements);
        } else {
          composition.addAll(topStatements);
          composition.addAll(bottomStatements);
        }

        composedCases.add(composition);
      }
    }

    return composedCases;
  }

  private Collection<Statement> copyTopStatements(final Statement stmt) {

    final Collection<Statement> topStatements = StatementUtils.getTopStatements(stmt);

    final Collection<Statement> copies = new LinkedHashSet<>();

    for (final Statement topStmt : topStatements) {
      copies.add(topStmt.deepCopy());
    }

    return copies;
  }

  private Collection<NodeValue> composeValues(
      final Collection<NodeValue> oldParts,
      final Collection<NodeValue> youngParts,
      final boolean revertedOrder) {

    final Collection<NodeValue> composed = new LinkedList<>();

    for (final NodeValue oldPart : oldParts) {

      for (final NodeValue youngPart : youngParts) {

        final Node concat =
            revertedOrder
                ? new NodeOperation(StandardOperation.BVCONCAT, youngPart, oldPart)
                : new NodeOperation(StandardOperation.BVCONCAT, oldPart, youngPart);

        composed.add((NodeValue) Transformer.reduce(concat));
      }
    }

    return composed;
  }
}
