/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.SourceLocation;
import org.zamia.ZamiaException;
import org.zamia.instgraph.IGDefaultOperationVisitor;
import org.zamia.instgraph.IGItem;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationAlias;
import org.zamia.instgraph.IGOperationAllocate;
import org.zamia.instgraph.IGOperationArrayAggregate;
import org.zamia.instgraph.IGOperationAttribute;
import org.zamia.instgraph.IGOperationBinary;
import org.zamia.instgraph.IGOperationDeref;
import org.zamia.instgraph.IGOperationError;
import org.zamia.instgraph.IGOperationIndex;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGOperationLiteral;
import org.zamia.instgraph.IGOperationObject;
import org.zamia.instgraph.IGOperationPhi;
import org.zamia.instgraph.IGOperationRange;
import org.zamia.instgraph.IGOperationRecordAggregate;
import org.zamia.instgraph.IGOperationRecordField;
import org.zamia.instgraph.IGOperationTypeConversion;
import org.zamia.instgraph.IGOperationTypeQualification;
import org.zamia.instgraph.IGOperationUnary;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGType;
import org.zamia.instgraph.IGTypeStatic;
import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.Radix;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.fortress.util.Pair;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.parser.vhdl.mapper.BitVectorOpCodeMapper;
import ru.ispras.retrascope.parser.vhdl.mapper.IgBinaryMapper;
import ru.ispras.retrascope.parser.vhdl.mapper.IgSubprogramMapper;
import ru.ispras.retrascope.parser.vhdl.mapper.IgUnaryMapper;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.TransformerUtils;

import java.math.BigInteger;
import java.util.LinkedList;
import java.util.List;

/**
 * IG operation objects parser.
 *
 * <p>Component that transforms IG (Instantiation Graph) objects
 * to internal representation.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgOperationVisitor extends IGDefaultOperationVisitor {

  /* The current result of object processing. */
  private IgOperationResult result;

  private IgStructureVisitor parent;

  static final String OFFSET = "offset";

  /**
   * Default constructor.
   */
  IgOperationVisitor() {
    super();
  }

  /**
   * Sets visitor for parent IG structure.
   *
   * @param visitor Parent structure visitor.
   */
  void setParentVisitor(final IgStructureVisitor visitor) {
    this.parent = visitor;
  }

  /**
   * Returns parent structure visitor.
   *
   * @return Parent structure visitor.
   */
  private IgStructureVisitor getStructureVisitor() {
    return this.parent;
  }

  /**
   * Returns result that is formed by object on the current argument.
   *
   * @see IgOperationResult
   *
   * @return Result of processing that is available at the current moment.
   */
  private IgOperationResult getResult() {
    return this.result;
  }

  private VariableBuilder getBuilder() {
    return getStructureVisitor().getBuilder();
  }

  @Override
  public void visitOperation(final IGOperation igOperation) {

    IGOperation operation = igOperation;
    InvariantChecks.checkNotNull(operation);
    final String opName = operation.toHRString();
    if (getStructureVisitor().hasFormalArg(opName)) {
      operation = getStructureVisitor().getActualArg(opName);
    }

    super.visitOperation(operation);
  }

  @Override
  public void visitStaticValue(final IGStaticValue igStaticValue) {

    this.result = new IgOperationResult(getNodeValue(igStaticValue));
  }

  private NodeValue getNodeValue(final IGStaticValue igStaticValue) {

    final IGType type = igStaticValue.getType();

    if (type == null) {

      return toBooleanTrue(igStaticValue);
    }

    if (type.isBit()) {

      return NodeValue.newBitVector(BitVector.valueOf(igStaticValue.isTrue()));

    } else if (type.isBool()) {

      return toBooleanTrue(igStaticValue);

    } else if (type.isInteger()) {

      return NodeValue.newInteger(igStaticValue.getInt());

    } else if (type.isReal()) {

      return NodeValue.newReal(igStaticValue.getReal().doubleValue());

    } else if (type.isLogic()) {

      return toBooleanVector(igStaticValue, type);

    } else if (type.isArray()) {

      return toArray(igStaticValue, type);

    } else if (type.isEnum()) {

      /* Enum fields are treated as integers. */
      return NodeValue.newInteger(igStaticValue.getEnumOrd());
    }
    throw new UnsupportedOperationException(igStaticValue.toHRString());
  }

  private NodeValue toArray(final IGStaticValue igStaticValue, final IGType type) {
    final SourceLocation location = igStaticValue.computeSourceLocation();
    try {
      final String arrayString = getSmtLibArrayString(igStaticValue, location);
      final IGType indexType = type.getIndexType();
      final DataType indexDataType = getBuilder().getDataType(indexType);
      final IGType elemType = type.getElementType();
      final DataType elemDataType = getBuilder().getDataType(elemType);
      final DataMap dataMapValue = DataMap.valueOf(arrayString, indexDataType, elemDataType);

      return new NodeValue(Data.newArray(dataMapValue));
    } catch (ZamiaException e) {
      throw new RetrascopeRuntimeException(e);
    }
  }

  private NodeValue toBooleanVector(final IGStaticValue igStaticValue, final IGType type) {

    //TODO: implement 9-sign VHDL logic
    final IGType indexType = type.getIndexType();
    BitVector initVector;

    if (indexType == null) {

      /* 1-bit size vector */
      final boolean one = igStaticValue.getCharLiteral() == '1';
      final boolean hyphen = igStaticValue.getCharLiteral() == '-';
      initVector = BitVector.valueOf(one || hyphen);

    } else {

      final int typeSize = getBuilder().getTypeSize((IGTypeStatic) indexType);
      initVector = BitVector.valueOf(0, typeSize);
    }
    return NodeValue.newBitVector(initVector);
  }

  private NodeValue toBooleanTrue(final IGStaticValue igStaticValue) {
    return NodeValue.newBoolean(igStaticValue.isTrue());
  }

  private String getSmtLibArrayString(
      final IGStaticValue igStaticValue,
      final SourceLocation location)
    throws ZamiaException {
    final int entryNum = igStaticValue.getNumArrayEntries(location);

    final StringBuilder builder = new StringBuilder("(");
    for (int i = 0; i < entryNum; i++) {
      final IGStaticValue entryValue = igStaticValue.getValue(i, location);
      final String key = Integer.toString(i + 1);
      final NodeValue value = getNodeValue(entryValue);
      builder.append(String.format("(%s:%s)", key, value));
    }
    builder.append(')');
    return builder.toString();
  }

  @Override
  public void visitRange(final IGRange igRange) {
    /* Empty. */
  }

  @Override
  public void visitUnary(final IGOperationUnary igUnary) {
    StandardOperation opId = IgUnaryMapper.getOperationCode(igUnary);
    final Node operandElem = getNode(igUnary.getA());
    this.result = new IgOperationResult(new NodeOperation(opId, operandElem));
  }

  @Override
  public void visitTypeQualification(final IGOperationTypeQualification igQualification) {
    throw new UnsupportedOperationException(igQualification.toHRString());
  }

  @Override
  public void visitTypeConversion(final IGOperationTypeConversion igConversion) {

    /* The conversion object contains exactly one operand. */
    visitOperation(igConversion.getOperand(0));
  }

  @Override
  public void visitRecordField(final IGOperationRecordField igRecordField) {
    throw new UnsupportedOperationException(igRecordField.toHRString());
  }

  @Override
  public void visitRecordAggregate(final IGOperationRecordAggregate igAggregate) {
    throw new UnsupportedOperationException(igAggregate.toHRString());
  }

  @Override
  public void visitOpRange(final IGOperationRange igOperationRange) {

    final Node operand = getNode(igOperationRange.getOperand());

    InvariantChecks.checkNotNull(operand);

    final IGOperation range = igOperationRange.getRange();

    if (range instanceof IGRange) {
      final IGOperation isAscending = ((IGRange) range).getAscending();

      final Node left = getNode(((IGRange) range).getLeft());
      final Node right = getNode(((IGRange) range).getRight());

      final boolean isTrue = ((IGStaticValue)isAscending).isTrue();
      final Node firstIdx = isTrue ? right : left;
      final Node secondIdx = isTrue ? left : right;

      /* Check bvextract argument shift */
      final VariableDescriptor desc = (VariableDescriptor) operand.getUserData();
      checkBitVectorMinIndex(igOperationRange.getOperand(), operand, desc);

      final boolean shifted = desc != null && desc.hasMetaInfo(OFFSET);

      final Node first =
          shifted ? new NodeOperation(StandardOperation.SUB, firstIdx, shift(desc)) : firstIdx;
      final Node second =
          shifted ? new NodeOperation(StandardOperation.SUB, secondIdx, shift(desc)) : secondIdx;

      final NodeOperation bvExtract =
          new NodeOperation(StandardOperation.BVEXTRACT, first, second, operand);
      final Node expr = TransformerUtils.standardize(bvExtract);

      this.result = new IgOperationResult(expr);
    } else {
      throw new UnsupportedOperationException(igOperationRange.toHRString());
    }
  }

  private static Node shift(final VariableDescriptor descriptor) {
    return NodeValue.newInteger(Integer.parseInt(descriptor.getMetaInfo(OFFSET)));
  }

  @Override
  public void visitPhi(final IGOperationPhi igPhi) {
    final Node condition = getNode(igPhi.getOperand(0));
    final Node firstVariant = getNode(igPhi.getOperand(1));
    final Node secondVariant = getNode(igPhi.getOperand(2));

    final Node expression =
        new NodeOperation(StandardOperation.ITE, condition, firstVariant, secondVariant);

    this.result = new IgOperationResult(expression);
  }

  @Override
  public void visitObject(final IGOperationObject igOperationObject) {

    Node node;
    final IGObject igObject = igOperationObject.getObject();
    final IGType opType = igOperationObject.getType();

    if (IgUtils.isVariable(igObject)) {

      /* Object keeps a variable. */

      final DataType dataType = getBuilder().getDataType(opType);
      final VariableType variableType = VariableBuilderUtils.getVariableType(igOperationObject);

      node = new NodeVariable(igObject.getId(), dataType);

      /* Set full type meta-info. */
      final VariableDescriptor data = new VariableDescriptor(variableType);
      final String fullType = VariableBuilder.getFullType(igObject);
      data.addMetaInfo(MetaInfoType.HDL_TYPE_STR, fullType);
      node.setUserData(data);
      getBuilder().fillVariableDescriptor(data, igOperationObject);

      checkBitVectorMinIndex(igOperationObject, node, data);

    } else if (IgUtils.isConstant(igObject)) {

      /* Object keeps a constant. */

      final IGOperation initValue = igObject.getInitialValue();
      node = getNode(initValue);
    } else {
      throw new IllegalArgumentException(igOperationObject.toHRString());
    }
    this.result = new IgOperationResult(node);
  }

  void checkBitVectorMinIndex(
      final IGObject varObject,
      final Node variable,
      final VariableDescriptor varDescriptor) {

    if (variable.isType(DataTypeId.BIT_VECTOR)) {
      final NodeValue minIndex = getMinIndex(varObject);

      /* Minimal index of vector's bit is not 0. */
      if (minIndex != null && !minIndex.equals(NodeValue.newInteger(0))) {

        varDescriptor.addMetaInfo(OFFSET, minIndex.toString());
      }
    }
  }

  void checkBitVectorMinIndex(
      final IGOperation varObject,
      final Node variable,
      final VariableDescriptor variableDescriptor) {

    if (variable.isType(DataTypeId.BIT_VECTOR)) {
      final NodeValue minIndex = getMinIndex(varObject);

      /* Minimal index of vector's bit is not 0. */
      if (minIndex != null && !minIndex.equals(NodeValue.newInteger(0))) {

        variableDescriptor.addMetaInfo(OFFSET, minIndex.toString());
      }
    }
  }

  private NodeValue getMinIndex(final IGObject object) {

    final IGType objType = object.getType();
    return getMinIndex(objType.getIndexType());
  }

  private NodeValue getMinIndex(final IGOperation operation) {

    final IGType opType = operation.getType();
    return getMinIndex(opType.getIndexType());
  }

  private NodeValue getMinIndex(final IGType indexType) {
    NodeValue minIndex = null;
    if (indexType instanceof IGTypeStatic) {
      final IGStaticValue staticRange = ((IGTypeStatic) indexType).getStaticRange();
      final boolean ascending = staticRange.getAscending().isTrue();
      final IGStaticValue left = staticRange.getLeft();
      final IGStaticValue right = staticRange.getRight();
      minIndex = getNodeValue(ascending ? left : right);
    }
    return minIndex;
  }

  private NodeValue getMaxIndex(final IGType indexType) {
    NodeValue maxIndex = null;
    if (indexType instanceof IGTypeStatic) {
      final IGStaticValue staticRange = ((IGTypeStatic) indexType).getStaticRange();
      final boolean ascending = staticRange.getAscending().isTrue();
      final IGStaticValue left = staticRange.getLeft();
      final IGStaticValue right = staticRange.getRight();
      maxIndex = getNodeValue(ascending ? right : left);
    }
    return maxIndex;
  }

  @Override
  public void visitLiteral(final IGOperationLiteral igLiteral) {

    NodeValue value;
    try {

      /* Compute static value. */
      final IGStaticValue opStaticValue = igLiteral.computeStaticValue(null);

      switch (opStaticValue.getStaticType().getCat()) {
        case INTEGER:
          value = new NodeValue(Data.newInteger(opStaticValue.getInt()));
          break;
        case REAL:
          value =  new NodeValue(Data.newReal(opStaticValue.getReal().doubleValue()));
          break;
        case ARRAY:

          // TODO: change when 'Z' values will be supported
          final String strVector = substituteUnsupported(opStaticValue.toString());

          value =
              new NodeValue(
                  Data.newBitVector(
                      strVector,
                      Radix.BIN.value(),
                      opStaticValue.toHRString().length()));
          break;
        default:
          throw new UnsupportedOperationException(igLiteral.toHRString());
      }
    } catch (ZamiaException e) {
      throw new RetrascopeRuntimeException(e);
    }
    this.result = new IgOperationResult(value);
  }

  @Override
  public void visitIndex(final IGOperationIndex igIndex) {

    Node index = getNode(igIndex.getIndex());
    final Node operand = getNode(igIndex.getOperand());

    Node resultNode;
    if (operand.isType(DataTypeId.BIT_VECTOR)) {

      final VariableDescriptor descriptor = (VariableDescriptor) operand.getUserData();

      if (descriptor == null) {

        resultNode =
            TransformerUtils.standardize(
                new NodeOperation(StandardOperation.BVEXTRACT, index, index, operand));
      } else {

        checkBitVectorMinIndex(igIndex.getOperand(), operand, descriptor);
        index = descriptor.hasMetaInfo(OFFSET)
            ? new NodeOperation(
                StandardOperation.SUB,
                index,
                NodeValue.newInteger(Integer.parseInt(descriptor.getMetaInfo(OFFSET))))
            : index;

        /* For parametric operations their parameters should go earlier than operands. */
        resultNode =
            TransformerUtils.standardize(
                new NodeOperation(StandardOperation.BVEXTRACT, index, index, operand));
      }
    } else if (operand.isType(DataTypeId.MAP)) {
      resultNode = new NodeOperation(StandardOperation.SELECT, operand, index);
    } else {
      throw new IllegalArgumentException(operand.toString());
    }
    this.result = new IgOperationResult(resultNode);
  }

  @Override
  public void visitError(final IGOperationError igError) {
    throw new UnsupportedOperationException(igError.toHRString());
  }

  @Override
  public void visitDeref(final IGOperationDeref igDeref) {
    throw new UnsupportedOperationException(igDeref.toHRString());
  }

  @Override
  public void visitAttribute(final IGOperationAttribute igAttribute) {

    /* Object keeps event-based attribute. */
    final IGItem item = igAttribute.getItem();

    if (item != null) {
      final VariableBuilder builder = getBuilder();
      final RangedVariable rangedVariable = builder.getRangedVariable((IGOperation) item);

      switch (igAttribute.getAttrOp()) {
        case EVENT:

          final Event event = new Event(rangedVariable);
          this.result = new IgOperationResult(event);
          break;
        case LENGTH:
          final NodeVariable variable = rangedVariable.getVariable();
          final Object descriptor = variable.getUserData();
          final Node length = ((VariableDescriptor)descriptor).getLength();
          this.result = new IgOperationResult(length);
          break;
        default:
          throw new UnsupportedOperationException(igAttribute.toHRString());
      }
    }
  }

  @Override
  public void visitAggregate(final IGOperationArrayAggregate igAggregate) {

    final IGType igType = igAggregate.getType();
    final IGType indexType = igType.getIndexType();
    final IGType elementType = igType.getElementType();
    final DataType keyType = getBuilder().getDataType(indexType);
    final DataType valueType = getBuilder().getDataType(elementType);
    final int operandNum = igAggregate.getNumOperands();

    final Node dataNode = isBitVector(keyType, valueType)
        ? getBitVector(igAggregate, indexType, operandNum)
        : getDataMap(igAggregate, indexType, keyType, valueType, operandNum);

    this.result = new IgOperationResult(dataNode);
  }

  private Node getBitVector(
      final IGOperationArrayAggregate igAggregate,
      final IGType indexType,
      final int operandNum) {

    final Node[] values;

    if (indexType instanceof IGTypeStatic) {

      final Pair<NodeValue, NodeValue> indexValuePair = getPair(indexType);

      final NodeValue min = indexValuePair.first;
      final BigInteger minInt = min.getInteger();
      final NodeValue max = indexValuePair.second;
      final BigInteger maxInt = max.getInteger();

      final int maxIndex = maxInt.intValue();
      final int minIndex = minInt.intValue();

      values = new Node[maxIndex - minIndex + 1];

      final IGOperation others = igAggregate.getOperand(0);
      final Data otherData = getValueData(others);

      /* Skip 'others' element if it is null. */
      int offset = minIndex;
      if (others == null) {
        offset++;
      }
      if (maxIndex > 0) {
        for (int i = offset; i <= maxIndex; i++) {

          final Data data =
              i >= (operandNum - offset) ? otherData : getValueData(igAggregate.getOperand(i));
          values[i - offset] = new NodeValue(data);
        }
      }

    } else {
      throw RetrascopeRuntimeException.makeException(
          LogMessage.UNSUPPORTED_TYPE,
          indexType.toHRString());
    }

    final NodeOperation concatOperation = new NodeOperation(StandardOperation.BVCONCAT, values);
    return Transformer.reduce(concatOperation);
  }

  private boolean isBitVector(final DataType keyType, final DataType valueType) {

    return keyType.equals(DataType.INTEGER) && valueType.equals(DataType.BIT_VECTOR(1));
  }

  private Node getDataMap(
      final IGOperationArrayAggregate igAggregate,
      final IGType indexType,
      final DataType keyType,
      final DataType valueType,
      final int operandNum) {

    final DataMap dataMap = new DataMap(keyType, valueType);

    if (indexType instanceof IGTypeStatic) {

      final Pair<NodeValue, NodeValue> indexValuePair = getPair(indexType);

      makeStaticDataMap(igAggregate, indexValuePair, operandNum, dataMap);

    } else {
      /* Skip 'others' element if exists. */
      int offset = 0;
      if (igAggregate.getOperand(0) == null) {
        offset = 1;
      }

      for (int i = offset; i < operandNum; i++) {

        final Node node = getNode(igAggregate.getOperand(i));
        final Node value = Transformer.reduce(node);

        if (value.getKind() == Node.Kind.VALUE) {
          dataMap.put(Data.newInteger(i - offset), ((NodeValue) value).getData());
        } else {
          throw new UnsupportedOperationException(value.toString());
        }
      }
    }
    return new NodeValue(Data.newArray(dataMap));
  }

  private Pair<NodeValue, NodeValue> getPair(final IGType indexType) {

    final NodeValue minIndex = getMinIndex(indexType);
    final NodeValue minIndexValue = (NodeValue) TransformerUtils.toInt(minIndex);

    final NodeValue maxIndex = getMaxIndex(indexType);
    final NodeValue maxIndexValue = (NodeValue) TransformerUtils.toInt(maxIndex);

    return new Pair<>(minIndexValue, maxIndexValue);
  }

  private void makeStaticDataMap(
      final IGOperationArrayAggregate igAggregate,
      final Pair<NodeValue, NodeValue> indexValuePair,
      final int operandNum, DataMap dataMap) {

    final NodeValue min = indexValuePair.first;
    final BigInteger minInt = min.getInteger();
    final NodeValue max = indexValuePair.second;
    final BigInteger maxInt = max.getInteger();

    final int maxIndex = maxInt.intValue();
    final int minIndex = minInt.intValue();

    final IGOperation others = igAggregate.getOperand(0);
    final Data otherData = getValueData(others);

      /* Skip 'others' element if it is null. */
    int offset = minIndex;
    if (others == null) {
      offset++;
    }
    if (maxIndex > 0) {
      for (int i = offset; i <= maxIndex; i++) {

        final Data index = Data.newInteger(i - offset);
        final Data mapData =
            i >= (operandNum - offset) ? otherData : getValueData(igAggregate.getOperand(i));
        dataMap.put(index, mapData);
      }
    }
  }

  private Data getValueData(final IGOperation operation) {

    if (operation == null) {
      return null;
    }

    final Node node = getNode(operation);
    final Node reduced = Transformer.reduce(node);
    if (reduced instanceof NodeValue) {
      return ((NodeValue) reduced).getData();
    } else {
      return null;
    }
  }

  @Override
  public void visitAllocate(final IGOperationAllocate igAllocate) {
    throw new UnsupportedOperationException(igAllocate.toHRString());
  }

  @Override
  public void visitAlias(final IGOperationAlias igAlias) {
    throw new UnsupportedOperationException(igAlias.toHRString());
  }

  @Override
  public void visitSubprogram(final IGOperationInvokeSubprogram igSubProgram) {

    final String name = IgUtils.getName(igSubProgram);

    if (IgSubprogramMapper.hasOperation(name)) {

      final Node[] nodes = getOperands(igSubProgram);

      final StandardOperation opCode = IgSubprogramMapper.getOperationCode(name);
      final StandardOperation resultOpCode = modifyOpCode(opCode, nodes);

      this.result = new IgOperationResult(new NodeOperation(resultOpCode, nodes));
    } else if (IgSubprogramMapper.hasMacro(name)) {

      final Node[] operands = getOperands(igSubProgram);
      final Node macro = IgSubprogramMapper.getMacroOperation(name, operands);
      this.result = new IgOperationResult(macro);
    } else {

      final String tag = IgUtils.getTag(igSubProgram);

      if (!getStructureVisitor().hasProgramElaborated(tag)) {
        throw new RetrascopeRuntimeException("Non-processed sub-program is detected: " + tag);
      }

      this.result = new IgOperationResult(getStructureVisitor().getReturnVar(tag));
    }
  }

  private Node[] getOperands(final IGOperationInvokeSubprogram igSubProgram) {
    final int operandNum = igSubProgram.getNumOperands();
    final List<Node> nodes = new LinkedList<>();
    for (int i = 0; i < operandNum; i++) {
      final Node node = getNode(igSubProgram.getOperand(i));
      if (node != null) {
        nodes.add(node);
      }
    }
    return nodes.toArray(new Node[nodes.size()]);
  }

  private static StandardOperation modifyOpCode(
      final StandardOperation opCode,
      final Node ... nodes) {

    StandardOperation resultOpCode = opCode;

    /* HACK: substitute binary operation identifiers to unary if needed. */
    if (nodes.length == 1) {
      if (opCode == StandardOperation.ADD) {
        resultOpCode = StandardOperation.PLUS;
      } else if (opCode == StandardOperation.SUB) {
        resultOpCode = StandardOperation.MINUS;
      }
    }

    /* Substitute operation to bit-vector-analogue if at least one operand is bit vector. */
    if (NodeUtils.hasBitVector(nodes) && BitVectorOpCodeMapper.hasOperation(resultOpCode)) {
      resultOpCode = BitVectorOpCodeMapper.getOperation(resultOpCode);
    }
    return resultOpCode;
  }

  @Override
  public void visitBinary(final IGOperationBinary igBinary) {
    final Enum<?> type = IgBinaryMapper.getOperationCode(igBinary);

    final Node left = getNode(igBinary.getA());
    final Node right = getNode(igBinary.getB());

    this.result = new IgOperationResult(new NodeOperation(type, left, right));
  }

  /**
   * Returns syntax node that is extracted from the specified IG object.
   *
   * @param object IG object to be processed.
   * @return Syntax node that is generated.
   */
  public Node getNode(final IGOperation object) {
    visitOperation(object);
    if (getResult().getNode() == null) {
      return getResult().getNode();
    } else {
      return TypeConversion.castConstants(getResult().getNode());
    }
  }

  /**
   * Returns data object that is extracted from th specified IG object.
   *
   * @param object IG object
   * @return data object
   */
  public Object getData(final IGOperation object) {
    visitOperation(object);
    return getResult().getData();
  }

  private String substituteUnsupported(final String vectorValue) {

    /* Substitutes 'Z'-values and '-'-values by zeros. */
    final char[] array = vectorValue.toCharArray();
    for (int i = 0; i < vectorValue.length(); i++) {
      if (array[i] == '-' || array[i] == 'Z') {
        array[i] = '0';
      }
    }
    return new String(array);
  }
}
