/*
 * Copyright 2014-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.SourceLocation;
import org.zamia.ZamiaException;
import org.zamia.instgraph.IGConcurrentStatement;
import org.zamia.instgraph.IGContainer;
import org.zamia.instgraph.IGContainerItem;
import org.zamia.instgraph.IGDefaultSequentialStatementVisitor;
import org.zamia.instgraph.IGItem;
import org.zamia.instgraph.IGManager;
import org.zamia.instgraph.IGMapping;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationAttribute;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGProcess;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialAssert;
import org.zamia.instgraph.IGSequentialAssignment;
import org.zamia.instgraph.IGSequentialExit;
import org.zamia.instgraph.IGSequentialIf;
import org.zamia.instgraph.IGSequentialLoop;
import org.zamia.instgraph.IGSequentialNext;
import org.zamia.instgraph.IGSequentialProcedureCall;
import org.zamia.instgraph.IGSequentialReport;
import org.zamia.instgraph.IGSequentialReturn;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGSequentialWait;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGStructure;
import org.zamia.instgraph.IGSubProgram;
import org.zamia.instgraph.IGType;
import org.zamia.instgraph.IGTypeStatic;
import org.zamia.instgraph.interpreter.IGInterpreterRuntimeEnv;
import org.zamia.util.PathName;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Declaration;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.model.cfg.CfgAssertStatement;
import ru.ispras.retrascope.model.cfg.CfgBlockStatement;
import ru.ispras.retrascope.model.cfg.CfgCaseStatement;
import ru.ispras.retrascope.model.cfg.CfgLoopStatement;
import ru.ispras.retrascope.model.cfg.CfgModule;
import ru.ispras.retrascope.model.cfg.CfgProcess;
import ru.ispras.retrascope.model.cfg.CfgSwitchStatement;
import ru.ispras.retrascope.model.cfg.CfgWaitStatement;
import ru.ispras.retrascope.parser.vhdl.mapper.IgSubprogramMapper;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.Deque;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * IG structure objects parser.
 *
 * <p>Component of VHDL parser that transforms IG (Instantiation Graph)
 * structure objects representing processes, modules, instances to
 * control flow graph {@link ru.ispras.retrascope.model.cfg.CfgModel} model representation.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgStructureVisitor extends IGDefaultSequentialStatementVisitor {

  /* The IG representation manager. */
  private final IGManager manager;

  /* The IG interpreter runtime environment. */
  private final IGInterpreterRuntimeEnv runtimeEnv;

  /* The parent top-level module. */
  private final CfgModule topLevel;

  /* The list of current structure statements. */
  private final List<Statement> statements;

  /* The current process. */
  private CfgProcess process;

  /* The builder for VHDL variables. */
  private VariableBuilder builder;

  /* The mapping from sub-program calls to the new variables that keep their return values. */
  private final Map<String, NodeVariable> subProgramReturnVarMap;

  /* The stack for sub-program tags that are processed right now. */
  private final Deque<String> subProgramTags;

  /* Formal-actual param mappings for the processed sub-program. */
  private Map<String, IGOperation> subProgramMappings;

  /* Constants that keep severity string values. */
  private static final String SEVERITY_NOTE = "NOTE";
  private static final String SEVERITY_WARN = "WARNING";
  private static final String SEVERITY_ERROR = "ERROR";
  private static final String SEVERITY_FAIL = "FAILURE";

  /**
   * Creates object with the specified IG manager, IG runtime environment and the top-level module.
   *
   * @param manager IG representation manager.
   * @param runtimeEnv IG runtime environment.
   * @param module Top-level module.
   *
   * @throws IllegalArgumentException when {@code module} argument is {@code null}.
   */
  IgStructureVisitor(
      final IGManager manager,
      final IGInterpreterRuntimeEnv runtimeEnv,
      final CfgModule module) {

    super(new IgOperationVisitor());
    InvariantChecks.checkNotNull(module);

    getOperationCfgVisitor().setParentVisitor(this);
    setBuilder(new VariableBuilder(this));

    this.manager = manager;
    this.runtimeEnv = runtimeEnv;
    this.topLevel = module;
    this.statements = new LinkedList<>();
    this.subProgramReturnVarMap = new LinkedHashMap<>();
    this.subProgramTags = new LinkedList<>();
    this.subProgramMappings = new LinkedHashMap<>();
  }

  /**
   * Creates the visitor object as a copy of the specified one.
   * @param visitor The object to be copied.
   */
  private IgStructureVisitor(final IgStructureVisitor visitor) {
    this(visitor.getManager(), visitor.getRuntimeEnv(), visitor.getTopLevelModule());

    this.process = visitor.getProcess();
    this.subProgramMappings = new LinkedHashMap<>(visitor.subProgramMappings);
  }

  /**
   * Returns sub-visitor for IGOperation objects.
   *
   * @return IGOperation objects sub-visitor.
   */
  IgOperationVisitor getOperationCfgVisitor() {

    return (IgOperationVisitor) this.fOperationVisitor;
  }

  /**
   * Returns top-level module.
   *
   * @return Top-level module that is processed at the current moment.
   */
  CfgModule getTopLevelModule() {
    return this.topLevel;
  }

  /**
   * Returns control flow graph model statements collection.
   *
   * @return Collection of statements that are constructed at the current moment.
   */
  public List<Statement> getStatements() {
    return this.statements;
  }

  /**
   * Returns IG interpreter runtime environment.
   *
   * @return Runtime environment.
   */
  IGInterpreterRuntimeEnv getRuntimeEnv() {
    return this.runtimeEnv;
  }

  /**
   * Returns IG manager.
   *
   * @return IG manager.
   */
  private IGManager getManager() {
    return this.manager;
  }

  /**
   * Returns the current process that is under construction.
   * @return The current process that is under construction.
   */
  public CfgProcess getProcess() {
    return this.process;
  }

  /**
   * Sets the specified process to be constructed by this visitor.
   *
   * @param process Process to be constructed.
   */
  private void setProcess(final CfgProcess process) {
    this.process = process;
  }

  /**
   * Returns inner variable and input\output signal builder.
   *
   * @return Variable builder.
   */
  public VariableBuilder getBuilder() {
    return this.builder;
  }

  private void setBuilder(final VariableBuilder builder) {
    this.builder = builder;
  }

  /**
   * Registers the specified mapping for arguments of the processed sub-program.
   * @param mapping The formal-to-actual mapping of the processed sub-program arguments.
   */
  private void registerMapping(final IGMapping mapping) {

    this.subProgramMappings.put(mapping.getFormal().toHRString(), mapping.getActual());
  }

  /**
   * Clears all the mappings for the processed sub-program.
   */
  void clearMappings() {
    this.subProgramMappings.clear();
  }

  /**
   * Checks whether this visitor processes a sub-program with the specified formal argument.
   * @param formalArg The name of the formal argument for the processed sub-program.
   * @return {@code true} when this visitor processes a sub-program with the specified formal
   *         argument, {@code false} otherwise.
   */
  boolean hasFormalArg(final String formalArg) {
    return this.subProgramMappings.containsKey(formalArg);
  }

  /**
   * Returns the actual argument for the specified formal one.
   * @param formalArg The formal argument name.
   * @return The actual argument for the specified formal one.
   */
  IGOperation getActualArg(final String formalArg) {
    return this.subProgramMappings.get(formalArg);
  }

  /**
   * Checks whether does the sub-program with the specified tag is already elaborated.
   * @param subProgramTag The sub-program tag.
   * @return {@code true} when sub-program with the specified tag has already been elaborated,
   *         {@code false} otherwise.
   */
  boolean hasProgramElaborated(final String subProgramTag) {
    return this.subProgramReturnVarMap.containsKey(subProgramTag);
  }

  /**
   * Returns the resulting value variable that is related to the sub-program with the specified tag.
   * @param subProgramTag The sub-program tag.
   * @return The resulting value variable that is related to the sub-program with the specified tag.
   */
  NodeVariable getReturnVar(final String subProgramTag) {
    return this.subProgramReturnVarMap.get(subProgramTag);
  }

  private void registerTag(final String subProgramTag, final NodeVariable tagVariable) {
    this.subProgramReturnVarMap.put(subProgramTag, tagVariable);
  }

  private void addSubProgramTag(final String subProgramTag) {
    this.subProgramTags.push(subProgramTag);
  }

  private void removeSubProgramTag() {
    this.subProgramTags.pop();
  }

  private String getCurrentSubProgramTag() {
    return this.subProgramTags.peek();
  }

  private void visitProcess(final IGProcess igProcess) {

    final CfgProcess newProcess = new CfgProcess();
    setProcess(newProcess);

    /* Get variable declarations. */
    final Map<NodeVariable, Declaration> declarations =
        getDeclarations(igProcess.getContainer());
    newProcess.declareVariables(declarations);

    /* Extract process sensitivity list events. */
    final EventList sensList = getBuilder().getProcessEventList(igProcess);
    newProcess.setEventList(sensList);

    /* Elaborate internal statements. */
    final IGSequenceOfStatements sos = igProcess.getSequenceOfStatements();
    visit(sos);

    final List<Statement> processNodes = new LinkedList<>(this.statements);
    newProcess.addAll(processNodes);

    /* Link process node with parent & child statements. */
    getTopLevelModule().add(newProcess);

    this.statements.clear();
    setProcess(null);
  }

  /**
   * Returns variables and their declarations from the specified container.
   * @param igContainer The container that keeps variable declarations.
   * @return The mapping from variables to their declarations.
   */
  Map<NodeVariable, Declaration> getDeclarations(final IGContainer igContainer) {

    final Map<NodeVariable, Declaration> map = new LinkedHashMap<>();

    for (int i = 0; i < igContainer.getNumLocalItems(); i++) {

      final IGContainerItem item = igContainer.getLocalItem(i);

      /* Elaborate variables (not-a-constants). */
      if (item instanceof IGObject && !IgUtils.isConstant((IGObject) item)) {

        final NodeVariable variable = getBuilder().getObjectVariable((IGObject) item);
        final Node initValue = getInitialValue((IGObject) item);
        final Range invariant = getBuilder().getInvariant(variable, ((IGObject) item).getType());

        map.put(variable, new Declaration(initValue, invariant));
      }
    }

    return map;
  }

  private Node getInitialValue(final IGObject igObject) {
    Node initValue;

    final IGOperation igValue = igObject.getInitialValue();

    /* Try to calculate LEFT expression since it is initial for VHDL variable. */
    if (igValue == null) {
      try {

        final IGType type = igObject.getType();
        final IGTypeStatic staticType = type.computeStaticType(getRuntimeEnv(), null, null);
        final SourceLocation sourceLocation = igObject.computeSourceLocation();
        final IGStaticValue staticLeft =
            staticType.getStaticRange() == null ? null : staticType.getStaticLeft(sourceLocation);
        final IGStaticValue leftValue =
            staticLeft == null ? IGStaticValue.generateZ(staticType, sourceLocation) : staticLeft;

        initValue = getOperationNode(leftValue);

      } catch (ZamiaException e) {
        throw new RetrascopeRuntimeException(e);
      }
    } else {
      initValue = getOperationNode(igValue);
    }

    if (initValue == null) {
      throw new RetrascopeRuntimeException(
          String.format("Can't calculate initial value of: %s.", igObject.toString()));
    }
    return initValue;
  }

  @Override
  public void visit(final IGStructure igStructure, final PathName pathName) {

    final Map<NodeVariable, Declaration> declarations =
        getDeclarations(igStructure.getContainer());

    for (int i = 0; i < igStructure.getNumStatements(); i++) {
      final IGConcurrentStatement stmt = igStructure.getStatement(i);

      if (stmt instanceof IGProcess) {
        visitProcess((IGProcess) stmt);
      } else if (stmt instanceof IGStructure) {
        visit((IGStructure) stmt, pathName);
      }
    }
    getTopLevelModule().declareVariables(declarations);
  }

  @Override
  public void visit(final IGSequentialStatement statement) {

    if (!(statement instanceof IGSequenceOfStatements) && statement != null) {

      final IGOperationInvokeSubprogram[] subPrograms = getSubPrograms(statement);

      for (final IGOperationInvokeSubprogram program : subPrograms) {

        final int mappingNum = program.getNumMappings();
        for (int i = 0; i < mappingNum; i++) {
          registerMapping(program.getMapping(i));
        }

        final String tagName = IgUtils.getTag(program);

        addSubProgramTag(tagName);

        final IGSubProgram programCode = program.getSub();
        final DataType returnType = getBuilder().getDataType(programCode.getReturnType());
        final NodeVariable returnVar = new NodeVariable(tagName, returnType);
        final IgSubprogramBuilder builder = new IgSubprogramBuilder(new IgStructureVisitor(this));
        final Statement[] statements = builder.build(program, tagName);

        registerTag(tagName, returnVar);
        removeSubProgramTag();
        addToThis(statements);
        clearMappings();
      }
    }
    super.visit(statement);
  }

  private IGOperationInvokeSubprogram[] getSubPrograms(final IGSequentialStatement statement) {

    final List<IGOperationInvokeSubprogram> subPrograms = new LinkedList<>();

    final int childNum = statement.getNumChildren();

    for (int i = 0; i < childNum; i++) {
      final IGItem child = statement.getChild(i);
      if (child != null) {
        findSubPrograms(subPrograms, child);
      }
    }

    return subPrograms.toArray(new IGOperationInvokeSubprogram[subPrograms.size()]);
  }

  private void findSubPrograms(
      final List<IGOperationInvokeSubprogram> subPrograms,
      final IGItem item) {

    if (item instanceof IGOperationInvokeSubprogram) {

      final String name = IgUtils.getName((IGOperationInvokeSubprogram) item);

      if (!IgSubprogramMapper.hasOperation(name) && !IgSubprogramMapper.hasMacro(name)) {
        subPrograms.add((IGOperationInvokeSubprogram) item);
      }
    } else {
      final int itemChildNum = item.getNumChildren();

      for (int i = 0; i < itemChildNum; i++) {
        final IGItem itemChild = item.getChild(i);

        if (itemChild != null) {
          findSubPrograms(subPrograms, itemChild);
        }
      }
    }
  }

  @Override
  public void visitWait(final IGSequentialWait igWait) {

    final IGOperation clause = igWait.getConditionClause();
    final Node condition = (clause == null) ? null : getOperationNode(clause);

    final EventList list = getBuilder().getEventList(igWait);
    final CfgWaitStatement waitNode =
        (condition == null) ? new CfgWaitStatement(list) : new CfgWaitStatement(list, condition);

    addToThis(waitNode);
  }

  @Override
  public void visitReturn(final IGSequentialReturn igReturn) {

    final IGItem igReturnValue = igReturn.getChild(0);
    final Node returnNode = getOperationNode((IGOperation) igReturnValue);

    final NodeVariable returnVar = getReturnVar(getCurrentSubProgramTag());
    final CfgBlockStatement returnBlock = new CfgBlockStatement(returnVar, returnNode);
    addToThis(returnBlock);
  }

  @Override
  public void visitReport(final IGSequentialReport igReport) {
    throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, igReport.toHRString());
  }

  @Override
  public void visitCall(final IGSequentialProcedureCall igCall) {
    throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, igCall.toHRString());
  }

  @Override
  public void visitNext(final IGSequentialNext igNext) {
    throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, igNext.toHRString());
  }

  @Override
  public void visitLoop(final IGSequentialLoop igLoop) {

    /* IG loop node children. */
    final IGRange loopRange = (IGRange)igLoop.getChild(0);
    final IGItem loopCounter = igLoop.getChild(1);
    final IGSequenceOfStatements loopBody = (IGSequenceOfStatements) igLoop.getChild(4);

    /* The range iteration variable. */
    final NodeVariable rangeVariable = getBuilder().getObjectVariable((IGObject) loopCounter);

    /* The loop condition. */
    final Node left = getOperationNode(loopRange.getLeft());
    final Node right = getOperationNode(loopRange.getRight());
    final Node ascending = getOperationNode(loopRange.getAscending());
    final boolean isAscending = ascending.equals(SolverUtils.TRUE);

    final Node range = createRange(rangeVariable, left, right, isAscending);

    /* Create the "i := 0" initialization block. */
    final Assignment init = new Assignment(rangeVariable, left);

    /* Create the incrementing "i++" basic block. */
    checkRangeParam(rangeVariable);

    final StandardOperation addOp = StandardOperation.ADD;
    final NodeValue one = NodeValue.newInteger(1);
    final NodeOperation incrementExpression = new NodeOperation(addOp, rangeVariable, one);
    final Assignment increment = new Assignment(rangeVariable, incrementExpression);

    final CfgLoopStatement loopStatement =
        new CfgLoopStatement(CfgLoopStatement.Type.FOR, range, init, increment);
    loopStatement.addAll(getSubGraph(loopBody));
    addToThis(loopStatement);
  }

  private void checkRangeParam(final NodeVariable rangeVariable) {
    if (!rangeVariable.isType(DataTypeId.LOGIC_INTEGER)) {
      throw RetrascopeRuntimeException.makeException(
          LogMessage.UNSUPPORTED_TYPE,
          rangeVariable.getDataType());
    }
  }

  @Override
  public void visitIf(final IGSequentialIf igIf) {

    final IGOperation ifCondition = igIf.getCond();
    final IGSequenceOfStatements elseSos = igIf.getElseSOS();
    final IGSequenceOfStatements thenSos = igIf.getThenSOS();

    final Node condition = getOperationNode(ifCondition);

    /* Transform operands & detect event-dependent part of "x'event && x='0'" condition .*/
    final EventList events = getEvents(ifCondition);

    final CfgSwitchStatement switchStatement = new CfgSwitchStatement(condition);

    final CfgCaseStatement trueCase = new CfgCaseStatement(events, SolverUtils.TRUE);
    switchStatement.add(trueCase);
    if (thenSos != null) {
      trueCase.addAll(getSubGraph(thenSos));
    }

    final CfgCaseStatement falseCase = new CfgCaseStatement(SolverUtils.FALSE);
    switchStatement.add(falseCase);
    if (elseSos != null) {
      falseCase.addAll(getSubGraph(elseSos));
    }

    addToThis(switchStatement);
  }

  private EventList getEvents(final IGOperation ifCondition) {

    final EventList list = new EventList();

    for (int i = 0; i < ifCondition.getNumOperands(); i++) {
      final IGOperation operand = ifCondition.getOperand(i);
      if (isEvent(operand)) {
        final Event event = (Event)getOperationCfgVisitor().getData(operand);
        list.add(event);
      }
    }

    return list;
  }

  private static boolean isEvent(final IGOperation operand) {
    return operand instanceof IGOperationAttribute
        && ((IGOperationAttribute)operand).getAttrOp().equals(IGOperationAttribute.AttrOp.EVENT);
  }

  @Override
  public void visitExit(final IGSequentialExit igExit) {
    throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, igExit.toHRString());
  }

  @Override
  public void visitAssert(final IGSequentialAssert igAssert) {

    final IGOperation assertOp = igAssert.getOp();
    final Node condition = getOperationNode(assertOp);
    final CfgAssertStatement.Severity severity = getSeverity(igAssert);
    final String report = igAssert.getReport().toHRString();

    final CfgAssertStatement assertStatement = new CfgAssertStatement(condition, severity, report);

    addToThis(assertStatement);
  }

  private CfgAssertStatement.Severity getSeverity(final IGSequentialAssert igAssert) {

    final String sevString = igAssert.getSeverity().toHRString();

    if (SEVERITY_NOTE.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.NOTE;
    } else if (SEVERITY_WARN.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.WARNING;
    } else if (SEVERITY_ERROR.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.ERROR;
    } else if (SEVERITY_FAIL.equalsIgnoreCase(sevString)) {

      return CfgAssertStatement.Severity.FAILURE;
    } else {
      Log.printWarn(LogMessage.CANT_PARSE + sevString + ".");
    }
    return null;
  }

  @Override
  public void visitAssignment(final IGSequentialAssignment igAssignment) {

    final IGOperation target = igAssignment.getTarget();
    final IGOperation value = igAssignment.getValue();

    addToThis(getBlock(target, getOperationNode(value)));
  }

  private Node getOperationNode(final IGOperation node) {
    return getOperationCfgVisitor().getNode(node);
  }

  private CfgBlockStatement getBlock(final IGOperation target, final Node valueExpr) {

    final RangedVariable rangedVariable = getBuilder().getRangedVariable(target);

    final Assignment assignment = new Assignment(rangedVariable, valueExpr);

    /* If the variable is of signal type, then basic block is concurrent. */
    final boolean isConcurrent = isSignal(rangedVariable);

    return new CfgBlockStatement(assignment, isConcurrent);
  }

  private boolean isSignal(final RangedVariable rangedVariable) {

    final NodeVariable variable = rangedVariable.getVariable();
    final Object userData = variable.getUserData();

    return VariableType.isSignal(((VariableDescriptor) userData).getVariableType());
  }

  private List<Statement> getSubGraph(final IGSequenceOfStatements stmt) {
    final IgStructureVisitor visitor = new IgStructureVisitor(this);
    visitor.setProcess(getProcess());
    visitor.visit(stmt);
    return visitor.getStatements();
  }

  private Node createRange(
      final NodeVariable var,
      final Node left,
      final Node right,
      final boolean isAscending) {

    /* Select operation identifiers depending on data types. */
    final boolean areBitVector = ExprUtils.isType(DataTypeId.BIT_VECTOR, left, right);

    final StandardOperation gtId =
        areBitVector ? StandardOperation.BVUGE : StandardOperation.GREATEREQ;
    final StandardOperation lessId =
        areBitVector ? StandardOperation.BVULE : StandardOperation.LESSEQ;

    Node expr;
    if (isAscending) {

      /* The result is: left <= var <= right. */
      expr = Node.AND(new NodeOperation(gtId, var, left), new NodeOperation(lessId, var, right));
    } else {

      /* The result is: left >= var >= right. */
      expr = Node.AND(new NodeOperation(lessId, var, left), new NodeOperation(gtId, var, right));
    }

    return expr;
  }

  private void addToThis(final Statement statement) {
    this.statements.add(statement);
  }

  private void addToThis(final Statement[] statements) {
    for (final Statement statement : statements) {
      addToThis(statement);
    }
  }
}
