/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.apache.commons.lang3.StringUtils;
import org.zamia.instgraph.IGItem;
import org.zamia.instgraph.IGMapping;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGOperationObject;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGSubProgram;

/**
 * Instantiation Graph (IG) related utility methods.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class IgUtils {

  private IgUtils() {
    /* Empty. */
  }

  /**
   * Returns name of the specified 'formal' parameter of binding.
   *
   * @param formal Formal parameter of binding.
   * @return Name of the specified parameter.
   */
  private static String getBindFormalName(final IGOperation formal) {
    return formal instanceof IGOperationObject
      ? ((IGOperationObject)formal).getObject().getId()
      : formal.getId();
  }

  /**
   * Returns the string name for the specified program object.
   * @param program The program object.
   * @return The string name for the specified program object.
   */
  public static String getName(final IGOperationInvokeSubprogram program) {
    final String subProgramName = program.getSub().getId();

    return getNonQuotedString(subProgramName);
  }

  private static String getNonQuotedString(final String opName) {
    final char dQuote = '"';
    final int dqIndex = opName.indexOf(dQuote);
    return dqIndex < 0 ? opName : opName.substring(dqIndex + 1, opName.lastIndexOf(dQuote));
  }

  /**
   * Returns the string representation of the specified program call.
   * <p>The tag is a string that is constructed from the program name and it's formal arguments
   * are separated with underscore symbols.</p>
   * @param program The program call which tag is created.
   * @return The string representation of the specified program call.
   */
  static String getTag(final IGOperationInvokeSubprogram program) {

    final String name = getName(program);

    final int mapNum = program.getNumMappings();
    final String[] argNames = new String[mapNum];

    for (int i = 0; i < mapNum; i++) {

      final IGMapping mapping = program.getMapping(i);
      argNames[i] = getBindFormalName(mapping.getFormal());
    }
    return String.format("%s_%s", name, StringUtils.join(argNames, "_"));
  }

  /**
   * Checks whether the specified program contains a recursion.
   *
   * @param program The program to be checked.
   * @return {@code true} when the specified program is a recursive one, {@code false} otherwise.
   */
  static boolean recursiveProgram(final IGOperationInvokeSubprogram program) {

    final IGSubProgram subProgram = program.getSub();
    final String id = subProgram.getId();
    final IGSequenceOfStatements code = subProgram.getCode();

    return code != null && hasRecursion(id, code);
  }

  private static boolean hasRecursion(
      final String programId,
      final IGSequenceOfStatements code) {

    for (int i = 0; i < code.getNumStatements(); i++) {
      final IGSequentialStatement stmt = code.getStatement(i);

      if (stmt instanceof IGSequenceOfStatements
          && hasRecursion(programId, (IGSequenceOfStatements) stmt)) {

        return true;
      }

      if (stmt.getNumChildren() > 0) {
        for (int j = 0; j < stmt.getNumChildren(); j++) {
          final IGItem child = stmt.getChild(j);
          if (child != null
              && child instanceof IGOperationInvokeSubprogram
              && ((IGOperationInvokeSubprogram) child).getSub().getId().equals(programId)) {

          /* Recursion */
            return true;
          }
        }
      }
    }
    return false;
  }

  /**
   * Checks whether the specified object is of constant category.
   * @param igObject The object to be checked.
   * @return {@code true} when the specified object is of constant category,
   *         {@code false} otherwise.
   */
  static boolean isConstant(final IGObject igObject) {

    final IGObject.IGObjectCat category = igObject.getCat();

    return category.equals(IGObject.IGObjectCat.CONSTANT);
  }

  /**
   * Checks whether the specified object is of variable category.
   * @param igObject The object to be checked.
   * @return {@code true} when the specified object is of variable category,
   *         {@code false} otherwise.
   */
  static boolean isVariable(final IGObject igObject) {

    final IGObject.IGObjectCat cat = igObject.getCat();

    return cat.equals(IGObject.IGObjectCat.SIGNAL) || cat.equals(IGObject.IGObjectCat.VARIABLE);
  }
}
