/*
 * Copyright 2015-2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.parser.vhdl;

import org.zamia.SourceLocation;
import org.zamia.ZamiaException;
import org.zamia.instgraph.IGObject;
import org.zamia.instgraph.IGOperation;
import org.zamia.instgraph.IGOperationIndex;
import org.zamia.instgraph.IGOperationInvokeSubprogram;
import org.zamia.instgraph.IGOperationObject;
import org.zamia.instgraph.IGOperationRange;
import org.zamia.instgraph.IGProcess;
import org.zamia.instgraph.IGRange;
import org.zamia.instgraph.IGSequenceOfStatements;
import org.zamia.instgraph.IGSequentialStatement;
import org.zamia.instgraph.IGSequentialWait;
import org.zamia.instgraph.IGStaticValue;
import org.zamia.instgraph.IGType;
import org.zamia.instgraph.IGTypeStatic;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.EventType;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.util.LogMessage;
import ru.ispras.retrascope.util.NodeUtils;
import ru.ispras.retrascope.util.SolverUtils;

import java.math.BigInteger;
import java.util.List;

/**
 * VHDL parser component that extracts control flow graph model nodes from IG objects.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class VariableBuilder {

  private final IgStructureVisitor parentVisitor;

  /**
   * Constructs object with the specified parent visitor of IG structure.
   *
   * @param visitor Parent IG structure visitor for this object.
   */
  VariableBuilder(final IgStructureVisitor visitor) {
    this.parentVisitor = visitor;
  }

  private IgStructureVisitor getParentVisitor() {
    return this.parentVisitor;
  }

  /**
   * Returns event list of the specified IG process object.
   *
   * <p>Returns {@link EventList} sensitivity list
   * of the specified IG (Instantiation Graph) process object.</p>
   *
   * @param process IG process to which the event list is extracted.
   * @return Event list of the specified process.
   */
  EventList getProcessEventList(final IGProcess process) {

    final EventList result = new EventList();

    final IGSequenceOfStatements sos = process.getSequenceOfStatements();

    final int maxStmtNum = sos.getNumStatements();

    for (int i = 0; i < maxStmtNum; i++) {

      final IGSequentialStatement sosStmt = sos.getStatement(i);

      if (sosStmt instanceof IGSequentialWait) {
        result.addEvents(getEventList((IGSequentialWait) sosStmt));
      }
    }
    return result;
  }

  /**
   * Returns event list of the specified IG wait object.
   *
   * <p>Returns {@link EventList} sensitivity list
   * of the specified IG (Instantiation Graph) wait node object.</p>
   *
   * @param wait IG wait node from which the event list is extracted.
   * @return Event list to be extracted from the specified node.
   */
  public EventList getEventList(final IGSequentialWait wait) {

    final EventList list = new EventList();

    final List<IGOperation> sensList = wait.getSensitivityList();

    if (sensList != null) {
      for (final IGOperation current : sensList) {
        list.add(getEvent(current));
      }
      final IGOperation conditionClause = wait.getConditionClause();
      if (conditionClause != null) {
        list.add(getEvent(conditionClause));
      }
    }
    return list;
  }

  private Event getEvent(final IGOperation current) {

    final RangedVariable rVar = getRangedVariable(current);
    EventType type = null;

    if (isType(current, EventType.POSITIVE_EDGE)) {
      type = EventType.POSITIVE_EDGE;
    } else if (isType(current, EventType.NEGATIVE_EDGE)) {
      type = EventType.NEGATIVE_EDGE;
    }
    return type == null ? new Event(rVar) : new Event(rVar, type);
  }

  private static boolean isType(
      final IGOperation current,
      final EventType type) {

    return current.toHRString().equalsIgnoreCase(type.name());
  }

  /**
   * Extracts ranged variable from IGOperation object.
   *
   * @param operation Object that contains data to be extracted.
   *
   * @return Ranged variable that is extracted from the specified object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeRuntimeException when argument type is unsupported.
   */
  RangedVariable getRangedVariable(final IGOperation operation) {
    InvariantChecks.checkNotNull(operation);
    RangedVariable var;

    if (operation instanceof IGOperationInvokeSubprogram) {

      var = getSubprogramVariable((IGOperationInvokeSubprogram) operation);

    } else if (operation instanceof IGOperationObject) {

      var = new RangedVariable(getObjectVariable(((IGOperationObject) operation).getObject()));

    } else if (operation instanceof IGOperationIndex) {

      var = getIndexVariable((IGOperationIndex) operation);

    } else if (operation instanceof IGOperationRange) {

      var = getRangeVariable((IGOperationRange) operation);

    } else {
      throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, operation.toHRString());
    }
    return var;
  }

  private RangedVariable getRangeVariable(final IGOperationRange opRange) {

    final IGOperation igOperand = opRange.getOperand();
    final Node operand = getOperationVisitor().getNode(igOperand);

    final IGOperation range = opRange.getRange();

    checkRangeVariableParams(operand, range);

    final String name = ((NodeVariable) operand).getName();
    final DataType dataType = ((NodeVariable) operand).getData().getType();

    final IGOperation leftIndex = range.getOperand(0);
    final IGOperation rightIndex = range.getOperand(2);

    final Node left = getOperationVisitor().getNode(leftIndex);
    final Node right = getOperationVisitor().getNode(rightIndex);

    Range operandRange = new Range(left, right);
    final NodeVariable variable = new NodeVariable(name, dataType);
    final VariableType variableType = VariableBuilderUtils.getVariableType(igOperand);
    final VariableDescriptor varData = new VariableDescriptor(variableType);
    varData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, opRange.getType().toHRString());
    variable.setUserData(varData);
    fillVariableDescriptor(varData, opRange);

    /* Subtract from range parameters the OFFSET value if it is determined. */
    getOperationVisitor().checkBitVectorMinIndex(igOperand, variable, varData);
    if (varData.hasMetaInfo(IgOperationVisitor.OFFSET)) {
      final String shift = varData.getMetaInfo(IgOperationVisitor.OFFSET);

      final Node shiftValue = NodeValue.newInteger(Integer.parseInt(shift));
      final Node young = operandRange.getYoung();
      final Node old = operandRange.getOld();
      operandRange =
          new Range(
              new NodeOperation(StandardOperation.SUB, young, shiftValue),
              new NodeOperation(StandardOperation.SUB, old, shiftValue));
    }

    return new RangedVariable(variable, operandRange);
  }

  private static void checkRangeVariableParams(
      final Node operand,
      final IGOperation range) {

    if (operand.getKind() != Node.Kind.VARIABLE || !(range instanceof IGRange)) {
      throw new UnsupportedOperationException(range.toHRString());
    }
  }

  private RangedVariable getIndexVariable(final IGOperationIndex opIdx) {
    final IGOperation igOperand = opIdx.getOperand();
    final Node operand = getOperationVisitor().getNode(igOperand);

    final VariableType type = VariableBuilderUtils.getVariableType(opIdx);
    final VariableDescriptor operandData = new VariableDescriptor(type);
    operandData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, igOperand.getType().toHRString());
    operand.setUserData(operandData);
    fillVariableDescriptor(operandData, opIdx);

    final RangedVariable var = new RangedVariable((NodeVariable) operand);
    final Node index = getOperationVisitor().getNode(opIdx.getIndex());

    var.setRange(new Range(index, index));

    return var;
  }

  /**
   * Extracts variable from the specified IG object.
   *
   * @param object IG object that contains data to be extracted.
   * @return Variable that is extracted from the specified object.
   */
  NodeVariable getObjectVariable(final IGObject object) {

    final IGType opType = object.getType();

    final VariableType variableType = VariableBuilderUtils.getVariableType(object);
    final DataType variableValueType = getDataType(opType);
    final NodeVariable variable = new NodeVariable(object.getId(), variableValueType);
    final VariableDescriptor variableDescriptor = new VariableDescriptor(variableType);
    final String fullType = getFullType(object);
    variableDescriptor.addMetaInfo(MetaInfoType.HDL_TYPE_STR, fullType);
    variable.setUserData(variableDescriptor);
    fillVariableDescriptor(variableDescriptor, object);

    getOperationVisitor().checkBitVectorMinIndex(object, variable, variableDescriptor);
    return variable;
  }

  private RangedVariable getSubprogramVariable(final IGOperationInvokeSubprogram sub) {
    RangedVariable var;
    final int opNum = sub.getNumOperands();
    if (opNum == 1) {
      var = getRangedVariable(sub.getOperand(0));
    } else {
      throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, sub.toHRString());
    }
    return var;
  }

  /**
   * Returns a full string representation of IG object type.
   *
   * @param object IG object that is processed by this method.
   * @return String representation of object type including base type and invariant.
   */
  static String getFullType(final IGObject object) {

    final IGType type = object.getType();
    final IGType elemType = type.getElementType();

    /* Bit vectors are treated as arrays in IG representation. */
    final boolean isBitVector = !type.isAtomic()
        && (type.isArray() && elemType.isLogic() || type.isCharEnum());

    /* The type representation. */
    final String typeId = isBitVector
        ? String.format("bit_vector(%s)", type.getIndexType().toHRString())
        : type.getId() == null ? type.getCat().name() : type.getId();

    /* Extract possible range (for integer variables only). */
    final boolean isRangedInteger = type.isInteger() && type.getRange() != null;
    final String range = isRangedInteger ? " range " +  type.getRange().toHRString() : "";

    return String.format("%s%s", typeId, range);
  }

  /**
   * Extracts data type of the specified IG type object.
   *
   * @param type IG object type.
   *
   * @return Data type that is extracted from the specified IG data type.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws UnsupportedOperationException when argument is not of expected category
   *         (INTEGER, REAL, ENUM, ARRAY) or is not of expected sub-category
   *         (ENUM-bit, ENUM-bool, ENUM-logic, ARRAY-bit, ARRAY-array, ARRAY-integer, ARRAY-logic).
   */
  public DataType getDataType(final IGType type) {
    InvariantChecks.checkNotNull(type);

    DataType dataType;
    switch (type.getCat()) {
      case INTEGER:
        dataType = DataType.INTEGER;
        break;
      case REAL:
        dataType = DataType.REAL;
        break;
      case ENUM:
        dataType = VariableBuilderUtils.getEnumDataType(type);
        break;
      case ARRAY:
        dataType = getArrayDataType(type);
        break;
      default:
        throw RetrascopeRuntimeException.makeException(
            LogMessage.UNSUPPORTED_TYPE, type.toHRString());
    }
    return dataType;
  }

  private DataType getArrayDataType(final IGType type) {

    DataType dataType;

    final IGType elemType = type.getElementType();
    final IGType indexType = type.getIndexType();

    if (elemType.isBit()) {

      /* Bit vector, calculate it's size. */
      if (indexType instanceof IGTypeStatic) {

        final int size = getTypeSize((IGTypeStatic) indexType);
        dataType = DataType.BIT_VECTOR(size);

      } else {
        throw new UnsupportedOperationException(type.toHRString());
      }

    } else if (elemType.isArray()) {

      dataType = DataType.MAP(DataType.INTEGER, getDataType(elemType));

    } else if (elemType.isInteger()) {

      dataType = DataType.MAP(DataType.INTEGER, DataType.INTEGER);

    } else if (elemType.isLogic()) {

      /* Bit vector, calculate it's size. */
      if (indexType instanceof IGTypeStatic) {

        final int size = getTypeSize((IGTypeStatic) indexType);
        dataType = DataType.BIT_VECTOR(size);

      } else {
        throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, type.toHRString());
      }

    } else {
      throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, type.toHRString());
    }
    return dataType;
  }

  /**
   * Returns size of the specified type.
   *
   * @param indexType Index type to be processed.
   *
   * @return Size of the specified index type.
   *
   * @throws RetrascopeRuntimeException when index left/right borders
   *         cannot be transformed into {@link ru.ispras.fortress.expression.NodeValue} objects.
   * @throws RetrascopeRuntimeException when index left/right borders
   *         are not of integer types.
   */
  int getTypeSize(final IGTypeStatic indexType) {

    final IGStaticValue range = indexType.getStaticRange();
    final Node left = getOperationVisitor().getNode(range.getLeft());
    final Node right = getOperationVisitor().getNode(range.getRight());

    checkRangeParams(left, right);

    final int leftValue = ((BigInteger) ((NodeValue) left).getValue()).intValue();
    final int rightValue = ((BigInteger) ((NodeValue) right).getValue()).intValue();

    return Math.abs(leftValue - rightValue + 1);
  }

  private static void checkRangeParams(
      final Node left,
      final Node right) {

    if (left.getKind() != Node.Kind.VALUE || right.getKind() != Node.Kind.VALUE) {
      throw new RetrascopeRuntimeException(
        String.format("%s%s, %s - should be of %s kind.",
          LogMessage.UNSUPPORTED_TYPE,
          left.getKind(),
          right.getKind(),
          Node.Kind.VALUE));
    }
    if (!ExprUtils.isType(DataType.INTEGER, left, right)) {
      throw new RetrascopeRuntimeException(
        String.format("%s%s, %s - should be of %s type.",
          LogMessage.UNSUPPORTED_TYPE,
          ((NodeValue) left).getData().getType().toString(),
          ((NodeValue) right).getData().getType().toString(),
          DataType.INTEGER));
    }
  }



  /**
   * Extracts invariant to the specified variable from the specified IGType object.
   *
   * @param variable Variable which invariant is to be extracted.
   * @param type Type which contains an invariant that is to be extracted.
   *
   * @return Invariant for the specified variable that is in the specified type object.
   *
   * @throws RetrascopeRuntimeException when argument is ranged
   *         and it's range is neither of {@code IGStaticValue}
   *         nor of {@code IGRange} type.
   */
  public Range getInvariant(
      final NodeVariable variable,
      final IGType type) {

    Range invariant;

    if (noTypeInvariant(variable, type)) {

      invariant = null;

    } else {

      final IGOperation range = type.getRange();

      if (range instanceof IGStaticValue) {
        invariant = getStaticRangeInvariant((IGStaticValue) range);
      } else if (range instanceof IGRange) {
        invariant = getVariableRangeInvariant((IGRange) range);
      } else {
        throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, range.getClass());
      }
    }
    return invariant;
  }

  private static boolean noTypeInvariant(
      final NodeVariable variable,
      final IGType type) {

    /* One-bit-sized vectors or booleans or non-ranged variables
     * do not have any type invariant. */
    return variable.isType(DataType.BIT_VECTOR(1))
      || variable.isType(DataType.BOOLEAN)
      || type.getRange() == null;
  }

  private Range getVariableRangeInvariant(final IGRange range) {

    final Node left = getOperationVisitor().getNode(range.getLeft());
    final Node right = getOperationVisitor().getNode(range.getRight());

    final Node ascending = getOperationVisitor().getNode(range.getAscending());

    final NodeValue trueNode =
        ascending.isType(DataTypeId.BIT_VECTOR)
            ? NodeValue.newBitVector(BitVector.TRUE) : SolverUtils.TRUE;

    final Node checkAscending = NodeUtils.getEquation(ascending, trueNode);

    /* Determine whether ascending node is true or false. */
    final boolean isAscending = SolverUtils.isSat(checkAscending);

    return Range.getRangeInvariant(left, right, isAscending);
  }

  private Range getStaticRangeInvariant(final IGStaticValue range) {

    /* The left/right borders. */
    final Node left = getOperationVisitor().getNode(range.getLeft());
    final Node right = getOperationVisitor().getNode(range.getRight());

    /* The flag that shows whether range is ascending or descending. */
    final boolean isAscending = range.getAscending().isTrue();

    return Range.getRangeInvariant(left, right, isAscending);
  }

  /**
   * Fill the specified descriptor with the specified IG operation object.
   *
   * @param descriptor Descriptor to be filled.
   * @param igOperation IG operation object to be analyzed.
   */
  void fillVariableDescriptor(final VariableDescriptor descriptor, final IGOperation igOperation) {

    if (igOperation instanceof IGOperationIndex) {
      return;
    }
    if (igOperation instanceof IGOperationObject) {

      fillVariableDescriptor(descriptor, ((IGOperationObject) igOperation).getObject());
      return;
    }
    if (igOperation instanceof IGOperationRange) {
      return;
    }
    throw RetrascopeRuntimeException.makeException(LogMessage.CANT_PARSE, igOperation.toHRString());
  }

  /**
   * Fill the specified descriptor with the data that are extracted from the specified
   * IG object.
   *
   * @param descriptor Descriptor to be filled.
   * @param object IG object to be analyzed.
   */
  private void fillVariableDescriptor(final VariableDescriptor descriptor, final IGObject object) {

    // BASE attribute
    final IGType igType = object.getType();
    descriptor.setBase(getDataType(igType));

    try {
      final SourceLocation srcLocation = object.computeSourceLocation();

      final IGTypeStatic staticType =
          igType.isArray()
          ? (IGTypeStatic) igType.getIndexType()
          : igType.computeStaticType(getParentVisitor().getRuntimeEnv(), null, null);

      final IGOperation range = staticType.getRange();

      if (range instanceof IGStaticValue) {
        // HIGH attribute
        final IGStaticValue high = staticType.getStaticHigh(srcLocation);
        descriptor.setHigh(getNode(high));

        // LOW attribute
        final IGStaticValue low = staticType.getStaticLow(srcLocation);
        descriptor.setLow(getNode(low));

        // ASCENDING attribute
        final IGStaticValue ascending = staticType.getStaticAscending(srcLocation);
        descriptor.setAscending(getNode(ascending));

        // LEFT attribute
        //      final IGStaticValue leftValue = IGStaticValue.generateZ(staticType, sourceLocation);
        descriptor.setLeft(isAscending(descriptor) ? descriptor.getLow() : descriptor.getHigh());

        // RIGHT attribute
        descriptor.setRight(isAscending(descriptor) ? descriptor.getHigh() : descriptor.getLow());

        if (igType.isArray()) {

          // RANGE attribute
          descriptor.setRange(
              isAscending(descriptor)
                  ? new Range(descriptor.getLow(), descriptor.getHigh())
                  : new Range(descriptor.getHigh(), descriptor.getLow()));

          // REVERSE_RANGE attribute
          descriptor.setReverseRange(
              isAscending(descriptor)
                  ? new Range(descriptor.getHigh(), descriptor.getLow())
                  : new Range(descriptor.getLow(), descriptor.getHigh()));

          // LENGTH attribute
          descriptor.setLength(
              new NodeOperation(
                  StandardOperation.ADD,
                      new NodeOperation(
                          StandardOperation.SUB,
                          descriptor.getHigh(),
                          descriptor.getLow()),
                      NodeValue.newInteger(1)));
        }
      } else if (range instanceof IGRange) {

        // LEFT attribute
        descriptor.setLeft(getNode((IGStaticValue) range.getRangeLeft(srcLocation)));

        // RIGHT attribute
        descriptor.setRight(getNode((IGStaticValue) range.getRangeRight(srcLocation)));

        // ASCENDING attribute
        descriptor.setAscending(getNode((IGStaticValue) ((IGRange)range).getAscending()));

        // HIGH attribute
        descriptor.setHigh(isAscending(descriptor) ? descriptor.getRight() : descriptor.getLeft());

        // LOW attribute
        descriptor.setLow(isAscending(descriptor) ? descriptor.getLeft() : descriptor.getRight());
      }
    } catch (ZamiaException e) {
      throw new RetrascopeRuntimeException(e);
    }
  }

  private static boolean isAscending(final VariableDescriptor descriptor) {
    return descriptor.getAscending().equals(SolverUtils.TRUE);
  }

  private IgOperationVisitor getOperationVisitor() {
    return getParentVisitor().getOperationCfgVisitor();
  }

  private Node getNode(final IGStaticValue value) {
    return getOperationVisitor().getNode(value);
  }
}
