/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl.mapper;

import org.zamia.instgraph.IGOperationBinary;

import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Mapping between IG binary operation identifiers
 * and standard operation identifiers that are supported.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class IgBinaryMapper {

  private static final Map<IGOperationBinary.BinOp, StandardOperation> MAP;

  private IgBinaryMapper() {
    /* Empty. */
  }

  static {
    MAP = new LinkedHashMap<>();

    MAP.put(IGOperationBinary.BinOp.ADD, StandardOperation.ADD);
    MAP.put(IGOperationBinary.BinOp.AND, StandardOperation.AND);
    MAP.put(IGOperationBinary.BinOp.CONCAT, StandardOperation.BVCONCAT);
    MAP.put(IGOperationBinary.BinOp.DIV, StandardOperation.DIV);
    MAP.put(IGOperationBinary.BinOp.EQUAL, StandardOperation.EQ);
    MAP.put(IGOperationBinary.BinOp.GREATER, StandardOperation.GREATER);
    MAP.put(IGOperationBinary.BinOp.GREATEREQ, StandardOperation.GREATEREQ);
    MAP.put(IGOperationBinary.BinOp.LESS, StandardOperation.LESS);
    MAP.put(IGOperationBinary.BinOp.LESSEQ, StandardOperation.LESSEQ);
    MAP.put(IGOperationBinary.BinOp.MAX, StandardOperation.MAX);
    MAP.put(IGOperationBinary.BinOp.MIN, StandardOperation.MIN);
    MAP.put(IGOperationBinary.BinOp.MOD, StandardOperation.MOD);
    MAP.put(IGOperationBinary.BinOp.MUL, StandardOperation.MUL);
    MAP.put(IGOperationBinary.BinOp.NAND, StandardOperation.BVNAND);
    MAP.put(IGOperationBinary.BinOp.NEQUAL, StandardOperation.NOTEQ);
    MAP.put(IGOperationBinary.BinOp.NOR, StandardOperation.BVNOR);
    MAP.put(IGOperationBinary.BinOp.OR, StandardOperation.OR);
    MAP.put(IGOperationBinary.BinOp.POWER, StandardOperation.POWER);
    MAP.put(IGOperationBinary.BinOp.REM, StandardOperation.REM);
    MAP.put(IGOperationBinary.BinOp.ROL, StandardOperation.BVROL);
    MAP.put(IGOperationBinary.BinOp.ROR, StandardOperation.BVROR);
    MAP.put(IGOperationBinary.BinOp.SLA, StandardOperation.BVASHL);
    MAP.put(IGOperationBinary.BinOp.SLL, StandardOperation.BVLSHL);
    MAP.put(IGOperationBinary.BinOp.SRA, StandardOperation.BVASHR);
    MAP.put(IGOperationBinary.BinOp.SRL, StandardOperation.BVLSHR);
    MAP.put(IGOperationBinary.BinOp.SUB, StandardOperation.SUB);
    MAP.put(IGOperationBinary.BinOp.XNOR, StandardOperation.BVXNOR);
    MAP.put(IGOperationBinary.BinOp.XOR, StandardOperation.XOR);
  }

  /**
   * Checks whether operation is supported.
   *
   * <p>If operation is supported, then it's identifier is in this mapping.</p>
   *
   * @param operation IG operation identifier.
   *
   * @return {@code true} if operation is supported, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  private static boolean hasOperation(final IGOperationBinary.BinOp operation) {
    InvariantChecks.checkNotNull(operation);

    return MAP.containsKey(operation);
  }

  /**
   * Gets operation code of the specified operation.
   *
   * @param binary IG operation
   *
   * @return Operation identifier.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws UnsupportedOperationException when operation does not
   *         have corresponding operation code in this mapping.
   */
  public static StandardOperation getOperationCode(final IGOperationBinary binary) {
    InvariantChecks.checkNotNull(binary);

    if (IgBinaryMapper.hasOperation(binary.getBinOp())) {
      return MAP.get(binary.getBinOp());
    } else {
      throw new UnsupportedOperationException(binary.getBinOp().name());
    }
  }
}
