/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl.mapper;

import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Mapping between IG subprogram identifiers and standard operation identifiers that are supported.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class IgSubprogramMapper {

  private static final Map<String, StandardOperation> OPERATION_MAP;
  private static final Map<String, Macro> MACRO_MAP;

  private IgSubprogramMapper() {
    /* Empty. */
  }

  private enum Macro {
    BV_INCREMENT,
    BV2INT
  }

  static {
    OPERATION_MAP = new LinkedHashMap<>();

    OPERATION_MAP.put("=", StandardOperation.EQ);
    OPERATION_MAP.put("<=", StandardOperation.LESSEQ);
    OPERATION_MAP.put("<", StandardOperation.LESS);
    OPERATION_MAP.put(">=", StandardOperation.GREATEREQ);
    OPERATION_MAP.put(">", StandardOperation.GREATER);
    OPERATION_MAP.put("+", StandardOperation.ADD);
    OPERATION_MAP.put("-", StandardOperation.SUB);
    OPERATION_MAP.put("AND", StandardOperation.AND);
    OPERATION_MAP.put("OR", StandardOperation.OR);
    OPERATION_MAP.put("NOT", StandardOperation.NOT);
    OPERATION_MAP.put("/=", StandardOperation.NOTEQ);
    OPERATION_MAP.put("&", StandardOperation.BVCONCAT);
    OPERATION_MAP.put("XOR", StandardOperation.BVXOR);
    OPERATION_MAP.put("*", StandardOperation.MUL);
    OPERATION_MAP.put("MOD", StandardOperation.MOD);
    OPERATION_MAP.put("/", StandardOperation.DIV);
    OPERATION_MAP.put("**", StandardOperation.POWER);
    OPERATION_MAP.put("NOR", StandardOperation.BVNOR);
    OPERATION_MAP.put("BV_NEGATE", StandardOperation.BVNEG);
    OPERATION_MAP.put("BV_ADDER", StandardOperation.BVADD);

    MACRO_MAP = new LinkedHashMap<>();
    MACRO_MAP.put("BV_INCREMENT", Macro.BV_INCREMENT);
    MACRO_MAP.put("CONV_INTEGER", Macro.BV2INT);
  }

  /**
   * Checks whether operation is supported.
   *
   * @param operation Operation identifier.
   *
   * @return {@code true} if operation is supported, {@code true} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static boolean hasOperation(final String operation) {
    InvariantChecks.checkNotNull(operation);

    return OPERATION_MAP.containsKey(operation);
  }

  /**
   * Checks whether this macro-definition is supported.
   *
   * @param name The macro-definition name.
   *
   * @return {@code true} if macro-definition is supported, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static boolean hasMacro(final String name) {
    InvariantChecks.checkNotNull(name);

    return MACRO_MAP.containsKey(name);
  }

  /**
   * Gets operation code of the operation with the specified name.
   *
   * @param opName Operation name.
   *
   * @return Operation code.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws UnsupportedOperationException when {@code IgSubprogramMapper}
   *         does not contain an operation with the specified string name.
   */
  public static StandardOperation getOperationCode(final String opName) {
    InvariantChecks.checkNotNull(opName);

    if (hasOperation(opName)) {
      return OPERATION_MAP.get(opName);
    }
    throw new UnsupportedOperationException(opName);
  }

  /**
   * Returns the syntax node for the macro-definition that is described by it's name and operands.
   *
   * @param name The macro-definition name.
   * @param operands Operands for the macro-definition.
   * @return The syntax node for the macro-definition that is described by it's name and operands.
   */
  public static Node getMacroOperation(final String name, final Node ... operands) {

    if (hasMacro(name)) {

      switch (MACRO_MAP.get(name)) {
        case BV_INCREMENT:
          return new NodeOperation(
              StandardOperation.BVADD,
              operands[0],
              NodeValue.newBitVector(BitVector.valueOf(1, operands[0].getDataType().getSize())));
        case BV2INT:
          return new NodeOperation(StandardOperation.BV2INT, operands);
        default:
          throw new UnsupportedOperationException(name);
      }
    }
    throw new UnsupportedOperationException(name);
  }
}
