/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.parser.vhdl.mapper;

import org.zamia.instgraph.IGOperationUnary;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Mapping between IG unary operation identifiers
 * and standard operation identifiers that are supported.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class IgUnaryMapper {

  private static final Map<IGOperationUnary.UnaryOp, StandardOperation> MAP;

  private IgUnaryMapper() {
    /* Empty. */
  }

  static {

    MAP = new LinkedHashMap<>();

    MAP.put(IGOperationUnary.UnaryOp.ABS, StandardOperation.ABS);
    MAP.put(IGOperationUnary.UnaryOp.NEG, StandardOperation.MINUS);
    MAP.put(IGOperationUnary.UnaryOp.BUF, StandardOperation.ADD);
    MAP.put(IGOperationUnary.UnaryOp.NOT, StandardOperation.NOT);
  }

  /**
   * Checks whether operation is supported.
   *
   * <p>If operation is supported, then it's identifier is in this mapping.</p>
   *
   * @param operation IG operation identifier.
   *
   * @return {@code true} if operation is supported, {@code false} otherwise.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  private static boolean hasOperation(final IGOperationUnary.UnaryOp operation) {
    InvariantChecks.checkNotNull(operation);

    return MAP.containsKey(operation);
  }

  /**
   * Gets operation code of the specified operation.
   *
   * @param unary IG operation
   *
   * @return Operation identifier.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws UnsupportedOperationException when operation does not
   *         have corresponding operation code in this mapping.
   */
  public static StandardOperation getOperationCode(final IGOperationUnary unary) {
    InvariantChecks.checkNotNull(unary);

    final IGOperationUnary.UnaryOp unaryOp = unary.getUnaryOp();
    if (hasOperation(unaryOp)) {
      return MAP.get(unaryOp);
    } else {
      throw new UnsupportedOperationException(unaryOp.name());
    }
  }
}
