/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.assertion;

import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.util.LogMessage;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Named container of assertion groups.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class AssertionContainer extends AssertionVariableContainer {

  private final String name;

  private final AssertionContainer parent;

  private NodeVariable clock;

  private final Map<String, AssertionContainer> nested;

  private final Map<String, AssertionGroup> groups;

  /**
   * Constructs the object with the specified parameters.
   *
   * @param decStatement The statement which variable declarations are to be copied.
   * @param name Name of this object.
   * @param parent Parent container for this object.
   *
   * @throws IllegalArgumentException when name argument is {@code null}.
   */
  public AssertionContainer(
      final DecStatement decStatement,
      final String name,
      final AssertionContainer parent) {
    super();
    InvariantChecks.checkNotNull(name);

    this.name = name;
    this.parent = parent;
    this.nested = new LinkedHashMap<>();
    this.groups = new LinkedHashMap<>();

    if (decStatement != null) {
      declareVariables(decStatement.getDeclarationMap());
    }
  }

  /**
   * Constructs the object with the specified parameters.
   *
   * @param name The object name.
   * @param parent The parent node for this object.
   */
  AssertionContainer(final String name, final AssertionContainer parent) {
    this(null, name, parent);
  }

  /**
   * Constructs the object with the specified name.
   * @param name The object name.
   */
  AssertionContainer(final String name) {
    this(null, name, null);
  }

  /**
   * Returns name of this object.
   *
   * @return Object name.
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns parent container.
   *
   * @return Parent container for this object.
   */
  public AssertionContainer getParent() {
    return this.parent;
  }

  /**
   * Returns clock-like variable of the top-level container.
   *
   * @return Clock-like variable for this object.
   */
  public NodeVariable getClock() {
    return this.clock;
  }

  /**
   * Sets the specified variable as a clock-like for the top-level container.
   *
   * @param variable Clock-like variable to be set for this object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeRuntimeException when this is not top-level container.
   * @throws RetrascopeRuntimeException when this container already has it's clock.
   */
  public void setClock(final NodeVariable variable) {
    InvariantChecks.checkNotNull(variable);

    if (this.parent != null) {
      throw new RetrascopeRuntimeException("Can't set clock to not top container.");
    }
    if (this.clock != null) {
      throw RetrascopeRuntimeException.makeException(LogMessage.CANT_REASSIGN, variable);
    }
    this.clock = variable;
  }

  /**
   * Adds the specified assertion container object.
   *
   * @param container Assertion container to be added to this object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addContainer(final AssertionContainer container) {
    InvariantChecks.checkNotNull(container);

    this.nested.put(container.getName(), container);
  }

  /**
   * Returns mapping of names to nested assertion containers.
   *
   * @return Mapping of names into assertion containers that this object contains.
   */
  Map<String, AssertionContainer> getContainerMap() {
    return Collections.unmodifiableMap(this.nested);
  }

  /**
   * Adds the specified assertion group object.
   *
   * @param group Assertion group to be added.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addGroup(final AssertionGroup group) {
    InvariantChecks.checkNotNull(group);

    this.groups.put(group.getName(), group);
  }

  /**
   * Returns mapping of names to nested assertion groups.
   *
   * @return Assertion groups that a stored into this object.
   */
  public Map<String, AssertionGroup> getGroupMap() {
    return Collections.unmodifiableMap(this.groups);
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }
    if (!super.equals(obj)) {
      return false;
    }

    final AssertionContainer that = (AssertionContainer) obj;

    return getName() == null
        ? that.getName() == null
        && (getParent() == null
            ? that.getParent() == null
                && (getClock() == null
                    ? that.getClock() == null
                        && (nested == null
                            ? that.nested == null
                                && (groups == null
                                    ? that.groups == null : groups.equals(that.groups))
                            : nested.equals(that.nested))
                    : getClock().equals(that.getClock()))
            : getParent().equals(that.getParent()))
        : getName().equals(that.getName());
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (getName() != null ? getName().hashCode() : 0);
    result = 31 * result + (getParent() != null ? getParent().hashCode() : 0);
    result = 31 * result + (getClock() != null ? getClock().hashCode() : 0);
    result = 31 * result + (nested != null ? nested.hashCode() : 0);
    result = 31 * result + (groups != null ? groups.hashCode() : 0);
    return result;
  }
}
