/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.assertion;

import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.DecStatement;
import ru.ispras.retrascope.model.efsm.GuardedAction;

import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;

/**
 * Named group of independent assertions.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class AssertionGroup extends AssertionVariableContainer {

  private final String name;

  private final AssertionContainer parentContainer;

  private final List<Node> assertions;

  private final Node initValueState;

  private final List<GuardedAction> rstGuardedActions;

  /**
   * Constructs object with specified parameters.
   *
   * @param decStatement Statement which variable declarations are to be copied.
   * @param name Name of this object.
   * @param parent Parent container for this object.
   * @param initValueState The init value state node.
   * @param rstGuardedActions The sequence of resetting guarded actions.
   *
   * @throws IllegalArgumentException when any of the arguments except resetting
   *         guarded action and declaration statement is {@code null}.
   */
  public AssertionGroup(
      final DecStatement decStatement,
      final String name,
      final AssertionContainer parent,
      final Node initValueState,
      final List<GuardedAction> rstGuardedActions) {
    super();
    InvariantChecks.checkNotNull(name);
    InvariantChecks.checkNotNull(parent);
    InvariantChecks.checkNotNull(initValueState);

    this.name = name;
    this.parentContainer = parent;
    this.initValueState = initValueState;
    this.rstGuardedActions = rstGuardedActions;
    this.assertions = new LinkedList<>();

    if (decStatement != null) {
      declareVariables(decStatement.getDeclarationMap());
    }
  }

  /**
   * Constructs new object with the specified parameters.
   *
   * @param name The object name.
   * @param parent The parent container for this object.
   * @param initValueState The initial value state node.
   * @param rstGuardedActions The sequence of resetting guarded actions.
   */
  AssertionGroup(
      final String name,
      final AssertionContainer parent,
      final Node initValueState,
      final List<GuardedAction> rstGuardedActions) {

    this(null, name, parent, initValueState, rstGuardedActions);
  }

  /**
   * Constructs object with specified parameters.
   *
   * @param decStatement The statement which variable declarations are to be copied.
   * @param name Name of this object.
   * @param parent Parent container for this object.
   * @param node The init value state node.
   *
   * @throws IllegalArgumentException when any of the arguments is {@code null}.
   */
  public AssertionGroup(
      final DecStatement decStatement,
      final String name,
      final AssertionContainer parent,
      final Node node) {
    this(decStatement, name, parent, node, null);
  }

  /**
   * Returns object's name.
   *
   * @return Name of this object.
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns parent container.
   *
   * @return Parent container for this object.
   */
  public AssertionContainer getParent() {
    return this.parentContainer;
  }

  /**
   * Returns node of the EFSM state that is compliant with initial values of state-like variables.
   * @return The EFSM state node that is compliant with initial values of state-like variables.
   */
  public Node getInitValueState() {
    return this.initValueState;
  }

  /**
   * Returns resetting guarded action.
   * @return The resetting guarded action.
   */
  public List<GuardedAction> getRstGuardedAction() {
    return this.rstGuardedActions;
  }

  /**
   * Adds the specified expression that uses variables
   * that are already declared at the parent container.
   *
   * @param assertion Assertion to be added to this object.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws RetrascopeRuntimeException when the specified expression
   *         contains at least one variable that is undeclared
   *         at the parent container.
   */
  public void addAssertion(final Node assertion) {
    InvariantChecks.checkNotNull(assertion);

    whetherVarsDeclared(assertion);

    this.assertions.add(assertion);
  }

  private void whetherVarsDeclared(final Node node) {
    InvariantChecks.checkNotNull(node);

    Collection<NodeVariable> variables = ExprUtils.getVariables(node);

    AssertionVariableContainer container = this;

    do {
      variables =
          CollectionUtils.complementSets(
              new LinkedHashSet<>(variables),
              new LinkedHashSet<>(container.getVariables()));

      if (container instanceof AssertionGroup) {

        container = ((AssertionGroup)container).getParent();

      } else {

        container = ((AssertionContainer)container).getParent();
      }
    } while (container != null && !variables.isEmpty());

    if (!variables.isEmpty()) {
      throw new RetrascopeRuntimeException(
          String.format(
              "Cannot add %s assertion: %s variables undeclared.",
              node.toString(),
              variables.toString()));
    }
  }

  /**
   * Returns a collection of assertions.
   *
   * @return Assertions of this object.
   */
  public List<Node> getAssertions() {
    return Collections.unmodifiableList(this.assertions);
  }

  /**
   * Returns the number of related assertions.
   * @return The number of related assertions.
   */
  public int getAssertionNum() {
    return getAssertions().size();
  }

  @Override
  public boolean equals(final Object obj) {

    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }
    if (!super.equals(obj)) {
      return false;
    }

    final AssertionGroup that = (AssertionGroup) obj;

    return getName() != null
        ? getName().equals(that.getName())
        : that.getName() == null
            && (parentContainer == null
                ? that.parentContainer == null
                    && getAssertions().equals(that.getAssertions())
                : parentContainer.equals(that.parentContainer));
  }

  @Override
  public int hashCode() {
    int result = super.hashCode();
    result = 31 * result + (getName() != null ? getName().hashCode() : 0);
    result = 31 * result + (parentContainer != null ? parentContainer.hashCode() : 0);
    result = 31 * result + getAssertions().hashCode();
    result = 31 * result + (getInitValueState() != null ? getInitValueState().hashCode() : 0);
    result = 31 * result + (getRstGuardedAction() != null ? getRstGuardedAction().hashCode() : 0);
    return result;
  }
}
