/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.result.cfginterface;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.Entity;
import ru.ispras.retrascope.basis.EntityType;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.DeclarationContainer;

import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Control flow graph model interface.
 *
 * <p>Stores the information about hardware modules interfaces.
 * Particularly, it contains module names and their input and output signals.</p>
 *
 * @see ru.ispras.retrascope.model.cfg.CfgModel
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgInterface extends Entity {

  /**
   * The entity type.
   */
  public static final EntityType TYPE = EntityType.get("cfg-iface");

  /* Mapping of module name to it's signal interface. */
  private final Map<String, DeclarationContainer> map = new LinkedHashMap<>();

  /**
   * Default constructor.
   */
  public CfgInterface() {
    super(CfgInterface.TYPE);
  }

  /**
   * Adds to interface storage the specified module's name and it's signal interface.
   *
   * @param moduleName Name of module which interface is added.
   * @param signalInterface Signal interface that is added.
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  public void addInterface(
      final String moduleName,
      final DeclarationContainer signalInterface) {

    InvariantChecks.checkNotNull(moduleName);
    InvariantChecks.checkNotNull(signalInterface);

    this.map.put(moduleName, signalInterface);
  }

  /**
   * Returns mapping of modules names and their signal interfaces.
   *
   * @return Mapping of modules names into their signal interfaces.
   */
  public Map<String, DeclarationContainer> getMapping() {
    return Collections.unmodifiableMap(this.map);
  }

  @Override
  public void add(final Entity entity) {

    if (entity instanceof CfgInterface) {

      this.map.putAll(((CfgInterface) entity).getMapping());

    } else {
      throw new RetrascopeRuntimeException(
          String.format(
              "Cannot add %s entity to %s entity.",
              entity.getClass().getName(),
              getClass().getName()));
    }
  }
}
