/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test;

import org.apache.commons.collections15.BidiMap;
import org.apache.commons.collections15.bidimap.DualHashBidiMap;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.basis.EventList;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;

/**
 * The representation of a test sequence. A sequence is consisted of a list of input vectors as
 * well as list of attendant events (for example clock or reset signal change).
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 *
 * @see Vector
 * @see EventList
 */
@XmlType(propOrder = {"vectors"}, name = "sequence")
@XmlRootElement
public final class Sequence implements Iterator<Vector> {

  private static final String LEFT_STRING_BOUND = "{";
  private static final String VECTOR_STRING_HEADER = "vector: ";
  private static final String VECTOR_STRING_DELIMITER = "; ";
  private static final String RIGHT_STRING_BOUND = "}";

  //@XmlElement(name = "vector")
  //private List<Vector> vectors;

  private Integer index;

  private final BidiMap<Vector, Integer> vectorMap;
  private final List<Integer> hashes;

  /**
   * Constructs an empty sequence.
   */
  public Sequence() {
    this.vectorMap = new DualHashBidiMap<>();
    this.hashes = new LinkedList<>();
    initIterator();
  }

  /**
   * Adds the specified vector to the end of this sequence.
   *
   * @param vector Vector to be added.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addVector(final Vector vector) {
    InvariantChecks.checkNotNull(vector);

    if (!this.vectorMap.containsKey(vector)) {
      this.vectorMap.put(vector, System.identityHashCode(vector));
    }
    this.hashes.add(this.vectorMap.get(vector));
  }

  /**
   * Adds the specified vectors to the end of this sequence. An original order is preserved.
   *
   * @param vectors a list of the vectors to be added
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public void addVectors(final List<Vector> vectors) {
    InvariantChecks.checkNotNull(vectors);

    for (final Vector vector : vectors) {
      addVector(vector);
    }
  }

  /**
   * Returns an amount of vectors in this sequence.
   *
   * @return the amount of vectors in this sequence
   */
  public int getLength() {
    return this.hashes.size();
  }

  /**
   * Prepares internal iterator for vectors.
   */
  public void initIterator() {
    this.index = 0;
  }

  @Override
  public boolean hasNext() {
    return this.hashes.size() > this.index;
  }

  @Override
  public Vector next() {
    final Vector next = this.vectorMap.getKey(this.hashes.get(this.index));
    this.index++;
    return next;
  }

  @Override
  public void remove() {
    /* Empty. */
  }

  @Override
  public String toString() {
    final StringBuilder stringBuilder = new StringBuilder(Sequence.LEFT_STRING_BOUND);
    final int preVectorsStringBuilderLength = stringBuilder.length();

    for (final Vector vector : this.vectorMap.keySet()) {
      if (stringBuilder.length() > preVectorsStringBuilderLength) {
        stringBuilder.append(Sequence.VECTOR_STRING_DELIMITER);
      }

      stringBuilder.append(Sequence.VECTOR_STRING_HEADER);
      stringBuilder.append(vector);
    }

    stringBuilder.append(Sequence.RIGHT_STRING_BOUND);

    return stringBuilder.toString();
  }

  @Override
  public boolean equals(final Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj == null || getClass() != obj.getClass()) {
      return false;
    }

    final Sequence sequence = (Sequence) obj;

    if (getLength() != sequence.getLength()) {
      return false;
    }

    final int thisIndex = this.index;
    final int sequenceIndex = sequence.index;
    initIterator();
    sequence.initIterator();

    while (hasNext() && sequence.hasNext()) {
      if (!next().equals(sequence.next())) {
        return false;
      }
    }
    this.index = thisIndex;
    sequence.index = sequenceIndex;

    return true;
  }

  @Override
  public int hashCode() {
    int result = 0;
    final int index = this.index;
    initIterator();
    while (hasNext()) {
      result = 31 * result + next().hashCode();
    }
    this.index = index;
    return result;
  }
}
