/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.result.test.efsm;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.model.efsm.Efsm;
import ru.ispras.retrascope.model.efsm.EfsmTransition;
import ru.ispras.retrascope.result.test.TestCoverage;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

/**
 * The representation of a transition-based coverage of an EFSM. In such metrics, an EFSM is
 * considered as fully covered if and only if all its transitions are covered.
 *
 * @author <a href="mailto:i.melnichenko@deltasolutions.ru">Igor Melnichenko</a>
 */
public class EfsmTransitionCoverage extends TestCoverage {
  private int defaultPrecision;

  private final Set<EfsmTransition> coveredTransitions;
  private final Set<EfsmTransition> uncoveredTransitions;

  private static final String TRANSITION_NOT_FOUND =
      "The specified transition hasn't been found among both covered and uncovered transitions";

  /**
   * Constructs an empty coverage.
   */
  public EfsmTransitionCoverage() {
    this.uncoveredTransitions = new HashSet<>();
    this.coveredTransitions = new HashSet<>();

    this.computeDefaultPrecision();
  }

  /**
   * Constructs a new coverage representation for the specified EFSM.
   *
   * <p>Initially, all transitions of the EFSM are considered as uncovered.</p>
   *
   * @param efsm the EFSM for coverage monitoring
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public EfsmTransitionCoverage(Efsm efsm) {
    InvariantChecks.checkNotNull(efsm);

    this.uncoveredTransitions = new HashSet<>(efsm.getTransitions());
    this.coveredTransitions = new HashSet<>(this.uncoveredTransitions.size());

    this.computeDefaultPrecision();
  }

  /**
   * Constructs a new coverage as a copy of the specified coverage.
   *
   * @param coverage the coverage to be copied
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public EfsmTransitionCoverage(EfsmTransitionCoverage coverage) {
    InvariantChecks.checkNotNull(coverage);

    this.uncoveredTransitions = new HashSet<>(coverage.uncoveredTransitions);
    this.coveredTransitions = new HashSet<>(coverage.coveredTransitions);

    this.defaultPrecision = coverage.defaultPrecision;
  }

  /**
   * Marks the specified transition as covered.
   *
   * @param coveredTransition the transition to be marked as covered
   *
   * @throws IllegalArgumentException if the specified transition is not found
   *         among the transitions extracted from the EFSM during initialisation
   */
  public void markTransitionAsCovered(EfsmTransition coveredTransition) {
    if (this.uncoveredTransitions.contains(coveredTransition)) {
      this.coveredTransitions.add(coveredTransition);
      this.uncoveredTransitions.remove(coveredTransition);
    } else {
      if (!this.coveredTransitions.contains(coveredTransition)) {
        throw new IllegalArgumentException(TRANSITION_NOT_FOUND);
      }
    }
  }

  /**
   * Checks if the specified transition is covered.
   *
   * @param transition the transition to be tested
   *
   * @return {@code true} if the transition is covered, otherwise returns {@code false}
   *
   * @throws IllegalArgumentException if the specified transition is not found
   *         among the transitions extracted from the EFSM during initialisation
   */
  public boolean isCovered(EfsmTransition transition) {
    if (this.coveredTransitions.contains(transition)) {
      return true;
    } else {
      if (this.uncoveredTransitions.contains(transition)) {
        return false;
      } else {
        throw new IllegalArgumentException(TRANSITION_NOT_FOUND);
      }
    }
  }

  /**
   * Returns a percentage representation of this test coverage computed with the default precision.
   *
   * <p>The returned value is computed as a ratio of amount of covered transitions to total amount
   * of transitions. It is guaranteed that the value is precise enough to reflect the smallest
   * possible change in a coverage information.</p>
   *
   * @return the percentage of this coverage
   */
  @Override
  public BigDecimal getPercentage() {
    return this.getPercentage(this.defaultPrecision);
  }

  /**
   * Returns a percentage representation of this test coverage computed with the specified
   * precision.
   *
   * <p>The returned value is computed as a ratio of amount of covered transitions to total
   * amount of transitions.</p>
   *
   * @param precision a number of decimal places in the result
   *
   * @return a percentage of this coverage
   */
  @Override
  public BigDecimal getPercentage(int precision) {
    BigDecimal result;

    if (this.uncoveredTransitions.isEmpty()) {
      result = BigDecimal.valueOf(100);
    } else {
      if (this.coveredTransitions.isEmpty()) {
        result = BigDecimal.ZERO;
      } else {
        result = new BigDecimal(this.coveredTransitions.size());
        BigDecimal totalTransitionsAmount = new BigDecimal(this.coveredTransitions.size()
            + this.uncoveredTransitions.size());
        result = result.divide(totalTransitionsAmount, precision + 3, RoundingMode.HALF_UP);
        result = result.multiply(BigDecimal.valueOf(100));
      }
    }

    return result.setScale(precision, RoundingMode.HALF_UP);
  }

  /**
   * Tests if this coverage is full, that is, there are no uncovered transitions.
   *
   * @return {@code true} if this coverage is full. Otherwise returns {@code false}
   */
  @Override
  public boolean isFull() {
    return this.uncoveredTransitions.isEmpty();
  }

  /**
   * Returns all covered transitions in this coverage.
   *
   * @return a set of covered transitions
   */
  public Set<EfsmTransition> getCoveredTransitions() {
    return Collections.unmodifiableSet(this.coveredTransitions);
  }

  /**
   * Returns all uncovered transitions in this coverage.
   *
   * @return a set of uncovered transitions
   */
  public Set<EfsmTransition> getUncoveredTransitions() {
    return Collections.unmodifiableSet(this.uncoveredTransitions);
  }

  @Override
  public void reset() {
    this.uncoveredTransitions.addAll(this.coveredTransitions);
    this.coveredTransitions.clear();
  }

  /**
   * Includes the specified coverage to this coverage.
   *
   * @param coverage the coverage to be included
   */
  public void include(EfsmTransitionCoverage coverage) {
    Set<EfsmTransition> newUncoveredTransitions = new HashSet<>(coverage.uncoveredTransitions);
    newUncoveredTransitions.removeAll(this.coveredTransitions);
    this.coveredTransitions.addAll(coverage.coveredTransitions);
    this.uncoveredTransitions.removeAll(coverage.coveredTransitions);
    this.uncoveredTransitions.addAll(newUncoveredTransitions);
  }

  private void computeDefaultPrecision() {
    int transitionsAmount = this.uncoveredTransitions.size();

    if (transitionsAmount > 100) {
      this.defaultPrecision = (int) Math.log10(transitionsAmount) - 1;
    } else {
      this.defaultPrecision = 0;
    }
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + this.coveredTransitions.hashCode();
    result = prime * result + this.uncoveredTransitions.hashCode();

    return result;
  }

  /**
   * Compares the specified object with this coverage for equality.
   *
   * <p>The object is considered as equal to this coverage if and only if
   * it is also an EFSM transition coverage and it contains
   * the same covered and covered transitions as this coverage does, or more formally:</p>
   * <p>
   * <ul>
   *   <li>it is also an instance of the {@code EfsmTransitionCoverage} class or its subclass;
   *   <li>it has the same amount of covered and uncovered transitions as this coverage does;
   *   <li>for each its covered and uncovered transition there is an equal one in this coverage.
   * </ul>
   * </p>
   *
   * @param obj the object to be compared for equality with this coverage
   *
   * @return {@code true} if the specified object is equal to this coverage. Otherwise returns
   * {@code false}
   */
  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }

    if ((obj == null) || !(obj instanceof EfsmTransitionCoverage)) {
      return false;
    }

    EfsmTransitionCoverage coverageToCompare = (EfsmTransitionCoverage) obj;

    return this.coveredTransitions.equals(coverageToCompare.coveredTransitions)
        && this.uncoveredTransitions.equals(coverageToCompare.uncoveredTransitions);
  }

  @Override
  public String toString() {
    int totalAmount = this.coveredTransitions.size() + this.uncoveredTransitions.size();

    StringBuilder stringBuilder = new StringBuilder().append(this.getPercentage()).append("% (");
    stringBuilder.append(this.coveredTransitions.size()).append(" of ").append(totalAmount);

    if (totalAmount == 1) {
      stringBuilder.append(" transition)");
    } else {
      stringBuilder.append(" transitions)");
    }

    return stringBuilder.toString();
  }
}
