/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorAlgorithm;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.AssignmentContainer;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.RangedVariableUtils;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

/**
 * Utility methods for models transformation.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class ModelTransformUtils {

  private ModelTransformUtils() {
    /* Empty. */
  }

  /**
   * Performs backward substitution on the specified nodes to the specified expression.
   *
   * <p>The substitution process moves in a backward direction (from last node of the list
   * to the first). The result is an expression where all the target variables that are
   * defined at least one {@link AssignmentContainer} node from the specified collection
   * will be substituted by the value expressions that are assigned to them.</p>
   *
   * @param blocks Blocks which should be used as a source of substitution.
   * @param node Expression to be transformed.
   *
   * @return Modified expression.
   */
  public static Node substituteBackward(
      final List<? extends AssignmentContainer> blocks,
      final Node node) {

    Node nodeCopy = node.deepCopy();

    final List<AssignmentContainer> concurrent = new LinkedList<>();
    final List<AssignmentContainer> blocking = new LinkedList<>();

    for (final AssignmentContainer block : blocks) {

      if (block.isConcurrent()) {
        concurrent.add(block);
      } else {
        blocking.add(block);
      }
    }

    final List<NodeBinding.BoundVariable> boundVariables = new LinkedList<>();

    for (final AssignmentContainer container : concurrent) {

      final List<Assignment> assignments = container.getAssignments();

      for (final Assignment assignment : assignments) {

        boundVariables.add(
            NodeBinding.bindVariable(
                assignment.getVariable(),
                assignment.getTarget().isRanged()
                    ? RangedVariableUtils.transformNodeWithRanged(
                        assignment.getTarget(),
                        assignment.getValue())
                    : assignment.getValue()));
      }
    }
    nodeCopy = Transformer.substituteAllBindings(new NodeBinding(nodeCopy, boundVariables));

    /* Elaborate "sequential" blocks in backward direction. */
    for (int i = blocking.size() - 1; i >= 0; i--) {

      final AssignmentContainer block = blocking.get(i);
      final List<Assignment> assignments = block.getAssignments();

      /* Elaborate assignments in backward direction. */
      for (int j = assignments.size() - 1; j >= 0; j--) {

        final Assignment assignment = assignments.get(j);
        final RangedVariable target = assignment.getTarget();
        final NodeVariable targetVariable = target.getVariable();
        final Node value = assignment.getValue();

        final Collection<NodeVariable> variables = ExprUtils.getVariables(nodeCopy);

        if (variables.contains(targetVariable)) {

          nodeCopy = substituteNode(nodeCopy, target, value);
        }
      }
    }

    /* Call the reducer to simplify the substitution result. */
    return Transformer.reduce(nodeCopy);
  }

  private static Node substituteNode(
      final Node node,
      final RangedVariable target,
      final Node value) {

    return target.isRanged()
        ? ModelTransformUtils.substituteRanged(node, target, value)
        : Transformer.substitute(node, target.getVariableName(), value);
  }

  private static Node substituteRanged(
      final Node input,
      final RangedVariable target,
      final Node value) {

    Node result;

    final NodeVariable targetVariable = target.getVariable();

    switch (targetVariable.getDataTypeId()) {

      case BIT_VECTOR:
        result = substituteBitVectors(input, target, value);
        break;

      case MAP:
        result = substituteArrays(input, target, value);
        break;

      default:
        throw RetrascopeRuntimeException.makeException(
            LogMessage.UNSUPPORTED_TYPE,
            targetVariable.getDataTypeId());
    }

    return result;
  }

  private static Node substituteArrays(
      final Node input,
      final RangedVariable target,
      final Node value) {

    /* The array case: substitute x by (STORE x i expr). */
    final NodeVariable targetVariable = target.getVariable();
    final String varName = targetVariable.getName();
    final Range targetRange = target.getRange();

    /* There is no difference here what range filed to select, because they're equal. */
    final Node index = targetRange.getYoung();

    final NodeOperation store =
        new NodeOperation(StandardOperation.STORE, targetVariable, index, value);

    return Transformer.substitute(input, varName, store);
  }

  private static Node substituteBitVectors(
      final Node input,
      final RangedVariable target,
      final Node value) {

    final NodeVariable targetVariable = target.getVariable();
    final Range targetRange = target.getRange();
    final Node targetOld = targetRange.getOld();
    final Node targetYoung = targetRange.getYoung();

    /* Select operation identifiers & constants depending on operand types. */
    final Enum<?> lsOp = choose(targetRange, StandardOperation.LESS, StandardOperation.BVULT);
    final Enum<?> gtOp = choose(targetRange, StandardOperation.GREATER, StandardOperation.BVUGT);
    final Enum<?> addOp = choose(targetRange, StandardOperation.ADD, StandardOperation.BVADD);
    final Enum<?> subOp = choose(targetRange, StandardOperation.SUB, StandardOperation.BVSUB);

    final NodeValue zero = chooseNodeValue(targetRange, 0);
    final NodeValue one = chooseNodeValue(targetRange, 1);
    final NodeValue max = createMaxValue(target);

    /* Create expressions that will substitute target's variable entries. */

    /* young == 0 && old == max */
    final Node fullRange =
        Node.AND(NodeUtils.getEquation(targetYoung, zero), NodeUtils.getEquation(targetOld, max));

    /* young == 0 && old < max */
    final Node youngRange =
        Node.AND(NodeUtils.getEquation(targetYoung, zero), new NodeOperation(lsOp, targetOld, max));

    /* young > 0 && old == max */
    final Node oldRange =
        Node.AND(new NodeOperation(gtOp, targetYoung, zero), NodeUtils.getEquation(targetOld, max));

    /* young > 0 && old < max */

    /* Check whether above conditions are satisfiable (using SAT solver). */
    final boolean satFullRange = SolverUtils.isSat(fullRange);
    final boolean satYoungRange = SolverUtils.isSat(youngRange);
    final boolean satOldRange = SolverUtils.isSat(oldRange);

    /* The formulas to which the bit vector ranged variable will be transformed
     * when the corresponding condition (and only this condition) is SAT. */

    /* BVEXTRACT parameters. */
    final Node maxToInt = TransformerUtils.toInt(max);
    final Node oldPlusOne = TransformerUtils.toInt(new NodeOperation(addOp, targetOld, one));
    final Node youngMinusOne = TransformerUtils.toInt(new NodeOperation(subOp, targetYoung, one));
    final Node zeroInt = TransformerUtils.toInt(zero);


    /* The case when no condition above is true. */
    final Node middleTransform = new NodeOperation(
        StandardOperation.BVCONCAT,
        new NodeOperation(StandardOperation.BVEXTRACT, maxToInt, oldPlusOne, targetVariable),
        value,
        new NodeOperation(StandardOperation.BVEXTRACT, youngMinusOne, zeroInt, targetVariable));

    final Node youngTransform = new NodeOperation(
        StandardOperation.BVCONCAT,
        new NodeOperation(StandardOperation.BVEXTRACT, maxToInt, oldPlusOne, targetVariable),
        value);

    final Node oldTransform = new NodeOperation(
        StandardOperation.BVCONCAT,
        value,
        new NodeOperation(StandardOperation.BVEXTRACT, youngMinusOne, zeroInt, targetVariable));

    Node transform;

    /* Elaborate all possible variants. */

    if (satFullRange) {

      transform = value;

    } else if (satYoungRange) {

      transform = youngTransform;

    } else if (satOldRange) {

      transform = oldTransform;

    } else {
      transform = middleTransform;
    }

    return Transformer.substitute(input, targetVariable.getName(), transform);
  }

  private static NodeValue createMaxValue(final RangedVariable rangedVariable) {

    NodeValue result;

    final Range range = rangedVariable.getRange();
    final Node old = range.getOld();
    final Node young = range.getYoung();
    final NodeVariable variable = rangedVariable.getVariable();

    if (ExprUtils.isType(DataType.INTEGER, old, young)) {

      result = NodeValue.newInteger(variable.getDataType().getSize() - 1);

    } else if (ExprUtils.isType(DataTypeId.BIT_VECTOR, old, young)) {

      final BitVector maxVector = BitVector.newEmpty(young.getDataType().getSize());
      BitVectorAlgorithm.fill(maxVector, (byte) 1);

      result = NodeValue.newBitVector(maxVector);

    } else {

      final DataType youngDataType = young.getDataType();
      final DataType oldDataType = old.getDataType();

      throw new RetrascopeRuntimeException(unsupportedTypesMsg(youngDataType, oldDataType));
    }
    return result;
  }

  private static NodeValue chooseNodeValue(
      final Range range,
      final int value) {

    NodeValue result;

    final Node old = range.getOld();
    final Node young = range.getYoung();
    final DataType oldDataType = old.getDataType();
    final DataType youngDataType = young.getDataType();
    final int youngSize = youngDataType.getSize();
    final int oldSize = oldDataType.getSize();
    final int vectorSize = youngSize > oldSize ? youngSize : oldSize;

    if (ExprUtils.isType(DataType.INTEGER, young, old)) {

      result = NodeValue.newInteger(value);

    } else if (ExprUtils.isType(DataTypeId.BIT_VECTOR, young, old)) {

      result = NodeValue.newBitVector(BitVector.valueOf(value, vectorSize));

    } else {

      throw new RetrascopeRuntimeException(unsupportedTypesMsg(youngDataType, oldDataType));
    }

    return result;
  }

  private static Enum<?> choose(
      final Range range,
      final Enum<?> standard,
      final Enum<?> bitVector) {

    Enum<?> result;

    final Node old = range.getOld();
    final Node young = range.getYoung();

    if (ExprUtils.isType(DataType.INTEGER, young, old)) {

      result = standard;

    } else if (ExprUtils.isType(DataTypeId.BIT_VECTOR, young, old)) {

      result = bitVector;
    } else {

      final DataType youngDataType = young.getDataType();
      final DataType oldDataType = old.getDataType();

      throw new RetrascopeRuntimeException(unsupportedTypesMsg(youngDataType, oldDataType));
    }

    return result;
  }

  private static String unsupportedTypesMsg(final DataType youngType, final DataType oldType) {

    return String.format("%s young=%s, old=%s.", LogMessage.UNSUPPORTED_TYPE, youngType, oldType);
  }
}
