/*
 * Copyright 2017 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import org.apache.commons.collections15.BidiMap;
import org.apache.commons.collections15.bidimap.DualHashBidiMap;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.logic.Clause;
import ru.ispras.fortress.logic.NormalForm;
import ru.ispras.fortress.logic.Orthogonalizer;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

/**
 * Component that performs an orthogonalization of nodes.
 * <p>Orthogonal nodes are mutually incompatible.</p>
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class NodeOrthogonalizer {

  private NodeOrthogonalizer() {
    /* Empty. */
  }

  /**
   * Orthogonalizes the specified collections of nodes into orthogonal nodes.
   * @param preStates Collections of nodes to be processed.
   * @return The collection of orthogonal nodes.
   */
  public static List<Node> orthogonalizeNodes(final Collection<Collection<Node>> preStates) {

    final List<Node> states = new LinkedList<>();

    if (preStates.isEmpty() || preStates.size() == 1) {

      states.add(SolverUtils.TRUE);

    } else {

      /* Bi-directional mapping for literals and expressions. */
      final BidiMap<Integer, Node> literalMap = new DualHashBidiMap<>();

      /* Transform pre-states into clauses. */
      final List<Clause> clauses = createClauses(preStates, literalMap);

      /* Make clauses orthogonal. */
      final NormalForm ortForm = orthogonalize(clauses, getConflicts(literalMap));

      /* Transform orthogonal clauses into EFSM states. */
      createNodes(states, literalMap, ortForm.getClauses());
    }
    return states;
  }

  private static List<Clause> createClauses(
      final Collection<Collection<Node>> preStates,
      final BidiMap<Integer, Node> literalMap) {

    final List<Clause> clauses = new LinkedList<>();

    for (final Collection<Node> preState : preStates) {

      final Clause.Builder clauseBuilder = new Clause.Builder();

      for (Node node : preState) {

        boolean negated = false;

        if (ExprUtils.isOperation(node, StandardOperation.NOT)) {

          node = ((NodeOperation) node).getOperand(0);
          negated = true;
        }

        int literal;

        if (literalMap.containsValue(node)) {

          literal = literalMap.getKey(node);

        } else {

          literal = literalMap.keySet().size();
          literalMap.put(literal, node);
        }
        clauseBuilder.add(literal, negated);
      }
      clauses.add(clauseBuilder.build());
    }
    return clauses;
  }

  private static NormalForm orthogonalize(
      final List<Clause> clauses,
      final Set<Clause> conflicts) {

    return Orthogonalizer.orthogonalize(new NormalForm(NormalForm.Type.DNF, clauses), conflicts);
  }

  private static void createNodes(
      final Collection<Node> states,
      final BidiMap<Integer, Node> literalMap,
      final Collection<Clause> ortClauses) {

    for (final Clause clause : ortClauses) {

      final Collection<Node> stateParts = new LinkedHashSet<>();
      final Set<Integer> clauseLiterals = clause.getVars();

      for (final Integer literal : clauseLiterals) {

        Node node = literalMap.get(literal);

        if (clause.getSign(literal)) {
          node = Node.NOT(node);
        }
        stateParts.add(node);
      }
      final Node state = NodeUtils.getConjunction(stateParts);
      states.add(state);
    }
  }

  private static Set<Clause> getConflicts(final BidiMap<Integer, Node> literalMap) {

    final Set<Clause> conflicts = new LinkedHashSet<>();
    final Set<Integer> literalSet = literalMap.keySet();
    final Integer[] variables = literalSet.toArray(new Integer[literalSet.size()]);

    for (int i = 0; i < variables.length - 1; i++) {

      for (int j = i + 1; j < variables.length; j++) {

        final int variableI = variables[i];
        final int variableJ = variables[j];

        final Node nodeI = literalMap.get(variableI);
        final Node nodeJ = literalMap.get(variableJ);

        InvariantChecks.checkFalse(nodeI.equals(nodeJ));

        for (final boolean signI : new boolean[] {false, true}) {

          final Node signedNodeI = signI ? Node.NOT(nodeI) : nodeI;

          for (final boolean signJ : new boolean[] {false, true}) {

            final Node signedNodeJ = signJ ? Node.NOT(nodeJ) : nodeJ;

            if (!SolverUtils.areCompatible(signedNodeI, signedNodeJ)) {

              final Clause.Builder conflictBuilder = new Clause.Builder();

              conflictBuilder.add(variableI, signI);
              conflictBuilder.add(variableJ, signJ);

              conflicts.add(conflictBuilder.build());
            }
          }
        }
      }
    }
    return conflicts;
  }
}
