/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorMath;
import ru.ispras.fortress.expression.ExprTreeVisitorDefault;
import ru.ispras.fortress.expression.ExprTreeWalker;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.util.CollectionUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Utility methods for {@code Node} objects.
 *
 * @see Node
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class NodeUtils {

  private NodeUtils() {
    /* Empty. */
  }

  /**
   * Returns an equation {@code (nodes[0] == ... == nodes[n - 1])} expression
   * of the specified nodes.
   *
   * @param nodes Expressions to be combined.
   *
   * @return An equation of the specified nodes.
   *
   * @throws IllegalArgumentException when argument is either {@code null} or empty.
   */
  public static Node getEquation(final Node ... nodes) {
    InvariantChecks.checkNotNull(nodes);
    InvariantChecks.checkNotEmpty(nodes);

    return new NodeOperation(StandardOperation.EQ, nodes);
  }

  /**
   * Returns a conjunction of the specified nodes.
   * @param nodes Nodes for conjunction.
   * @return Conjunction of specified nodes.
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static Node getConjunction(final Node ... nodes) {
    InvariantChecks.checkNotNull(nodes);

    final Set<Node> nodeSet = new LinkedHashSet<>(Arrays.asList(nodes));
    return getConjunction(nodeSet);
  }

  /**
   * Returns a conjunction of non-empty collection of nodes.
   * <p>Repeated nodes will not appear in the result.</p>
   * @param nodes Nodes for conjunction.
   * @return Conjunction of nodes that are in the specified collection.
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalArgumentException when argument is empty.
   */
  public static Node getConjunction(final Collection<Node> nodes) {
    InvariantChecks.checkNotNull(nodes);
    InvariantChecks.checkNotEmpty(nodes);

    /* Do not put same nodes in conjunction. */
    final Set<Node> nodeSet = new LinkedHashSet<>();

    for (final Node node : nodes) {

      if (ExprUtils.isOperation(node, StandardOperation.AND)) {

        nodeSet.addAll(splitConjunction(node));

      } else {
        nodeSet.add(node);
      }
    }

    return ExprUtils.getConjunction(nodeSet.toArray(new Node[nodeSet.size()]));
  }

  /**
   * Returns a disjunction of non-empty collection of nodes.
   * @param nodes Nodes for disjunction.
   * @return Disjunction of nodes that are in the specified collection.
   * @throws IllegalArgumentException when argument is {@code null}.
   * @throws IllegalArgumentException when argument is empty.
   */
  public static Node getDisjunction(final Collection<Node> nodes) {
    InvariantChecks.checkNotNull(nodes);
    InvariantChecks.checkNotEmpty(nodes);

    return ExprUtils.getDisjunction(nodes.toArray(new Node[nodes.size()]));
  }

  /**
   * Returns a collection of names for variables from the specified collection.
   *
   * @param nodes Collection of nodes.
   *
   * @return Collection of names for variable nodes from the specified collection.
   *
   * @throws IllegalArgumentException when argument is {@code null}.
   */
  public static Collection<String> getVariableNames(final Collection<? extends Node> nodes) {
    InvariantChecks.checkNotNull(nodes);

    final Collection<String> names = new LinkedList<>();

    for (final Node node : nodes) {

      if (ExprUtils.isKind(Node.Kind.VARIABLE, node)) {
        names.add(((NodeVariable)node).getName());
      }
    }

    return names;
  }

  /**
   * Casts the specified node to the specified data type.
   *
   * @param node Expression to be casted.
   * @param dataType Type to which the expression to be casted to.
   *
   * @return An expression which is equivalent to the specified node and has the specified type.
   *
   * @throws IllegalArgumentException when any of arguments is {@code null}.
   */
  static Node castToDataType(final Node node, final DataType dataType) {
    InvariantChecks.checkNotNull(node);
    InvariantChecks.checkNotNull(dataType);

    if (node.getDataType() == dataType) {
      return node;
    }

    switch (dataType.getTypeId()) {

      case BIT_VECTOR:
        return castToBitVector(node, dataType.getSize());

      case LOGIC_BOOLEAN:
        return castToBool(node);

      case LOGIC_INTEGER:
        return castToInteger(node);

      default:
        unsupported(dataType.getTypeId());
    }
    return null;
  }

  private static Node castToInteger(final Node node) {

    final Node zeroValue = NodeValue.newInteger(0);
    final Node oneValue = NodeValue.newInteger(1);

    switch (node.getDataTypeId()) {

      case BIT_VECTOR:
        final int bvSize = node.getDataType().getSize();
        final BitVector oneBitVector = BitVector.valueOf(1, bvSize);
        Node intExpression = zeroValue;
        BitVector current = oneBitVector;

        for (int i = 1; i < Math.pow(2, bvSize); i++) {

          final Node newInt = NodeValue.newInteger(i);
          intExpression =
              ifThenElse(getEquation(node, NodeValue.newBitVector(current)), newInt, intExpression);
          current = BitVectorMath.add(current, oneBitVector);
        }
        return intExpression;

      case LOGIC_BOOLEAN:
        return ifThenElse(getEquation(node, SolverUtils.FALSE), zeroValue, oneValue);

      default:
        unsupported(node);
    }
    return null;
  }

  private static Node castToBool(final Node node) {

    switch (node.getDataTypeId()) {

      case BIT_VECTOR:
        final Node zeroBitVector =
            NodeValue.newBitVector(BitVector.valueOf(0, node.getDataType().getSize()));
        return ifThenElse(getEquation(node, zeroBitVector), SolverUtils.FALSE, SolverUtils.TRUE);

      case LOGIC_INTEGER:
        final Node zeroInt = NodeValue.newInteger(0);
        return ifThenElse(getEquation(node, zeroInt), SolverUtils.FALSE, SolverUtils.TRUE);

      default:
        unsupported(node.getDataTypeId());
    }
    return null;
  }

  private static Node castToBitVector(
      final Node node,
      final int size) {

    final Node trueValue = NodeValue.newBitVector(BitVector.TRUE);
    final Node falseValue = NodeValue.newBitVector(BitVector.FALSE);

    switch (node.getDataTypeId()) {

      case LOGIC_BOOLEAN:
        return ifThenElse(getEquation(node, SolverUtils.FALSE), falseValue, trueValue);

      case LOGIC_INTEGER:
        Node bitVector = NodeValue.newBitVector(BitVector.valueOf(0, size));
        for (int i = 1; i < Math.pow(2, size); i++) {
          final Node intValue = NodeValue.newInteger(i);
          final Node bvValue = NodeValue.newBitVector(BitVector.valueOf(i, size));
          bitVector = ifThenElse(getEquation(node, intValue), bvValue, bitVector);
        }
        return bitVector;

      default:
        unsupported(node.getDataTypeId());
    }
    return null;
  }

  private static Node ifThenElse(final Node ... nodes) {
    return new NodeOperation(StandardOperation.ITE, nodes);
  }

  private static void unsupported(final Object obj) {
    throw RetrascopeRuntimeException.makeException(LogMessage.UNSUPPORTED_TYPE, obj);
  }

  private static boolean dependExactlyOnVariables(
      final Collection<NodeVariable> exprVariables,
      final Collection<NodeVariable> variables) {

    if (exprVariables.isEmpty()) {
      return false;
    } else {
      exprVariables.removeAll(variables);
      return exprVariables.isEmpty();
    }
  }

  /**
   * Splits the specified node into two parts - one that depends exactly from the specified
   * variables and another that does not depend.
   * @param node Node to be split.
   * @param variables Variables dependence to which is a splitting criterion.
   * @return Two nodes that being united (with AND) form a node that is equal to the specified one.
   *         The first one depends exactly from the specified variables the second does not depend
   *         from all of them.
   */
  public static Pair<Node, Node> splitByVar(
      final Node node,
      final Collection<NodeVariable> variables) {

    class DepSplitVisitor extends ExprTreeVisitorDefault {
      private Collection<Node> dependOnVar;
      private Collection<Node> freeFromVar;
      private Collection<NodeVariable> variables;

      private DepSplitVisitor(final Collection<NodeVariable> variables,
          final Collection<Node> dependNodes, final Collection<Node> freeNodes) {
        InvariantChecks.checkNotNull(variables);

        this.variables = variables;
        this.dependOnVar = dependNodes;
        this.freeFromVar = freeNodes;
      }

      @Override
      public void onOperationBegin(final NodeOperation node) {
        if (!ExprUtils.isOperation(node, StandardOperation.AND)) {

          if (dependExactlyOnVariables(ExprUtils.getVariables(node), this.variables)) {
            this.dependOnVar.add(node);
          } else {
            this.freeFromVar.add(node);
          }

          setStatus(Status.SKIP);
        }
      }

      @Override
      public void onOperationEnd(final NodeOperation node) {
        if (getStatus() == Status.SKIP) {
          setStatus(Status.OK);
        }
      }
    }

    final List<Node> varDepends = new LinkedList<>();
    final List<Node> varFreeNodes = new LinkedList<>();

    final DepSplitVisitor visitor = new DepSplitVisitor(variables, varDepends, varFreeNodes);
    final ExprTreeWalker walker = new ExprTreeWalker(visitor);
    walker.visit(node);

    return new ImmutablePair<>(
        varDepends.isEmpty() ? null : NodeUtils.getConjunction(varDepends),
        varFreeNodes.isEmpty() ? null : NodeUtils.getConjunction(varFreeNodes));
  }

  /**
   * Splits the specified array of nodes (conjunctions, probably) into the collection
   * of collections of nodes that depend on same variables.
   * @param nodes The nodes to be grouped by variable from that they are depend.
   * @return The collection of collections of nodes, such that every collection contains
   *         only the nodes that depend on common set of variables.
   */
  static Collection<Collection<Node>> splitByVar(final Node... nodes) {

    final Map<Set<NodeVariable>, Collection<Node>> varNodeMap = new HashMap<>();

    for (final Node node : nodes) {

      final Collection<Node> elementaryOps = splitConjunction(node);

      for (final Node element : elementaryOps) {

        final Set<NodeVariable> vars = new HashSet<>(ExprUtils.getVariables(element));

        final Collection<Set<NodeVariable>> intersected =
            getIntersected(varNodeMap.keySet(), vars);

        if (intersected.isEmpty()) {

          final Collection<Node> useNodes = new HashSet<>();
          useNodes.add(element);
          varNodeMap.put(vars, useNodes);

        } else if (intersected.size() == 1) {

          varNodeMap.get(intersected.iterator().next()).add(element);
        } else {

          final Set<NodeVariable> union = new HashSet<>();
          final Collection<Node> useNodes = new HashSet<>();

          for (final Set<NodeVariable> key : intersected) {
            union.addAll(key);
            useNodes.addAll(varNodeMap.get(key));
            varNodeMap.remove(key);
          }

          union.addAll(vars);
          useNodes.add(element);
          varNodeMap.put(union, useNodes);
        }
      }
    }

    return new HashSet<>(varNodeMap.values());
  }

  private static Collection<Set<NodeVariable>> getIntersected(
      final Set<Set<NodeVariable>> collections,
      final Set<NodeVariable> vars) {

    final Collection<Set<NodeVariable>> intersected = new HashSet<>();

    for (final Set<NodeVariable> collection : collections) {

      if (CollectionUtils.areIntersectedSets(collection, vars)) {
        intersected.add(collection);
      }
    }

    return intersected;
  }

  /**
   * Splits conjunction node into non-conjunction nodes.
   * @param node Node to be split.
   * @return Collection of nodes that being united by AND operation (a conjunction)
   *         forms an expression that is equivalent to the specified one.
   */
  private static Collection<Node> splitConjunction(final Node node) {

    class ConjunctionSplitVisitor extends ExprTreeVisitorDefault {

      private Collection<Node> nodes;

      private ConjunctionSplitVisitor() {
        this.nodes = new LinkedHashSet<>();
      }

      Collection<Node> getNodes() {
        return this.nodes;
      }

      @Override
      public void onOperationBegin(final NodeOperation operation) {
        elaborate(operation);
      }

      private void elaborate(final NodeOperation operation) {

        if (ExprUtils.isOperation(operation, StandardOperation.AND)) {

          for (int i = 0; i < operation.getOperandCount(); i++) {

            final Node operand = operation.getOperand(i);

            if (ExprUtils.isKind(Node.Kind.OPERATION, operand)) {

              elaborate((NodeOperation) operand);

            } else {
              this.nodes.add(operand);
            }
          }
        } else {
          this.nodes.add(operation);
          setStatus(Status.SKIP);
        }
      }

      @Override
      public void onOperationEnd(final NodeOperation node) {

        if (getStatus() == Status.SKIP) {
          setStatus(Status.OK);
        }
      }
    }

    InvariantChecks.checkNotNull(node);

    final ConjunctionSplitVisitor visitor = new ConjunctionSplitVisitor();
    final ExprTreeWalker walker = new ExprTreeWalker(visitor);
    walker.visit(node);
    return visitor.getNodes();
  }

  /**
   * Checks rather the specified node is a {@code NodeOperation} object with identifier
   * from the specified list.
   * @param node Node to be checked.
   * @param group Group of identifiers.
   * @return {@code true} if node is a {@code NodeOperation} object which identifier
   *         is in the specified list, {@code false} otherwise.
   */
  public static boolean isOperation(
      final Node node,
      final Collection<StandardOperation> group) {

    for (final StandardOperation id : group) {

      if (ExprUtils.isOperation(node, id)) {
        return true;
      }
    }

    return false;
  }

  /**
   * Checks whether any of the specified nodes is of bit vector type.
   * @param nodes The nodes to be checked.
   * @return {@code true} when some of the specified nodes is of bit vector type,
   *         {@code false} otherwise.
   */
  public static boolean hasBitVector(final Node ... nodes) {

    boolean hasBvOperand = false;

    for (final Node node : nodes) {

      if (node.isType(DataTypeId.BIT_VECTOR)) {

        hasBvOperand = true;
        break;
      }
    }
    return hasBvOperand;
  }

  /**
   * Creates bit vector of the specified size object from the specified value.
   * @param value The bit vector value.
   * @param size The bit vector size.
   * @return The bit vector object the has the specified size and the specified value.
   */
  public static NodeValue newBitVector(final int value, final int size) {

    return NodeValue.newBitVector(BitVector.valueOf(value, size));
  }

  /**
   * Substitutes variables that are used in the node by the specified bound variables.
   *
   * @param node The node to be performed.
   * @param boundVariables The substitutions.
   * @return The performed node, some of that variables are substituted by bound ones.
   */
  public static Node substituteAllBindings(
      final Node node,
      final List<NodeBinding.BoundVariable> boundVariables) {

    return Transformer.substituteAllBindings(new NodeBinding(node, boundVariables));
  }
}
