/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TransformerRule;
import ru.ispras.fortress.transformer.TypeConversion;
import ru.ispras.fortress.util.InvariantChecks;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Set of expression transformation rules.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
final class TransformerRuleSet {

  /* Operations that produce boolean result. */
  private static final List<StandardOperation> BOOL_GROUP = Arrays.asList(
      StandardOperation.AND,
      StandardOperation.OR,
      StandardOperation.NOT,
      StandardOperation.XOR,
      StandardOperation.IMPL);

  /* Operations that compare same type arguments. */
  private static final List<StandardOperation> COMPARE_GROUP = Arrays.asList(
      StandardOperation.EQ,
      StandardOperation.NOTEQ,
      StandardOperation.GREATER,
      StandardOperation.GREATEREQ,
      StandardOperation.LESS,
      StandardOperation.LESSEQ);

  private static final List<StandardOperation> ARITHM_GROUP = Arrays.asList(
      StandardOperation.ADD,
      StandardOperation.SUB);


  static class CompareOpDescriptor {
    NodeValue rhs;
    NodeValue lhs;
    NodeValue module;
    NodeVariable var;

    CompareOpDescriptor() {
      /* Empty. */
    }

    NodeValue getRightHandSide() {
      return (NodeValue) this.rhs.deepCopy();
    }

    void setRightHandSide(final NodeValue value) {
      this.rhs = value;
    }

    NodeValue getLeftHandSide() {
      return (NodeValue) this.lhs.deepCopy();
    }

    void setLeftHandSide(final NodeValue value) {
      this.lhs = value;
    }

    NodeValue getModule() {
      return (NodeValue) this.module.deepCopy();
    }

    void setModule(final NodeValue value) {
      this.module = value;
    }

    NodeVariable getVariable() {
      return (NodeVariable) this.var.deepCopy();
    }

    void setVariable(final  NodeVariable variable) {
      this.var = variable;
    }
  }

  /**
   * Returns set of transformation rules that are applied while standardization process.
   *
   * @return Mapping from operation identifiers to transformation rules.
   */
  static NodeTransformer getRuleSet() {

    final NodeTransformer transformer = new NodeTransformer();

    final TransformerRule booleanGroupRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        if (ExprUtils.isKind(Node.Kind.OPERATION, node)) {

          final NodeOperation operation = (NodeOperation) node;

          InvariantChecks.checkTrue(operation.getOperationId() instanceof StandardOperation);

          StandardOperation opId = (StandardOperation) operation.getOperationId();

          return BOOL_GROUP.contains(opId);
        }
        return false;
      }

      @Override
      public Node apply(final Node node) {

        final NodeOperation operation = (NodeOperation) node;
        final List<Node> args = new ArrayList<>();

        for (final Node child : operation.getOperands()) {

          final Node converted = TypeConversion.coerce(child, DataType.BOOLEAN);

          // not integral node, maybe throw exception?
          if (converted == null) {
            return node;
          }
          args.add(converted);
        }
        return new NodeOperation(operation.getOperationId(), args);
      }
    };

    final TransformerRule iteRule = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.ITE);
      }

      @Override
      public Node apply(final Node node) {

        final NodeOperation ite = (NodeOperation) node;
        final List<Node> args = new ArrayList<>(ite.getOperands());
        final Node condition = TypeConversion.coerce(args.get(0), DataType.BOOLEAN);

        // not integral node, maybe throw exception?
        if (condition == null) {
          return node;
        }
        args.set(0, condition);

        return new NodeOperation(StandardOperation.ITE, args);
      }
    };

    final TransformerRule expandEq = new TransformerRule() {

      @Override
      public boolean isApplicable(final Node node) {

        return ExprUtils.isOperation(node, StandardOperation.EQ)
            && ((NodeOperation) node).getOperandCount() > 2;
      }

      @Override
      public Node apply(final Node node) {

        final List<Node> operands = ((NodeOperation)node).getOperands();
        final Node[] equations = new NodeOperation[operands.size() - 1];

        for (int i = 1; i < operands.size(); i++) {
          equations[i - 1] = NodeUtils.getEquation(operands.get(i), operands.get(0));
        }

        return ExprUtils.getConjunction(equations);
      }
    };

    final TransformerRule addRule = new TransformerRule() {

      CompareOpDescriptor descriptor;

      @Override
      public boolean isApplicable(final Node node) {

        this.descriptor = new CompareOpDescriptor();

        return NodeUtils.isOperation(node, COMPARE_GROUP)
            && hasAddSubOperand(this.descriptor, (NodeOperation) node, StandardOperation.ADD);
      }

      @Override
      public Node apply(final Node node) {

        return new NodeOperation(
            ((NodeOperation)node).getOperationId(),
            this.descriptor.getVariable(),
            Transformer.reduce(
                new NodeOperation(
                    StandardOperation.SUB,
                    this.descriptor.getRightHandSide(),
                    this.descriptor.getLeftHandSide())));
      }
    };

    final TransformerRule addByModuleRule = new TransformerRule() {

      CompareOpDescriptor descriptor;

      @Override
      public boolean isApplicable(final Node node) {

        this.descriptor = new CompareOpDescriptor();

        return NodeUtils.isOperation(node, COMPARE_GROUP)
            && hasAddByModuleOperand(this.descriptor, (NodeOperation) node, StandardOperation.ADD);
      }

      @Override
      public Node apply(final Node node) {

        return new NodeOperation(
            ((NodeOperation)node).getOperationId(),
            this.descriptor.getVariable(),
            Transformer.reduce(
                new NodeOperation(
                    StandardOperation.MOD,
                    new NodeOperation(
                        StandardOperation.SUB,
                        this.descriptor.getRightHandSide(),
                        this.descriptor.getLeftHandSide()),
                this.descriptor.getModule())));
      }
    };

    final TransformerRule subRule = new TransformerRule() {

      CompareOpDescriptor descriptor;

      @Override
      public boolean isApplicable(final Node node) {

        this.descriptor = new CompareOpDescriptor();

        return NodeUtils.isOperation(node, COMPARE_GROUP)
            && hasAddSubOperand(this.descriptor, (NodeOperation) node, StandardOperation.SUB);
      }

      @Override
      public Node apply(final Node node) {

        return new NodeOperation(
            ((NodeOperation)node).getOperationId(),
            this.descriptor.getVariable(),
            Transformer.reduce(
                new NodeOperation(
                    StandardOperation.ADD,
                    this.descriptor.getRightHandSide(),
                    this.descriptor.getLeftHandSide())));
      }
    };

    final TransformerRule subByModuleRule = new TransformerRule() {

      CompareOpDescriptor descriptor;

      @Override
      public boolean isApplicable(final Node node) {

        this.descriptor = new CompareOpDescriptor();

        return NodeUtils.isOperation(node, COMPARE_GROUP)
            && hasAddByModuleOperand(this.descriptor, (NodeOperation) node, StandardOperation.SUB);
      }

      @Override
      public Node apply(final Node node) {

        return new NodeOperation(
            ((NodeOperation)node).getOperationId(),
            this.descriptor.getVariable(),
            Transformer.reduce(
                new NodeOperation(
                    StandardOperation.MOD,
                    new NodeOperation(
                        StandardOperation.ADD,
                        this.descriptor.getRightHandSide(),
                        this.descriptor.getLeftHandSide()),
                    this.descriptor.getModule())));
      }
    };

    for (final Enum<?> op : BOOL_GROUP) {
      transformer.addRule(op, booleanGroupRule);
    }

    transformer.addRule(StandardOperation.ITE, iteRule);

    for (final Enum<?> op : COMPARE_GROUP) {

      transformer.addRule(op, addRule);
      transformer.addRule(op, subRule);
      transformer.addRule(op, addByModuleRule);
      transformer.addRule(op, subByModuleRule);
    }
    transformer.addRule(StandardOperation.EQ, expandEq);

    return transformer;
  }

  private static boolean hasAddByModuleOperand(
      final CompareOpDescriptor descriptor,
      final NodeOperation node,
      final StandardOperation lowOp) {

    final int opCount = node.getOperandCount();

    if (opCount != 2) {
      return false;
    }

    boolean constOp = false;
    boolean addByModOp = false;

    for (int i = 0; i < opCount; i++) {

      final Node operand = node.getOperand(i);

      if (!addByModOp && ExprUtils.isOperation(operand, StandardOperation.MOD)) {

        final int count = ((NodeOperation)operand).getOperandCount();

        if (count != 2) {
          return false;
        }

        boolean hasConstOperand = false;
        boolean hasVarOperand = false;

        for (int j = 0; j < count; j++) {

          final Node subOperand = ((NodeOperation)operand).getOperand(j);

          if (ExprUtils.isOperation(subOperand, lowOp) && !hasVarOperand) {

            boolean hasVarArg = false;
            boolean hasConstArg = false;

            for (int k = 0; k < ((NodeOperation) subOperand).getOperandCount(); k++) {

              final Node subSubOperand = ((NodeOperation) subOperand).getOperand(k);

              if (!hasVarArg && subSubOperand instanceof NodeVariable) {
                hasVarArg = true;
                descriptor.setVariable((NodeVariable) subSubOperand);
              }

              if (!hasConstArg && subSubOperand instanceof NodeValue) {
                hasConstArg = true;
                descriptor.setLeftHandSide((NodeValue) subSubOperand);
              }
            }
            hasVarOperand = hasVarArg && hasConstArg;
          }

          if (subOperand instanceof NodeValue && !hasConstOperand) {
            hasConstOperand = true;
            descriptor.setModule((NodeValue)subOperand);
          }
        }

        addByModOp = hasConstOperand && hasVarOperand;
      }
      if (operand instanceof NodeValue && !constOp) {
        constOp = true;
        descriptor.setRightHandSide((NodeValue) operand);
      }
    }

    return constOp && addByModOp;
  }

  private static boolean hasAddSubOperand(
      final CompareOpDescriptor descriptor,
      final NodeOperation node,
      final StandardOperation opCode) {

    final int opCount = node.getOperandCount();

    if (opCount != 2) {
      return false;
    }

    boolean constOperand = false;
    boolean addSubOperand = false;

    for (int i = 0; i < opCount; i++) {

      final Node operand = node.getOperand(i);

      if (!addSubOperand && NodeUtils.isOperation(operand, ARITHM_GROUP)) {

        final int subOpCount = ((NodeOperation)operand).getOperandCount();

        if (subOpCount != 2) {
          return false;
        }

        boolean hasVarArg = false;
        boolean hasConstArg = false;

        for (int j = 0; j < subOpCount; j++) {

          final Node subOperand = ((NodeOperation)operand).getOperand(j);

          if (!hasVarArg && subOperand instanceof NodeVariable) {
            hasVarArg = true;
            descriptor.setVariable((NodeVariable) subOperand);
          }
          if (!hasConstArg && subOperand instanceof NodeValue) {
            hasConstArg = true;
            descriptor.setLeftHandSide((NodeValue) subOperand);
          }
        }
        addSubOperand = hasVarArg && hasConstArg && ExprUtils.isOperation(operand, opCode);
      }
      if (operand instanceof NodeValue && !constOperand) {
        constOperand = true;
        descriptor.setRightHandSide((NodeValue) operand);
      }
    }

    return constOperand && addSubOperand;
  }
}
