/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.transformer.NodeTransformer;
import ru.ispras.fortress.transformer.Transformer;
import ru.ispras.fortress.transformer.TypeConversion;

import java.util.List;

/**
 * Utility methods for transformation of expressions.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class TransformerUtils {

  private TransformerUtils() {
    /* Empty. */
  }

  /**
   * Replace operations in expression with standard counterparts.
   * <p>Transforms nodes of the following kinds:
   * <ul>
   *   <li> EQ operation with more than 2 operands - to conjunction of 2-operand EQs;
   * </ul>
   * Wrapping method for {@code Transformer.standardize} method.
   * Uses it's internal rules for additional replacements.</p>
   *
   * @see Transformer
   * @param node Expression to be transformed.
   * @return Expression with non-standard operations being replaced.
   */
  public static Node standardize(final Node node) {

    final Node standardizedNode = Transformer.standardize(node);

    final NodeTransformer transformer = TransformerRuleSet.getRuleSet();
    transformer.walk(standardizedNode);
    return transformer.getResult().iterator().next();
  }

  /**
   * Reduces the specified expression and tries to convert it to integer data type.
   * <p>It is recommended to use this method for non-variable-dependent expressions
   * containing constants and operators.</p>
   * @param node Expression to be simplified and converted.
   * @return Expression that is a simplified form of the specified node and is converted
   *         to integer type.
   */
  public static Node toInt(final Node node) {

    return TypeConversion.coerce(Transformer.reduce(node), DataType.INTEGER);
  }

  /**
   * Substitutes the given term for variables with the specified name at the specified expression
   * and simplifies the result.
   *
   * @param expression The expression in which substitution and simplification take place.
   * @param name The name of variable to be substituted.
   * @param term The term to which variables with the specified name should be substituted by.
   * @return The simplified expression that is produced by the substitution
   *         of all the variables are used in the specified expression and
   *         having the specified name by the specified term.
   */
  public static Node substitute(final Node expression, final String name, final Node term) {
    return Transformer.reduce(Transformer.substitute(expression, name, term));
  }


  /**
   * Substitutes bindings from the specified collection on the specified node.
   * @param node The node that should be performed.
   * @param bindings Bindings that should be used for substitution.
   * @return The node at which all the specified bindings are applied.
   */
  static Node substituteBindings(final Node node, final List<NodeBinding.BoundVariable> bindings) {

    return Transformer.reduce(
        Transformer.substituteAllBindings(new NodeBinding(node, bindings)));
  }
}
