/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope;

import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.parser.verilog.VerilogParser;
import ru.ispras.retrascope.parser.vhdl.VhdlParser;
import ru.ispras.retrascope.util.LogMessage;

import java.io.File;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;

/**
 * Utility methods for testing at module-level.
 * Can be applied for HDL modules written in languages are supported by the project.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public final class HdlTestUtils {

  private HdlTestUtils() {
    /* Empty. */
  }

  /**
   * Runs the tool on the specified mapping of HDL files with the specified target and options.
   *
   * @param map Mapping of HDL files to their additional data (for example, toplevel module names).
   * @param target Tool target entity.
   * @param additionalOptions Additional command line options for the tool.
   *
   * @throws RetrascopeRuntimeException when it encounters at the file with unknown extension.
   */
  public static void runHdl(
      final Map<String, String> map,
      final String target,
      final String ... additionalOptions) {

    for (final Map.Entry<String, String> entry : map.entrySet()) {

      final String filePath = entry.getKey();
      final Map<String, String> oneElemMap = new LinkedHashMap<>(1);
      oneElemMap.put(filePath, entry.getValue());

      final String ext = FilenameUtils.getExtension(filePath);

      if ("vhd".equalsIgnoreCase(ext)) {
        runVhdl(oneElemMap, target, additionalOptions);
      } else if ("v".equalsIgnoreCase(ext)) {
        runVerilog(oneElemMap, target, additionalOptions);
      } else {
        throw new RetrascopeRuntimeException(
            LogMessage.makeMessage(LogMessage.UNSUPPORTED_TYPE, ext));
      }
    }
  }

  /**
   * Runs the tool on all the HDL test projects (excluding files which path suffixes are in the
   * specified set) with the specified target and additional options.
   *
   * @param target Tool target.
   * @param ignored Path suffixes for files are to be ignored.
   * @param additionalOptions Additional options for tool.
   */
  private static void runHdl(
      final String target,
      final Set<String> ignored,
      final String... additionalOptions) {
    runVhdl(target, ignored, additionalOptions);
    runVerilog(target, ignored, additionalOptions);
  }

  /**
   * Runs tool on HDL modules with the specified target and options.
   *
   * @param target Tool target.
   * @param additionalOptions Additional options for tool.
   */
  public static void runHdl(
      final String target,
      final String ... additionalOptions) {
    runHdl(target, null, additionalOptions);
  }

  /**
   * Runs tool on all the VHDL projects from the test suite with the specified options.
   * <p>Runs the tool on all the VHDL test projects (excluding files which path suffixes are in the
   * specified set) with the specified target and additional options.</p>
   * @param target Tool target.
   * @param ignored Path suffixes for files are to be ignored.
   * @param additionalOptions Additional options for tool.
   */
  public static void runVhdl(
      final String target,
      final Set<String> ignored,
      final String ... additionalOptions) {

    /* Run Retrascope on each module of each VHDL project. */

    final Collection<Map<String, String>> vhdlFiles = HdlTestFiles.getVhdlProjects().values();

    for (final Map<String, String> projectMap : vhdlFiles) {

      for (final Map.Entry<String, String> entry : projectMap.entrySet()) {

        final String fullFileName = entry.getKey();

        if (ignored == null || notContain(ignored, entry.getKey())) {
          final Map<String, String> onlyFile = new LinkedHashMap<>();
          onlyFile.put(fullFileName, entry.getValue());

          runVhdl(onlyFile, target, additionalOptions);
        }
      }
    }
  }

  /**
   * Runs tool on VHDL modules with the specified target and options.
   *
   * @param target Tool target.
   * @param additionalOptions Additional options for tool.
   */
  public static void runVhdl(
      final String target,
      final String ... additionalOptions) {

    runVhdl(target, null, additionalOptions);
  }

  /**
   * Runs Retrascope on the specified VHDL files with the specified target and additional options.
   *
   * @param map VHDL projects-to-files mapping that are to be processed.
   * @param target Tool target.
   * @param additionalOptions Additional options for tool.
   */
  public static void runVhdl(
      final Map<String, String> map,
      final String target,
      final String... additionalOptions) {

    final String[] fileNames = getFileNames(map.keySet());
    final String topLevelNames = getTopLevel(map);

    runRetrascope(fileNames, createCommandLine(target, topLevelNames, additionalOptions));
  }

  /**
   * Runs tool on all the Verilog projects from the test suite with the specified options.
   * <p>Runs the tool on all the Verilog test projects (excluding files which path suffixes
   * are in the specified set) with the specified target and additional options.</p>
   * @param target Tool target.
   * @param ignored Path suffixes for files are to be ignored.
   * @param additionalOptions Additional options for tool.
   */
  public static void runVerilog(
      final String target,
      final Set<String> ignored,
      final String ... additionalOptions) {

    /* Run Retrascope on each module of each Verilog project. */

    final Collection<Map<String, String>> verilogFiles = HdlTestFiles.getVerilogProjects().values();

    for (final Map<String, String> projectMap : verilogFiles) {

      for (final Map.Entry<String, String> entry : projectMap.entrySet()) {

        final String fullFileName = entry.getKey();

        if (ignored == null || notContain(ignored, entry.getKey())) {
          final Map<String, String> onlyFile = new LinkedHashMap<>();
          onlyFile.put(fullFileName, entry.getValue());

          runVerilog(onlyFile, target, additionalOptions);
        }
      }
    }
  }

  /**
   * Runs tool on Verilog modules with the specified target.
   *
   * @param target Tool target.
   */
  public static void runVerilog(final String target) {

    runVerilog(target, null);
  }

  /**
   * Runs Retrascope on Verilog file with the specified options.
   *
   * @param map Map of projects to files to be processed.
   * @param target Value of command line target option.
   * @param additionalOptions Additional command line options for tool.
   */
  public static void runVerilog(
      final Map<String, String> map,
      final String target,
      final String... additionalOptions) {

    final String[] fileNames = getFileNames(map.keySet());
    final String includes = getIncludes(map.values());

    runRetrascope(fileNames, createIncludeCommandLine(target, includes, additionalOptions));
  }

  private static String getIncludes(final Collection<String> values) {
    return StringUtils.join(values, File.pathSeparator);
  }

  private static boolean notContain(
      final Set<String> strings,
      final String key) {

    for (final String substring : strings) {
      if (key.endsWith(substring)) {
        return false;
      }
    }
    return true;
  }

  /**
   * Create "TopLevel" parameter for command line.
   *
   * @param map VHDL files for processing.
   */
  private static String getTopLevel(final Map<String, String> map) {

    final StringBuilder topLevelNames = new StringBuilder();

    int index = 0;
    for (final String fileName : map.values()) {
      topLevelNames.append(fileName);

      if (index < map.size() - 1) {
        topLevelNames.append(File.pathSeparator);
      }
      index++;
    }

    return topLevelNames.toString();
  }

  /**
   * Creates command line with the specified parameters.
   *
   * @param target Value of command line "target" option.
   * @param topLevelNames Value of command line "toplevel" option.
   * @param additionalOptions Additional options for tool.
   *
   * @return Command line for tool.
   */
  private static String[] createCommandLine(
      final String target,
      final String topLevelNames,
      final String ... additionalOptions) {

    return (String[]) ArrayUtils.addAll(
        new String[] {
            "--" + Retrascope.TARGET_PARAM.getLongName(),
            target,
            "--" + VhdlParser.TOP_LEVEL.getLongName(),
            topLevelNames},
        additionalOptions);
  }

  private static String[] createIncludeCommandLine(
      final String target,
      final String includes,
      final String ... additionalOptions) {

    return (String[]) ArrayUtils.addAll(
        new String[] {
            "--" + Retrascope.TARGET_PARAM.getLongName(),
            target,
            "--" + VerilogParser.INCLUDE_PATH.getLongName(),
            includes},
      additionalOptions);
  }

  /**
   * Runs Retrascope.
   *
   * @param fileNames Names of VHDL files for processing.
   * @param options Command line options for the tool.
   */
  private static void runRetrascope(
      final String[] fileNames,
      final String ... options) {

    Retrascope.main((String[]) ArrayUtils.addAll(fileNames, options));
  }

  /**
   * Creates an array of file names for the command line.
   *
   * @param files HDL files.
   */
  private static String[] getFileNames(final Set<String> files) {

    final String[] fileNames = new String[files.size()];

    int index = 0;
    for (final String fileName : files) {
      fileNames[index] = (new File(fileName)).getAbsolutePath();
      index++;
    }

    return fileNames;
  }
}
