/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope;

import org.apache.commons.io.FilenameUtils;
import org.junit.Assert;
import org.junit.Test;

import ru.ispras.retrascope.basis.Engine;
import ru.ispras.retrascope.engine.cgaa.transformer.efsm.CgaaEfsmTransformer;
import ru.ispras.retrascope.engine.hldd.printer.smv.assertion.HlddAssertSmvPrinter;
import ru.ispras.retrascope.engine.smv.launcher.SmvModelCheckerLauncher;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * This class represents the base class for test cases which use the SMV model checker.
 *
 * @author <a href="mailto:lebedev@ispras.ru">Mikhail Lebedev</a>
 */
public abstract class SmvUtilTest {

  private static final String BOUNDED_MODEL_CHECKING = "bmc";
  private static final String SYMBOLIC_MODEL_CHECKING = "bdd";

  private static final List<String> EXCLUDED;
  private static final Map<String, String> METHOD = new HashMap<>();
  private static final Map<String, Integer> EXPECTED_COUNTEREXAMPLE_COUNT = new HashMap<>();

  static {
    EXCLUDED = new ArrayList<>();
    // EXCLUDED.add(Paths.get("b03", "b03.vhd").toString());
    // EXCLUDED.add(Paths.get("b04", "b04.vhd").toString());
    EXCLUDED.add(Paths.get("b05", "b05.vhd").toString()); // nuSMV works too long
    // EXCLUDED.add(Paths.get("b07", "b07.vhd").toString());
    // EXCLUDED.add(Paths.get("b08", "b08.vhd").toString());
    // EXCLUDED.add(Paths.get("b10", "b10.vhd").toString());
    // EXCLUDED.add(Paths.get("b11", "b11.vhd").toString());
    EXCLUDED.add(Paths.get("b12", "b12.vhd").toString());
    EXCLUDED.add(Paths.get("b13", "b13.vhd").toString());
    EXCLUDED.add(Paths.get("fifo_mod", "fifo.v").toString()); // assignment type mismatch
    EXCLUDED.add(Paths.get("plasma", "reg_bank.vhd").toString()); // array store operations
    


    METHOD.put("b01", SYMBOLIC_MODEL_CHECKING);
    METHOD.put("b02", SYMBOLIC_MODEL_CHECKING);
    METHOD.put("b03", SYMBOLIC_MODEL_CHECKING);
    METHOD.put("b04", BOUNDED_MODEL_CHECKING);
    METHOD.put("b06", SYMBOLIC_MODEL_CHECKING);
    METHOD.put("b07", SYMBOLIC_MODEL_CHECKING);
    METHOD.put("b08", SYMBOLIC_MODEL_CHECKING);
    METHOD.put("b09", SYMBOLIC_MODEL_CHECKING);
    METHOD.put("b10", SYMBOLIC_MODEL_CHECKING);
    METHOD.put("b11", BOUNDED_MODEL_CHECKING);

   /* EXPECTED_COUNTEREXAMPLE_COUNT.put("b01", 29);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b02", 48);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b03", 41);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b04", 46);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b06", 75);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b07", 52);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b08", 63);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b09", 32);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b10", 95);
    EXPECTED_COUNTEREXAMPLE_COUNT.put("b11", 78);*/

  }

  public SmvUtilTest() {}

  @Test
  public void test() {
    runTest();
  }

  private void runTest() {
    final Map<String, Map<String, String>> files = getDesigns();

    for (Entry<String, Map<String, String>> project : files.entrySet()) {
      for (Entry<String, String> file : project.getValue().entrySet()) {

        final String filePath = file.getKey();

        final String folderName = ReadWriteUtils.getParentDirName(filePath);
        final String designName = FilenameUtils.getBaseName(filePath);

        final Map<String, String> onlyFile = new LinkedHashMap<>();
        onlyFile.put(filePath, file.getValue());

        if (!isExcluded(folderName)) {
          HdlTestUtils.runHdl(
              onlyFile,
              getMainTarget(),
              getAdditionalTargets(designName));

          checkResults(designName);
        }
      }
    }
  }

  /**
   * The main target of the launch.
   *
   * @return The main target.
   */
  protected abstract String getMainTarget();

  /**
   * Additional parameters of the launch.
   *
   * @param designName - the specified design name.
   * @return The array of parameters.
   */
  protected abstract String[] getAdditionalTargets(final String designName);

  /**
   * HDL designs used by the test.
   *
   * @return The map of designs.
   */
  protected abstract Map<String, Map<String, String>> getDesigns();

  /**
   * Indicates whether the specified design path is not supported by the checker.
   *
   * @param path - the specified design path (in the form {@code "folder/file"}).
   * @return {@code True} if the design is NOT supported, {@code false} otherwise.
   */
  private boolean isExcluded(final String path) {
    return EXCLUDED.contains(path);
  }

  /**
   * Returns the preferred method of the specified design model checking.
   *
   * @param designName - the specified design name (filename without extension).
   * @return {@code bdd} for symbolic model checking or {@code bmc} for bounded model checking.
   *         Bounded model checking is the default value.
   */
  private String getCheckMethod(final String designName) {
    return METHOD.containsKey(designName) ? METHOD.get(designName) : BOUNDED_MODEL_CHECKING;
  }

  /**
   * Targets used by any SMV-based test.
   *
   * @param designName - the specified design name.
   * @return The list of targets.
   */
  protected List<String> commonTargets(final String designName) {
    final List<String> targets = new ArrayList<>();
    final String smvFileName = designName + ".smv";
    targets.add("--" + HlddAssertSmvPrinter.SMV_FILE_NAME.getLongName());
    targets.add(smvFileName);
    targets.add("--" + HlddAssertSmvPrinter.CHECK_METHOD.getLongName());
    targets.add(getCheckMethod(designName));
    if (designName.equals("b07") || designName.equals("b11") || "b08".equals(designName)) {
      targets.add("--" + CgaaEfsmTransformer.STATE_VARIABLES.getLongName());
      targets.add("STATO");
    }
    return targets;
  }

  /*
   * Checks nuXmv run results.
   */
  private void checkResults(final String designName) {
    if (EXPECTED_COUNTEREXAMPLE_COUNT.containsKey(designName)) {
      final Engine launcher =
          Retrascope.getEngineRegistry().getEngine(SmvModelCheckerLauncher.ENGINE_ID);
      final int actual = ((SmvModelCheckerLauncher) launcher).getCounterexampleCount();
      final int expected = EXPECTED_COUNTEREXAMPLE_COUNT.get(designName);
      Assert.assertEquals(expected, actual);
    }
  }
}
