/*
 * Copyright 2014-2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.retrascope.basis;

import org.junit.Assert;
import org.junit.Test;
import ru.ispras.retrascope.util.Log;
import ru.ispras.retrascope.util.LogLevel;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public class EngineRegistryTestCase {

  private static class TestEntity extends Entity {
    TestEntity(final EntityType entityType) {
      super(entityType);
    }

    @Override
    public void add(final Entity entity) {
      Log.getLogger().log(LogLevel.INFO, "Merging " + entity.getEntityType());
    }
  }

  private static class TestEngine extends Engine {

    TestEngine(final String engineId, final String inputType, final String outputType) {
      super(engineId);

      addInputType(EntityType.get(inputType));
      setOutputType(EntityType.get(outputType));
    }

    TestEngine(final String engineId,
        final String inputType1,
        final String inputType2,
        final String outputType) {

      super(engineId);

      addInputType(EntityType.get(inputType1));
      addInputType(EntityType.get(inputType2));
      setOutputType(EntityType.get(outputType));
    }

    @Override
    public Entity start(final Map<EntityType, Entity> inputs) {
      Log.getLogger().log(LogLevel.INFO, "Starting " + this);
      return new TestEntity(getOutputType());
    }
  }

  @Test
  public void runTest() {
    final EngineRegistry config = new EngineRegistry();

    config.add(new TestEngine( "a1",    "i1",           "o1"    ));
    config.add(new TestEngine( "b1",    "I1",           "o1"    ));
    config.add(new TestEngine( "c1",    "i1",           "o2"    ));
    config.add(new TestEngine( "d1",    "o1",           "o2"    ));
    config.add(new TestEngine( "a12",   "i1",   "i2",   "o12"   ));
    config.add(new TestEngine( "a23",   "i2",   "i3",   "o23"   ));
    config.add(new TestEngine( "a34",   "i3",   "i4",   "o34"   ));
    config.add(new TestEngine( "a123",  "o1",   "o23",  "o123"  ));
    config.add(new TestEngine( "b123",  "o12",  "o23",  "o123"  ));
    config.add(new TestEngine( "a123",  "o23",  "o34",  "o234"  ));
    config.add(new TestEngine( "a1234", "o123", "o34",  "o1234" ));
    config.add(new TestEngine( "b1234", "o12",  "o234", "o1234" ));

    final Set<EntityType> targets = new HashSet<>();
    final Set<Engine> engines     = new HashSet<>();
    final Set<EntityType> sources = new HashSet<>();

    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      runTest("{} -> {}", config, targets, engines, sources, true);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      engines.add(config.getEngine("a1"));

      // Engine a1 requires i1.
      runTest("{} -> {} using {a1}", config, targets, engines, sources, false);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      engines.add(config.getEngine("a1"));
      sources.add(EntityType.get("i1"));

      runTest("{i1} -> {} using {a1}", config, targets, engines, sources, true);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("undef"));
      sources.add(EntityType.get("i1"));

      // Unknown target.
      runTest("{i1} -> {undef}", config, targets, engines, sources, false);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("undef"));
      sources.add(EntityType.get("o1"));

      // Unknown source.
      runTest("{undef} -> {o1}", config, targets, engines, sources, false);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o1"));

      // No input.
      runTest("{} -> {o1}", config, targets, engines, sources, false);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o1"));
      sources.add(EntityType.get("i1"));

      runTest("{i1} -> {o1}", config, targets, engines, sources, true);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o1"));
      sources.add(EntityType.get("i1"));
      sources.add(EntityType.get("I1"));

      runTest("{i1, I1} -> {o1}", config, targets, engines, sources, true);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o2"));
      sources.add(EntityType.get("i1"));
      sources.add(EntityType.get("I1"));

      runTest("{i1, I1} -> {o2}", config, targets, engines, sources, true);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o23"));
      sources.add(EntityType.get("i2"));

      // No input.
      runTest("{i2} -> {o23}", config, targets, engines, sources, false);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o23"));
      sources.add(EntityType.get("i2"));
      sources.add(EntityType.get("i3"));

      runTest("{i2, i3} -> {o23}", config, targets, engines, sources, true);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o23"));
      targets.add(EntityType.get("o34"));
      sources.add(EntityType.get("i2"));
      sources.add(EntityType.get("i4"));

      // No input.
      runTest("{i2, i4} -> {o23, o34}", config, targets, engines, sources, false);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o23"));
      targets.add(EntityType.get("o34"));
      sources.add(EntityType.get("i2"));
      sources.add(EntityType.get("i3"));
      sources.add(EntityType.get("i4"));

      runTest("{i2, i3, i4} -> {o23, o34}", config, targets, engines, sources, true);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o123"));
      sources.add(EntityType.get("i1"));
      sources.add(EntityType.get("i2"));
      sources.add(EntityType.get("i3"));

      runTest("{i1, i2, i3} -> {o123}", config, targets, engines, sources, true);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o1"));
      targets.add(EntityType.get("o23"));
      targets.add(EntityType.get("o123"));
      sources.add(EntityType.get("i1"));
      sources.add(EntityType.get("i2"));
      sources.add(EntityType.get("i3"));
      sources.add(EntityType.get("i4"));

      runTest("{i1, i2, i3, i4} -> {o1, o23, o123}", config, targets, engines, sources, true);
    }
    // ---------------------------------------------------------------------------------------------
    {
      targets.clear();
      engines.clear();
      sources.clear();

      targets.add(EntityType.get("o1234"));
      sources.add(EntityType.get("i1"));
      sources.add(EntityType.get("i2"));
      sources.add(EntityType.get("i3"));
      sources.add(EntityType.get("i4"));

      // Conflict.
      runTest("{i1, i2, i3, i4} -> {o1234}", config, targets, engines, sources, false);
    }
  }

  private void runTest(
      final String testCase,
      final EngineRegistry config,
      final Set<EntityType> targets,
      final Set<Engine> engines,
      final Set<EntityType> sources,
      final boolean positive) {

    Log.getLogger().log(LogLevel.INFO, "Running test case " + testCase);

    final Engine engine = config.getEngine(targets, engines, sources);

    if (engine != null) {
      final Map<EntityType, Entity> inputs = new HashMap<>();
      for (final EntityType source : sources) {
        inputs.put(source, new TestEntity(source));
      }

      engine.start(inputs);
    }

    if (positive == (engine == null)) {
      Assert.fail();
    }

    Log.getLogger().log(LogLevel.INFO, "");
  }
}
