/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cfg.generator.test.random;

import org.junit.Test;
import ru.ispras.retrascope.HdlTestFiles;
import ru.ispras.retrascope.HdlTestUtils;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.result.test.TestModel;

import java.nio.file.Paths;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Test for Random test generator based on Control Flow Graph Model.
 * The test runs test generator on every separate HDL file and the engine produces
 * {@link TestModel} representation of test.
 *
 * @see CfgRandomTestGenerator
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CfgRandomTestGeneratorTestCase {

  static final int TEST_LENGTH = 10;
  private static final String ITC99 = "itc99";
  private static final String MINIMIPS = "minimips";
  private static final String PLASMA = "plasma";
  private static final String CLK = "CLK";
  private static final String RESET = "RESET";
  private static final String RESET_IN = "RESET_IN";
  private static final String CLOCK = "CLOCK";
  private static final String CLK_LOW = "clk";
  private static final String RST_LOW = "rst";
  private static final Map<String, FileDescriptor> MAP = new LinkedHashMap<>();

  static class FileDescriptor {
    private final String clkName;
    private final String rstName;
    private final String inputValues;

    FileDescriptor(
        final String clkName,
        final String rstName) {
      this.clkName = clkName;
      this.rstName = rstName;
      this.inputValues = "";
    }

    FileDescriptor(
        final String clkName,
        final String rstName,
        final String inputValues) {
      this.clkName = clkName;
      this.rstName = rstName;
      this.inputValues = inputValues;
    }

    String getClkName() {
      return this.clkName;
    }

    String getRstName() {
      return this.rstName;
    }

    String getInputValues() {
      return this.inputValues;
    }
  }

  static {

    /* === VHDL test projects === */

    /* sample */
    MAP.put(Paths.get("bvsample", "bvsample.vhd").toString(), new FileDescriptor(CLK, ""));

    /* sample */
    MAP.put(Paths.get("sample", "sample.vhd").toString(), new FileDescriptor(CLK, "RST"));

    /* count */
    MAP.put(Paths.get("count", "count.vhd").toString(), new FileDescriptor(CLK, RESET));

    /* example */
    MAP.put(Paths.get("example", "test.vhd").toString(), new FileDescriptor(CLK, RESET));
    MAP.put(Paths.get("example", "driver.vhd").toString(), new FileDescriptor(CLK, RESET));
    MAP.put(Paths.get("example", "example.vhd").toString(), new FileDescriptor(CLK, RESET));

    /* ITC'99  benchmark */
    MAP.put(Paths.get(ITC99, "b01", "b01.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b02", "b02.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b03", "b03.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b04", "b04.vhd").toString(),
        new FileDescriptor(CLOCK, RESET, "DATA_IN=[-128,127,10]"));
    MAP.put(Paths.get(ITC99, "b05", "b05.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b06", "b06.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b07", "b07.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b08", "b08.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b09", "b09.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b10", "b10.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b11", "b11.vhd").toString(),
        new FileDescriptor(CLOCK, RESET, "X_IN=[0,63,10]"));
    MAP.put(Paths.get(ITC99, "b12", "b12.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b13", "b13.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b14", "b14.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b14_1", "b14_1.vhd").toString(),
        new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b15", "b15.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b15_1", "b15_1.vhd").toString(),
        new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b17", "b17.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b17_1", "b17_1.vhd").toString(),
        new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b18", "b18.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b18_1", "b18_1.vhd").toString(),
        new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b19", "b19.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b19_1", "b19_1.vhd").toString(),
        new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b20", "b20.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b20_1", "b20_1.vhd").toString(),
        new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b21", "b21.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b21_1", "b21_1.vhd").toString(),
        new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b22", "b22.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(ITC99, "b22_1", "b22_1.vhd").toString(),
        new FileDescriptor(CLOCK, RESET));

    /* MiniMIPS */
    MAP.put(Paths.get(MINIMIPS, "alu.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(MINIMIPS, "banc.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(MINIMIPS, "bus_ctrl.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(MINIMIPS, "minimips.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(MINIMIPS, "pps_di.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(MINIMIPS, "pps_ei.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(MINIMIPS, "pps_ex.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(MINIMIPS, "pps_mem.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(MINIMIPS, "pps_pf.vhd").toString(), new FileDescriptor(CLOCK, RESET));
    MAP.put(Paths.get(MINIMIPS, "ram.vhd").toString(), new FileDescriptor("", ""));
    MAP.put(Paths.get(MINIMIPS, "renvoi.vhd").toString(), new FileDescriptor("", ""));
    MAP.put(Paths.get(MINIMIPS, "rom.vhd").toString(), new FileDescriptor("", ""));
    MAP.put(Paths.get(MINIMIPS, "syscop.vhd").toString(), new FileDescriptor(CLOCK, RESET));

    /* Plasma */
    MAP.put(Paths.get(PLASMA, "alu.vhd").toString(), new FileDescriptor("", ""));
    MAP.put(Paths.get(PLASMA, "bus_mux.vhd").toString(), new FileDescriptor("", ""));
    MAP.put(Paths.get(PLASMA, "control.vhd").toString(), new FileDescriptor("", ""));
    MAP.put(Paths.get(PLASMA, "mem_ctrl.vhd").toString(), new FileDescriptor(CLK, RESET_IN));
    MAP.put(Paths.get(PLASMA, "mlite2sram.vhd").toString(), new FileDescriptor(CLK, ""));
    MAP.put(Paths.get(PLASMA, "mlite2uart.vhd").toString(), new FileDescriptor(CLK, RESET));
    MAP.put(Paths.get(PLASMA, "mlite_cpu.vhd").toString(), new FileDescriptor(CLK, RESET_IN));
    MAP.put(Paths.get(PLASMA, "mult.vhd").toString(), new FileDescriptor(CLK, ""));
    MAP.put(Paths.get(PLASMA, "pc_next.vhd").toString(), new FileDescriptor(CLK, RESET_IN));
    MAP.put(Paths.get(PLASMA, "pipeline.vhd").toString(), new FileDescriptor(CLK, RESET));
    MAP.put(Paths.get(PLASMA, "plasma.vhd").toString(),
        new FileDescriptor("BOARD_CLK", "BOARD_RES"));
    MAP.put(Paths.get(PLASMA, "ram.vhd").toString(), new FileDescriptor(CLK, ""));
    MAP.put(Paths.get(PLASMA, "reg_bank.vhd").toString(), new FileDescriptor(CLK, RESET_IN));
    MAP.put(Paths.get(PLASMA, "shifter.vhd").toString(), new FileDescriptor("", ""));
    MAP.put(Paths.get(PLASMA, "sram2mlite.vhd").toString(), new FileDescriptor(CLK, ""));
    MAP.put(Paths.get(PLASMA, "uart.vhd").toString(), new FileDescriptor(CLK, RESET));

    /* === Verilog test projects === */

    /* adder */
    MAP.put(Paths.get("adder", "adder4.v").toString(), new FileDescriptor("", ""));
    MAP.put(Paths.get("adder", "adder4_testbench.v").toString(), new FileDescriptor("", ""));

    /* empty */
    MAP.put(Paths.get("empty", "empty.v").toString(), new FileDescriptor("", ""));

    /* fifo */
    MAP.put(Paths.get("fifo", "fifo.v").toString(), new FileDescriptor(CLK_LOW, RST_LOW));
    MAP.put(Paths.get("fifo", "fifo_testbench.v").toString(), new FileDescriptor(CLK_LOW, RST_LOW));

    /* fifo_mod */
    MAP.put(Paths.get("fifo_mod", "fifo.v").toString(), new FileDescriptor(CLK_LOW, RST_LOW));

    /* for_loop */
    MAP.put(Paths.get("for_loop", "for_loop.v").toString(), new FileDescriptor(CLK_LOW, ""));

    /* incompat_cond */
    MAP.put(Paths.get("incompat_cond", "incompat_cond.v").toString(),
        new FileDescriptor(CLK_LOW, ""));

    /* fifo0 */
    MAP.put(Paths.get("fifo0", "mem_2p.v").toString(), new FileDescriptor("rd_clk", ""));
    MAP.put(Paths.get("fifo0", "mul_fifo.v").toString(), new FileDescriptor(CLK_LOW, RST_LOW));
    MAP.put(Paths.get("fifo0", "myfifo.v").toString(), new FileDescriptor(CLK_LOW, RST_LOW));

    /* mux */
    MAP.put(Paths.get("mux", "mux4_1.v").toString(), new FileDescriptor("", ""));
    MAP.put(Paths.get("mux", "mux_1_testbench.v").toString(), new FileDescriptor("", ""));

    /* ram */
    MAP.put(Paths.get("ram", "ram.v").toString(), new FileDescriptor(CLK_LOW, RST_LOW));
    MAP.put(Paths.get("ram", "ram_testbench.v").toString(), new FileDescriptor(CLK_LOW, RST_LOW));
  }

  @Test
  public void runTest() {
    final Map<String, Map<String, String>> hdlProjects = HdlTestFiles.getProjects();

    for (final Map.Entry<String, Map<String, String>> prjEntry : hdlProjects.entrySet()) {

      for (final Map.Entry<String, String> entry : prjEntry.getValue().entrySet()) {

        final String filePath = entry.getKey();
        final Map<String, String> oneFile = new LinkedHashMap<>();
        oneFile.put(filePath, entry.getValue());
        final FileDescriptor descriptor = getDescriptor(filePath);

        HdlTestUtils.runHdl(
            oneFile,
            TestModel.TYPE.getId(),
            "--" + Retrascope.ENGINE_PARAM.getLongName(),
            CfgRandomTestGenerator.ENGINE_ID,
            "--" + CfgRandomTestGenerator.TEST_LEN.getLongName(),
            String.valueOf(TEST_LENGTH),
            "--" + CfgRandomTestGenerator.CLK_NAME.getLongName(),
            descriptor.getClkName(),
            "--" + CfgRandomTestGenerator.RST_NAME.getLongName(),
            descriptor.getRstName(),
            "--" + CfgRandomTestGenerator.INPUT_VALUES.getLongName(),
            descriptor.getInputValues());
      }
    }
  }

  FileDescriptor getDescriptor(final String filePath) {
    for (final Map.Entry<String, FileDescriptor> entry : MAP.entrySet()) {
      if (filePath.endsWith(entry.getKey())) {
        return entry.getValue();
      }
    }
    return new FileDescriptor("", "");
  }
}
