/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.engine.cgaa.transformer.efsm;

import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Test;
import ru.ispras.retrascope.HdlTestUtils;
import ru.ispras.retrascope.Retrascope;
import ru.ispras.retrascope.basis.exception.RetrascopeRuntimeException;
import ru.ispras.retrascope.model.efsm.EfsmModel;
import ru.ispras.retrascope.util.ReadWriteUtils;

import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Test for transformer from clocked guarded atomic actions model
 * {@link ru.ispras.retrascope.model.cgaa.CgaaModel} to extended finite state machine model
 * {@link ru.ispras.retrascope.model.efsm.EfsmModel}.
 *
 * <p>Test runs the transformer on every top-level module
 * of every HDL file from project test suite.
 * After that it compares statistical info that is stored in debug info file
 * with the reference one.</p>
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class CgaaEfsmTransformerTestCase {

  private static final String STAT_FMT = "states=%d transitions=%d efsms=%d";

  static class EfsmStatistics {
    private final int stateNum;
    private final int transNum;
    private final int modelNum;

    EfsmStatistics(final int stateNum, final int transNum, final int modelNum) {
      this.stateNum = stateNum;
      this.transNum = transNum;
      this.modelNum = modelNum;
    }

    @Override
    public String toString() {
      return String.format(STAT_FMT, this.stateNum, this.transNum, this.modelNum);
    }
  }

  private final Map<String, EfsmStatistics> refDebug = new LinkedHashMap<>();

  private void putToMap(
      final String name,
      final int stateNum,
      final int transNum,
      final int modelNum) {
    this.refDebug.put(name, new EfsmStatistics(stateNum, transNum, modelNum));
  }

  @Test
  public void runTest() {

    final String debugFileName = CgaaEfsmTransformer.ENGINE_ID + ".log";

    putToMap("bvsample", 5, 9, 1);
    putToMap("sample", 8, 12, 1);
    putToMap("test", 6, 11, 1);
    putToMap("driver", 3, 4, 1);
    putToMap("example", 11, 18, 3);
    putToMap("b01", 40, 64, 1);
    putToMap("b02", 31, 48, 1);
    putToMap("b03", 22, 41, 1);
    putToMap("b04", 26, 46, 1);
    putToMap("b05", 166, 284, 3);
    putToMap("b06", 48, 75, 1);
    putToMap("b07", 63, 99, 1);
    putToMap("b08", 153, 234, 1);
    putToMap("b09", 20, 32, 1);
    putToMap("b10", 60, 98, 1);
    putToMap("b11", 96, 155, 1);
//      "states=47 transitions=235 efsms=4", // b12
//      "states=30 transitions=96 efsms=5", // b13
    putToMap("bus_ctrl", 22, 37, 7);
    putToMap("pps_ei", 8, 14, 2);
    putToMap("pps_mem", 11, 16, 6);
    putToMap("pps_pf", 13, 22, 4);
    //putToMap("alu", 8, 19, 2);
    putToMap("bus_mux", 12, 39, 4);
    putToMap("control", 113, 213, 1);
    putToMap("mem_ctrl", 56, 143, 1);
    putToMap("mlite2sram", 34, 65, 3);
    putToMap("pc_next", 10, 20, 1);
    putToMap("pipeline", 36, 68, 1);
    putToMap("reg_bank", 19, 41, 2);
    putToMap("shifter", 32, 58, 1);
    putToMap("sram2mlite", 9, 12, 6);
    putToMap("adder4", 5, 5, 5);
    putToMap("blocks", 3, 3, 3);
    putToMap("bv_non_zero_idx", 5, 8, 2);
    putToMap("empty", 0, 0, 0);
    putToMap("fifo", 12, 32, 4);
    putToMap("fifo_testbench", 5, 5, 4);
//      "states=22 transitions=978 efsms=7", // fifo_mod
    putToMap("for_loop", 1, 1, 1);
    putToMap("incompat_cond", 5, 13, 1);
    putToMap("mux4_1", 1, 1, 1);
    putToMap("ram", 9, 19, 4);
    putToMap("ram_testbench", 5, 5, 4);
    putToMap("data_mem", 3, 4, 2);
//      "states=1 transitions=1 efsms=1", // EX_stage.v
    putToMap("hazard_detection_unit", 4, 7, 1);
//      "states=1 transitions=1 efsms=1", // ID_stage.v
    putToMap("if_stage", 4, 7, 1);
//      "states=1 transitions=1 efsms=1", // instruction_mem.v
//      "states=1 transitions=1 efsms=1", // MEM_stage.v
    putToMap("mips_16_core_top", 0, 0, 0);
    putToMap("mips_16_defs", 0, 0, 0);
//      "states=3 transitions=3 efsms=3" // register_file.v

    /* If reference debug file exists, clear it. */
    final File debugFile = new File(debugFileName);
    if (debugFile.exists()) {
      ReadWriteUtils.clear(debugFile);
    }

    HdlTestUtils.runHdl(EfsmModel.TYPE.getId(),
        "--" + Retrascope.ENGINE_PARAM.getLongName(),
        CgaaEfsmTransformer.ENGINE_ID,
        "--" + Retrascope.TOOL_DEBUG_PARAM.getLongName(),
        debugFileName);

    try {
      final FileInputStream inputStream = new FileInputStream(new File(debugFileName));
      final DataInputStream dataStream = new DataInputStream(inputStream);

      try (final BufferedReader reader = ReadWriteUtils.newBufferedReader(dataStream)) {

        String line;
        while ((line = reader.readLine()) != null) {

          final String topLevelNamesWithPrefix = StringUtils.split(line)[0].toLowerCase();
          final String[] splitByEq = StringUtils.split(topLevelNamesWithPrefix, '=');

          /* NOTE: no-name model has "empty" name. */
          final String topLevelNames = splitByEq.length < 2 ? "empty" : splitByEq[1];

          final String noInfoMsg = String.format("Can't find reference info for %s", topLevelNames);
          Assert.assertTrue(noInfoMsg, this.refDebug.containsKey(topLevelNames));

          final String refString = this.refDebug.get(topLevelNames).toString();
          final String wrongInfoMsg =
              String.format("%s should be equal to %s [%s]", line, refString, topLevelNames);
          Assert.assertTrue(wrongInfoMsg, line.contains(refString));
        }
      }
    } catch (final IOException e) {
      throw new RetrascopeRuntimeException(e);
    }
  }
}
