/*
 * Copyright 2015-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.basis;

import org.junit.Assert;
import org.junit.Test;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.model.cfg.CfgProcess;

import java.util.LinkedHashSet;
import java.util.Set;

/**
 * Test for {@link DeclarationContainer} variable container class.
 *
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class DeclarationContainerTestCase {

  private CfgProcess process;
  private String input1Name;
  private String output2Name;
  private String regName;
  private NodeVariable input1;
  private NodeVariable output2;

  @Test
  public void runTest() {

    getDeclarationTest();
    getNamesTest();
    getContainsTest();
  }

  private void getDeclarationTest() {
    process = new CfgProcess();
    input1Name = "input1";
    output2Name = "output2";
    input1 = NodeVariable.newInteger(input1Name);
    input1.setUserData(new VariableDescriptor(VariableType.IN));
    output2 = NodeVariable.newInteger(output2Name);
    output2.setUserData(new VariableDescriptor(VariableType.OUT));

    process.declareVariable(input1, NodeValue.newInteger(0), new Range(0, 1));
    process.declareVariable(output2);

    Assert.assertTrue(process.getDataType(input1Name) == DataType.INTEGER);

    Assert.assertFalse(process.getInitialValue(input1Name) == null
        || process.getInitialValue(output2Name) != null);

    Assert.assertFalse(process.getInvariant(input1Name) == null
        || process.getInvariant(output2Name) != null);
  }

  /**
   * Test for methods checking whether object contains variable with the specified name or not.
   */
  private void getContainsTest() {
    process = new CfgProcess();
    input1Name = "input1";
    output2Name = "output2";
    regName = "reg";
    input1 = NodeVariable.newInteger(input1Name);
    input1.setUserData(new VariableDescriptor(VariableType.IN));
    output2 = NodeVariable.newInteger(output2Name);
    output2.setUserData(new VariableDescriptor(VariableType.OUT));

    process.declareVariable(input1);
    process.declareVariable(output2);

    Assert.assertFalse(!process.declares(input1Name) && process.declares(regName));
    Assert.assertTrue(process.declaresInput(input1Name));

    Assert.assertTrue(process.declaresOutput(output2Name));

    Assert.assertFalse(process.declaresRegister(regName));
  }

  /**
   * Test for methods returning variable names.
   */
  private void getNamesTest() {

    process = new CfgProcess();
    input1Name = "input1";
    output2Name = "output2";
    regName = "reg1";
    input1 = NodeVariable.newInteger(input1Name);
    input1.setUserData(new VariableDescriptor(VariableType.IN));
    final String input2Name = "input2";
    final NodeVariable input2 = NodeVariable.newInteger(input2Name);
    input2.setUserData(new VariableDescriptor(VariableType.INOUT));
    final String output1Name = "output1";
    final NodeVariable output1 = NodeVariable.newInteger(output1Name);
    output1.setUserData(new VariableDescriptor(VariableType.INOUT));
    output2 = NodeVariable.newInteger(output2Name);
    output2.setUserData(new VariableDescriptor(VariableType.OUT));
    final NodeVariable reg1 = NodeVariable.newInteger(regName);
    reg1.setUserData(new VariableDescriptor(VariableType.REG));

    process.declareVariable(input1);
    process.declareVariable(input2);
    process.declareVariable(output1);
    process.declareVariable(output2);
    process.declareVariable(reg1);

    final Set<String> variableNames = new LinkedHashSet<>();
    variableNames.add(input1Name);
    variableNames.add(input2Name);
    variableNames.add(output1Name);
    variableNames.add(output2Name);
    variableNames.add(regName);

    final Set<String> inputNames = new LinkedHashSet<>();
    inputNames.add(input1Name);
    inputNames.add(input2Name);
    inputNames.add(output1Name);

    final Set<String> outputNames = new LinkedHashSet<>();
    outputNames.add(input2Name);
    outputNames.add(output1Name);
    outputNames.add(output2Name);

    final Set<String> regNames = new LinkedHashSet<>();
    regNames.add(regName);

    Assert.assertTrue(process.getVariableNames().equals(variableNames));
    Assert.assertTrue(process.getInputNames().equals(inputNames));
    Assert.assertTrue(process.getOutputNames().equals(outputNames));
    Assert.assertTrue(process.getRegisterNames().equals(regNames));
  }
}
