/*
 * Copyright 2014-2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.model.cfg;

import org.junit.Assert;
import org.junit.Test;
import ru.ispras.fortress.expression.ExprUtils;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.retrascope.basis.MetaInfoType;
import ru.ispras.retrascope.basis.MetaInfoValue;
import ru.ispras.retrascope.model.basis.Assignment;
import ru.ispras.retrascope.model.basis.Event;
import ru.ispras.retrascope.model.basis.EventList;
import ru.ispras.retrascope.model.basis.Range;
import ru.ispras.retrascope.model.basis.RangedVariable;
import ru.ispras.retrascope.model.basis.Statement;
import ru.ispras.retrascope.model.basis.VariableDescriptor;
import ru.ispras.retrascope.model.basis.VariableType;
import ru.ispras.retrascope.util.SolverUtils;

import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

/**
 * The test constructs control graph model {@link ru.ispras.retrascope.model.cfg.CfgModel}
 * of the module that is described in the following VHDL files:

driver.vhd:
------------------------------

library ieee;
use ieee.std_logic_1164.all;

entity Driver is
port( clk: in std_logic;
reset: in std_logic;
z: out std_logic
);
end Driver;

architecture behav of Driver is
begin
  process(clk)
  begin
    if (clk'event and clk='1') then
      z <= '1';
    else
      assert(reset='1')
      severity ERROR;
    end if;
  end process;
end behav;
-----------------------------------

test.vhd
-----------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity test is
port (clk : in std_logic;
reset :in std_logic
);
end test;

architecture Behavioral of test is
signal c : std_logic_vector(3 downto 0);
begin
  process(clk,reset)
  begin
    if(clk'event and clk='1') then
      if(reset = '0') then
        c(0) <='1';
      end if;
    end if;
    if(reset='1') then
      c(0) <='0';
    end if;
  end process;
end Behavioral;
-----------------------------------

example.vhd
-----------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity example is
port(   clk: in std_logic;
reset: in std_logic;
z: out std_logic;
w: out std_logic
);
end example;

architecture behavior of example is

component driver
port(clk: in std_logic;
reset: in std_logic;
z: out std_logic
);
end component;

component test
port (clk: in std_logic;
reset :in std_logic
);
end component;

begin
  test1: test port map(
  clk => clk,
  reset => reset
  );

  driver1: driver port map(
  clk => clk,
  reset => reset,
  z => z
  );

  process(clk)
  begin
    if (reset='1') then
      w <= reset;
    else
      w <= not reset;
    end if;
  end process;
end behavior;

 @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>

 **/
public class ExampleCfgModelTestCase {

  private static final String STD_VECTOR = "std_logic";
  private static final String STD_3_VECTOR = "std_logic_vector(3 downto 0)";
  private static final String CLK = "clk";
  private static final String RESET = "reset";

  @Test
  public void runTest() {

    final CfgModel model = new CfgModel();

    final CfgModule example = new CfgModule("example");
    example.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);

    /* Variable declarations */
    final NodeVariable clk = NodeVariable.newBoolean(CLK);
    final VariableDescriptor clkData = new VariableDescriptor(VariableType.IN);
    clkData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_VECTOR);
    clk.setUserData(clkData);

    final NodeVariable reset = NodeVariable.newBoolean(RESET);
    final VariableDescriptor resetData = new VariableDescriptor(VariableType.IN);
    resetData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_VECTOR);
    reset.setUserData(resetData);

    final NodeVariable zVariable = NodeVariable.newBoolean("z");
    final VariableDescriptor zData = new VariableDescriptor(VariableType.OUT);
    zData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_VECTOR);
    zVariable.setUserData(zData);

    final NodeVariable wVariable = NodeVariable.newBoolean("w");
    final VariableDescriptor wData = new VariableDescriptor(VariableType.OUT);
    wData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_VECTOR);
    wVariable.setUserData(wData);

    example.declareVariable(clk);
    example.declareVariable(reset);
    example.declareVariable(zVariable);
    example.declareVariable(wVariable);

    /*
     * "example" module process
     */

    final CfgProcess exampleProcess = new CfgProcess();

    final List<Statement> exampleCfg = new LinkedList<>();
    final CfgSwitchStatement eSwitch = new CfgSwitchStatement(reset);
    final CfgCaseStatement c1Case = new CfgCaseStatement(SolverUtils.TRUE);

    final CfgBlockStatement bb1 = new CfgBlockStatement(true);
    bb1.addAssignment(new Assignment(wVariable, reset));

    final CfgCaseStatement c2Case = new CfgCaseStatement(SolverUtils.FALSE);

    final CfgBlockStatement bb2 = new CfgBlockStatement(true);
    bb2.addAssignment(new Assignment(wVariable, ExprUtils.getNegation(reset)));

    exampleCfg.add(eSwitch);
    eSwitch.add(c1Case);
    eSwitch.add(c2Case);
    c1Case.add(bb1);
    c2Case.add(bb2);

    exampleProcess.addAll(exampleCfg);

    /*
     * "test" module instance
     */

    final CfgModule test1 = new CfgModule("test");
    test1.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);

    final NodeVariable testClk = NodeVariable.newBoolean(CLK);
    final VariableDescriptor testClkData = new VariableDescriptor(VariableType.IN);
    testClkData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_VECTOR);
    testClk.setUserData(testClkData);

    final NodeVariable testReset = NodeVariable.newBoolean(RESET);
    final VariableDescriptor testResetData = new VariableDescriptor(VariableType.IN);
    testResetData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_VECTOR);
    testReset.setUserData(testResetData);

    final NodeVariable test1c = NodeVariable.newBitVector("c", 4);
    final VariableDescriptor test1cData = new VariableDescriptor(VariableType.REG);
    test1cData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_3_VECTOR);
    test1c.setUserData(test1cData);

    test1.declareVariable(testClk);
    test1.declareVariable(testReset);
    test1.declareVariable(test1c);

    final Map<String, Node> test1Bindings = new LinkedHashMap<>();
    test1Bindings.put(CLK, clk);
    test1Bindings.put(RESET, reset);
    test1.instantiate("test1", test1Bindings);

    final CfgProcess t1Process = new CfgProcess();
    final EventList t1List = new EventList();
    t1List.add(new Event(new RangedVariable(testClk)));
    t1List.add(new Event(new RangedVariable(testReset)));
    t1Process.setEventList(t1List);

    final List<Statement> t1Cfg = new LinkedList<>();

    final CfgSwitchStatement test1s1 = new CfgSwitchStatement(testClk);
    final CfgCaseStatement test1c1 = new CfgCaseStatement(SolverUtils.TRUE);
    final CfgCaseStatement test1c2 = new CfgCaseStatement(SolverUtils.FALSE);

    final CfgSwitchStatement test1s2 = new CfgSwitchStatement(testReset);
    final CfgCaseStatement test1c3 = new CfgCaseStatement(SolverUtils.FALSE);

    final CfgCaseStatement test1c4 = new CfgCaseStatement(SolverUtils.TRUE);
    final CfgBlockStatement test1bb1 = new CfgBlockStatement(true);
    final NodeValue zero = NodeValue.newInteger(0);
    test1bb1.addAssignment(
        new Assignment(
            new RangedVariable(test1c, new Range(zero, zero)),
            SolverUtils.TRUE));
    final CfgSwitchStatement test1s3 = new CfgSwitchStatement(testReset);
    final CfgCaseStatement test1c5 = new CfgCaseStatement(SolverUtils.TRUE);
    final CfgCaseStatement test1c6 = new CfgCaseStatement(SolverUtils.FALSE);
    final CfgBlockStatement test1bb2 = new CfgBlockStatement(true);
    test1bb2.addAssignment(
        new Assignment(
            new RangedVariable(
                test1c,
                new Range(zero, zero)),
            SolverUtils.FALSE));

    t1Cfg.add(test1s1);
    test1s1.add(test1c1);
    test1s1.add(test1c2);
    test1c1.add(test1s2);
    test1s2.add(test1c3);
    test1s2.add(test1c4);
    test1c3.add(test1bb1);
    t1Cfg.add(test1s3);
    test1s3.add(test1c5);
    test1s3.add(test1c6);
    test1c5.add(test1bb2);

    test1.add(t1Process);
    t1Process.addAll(t1Cfg);
    example.add(test1);

    /* Check 'getStatement(String id)' method. */
    final Statement stmt = CfgUtils.getStatement(test1, test1bb1.getId());
    Assert.assertEquals(test1bb1, stmt);

    /*
     * "driver" module instance
     */
    final CfgModule d1Module = new CfgModule("driver");
    d1Module.addMetaInfo(MetaInfoType.HDL_LANG, MetaInfoValue.LANG_VHDL);

    final NodeVariable d1Clk = NodeVariable.newBoolean(CLK);
    final VariableDescriptor d1ClkData = new VariableDescriptor(VariableType.IN);
    d1ClkData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_VECTOR);
    d1Clk.setUserData(d1ClkData);

    final NodeVariable d1Reset = NodeVariable.newBoolean(RESET);
    final VariableDescriptor d1ResetData = new VariableDescriptor(VariableType.IN);
    d1ResetData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_VECTOR);
    d1Reset.setUserData(d1ResetData);

    final NodeVariable d1Z = NodeVariable.newBoolean("z");
    final VariableDescriptor d1ZData = new VariableDescriptor(VariableType.OUT);
    d1ZData.addMetaInfo(MetaInfoType.HDL_TYPE_STR, STD_VECTOR);
    d1Z.setUserData(d1ZData);

    d1Module.declareVariable(d1Clk);
    d1Module.declareVariable(d1Reset);
    d1Module.declareVariable(d1Z);

    /* Sub-module instantiation. */
    final Map<String, Node> d1b = new LinkedHashMap<>();
    d1b.put(CLK, clk);
    d1b.put(RESET, reset);
    d1b.put(RESET, zVariable);
    d1Module.instantiate("driver1", d1b);

    final CfgProcess d1p = new CfgProcess();
    d1Module.add(d1p);
    final EventList d1pEvents = new EventList();
    d1pEvents.add(new Event(new RangedVariable(d1Clk)));

    final List<Statement> d1pCfg = new LinkedList<>();

    final CfgSwitchStatement d1s1 = new CfgSwitchStatement(d1Clk);
    final CfgCaseStatement d1c1 = new CfgCaseStatement(SolverUtils.TRUE);
    final CfgCaseStatement d1c2 = new CfgCaseStatement(true);
    final CfgBlockStatement d1bb1 = new CfgBlockStatement(true);
    d1bb1.addAssignment(new Assignment(d1Z, SolverUtils.TRUE));
    final CfgSwitchStatement d1s2 = new CfgSwitchStatement(d1Reset);
    final CfgCaseStatement d1c3 = new CfgCaseStatement(SolverUtils.TRUE);
    final CfgCaseStatement d1c4 = new CfgCaseStatement(SolverUtils.FALSE);

    d1pCfg.add(d1s1);
    d1s1.add(d1c1);
    d1s1.add(d1c2);
    d1c1.add(d1bb1);
    d1c2.add(d1s2);
    d1s2.add(d1c3);
    d1s2.add(d1c4);

    d1p.addAll(d1pCfg);
    example.add(d1Module);
    model.addModule(example);
  }
}
