/*
 * Copyright 2016 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ru.ispras.retrascope.util;

import org.junit.Assert;
import org.junit.Test;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.retrascope.SolverUtilTest;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

/**
 * Test for {@link NodeUtils} class.
 * @author <a href="mailto:smolov@ispras.ru">Sergey Smolov</a>
 */
public class NodeUtilTestCase extends SolverUtilTest {

  private static final Node INT_ONE = NodeValue.newInteger(1);
  private static final Node INT_SEVEN = NodeValue.newInteger(7);
  private static final Node BV_SEVEN = NodeValue.newBitVector(BitVector.valueOf("111"));
  private static final Node BV_ONE = NodeValue.newBitVector(BitVector.TRUE);

  private static final Node X_VAR = NodeVariable.newBoolean("x");
  private static final Node Y_VAR = NodeVariable.newBoolean("y");
  private static final Node Z_VAR = NodeVariable.newBoolean("z");
  private static final Node A_VAR = NodeVariable.newBoolean("a");
  private static final Node I_VAR = NodeVariable.newInteger("i");
  private static final Node J_VAR = NodeVariable.newInteger("j");
  private static final Node K_VAR = NodeVariable.newInteger("k");

  @Test
  public void runTest() {

    runCastMethodTest();
    runSplitConjunctionTest();
    runSplitByVarTest();
  }

  private void runSplitByVarTest() {

    final Node iNode = new NodeOperation(StandardOperation.EQ, I_VAR, INT_ONE);
    final Node jNode = new NodeOperation(StandardOperation.GREATER, J_VAR, INT_ONE);
    final Node kNode = new NodeOperation(StandardOperation.LESS, K_VAR, INT_ONE);
    final Node ijNode =
          new NodeOperation(
                StandardOperation.EQ,
                new NodeOperation(StandardOperation.ADD, I_VAR, J_VAR), INT_SEVEN);

    final Collection<Collection<Node>> splitByVarNodes =
        NodeUtils.splitByVar(iNode, jNode, kNode, ijNode);

    final Collection<Collection<Node>> result = new HashSet<>();
    final Collection<Node> ijNodes = new HashSet<>();
    Collections.addAll(ijNodes, iNode, jNode, ijNode);
    result.add(ijNodes);
    final Collection<Node> kNodes = new HashSet<>();
    kNodes.add(kNode);
    result.add(kNodes);

    Assert.assertEquals(splitByVarNodes, result);
  }

  private void runSplitConjunctionTest() {
    final Node multiConjunction =
        Node.AND(
            X_VAR,
            new NodeOperation(
                StandardOperation.AND,
                Y_VAR,
                NodeUtils.getEquation(A_VAR, INT_ONE)));

    final Node repeatNodeExpr = Node.AND(X_VAR, Z_VAR);

    final Node result =
        new NodeOperation(
            StandardOperation.AND,
            X_VAR,
            Y_VAR,
            NodeUtils.getEquation(A_VAR, INT_ONE),
            Z_VAR);

    Assert.assertEquals(result, NodeUtils.getConjunction(multiConjunction, repeatNodeExpr));
  }

  private void runCastMethodTest() {

    /* integer -> boolean */
    Assert.assertTrue(
        isSatEquation(SolverUtils.TRUE, NodeUtils.castToDataType(INT_SEVEN, DataType.BOOLEAN)));

    /* bool -> integer */
    Assert.assertTrue(
        isSatEquation(INT_ONE, NodeUtils.castToDataType(SolverUtils.TRUE, DataType.INTEGER)));

    /* bit vector -> boolean */
    Assert.assertTrue(
        isSatEquation(SolverUtils.TRUE, NodeUtils.castToDataType(BV_SEVEN, DataType.BOOLEAN)));

    /* boolean -> bit vector */
    Assert.assertTrue(
        isSatEquation(BV_ONE, NodeUtils.castToDataType(SolverUtils.TRUE, DataType.BIT_VECTOR(1))));

    /* integer -> bit vector */
    Assert.assertTrue(
        isSatEquation(BV_SEVEN, NodeUtils.castToDataType(INT_SEVEN, DataType.BIT_VECTOR(3))));

    /* bit vector -> integer */
    Assert.assertTrue(
        isSatEquation(INT_SEVEN, NodeUtils.castToDataType(BV_SEVEN, DataType.INTEGER)));
  }

  private boolean isSatEquation(final Node ... exprs) {
    return SolverUtils.isSat(NodeUtils.getEquation(exprs));
  }
}
